// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Net.Imap4.Protocol.Transmission;

namespace Smdn.Net.Imap4.Client {
  partial class ImapClient {
    internal const bool DefaultSubscription = false;
    internal const bool DefaultSelectReadOnly = false;

    [Obsolete("use OpeningMailbox instead")]
    public ImapOpenMailbox OpenedMailbox {
      get { return OpeningMailbox; }
    }

    public ImapOpenMailbox OpeningMailbox {
      get { ThrowIfNotConnected(); return openMailbox; }
    }

    public bool IsIdling {
      get {
        ThrowIfNotConnected();

        if (openMailbox == null)
          return false;
        else
          return openMailbox.IsIdling;
      }
    }

    public ISet<ImapCapability> DisabledCapabilities => disabledCapabilities;

    /*
     * LIST/LSUB
     */
    public ImapMailbox GetInbox(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
      => GetMailbox(ImapMailboxNames.Inbox, options);

    public Task<ImapMailbox> GetInboxAsync(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
      => GetMailboxAsync(ImapMailboxNames.Inbox, options);

    public ImapMailbox GetMailbox(string mailboxName,
                                  ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return RunExclusively(() => GetMailboxCore(mailboxName, options, true, false));
    }

    public Task<ImapMailbox> GetMailboxAsync(string mailboxName,
                                             ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return StartTaskExclusively(() => GetMailboxCore(mailboxName, options, true, false),
                                  CancellationToken.None);
    }

    /// <param name="mailboxName">must be validated by caller</param>
    private ImapMailbox GetMailboxCore(string mailboxName,
                                       ImapMailboxListOptions options,
                                       bool throwIfNotFound,
                                       bool throwIfNonExistentOrUnselectable)
    {
      var mailbox = GetMailboxesCore(mailboxName,
                                     false,
                                     options,
                                     null,
                                     ListMethod.SelectAppropriate,
                                     (m) => ImapMailboxNames.EqualityComparer.Equals(m.Name, mailboxName)).FirstOrDefault();

      if (mailbox == null && throwIfNotFound)
        throw new ImapMailboxNotFoundException(mailboxName);

      if (mailbox != null && throwIfNonExistentOrUnselectable)
        mailbox.ThrowIfNonExistentOrUnselectable();

      return mailbox;
    }

    public ImapMailbox GetMailbox(ImapSpecialMailbox specialMailbox,
                                  ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return RunExclusively(() => GetMailboxCore(specialMailbox, options, true, false));
    }

    public Task<ImapMailbox> GetMailboxAsync(ImapSpecialMailbox specialMailbox,
                                             ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return StartTaskExclusively(() => GetMailboxCore(specialMailbox, options, true, false),
                                  CancellationToken.None);
    }

    private ImapMailbox GetMailboxCore(ImapSpecialMailbox specialMailbox,
                                       ImapMailboxListOptions options,
                                       bool throwIfNotFound,
                                       bool throwIfNonExistentOrUnselectable)
    {
      var specialUseAvailable =
        IsCapabilityAvailable(ImapCapability.SpecialUse) ||
        IsCapabilityAvailable(ImapCapability.GimapXlist); // XXX: Gmail supports SPECIAL-USE but don't list it

      if (!specialUseAvailable)
        throw new ImapIncapableException(ImapCapability.SpecialUse);

      var expectedUseFlag = TranslateSpecialMailboxUseFlag(specialMailbox);
      var selectionOptions = specialUseAvailable ? ImapListSelectionOptions.SpecialUse : null;
      var listMethod = IsCapabilityAvailable(ImapCapability.ListExtended) ? ListMethod.ListExtended : ListMethod.ListOrListRemote;

      var mailbox = GetMailboxesCore(string.Empty,
                                     true,
                                     options,
                                     selectionOptions,
                                     listMethod,
                                     (m) => m.Flags.Contains(expectedUseFlag)).FirstOrDefault();

      if (mailbox == null && throwIfNotFound)
        throw new ImapMailboxNotFoundException(expectedUseFlag.Value);

      if (mailbox != null && throwIfNonExistentOrUnselectable)
        mailbox.ThrowIfNonExistentOrUnselectable();

      return mailbox;
    }

    private static ImapMailboxFlag TranslateSpecialMailboxUseFlag(ImapSpecialMailbox specialMailbox)
    {
      switch (specialMailbox) {
        case ImapSpecialMailbox.All:      return ImapMailboxFlag.All;
        case ImapSpecialMailbox.Archive:  return ImapMailboxFlag.Archive;
        case ImapSpecialMailbox.Drafts:   return ImapMailboxFlag.Drafts;
        case ImapSpecialMailbox.Flagged:  return ImapMailboxFlag.Flagged;
        case ImapSpecialMailbox.Junk:     return ImapMailboxFlag.Junk;
        case ImapSpecialMailbox.Sent:     return ImapMailboxFlag.Sent;
        case ImapSpecialMailbox.Trash:    return ImapMailboxFlag.Trash;

        default:
          throw ExceptionUtils.CreateArgumentMustBeValidEnumValue(nameof(specialMailbox), specialMailbox);
      }
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public ImapMailbox GetOrCreateMailbox(string mailboxName,
                                          ImapMailboxListOptions options)
    {
      return GetOrCreateMailbox(mailboxName,
                                DefaultSubscription,
                                options);
    }

    public ImapMailbox GetOrCreateMailbox(string mailboxName,
                                          bool subscribeIfCreated = DefaultSubscription,
                                          ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return RunExclusively(() => GetOrCreateMailboxCore(mailboxName, subscribeIfCreated, options));
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<ImapMailbox> GetOrCreateMailboxAsync(string mailboxName,
                                                     ImapMailboxListOptions options)
    {
      return GetOrCreateMailboxAsync(mailboxName,
                                     DefaultSubscription,
                                     options);
    }

    public Task<ImapMailbox> GetOrCreateMailboxAsync(string mailboxName,
                                                     bool subscribeIfCreated = DefaultSubscription,
                                                     ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return StartTaskExclusively(() => GetOrCreateMailboxCore(mailboxName, subscribeIfCreated, options),
                                  CancellationToken.None);
    }

    private ImapMailbox GetOrCreateMailboxCore(string mailboxName,
                                               bool subscribeIfCreated,
                                               ImapMailboxListOptions options)
    {
      var mailbox = GetMailboxCore(mailboxName, options, false, false);

      if (mailbox == null)
        return CreateMailboxCore(mailboxName, subscribeIfCreated);
      else
        return mailbox;
    }

#if IMAP_CREATE_SPECAIL_MAILBOX
    public ImapMailbox GetOrCreateMailbox(ImapSpecialMailbox mailbox,
                                          ImapMailboxListOptions options)
    {
      return GetOrCreateMailbox(mailbox,
                                DefaultSubscription,
                                options);
    }

    public ImapMailbox GetOrCreateMailbox(ImapSpecialMailbox mailbox,
                                          bool subscribeIfCreated = DefaultSubscription,
                                          ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      var m = GetMailboxNoException(mailbox, options);

      if (m == null)
        return CreateMailbox(mailbox, subscribeIfCreated);
      else
        return m;
    }
#endif

    public IReadOnlyList<ImapMailbox> GetMailboxes(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return RunExclusively(() => {
        return (IReadOnlyList<ImapMailbox>)GetMailboxesCore(string.Empty,
                                                            true,
                                                            options,
                                                            null,
                                                            ListMethod.SelectAppropriate,
                                                            null).ToList();
      });
    }

    public Task<IReadOnlyList<ImapMailbox>> GetMailboxesAsync(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return StartTaskExclusively(() => {
        return (IReadOnlyList<ImapMailbox>)GetMailboxesCore(string.Empty,
                                                            true,
                                                            options,
                                                            null,
                                                            ListMethod.SelectAppropriate,
                                                            null).ToList();
      }, CancellationToken.None);
    }

    internal IReadOnlyList<ImapMailbox> GetMailboxesCore
    (
      ImapMailboxStatus mailbox,
      ImapMailboxListOptions options
    )
    {
      if (mailbox.Flags.Contains(ImapMailboxFlag.NoInferiors))
        // non hierarchical (ignore \HasNoChildren)
#if NET45
        return ArrayExtensions.Empty<ImapMailbox>();
#else
        return Array.Empty<ImapMailbox>();
#endif
      else if (string.IsNullOrEmpty(mailbox.HierarchyDelimiter))
        // non hierarchical
#if NET45
        return ArrayExtensions.Empty<ImapMailbox>();
#else
        return Array.Empty<ImapMailbox>();
#endif
      else
        return GetMailboxesCore(mailbox.Name + mailbox.HierarchyDelimiter,
                                true,
                                options,
                                null,
                                ListMethod.SelectAppropriate,
                                null).ToList();
    }

    private enum ListMethod {
      SelectAppropriate,
      ListOrListRemote,
      ListExtended,
      ListRemote,
      List,
    }

    private IEnumerable<ImapMailbox> GetMailboxesCore(string mailboxName,
                                                      bool patternMatching,
                                                      ImapMailboxListOptions options,
                                                      ImapListSelectionOptions selectionOptions,
                                                      ListMethod listMethod,
                                                      Predicate<ImapMailboxStatus> match)
    {
      string referenceName, mailboxNamePattern;

      if (patternMatching) {
        referenceName = mailboxName;
        mailboxNamePattern = ((int)(options & ImapMailboxListOptions.TopLevelOnly) != 0)
          ? "%"
          : "*";
      }
      else {
        referenceName = string.Empty;
        mailboxNamePattern = mailboxName;
      }

      TranslateListOptions(options, out var subscribedOnly, out var listRemote, out var requestStatus);

      IReadOnlyList<ImapMailboxStatus> mailboxes;

      switch (listMethod) {
        case ListMethod.SelectAppropriate:
          if (IsCapabilityAvailable(ImapCapability.ListExtended))
            goto case ListMethod.ListExtended;
          else if (listRemote && IsCapabilityAvailable(ImapCapability.MailboxReferrals))
            goto case ListMethod.ListRemote;
          else
            goto case ListMethod.List;

        case ListMethod.ListOrListRemote:
          if (listRemote)
            goto case ListMethod.ListRemote;
          else
            goto case ListMethod.List;

        case ListMethod.ListExtended: {
          var returnOptions = ImapListReturnOptions.Children;

          if (selectionOptions == null)
            selectionOptions = ImapListSelectionOptions.Empty;

          if (subscribedOnly)
            selectionOptions += ImapListSelectionOptions.Subscribed;

          if (listRemote)
            selectionOptions += ImapListSelectionOptions.Remote;

          if (requestStatus && IsCapabilityAvailable(ImapCapability.ListStatus)) {
            returnOptions += ImapListReturnOptions.StatusDataItems(GetAvailableStatusDataItems());
            requestStatus = false; // no need to request STATUS
          }

          ProcessResult(Session.ListExtended(referenceName,
                                             Singleton.CreateList(mailboxNamePattern),
                                             selectionOptions,
                                             returnOptions,
                                             out mailboxes));

          break;
        }

        case ListMethod.ListRemote: {
          if (subscribedOnly)
            ProcessResult(Session.RLsub(referenceName,
                                        mailboxNamePattern,
                                        out mailboxes));
          else
            ProcessResult(Session.RList(referenceName,
                                        mailboxNamePattern,
                                        out mailboxes));

          break;
        }

        case ListMethod.List: {
          if (subscribedOnly)
            ProcessResult(Session.Lsub(referenceName,
                                       mailboxNamePattern,
                                       out mailboxes));
          else
            ProcessResult(Session.List(referenceName,
                                       mailboxNamePattern,
                                       out mailboxes));

          break;
        }

        default:
          throw ExceptionUtils.CreateArgumentMustBeValidEnumValue(nameof(listMethod), listMethod);
      }

      var selectedMailboxName = openMailbox == null ? null : openMailbox.FullName;

      foreach (var m in mailboxes) {
        if (match != null && !match(m))
          continue;

        if (requestStatus && !m.IsUnselectable && !m.NameEquals(selectedMailboxName))
          ProcessResult(Session.Status(m, GetAvailableStatusDataItems()));

        yield return new ImapMailbox(this, m);
      }
    }

    private static void TranslateListOptions(ImapMailboxListOptions options,
                                             out bool subscribedOnly,
                                             out bool listRemote,
                                             out bool requestStatus)
    {
      subscribedOnly  = 0 != (int)(options & ImapMailboxListOptions.SubscribedOnly);
      listRemote      = 0 != (int)(options & ImapMailboxListOptions.Remote);
      requestStatus   = 0 != (int)(options & ImapMailboxListOptions.RequestStatus);
    }

    /*
     * CREATE
     */
    public ImapMailbox CreateMailbox(string mailboxName,
                                     bool subscribeIfCreated = DefaultSubscription)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return RunExclusively(() => CreateMailboxCore(mailboxName, subscribeIfCreated));
    }

    public Task<ImapMailbox> CreateMailboxAsync(string mailboxName,
                                                bool subscribeIfCreated = DefaultSubscription)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return StartTaskExclusively(() => CreateMailboxCore(mailboxName, subscribeIfCreated),
                                  CancellationToken.None);
    }

    private ImapMailbox CreateMailboxCore(string mailboxName,
                                          bool subscribeIfCreated)
    {
      ProcessResult(Session.Create(mailboxName, out var createdMailbox));

      if (subscribeIfCreated)
        ProcessResult(Session.Subscribe(createdMailbox));

      return CreateFlagsRetrieved(createdMailbox);
    }

    internal ImapMailbox CreateFlagsRetrieved(ImapMailboxStatus mailbox)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      // retrieve mailbox flags
      ProcessResult(Session.List(mailbox));

      return new ImapMailbox(this, mailbox);
    }

#if IMAP_CREATE_SPECAIL_MAILBOX
    public ImapMailbox CreateMailbox(ImapSpecialMailbox mailbox,
                                     bool subscribeIfCreated = DefaultSubscription)
    {
      throw new NotImplementedException();
    }
#endif

    /*
     * SELECT/EXAMINE
     */
    public ImapOpenMailbox OpenInbox(bool asReadOnly = DefaultSelectReadOnly)
      => OpenMailbox(ImapMailboxNames.Inbox, asReadOnly: asReadOnly, options: ImapMailboxListOptions.Default);

    public Task<ImapOpenMailbox> OpenInboxAsync(bool asReadOnly = DefaultSelectReadOnly)
      => OpenMailboxAsync(ImapMailboxNames.Inbox, asReadOnly: asReadOnly, options: ImapMailboxListOptions.Default);

    public ImapOpenMailbox OpenMailbox(ImapMailbox mailbox,
                                       bool asReadOnly = DefaultSelectReadOnly)
    {
      if (mailbox == null)
        throw new ArgumentNullException(nameof(mailbox));
      if (mailbox.Client != this)
        throw new ArgumentException("can't open mailbox of different session", nameof(mailbox));

      mailbox.ThrowIfNonExistentOrUnselectable();

      return RunExclusively(() => OpenMailboxCore(mailbox, asReadOnly));
    }

    [Obsolete("use OpenInboxAsync() instead")] public Task<ImapOpenMailbox> OpenMailboxAsync() => OpenInboxAsync();
    [Obsolete("use OpenInboxAsync() instead")] public Task<ImapOpenMailbox> OpenMailboxAsync(bool asReadOnly) => OpenInboxAsync(asReadOnly);

    public Task<ImapOpenMailbox> OpenMailboxAsync(ImapMailbox mailbox,
                                                  bool asReadOnly = DefaultSelectReadOnly)
    {
      if (mailbox == null)
        throw new ArgumentNullException(nameof(mailbox));
      if (mailbox.Client != this)
        throw new ArgumentException("can't open mailbox of different session", nameof(mailbox));

      mailbox.ThrowIfNonExistentOrUnselectable();

      return StartTaskExclusively(() => OpenMailboxCore(mailbox, asReadOnly),
                                  CancellationToken.None);
    }

    /// <param name="mailbox">mailbox must exist and be selectable. must be checked by caller.</param>
    private ImapOpenMailbox OpenMailboxCore(ImapMailbox mailbox,
                                            bool asReadOnly)
    {
      if (IsCapabilityAvailable(ImapCapability.CondStore))
        ProcessResult(asReadOnly
                      ? Session.ExamineCondStore(mailbox.FullName)
                      : Session.SelectCondStore(mailbox.FullName));
      else
        ProcessResult(asReadOnly
                      ? Session.Examine(mailbox.FullName)
                      : Session.Select(mailbox.FullName));

      return SetOpenMailbox(new ImapOpenMailbox(this, Session.SelectedMailbox));
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public ImapOpenMailbox OpenMailbox(string mailboxName,
                                       ImapMailboxListOptions options)
    {
      return OpenMailbox(mailboxName,
                         DefaultSelectReadOnly,
                         options);
    }

    public ImapOpenMailbox OpenMailbox(string mailboxName,
                                       bool asReadOnly = DefaultSelectReadOnly,
                                       ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return RunExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(mailboxName, options, true, true),
                               asReadOnly);
      });
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<ImapOpenMailbox> OpenMailboxAsync(string mailboxName,
                                                  ImapMailboxListOptions options)
    {
      return OpenMailboxAsync(mailboxName,
                              DefaultSelectReadOnly,
                              options);
    }

    public Task<ImapOpenMailbox> OpenMailboxAsync(string mailboxName,
                                                  bool asReadOnly = DefaultSelectReadOnly,
                                                  ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return StartTaskExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(mailboxName, options, true, true),
                               asReadOnly);
      }, CancellationToken.None);
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public ImapOpenMailbox OpenMailbox(ImapSpecialMailbox specialMailbox,
                                       ImapMailboxListOptions options)
    {
      return OpenMailbox(specialMailbox,
                         DefaultSelectReadOnly,
                         options);
    }

    public ImapOpenMailbox OpenMailbox(ImapSpecialMailbox specialMailbox,
                                       bool asReadOnly = DefaultSelectReadOnly,
                                       ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return RunExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(specialMailbox, options, true, true),
                               asReadOnly);
      });
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<ImapOpenMailbox> OpenMailboxAsync(ImapSpecialMailbox specialMailbox,
                                                  ImapMailboxListOptions options)
    {
      return OpenMailboxAsync(specialMailbox,
                              DefaultSelectReadOnly,
                              options);
    }

    public Task<ImapOpenMailbox> OpenMailboxAsync(ImapSpecialMailbox specialMailbox,
                                                  bool asReadOnly = DefaultSelectReadOnly,
                                                  ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return StartTaskExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(specialMailbox, options, true, true),
                               asReadOnly);
      }, CancellationToken.None);
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public ImapOpenMailbox OpenOrCreateMailbox(string mailboxName,
                                               ImapMailboxListOptions options)
    {
      return OpenOrCreateMailbox(mailboxName,
                                 DefaultSubscription,
                                 DefaultSelectReadOnly,
                                 options);
    }

    public ImapOpenMailbox OpenOrCreateMailbox(string mailboxName,
                                               bool subscribeIfCreated = DefaultSubscription, // cannot swap position of subscribeIfCreated and asReadOnly for compatibility
                                               bool asReadOnly = DefaultSelectReadOnly,
                                               ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return RunExclusively(() => OpenOrCreateMailboxCore(mailboxName,
                                                          subscribeIfCreated: subscribeIfCreated,
                                                          asReadOnly: asReadOnly,
                                                          options: options));
    }

    [Obsolete(ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<ImapOpenMailbox> OpenOrCreateMailboxAsync(string mailboxName,
                                                          ImapMailboxListOptions options)
    {
      return OpenOrCreateMailboxAsync(mailboxName,
                                      DefaultSubscription,
                                      DefaultSelectReadOnly,
                                      options);
    }

    public Task<ImapOpenMailbox> OpenOrCreateMailboxAsync(string mailboxName,
                                                          bool subscribeIfCreated = DefaultSubscription, // cannot swap position of subscribeIfCreated and asReadOnly for compatibility
                                                          bool asReadOnly = DefaultSelectReadOnly,
                                                          ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      mailboxName = ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      return StartTaskExclusively(() => OpenOrCreateMailboxCore(mailboxName,
                                                                subscribeIfCreated: subscribeIfCreated,
                                                                asReadOnly: asReadOnly,
                                                                options: options),
                                  CancellationToken.None);
    }

    private ImapOpenMailbox OpenOrCreateMailboxCore(string mailboxName,
                                                    bool subscribeIfCreated,
                                                    bool asReadOnly,
                                                    ImapMailboxListOptions options)
    {
      options &= (~ImapMailboxListOptions.RequestStatus); // no need to request status of mailbox

      var mailbox = GetMailboxCore(mailboxName, options, false, false);

      if (mailbox == null)
        mailbox = CreateMailboxCore(mailboxName, subscribeIfCreated);

      mailbox.ThrowIfNonExistentOrUnselectable();

      return OpenMailboxCore(mailbox, asReadOnly);
    }

#if IMAP_CREATE_SPECAIL_MAILBOX
    public ImapOpenMailbox OpenOrCreateMailbox(ImapSpecialMailbox mailbox,
                                               ImapMailboxListOptions options)
    {
      return OpenOrCreateMailbox(mailbox,
                                 DefaultSubscription,
                                 DefaultSelectReadOnly,
                                 options);
    }

    public ImapOpenMailbox OpenOrCreateMailbox(ImapSpecialMailbox mailbox,
                                               bool subscribeIfCreated = DefaultSubscription, // cannot swap position of subscribeIfCreated and asReadOnly for compatibility
                                               bool asReadOnly = DefaultSelectReadOnly,
                                               ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      options &= (~ImapMailboxListOptions.RequestStatus); // no need to request status of mailbox

      var m = GetMailboxNoException(mailbox, options);

      if (m == null)
        m = CreateMailbox(mailbox, subscribeIfCreated);

      return OpenMailbox(m, asReadOnly);
    }
#endif

    /*
     * CLOSE
     */
    public void CloseMailbox()
    {
      RunExclusively(CloseMailboxCore);
    }

    public Task CloseMailboxAsync()
    {
      return StartTaskExclusively(CloseMailboxCore, CancellationToken.None);
    }

    private void CloseMailboxCore()
    {
      /*
       *       The CLOSE command permanently removes all messages that have the
       *       \Deleted flag set from the currently selected mailbox, and returns
       *       to the authenticated state from the selected state.  No untagged
       *       EXPUNGE responses are sent.
       * 
       *       No messages are removed, and no error is given, if the mailbox is
       *       selected by an EXAMINE command or is otherwise selected read-only.
       */
      if (Session.SelectedMailbox != null)
        ProcessResult(Session.Close());

      SetOpenMailbox(null);
    }

    internal void CloseMailboxIfSelected(ImapMailbox mailbox)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      if (Session.SelectedMailbox != null && Session.SelectedMailbox.NameEquals(mailbox.Mailbox)) {
        ProcessResult(Session.Close());

        SetOpenMailbox(null);
      }
    }

    /*
     * NOOP
     */
    public void Refresh()
    {
      RunExclusively(RefreshCore);
    }

    public Task RefreshAsync()
    {
      return StartTaskExclusively(RefreshCore, CancellationToken.None);
    }

    private void RefreshCore()
    {
      if (openMailbox == null)
        ProcessResult(Session.NoOp());
      else
        openMailbox.ProcessResult(Session.NoOp());
    }

    /*
     * LIST (root)
     */
    /// <remarks><paramref name="rootMailboxName"/> must be validated by caller</remarks>
    private string GetHierarchyDelimiterOf(string rootMailboxName)
    {
      if (Session.HierarchyDelimiters.TryGetValue(rootMailboxName, out var hierarchyDelimiter)) {
        return hierarchyDelimiter;
      }
      else {
        return RunExclusively(() => {
          ProcessResult(Session.ListRoot(rootMailboxName, out var mailboxList));

          return mailboxList.HierarchyDelimiter;
        });
      }
    }

    public string CombineMailboxName(string name1, string name2)
    {
      ImapMailboxNames.GetValidated(name1, nameof(name1));
      ImapMailboxNames.GetValidated(name2, nameof(name2));

      return string.Join(GetHierarchyDelimiterOf(name1), name1, name2);
    }

    public string CombineMailboxName(string name1, string name2, string name3)
    {
      ImapMailboxNames.GetValidated(name1, nameof(name1));
      ImapMailboxNames.GetValidated(name2, nameof(name2));
      ImapMailboxNames.GetValidated(name3, nameof(name3));

      return string.Join(GetHierarchyDelimiterOf(name1), name1, name2, name3);
    }

    public string CombineMailboxName(string name1, string name2, string name3, string name4)
    {
      ImapMailboxNames.GetValidated(name1, nameof(name1));
      ImapMailboxNames.GetValidated(name2, nameof(name2));
      ImapMailboxNames.GetValidated(name3, nameof(name3));
      ImapMailboxNames.GetValidated(name4, nameof(name4));

      return string.Join(GetHierarchyDelimiterOf(name1), name1, name2, name3, name4);
    }

    public string CombineMailboxName(IEnumerable<string> names)
    {
      return CombineMailboxName((IReadOnlyList<string>)names.ToList());
    }

    public string CombineMailboxName(params string[] names)
    {
      return CombineMailboxName((IReadOnlyList<string>)names);
    }

    private string CombineMailboxName(IReadOnlyList<string> names)
    {
      if (names == null)
        throw new ArgumentNullException(nameof(names));
      if (names.Count == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyArray(nameof(names));

      foreach (var n in names)
        ImapMailboxNames.GetValidated(n, nameof(names));

      if (names.Count == 1)
        return names[0];
      else
        return string.Join(GetHierarchyDelimiterOf(names[0]), names);
    }

    /*
     * GETQUOTA
     */
    public ImapQuota GetQuota(string quotaRoot)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));

      if (!IsCapabilityAvailable(ImapCapability.Quota))
        return null; // as default

      return RunExclusively(() => {
        ProcessResult(Session.GetQuota(quotaRoot, out var quota));

        return quota;
      });
    }

    public Task<ImapQuota> GetQuotaAsync(string quotaRoot)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));

      if (!IsCapabilityAvailable(ImapCapability.Quota)) {
        var tcs = new TaskCompletionSource<ImapQuota>();

        tcs.SetResult(null); // as default

        return tcs.Task;
      }

      return StartTaskExclusively(() => {
        ProcessResult(Session.GetQuota(quotaRoot, out var quota));

        return quota;
      }, CancellationToken.None);
    }

    public double GetQuotaUsage(string quotaRoot, string resourceName)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));
      if (resourceName == null)
        throw new ArgumentNullException(nameof(resourceName));
      if (resourceName.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(resourceName));

      return GetQuotaUsage(GetQuota(quotaRoot), resourceName);
    }

    public Task<double> GetQuotaUsageAsync(string quotaRoot, string resourceName)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));
      if (resourceName == null)
        throw new ArgumentNullException(nameof(resourceName));
      if (resourceName.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(resourceName));

      return GetQuotaAsync(quotaRoot).ContinueWith(t => GetQuotaUsage(t.Result, resourceName));
    }

    private double GetQuotaUsage(ImapQuota quota, string resourceName)
    {
      if (quota == null)
        return 0.0; // as default

      var resource = quota.Resources.FirstOrDefault(r => string.Equals(r.Name, resourceName, StringComparison.OrdinalIgnoreCase));

      if (resource.Name == null /*resource == default(ImapQuotaResourceUsage)*/)
        throw new ImapProtocolViolationException(string.Format("no such resource name: '{0}'", resourceName));

      return resource.UsageRatio;
    }

    /*
     * utilities
     */
    private ImapOpenMailbox SetOpenMailbox(ImapOpenMailbox newlyOpenedMailbox)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      if (openMailbox != null)
        openMailbox.DetachAllMailboxObjects();

      openMailbox = newlyOpenedMailbox;

      return openMailbox;
    }

    internal protected virtual ImapCommandResult ProcessResult(ImapCommandResult result,
                                                               Func<ImapTaggedStatusResponse, bool> throwIfErrorResponse = null)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      var alert = result.GetStatusResponseOfCode(ImapResponseCode.Alert);

      if (alert != null)
        RaiseEvent(EventType.AlertReceived,
                   new ImapAlertReceivedEventArgs(alert.ResponseText.Text, alert.Condition),
                   false);

      if (result.IsSuccess)
        return result;

      if (throwIfErrorResponse != null && result.TaggedStatusResponse != null)
        if (!throwIfErrorResponse(result.TaggedStatusResponse))
          return result;

      if (result.IsFinalResponseCondition(ImapResponseCondition.Bye))
        throw new ImapConnectionException(string.Concat("disconnected from server: ", result.Description));
      else
        throw new ImapErrorResponseException(result);
    }

    internal bool IsCapabilityAvailable(ImapCapability requiredCapability)
    {
      if (disabledCapabilities.Contains(requiredCapability))
        return false;

      return ServerCapabilities.Contains(requiredCapability);
    }

    internal void ThrowIfCapabilityUnavailable(IImapExtension extension)
    {
      foreach (var requiredCapability in extension.RequiredCapabilities) {
        ThrowIfCapabilityUnavailable(requiredCapability);
      }
    }

    internal void ThrowIfCapabilityUnavailable(ImapCapability requiredCapability)
    {
      if (!IsCapabilityAvailable(requiredCapability))
        throw new ImapIncapableException(requiredCapability);
    }

    internal ImapStatusDataItems GetAvailableStatusDataItems()
    {
      if (IsCapabilityAvailable(ImapCapability.CondStore))
        return ImapStatusDataItems.StandardAll + ImapStatusDataItems.HighestModSeq;
      else
        return ImapStatusDataItems.StandardAll;
    }

    private ImapOpenMailbox openMailbox;
    private readonly ImapCapabilitySet disabledCapabilities = new ImapCapabilitySet(8);
  }
}
