// 
// Copyright (c) 2019 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.IO;
using Smdn.IO.Streams;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Text;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  public sealed class ImapMessageBody : ImapMessageBodyBase {
    internal ImapMessageBody(ImapMessage message)
      : base(message)
    {
    }

    internal ImapMessageBody(ImapMessagePart messagePart)
      : base(messagePart)
    {
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Message), Message),
        (nameof(PartSpecifier), PartSpecifier),
      });
    }

    private void TranslateFetchSection(
      ImapMessageFetchBodyOptions options,
      out string fetchSection
    )
    {
      // https://tools.ietf.org/html/rfc3501#section-6.4.5
      // 6.4.5.  FETCH Command
      //          The HEADER, HEADER.FIELDS, HEADER.FIELDS.NOT, and TEXT part
      //          specifiers can be the sole part specifier or can be prefixed by
      //          one or more numeric part specifiers, provided that the numeric
      //          part specifier refers to a part of type MESSAGE/RFC822.
      // 
      // 6.4.5.  FETCH Command
      //          The TEXT part specifier refers to the text body of the message,
      //          omitting the [RFC-2822] header.
      //
      // 9.      Formal Syntax
      // section         = "[" [section-spec] "]"
      // section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
      //                   "TEXT"
      //                     ; top-level or MESSAGE/RFC822 part
      // section-part    = nz-number *("." nz-number)
      //                     ; body part nesting
      // section-spec    = section-msgtext / (section-part ["." section-text])
      // section-text    = section-msgtext / "MIME"
      //                     ; text other than actual body part (headers, etc.)
      if ((int)(options & ImapMessageFetchBodyOptions.OmitHeader) == 0) {
        // fetch entire message or specific message part
        fetchSection = PartSpecifier;
      }
      else {
        // fetch body of message
        if (IsOfTopLevel)
          fetchSection = "TEXT";
        //else if (IsOfEncapsulatedMessagePart)
        //  fetchSection = MessagePart.ParentPart.Specifier + ".TEXT";
        else
          fetchSection = MessagePart.Specifier;
          // throw new ImapProtocolViolationException("TEXT can be specified only with top-level or message/rfc822 part");
      }
    }

    private static TResult ReadAsConverter<TArg, TResult>(
      Stream bodyStream,
      Encoding contentCharset,
      ImapMessageAttributes _,
      (Func<Stream, Encoding, TArg, TResult> converter, TArg arg) args
    )
      => args.converter(bodyStream, contentCharset, args.arg);

    public TResult ReadAs<TArg, TResult>(
      Func<Stream, Encoding, TArg, TResult> converter,
      TArg arg,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    )
    {
      if (converter == null)
        throw new ArgumentNullException(nameof(converter));

      TranslateFetchSection(options, out var fetchSection);

      return ReadMessageBody(
        setBusy: true,
        messageStreamConverter: ReadAsConverter,
        messageStreamConverterArgument: (converter, arg),
        options: options,
        selectFallbackCharset: selectFallbackCharset,
        forceCharsetTo: forceCharsetTo,
        fetchSection: fetchSection,
        fetchRange: fetchRange,
        cancellationToken: cancellationToken,
        progress: progress
      );
    }

    public Task<TResult> ReadAsAsync<TResult, TArg>(
      Func<Stream, Encoding, TArg, Task<TResult>> converter,
      TArg arg,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    )
    {
      if (converter == null)
        throw new ArgumentNullException(nameof(converter));

      TranslateFetchSection(options, out var fetchSection);

      return ReadMessageBodyAsync(
        setBusy: true,
        messageStreamConverter: ReadAsConverter,
        messageStreamConverterArgument: (converter, arg),
        options: options,
        selectFallbackCharset: selectFallbackCharset,
        forceCharsetTo: forceCharsetTo,
        fetchSection: fetchSection,
        fetchRange: fetchRange,
        cancellationToken: cancellationToken,
        progress: progress
      );
    }

    /*
     * methods for reading body as binary
     */
    private static int ReadConverter(Stream bodyStream, Encoding _, Action<Stream> read)
    {
      read(bodyStream);

      return default;
    }

    private static async Task<int> ReadAsyncConverter(Stream bodyStream, Encoding _, Func<Stream, Task> read)
    {
      await read(bodyStream).ConfigureAwait(false);

      return default;
    }

    public void Read(
      Action<Stream> read,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadConverter,
      arg: read ?? throw new ArgumentNullException(nameof(read)),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task ReadAsync(
      Func<Stream, Task> read,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsyncConverter,
      arg: read ?? throw new ArgumentNullException(nameof(read)),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static TResult ReadAsConverter<TResult>(Stream bodyStream, Encoding _, Converter<Stream, TResult> converter)
      => converter(bodyStream);

    public TResult ReadAs<TResult>(
      Converter<Stream, TResult> converter,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadAsConverter,
      arg: converter ?? throw new ArgumentNullException(nameof(converter)),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<TResult> ReadAsAsync<TResult>(
      Converter<Stream, Task<TResult>> converter,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsConverter,
      arg: converter ?? throw new ArgumentNullException(nameof(converter)),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static TResult ReadAsConverter<TArg, TResult>(Stream bodyStream, Encoding _, (Func<Stream, TArg, TResult> converter, TArg arg) args)
      => args.converter(bodyStream, args.arg);

    public TResult ReadAs<TArg, TResult>(
      Func<Stream, TArg, TResult> converter,
      TArg arg,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadAsConverter,
      arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<TResult> ReadAsAsync<TArg, TResult>(
      Func<Stream, TArg, Task<TResult>> converter,
      TArg arg,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsConverter,
      arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static byte[] ReadAsByteArrayConverter(Stream bodyStream) => bodyStream.ReadToEnd();
    private static Task<byte[]> ReadAsByteArrayAsyncConverter(Stream bodyStream) => bodyStream.ReadToEndAsync();

    public byte[] ReadAsByteArray(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadAsByteArrayConverter,
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<byte[]> ReadAsByteArrayAsync(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsByteArrayAsyncConverter,
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    [Obsolete]
    internal static Stream Download(Stream bodyStream) => DownloadConverter(bodyStream); // TODO: remove me

    [Obsolete]
    internal static Task<Stream> DownloadAsync(Stream bodyStream) => DownloadAsyncConverter(bodyStream); // TODO: remove me

    private static Stream DownloadConverter(Stream bodyStream)
    {
      var stream = new ChunkedMemoryStream();

      CopyFetchMessageBodyStream(bodyStream, stream);

      stream.Position = 0L;

      return stream;
    }

    private static async Task<Stream> DownloadAsyncConverter(Stream bodyStream)
    {
      var stream = new ChunkedMemoryStream();

      await CopyFetchMessageBodyStreamAsync(bodyStream, stream).ConfigureAwait(false);

      stream.Position = 0L;

      return stream;
    }

    public Stream Download(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: DownloadConverter,
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<Stream> DownloadAsync(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: DownloadAsyncConverter,
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static int SaveConverter(Stream bodyStream, string path)
    {
      using (var fileStream = new FileStream(path, FileMode.Create, FileAccess.Write)) {
        CopyFetchMessageBodyStream(bodyStream, fileStream);

        return default;
      }
    }

    private static async Task<int> SaveAsyncConverter(Stream bodyStream, string path)
    {
      using (var fileStream = new FileStream(path, FileMode.Create, FileAccess.Write)) {
        await CopyFetchMessageBodyStreamAsync(bodyStream, fileStream).ConfigureAwait(false);

        return default;
      }
    }

    public void Save(
      string path,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    )
    {
      if (path == null)
        throw new ArgumentNullException(nameof(path));
      if (path.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(path));

      ReadAs(
        converter: SaveConverter,
        arg: path,
        options: options,
        fetchRange: fetchRange,
        cancellationToken: cancellationToken,
        progress: progress
      );
    }

    public Task SaveAsync(
      string path,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    )
    {
      if (path == null)
        throw new ArgumentNullException(nameof(path));
      if (path.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(path));

      return ReadAsAsync(
        converter: SaveAsyncConverter,
        arg: path,
        options: options,
        fetchRange: fetchRange,
        cancellationToken: cancellationToken,
        progress: progress
      );
    }

    public void WriteTo(
      Stream stream,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: CopyFetchMessageBodyStream,
      arg: ThrowIfStreamIsNullOrNotWritable(stream),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task WriteToAsync(
      Stream stream,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: CopyFetchMessageBodyStreamAsync,
      arg: ThrowIfStreamIsNullOrNotWritable(stream),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public void WriteTo(
      BinaryWriter writer,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: CopyFetchMessageBodyStream,
      arg: writer ?? throw new ArgumentNullException(nameof(writer)),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task WriteToAsync(
      BinaryWriter writer,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      ImapFetchRange fetchRange = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: CopyFetchMessageBodyStreamAsync,
      arg: writer ?? throw new ArgumentNullException(nameof(writer)),
      options: options,
      fetchRange: fetchRange,
      cancellationToken: cancellationToken,
      progress: progress
    );

    /*
     * methods for reading body as text
     */
    private static int ReadConverter(Stream bodyStream, Encoding contentCharset, Action<StreamReader> read)
    {
      read(new StreamReader(bodyStream, contentCharset ?? DefaultCharset));

      return default;
    }

    private static async Task<int> ReadAsyncConverter(Stream bodyStream, Encoding contentCharset, Func<StreamReader, Task> read)
    {
      await read(new StreamReader(bodyStream, contentCharset ?? DefaultCharset)).ConfigureAwait(false);

      return default;
    }

    public void Read(
      Action<StreamReader> read,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadConverter,
      arg: read ?? throw new ArgumentNullException(nameof(read)),
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task ReadAsync(
      Func<StreamReader, Task> read,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsyncConverter,
      arg: read ?? throw new ArgumentNullException(nameof(read)),
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static TResult ReadAsConverter<TResult>(Stream bodyStream, Encoding contentCharset, Converter<StreamReader, TResult> converter)
      => converter(new StreamReader(bodyStream, contentCharset ?? DefaultCharset));

    public TResult ReadAs<TResult>(
      Converter<StreamReader, TResult> converter,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadAsConverter,
      arg: converter ?? throw new ArgumentNullException(nameof(converter)),
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<TResult> ReadAsAsync<TResult>(
      Converter<StreamReader, Task<TResult>> converter,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsConverter,
      arg: converter ?? throw new ArgumentNullException(nameof(converter)),
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static TResult ReadAsConverter<TArg, TResult>(Stream bodyStream, Encoding contentCharset, (Func<StreamReader, TArg, TResult> converter, TArg arg) args)
      => args.converter(new StreamReader(bodyStream, contentCharset ?? DefaultCharset), args.arg);

    public TResult ReadAs<TArg, TResult>(
      Func<StreamReader, TArg, TResult> converter,
      TArg arg,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadAsConverter,
      arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg),
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<TResult> ReadAsAsync<TArg, TResult>(
      Func<StreamReader, TArg, Task<TResult>> converter,
      TArg arg,
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsConverter,
      arg: (converter ?? throw new ArgumentNullException(nameof(converter)), arg),
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static IReadOnlyList<string> ReadAsStringLinesConverter(StreamReader reader) => reader.ReadAllLines();
    private static Task<IReadOnlyList<string>> ReadAsStringLinesAsyncConverter(StreamReader reader) => reader.ReadAllLinesAsync();

    public IReadOnlyList<string> ReadAsStringLines(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadAsStringLinesConverter,
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<IReadOnlyList<string>> ReadAsStringLinesAsync(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsStringLinesAsyncConverter,
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    private static string ReadAsStringConverter(StreamReader reader) => reader.ReadToEnd();
    private static Task<string> ReadAsStringAsyncConverter(StreamReader reader) => reader.ReadToEndAsync();

    public string ReadAsString(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAs(
      converter: ReadAsStringConverter,
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );

    public Task<string> ReadAsStringAsync(
      ImapMessageFetchBodyOptions options = ImapMessageFetchBodyOptions.Default,
      EncodingSelectionCallback selectFallbackCharset = null,
      Encoding forceCharsetTo = null,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = null
    ) =>
    ReadAsAsync(
      converter: ReadAsStringAsyncConverter,
      options: options,
      selectFallbackCharset: selectFallbackCharset,
      forceCharsetTo: forceCharsetTo,
      cancellationToken: cancellationToken,
      progress: progress
    );
  }
}
