// 
// Copyright (c) 2019 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Formats.Mime;
using Smdn.IO;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  public abstract partial class ImapMessageBodyBase {
    public static readonly Encoding DefaultCharset = Encoding.GetEncoding("ISO-8859-1");

    internal static Stream CopyFetchMessageBodyStream(Stream bodyStream, Stream destination)
    {
      if (bodyStream is ImapFetchMessageBodyStream fetchBodyStream)
        fetchBodyStream.ReadToEnd(destination);
      else
        bodyStream.CopyTo(destination);

      return bodyStream;
    }

    private protected static Task<Stream> CopyFetchMessageBodyStreamAsync(Stream bodyStream, Stream destination)
      => Task.FromResult(CopyFetchMessageBodyStream(bodyStream, destination)); // TODO: async

    private protected static Stream CopyFetchMessageBodyStream(Stream bodyStream, BinaryWriter destination)
    {
      if (bodyStream is ImapFetchMessageBodyStream fetchBodyStream)
        fetchBodyStream.ReadToEnd(destination);
      else
        bodyStream.CopyTo(destination);

      return bodyStream;
    }

    private protected static Task<Stream> CopyFetchMessageBodyStreamAsync(Stream bodyStream, BinaryWriter destination)
      => Task.FromResult(CopyFetchMessageBodyStream(bodyStream, destination)); // TODO: async

    private protected static Stream ThrowIfStreamIsNullOrNotWritable(Stream stream)
    {
      if (stream == null)
        throw new ArgumentNullException(nameof(stream));
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream(nameof(stream));

      return stream;
    }

    /*
     * instance members
     */
    public ImapMessage Message { get; }
    public ImapMessagePart MessagePart { get; }

    public bool IsOfTopLevel => MessagePart == null;
    public bool IsOfEncapsulatedMessagePart => MessagePart != null && MessagePart.IsEncapsulatedMessagePart;

    public string PartSpecifier => MessagePart?.Specifier;

    private protected ImapMessageBodyBase(ImapMessage message)
    {
      Message = message;
      MessagePart = null;
    }

    private protected ImapMessageBodyBase(ImapMessagePart messagePart)
    {
#if DEBUG
      if (messagePart.IsRootPart)
        throw new ArgumentException("use .ctor(ImapMessage)");
#endif

      Message = messagePart.Message;
      MessagePart = messagePart;
    }

    private void ValidateAndTranslateFetchOptions(ImapFetchRange fetchRange,
                                                  ImapMessageFetchBodyOptions options,
                                                  EncodingSelectionCallback selectFallbackCharset,
                                                  Encoding forceContentCharsetTo,
                                                  out bool fetchPeek,
                                                  out string contentTransferEncoding,
                                                  out Encoding contentCharset)
    {
      contentTransferEncoding = null;
      contentCharset = null;

      fetchPeek = ((int)(options & ImapMessageFetchBodyOptions.SetSeen) == 0);

      if ((int)(options & ImapMessageFetchBodyOptions.DecodeContent) == 0) {
        contentCharset = forceContentCharsetTo ?? DefaultCharset;
      }
      else { // if decode content
        if (IsOfTopLevel)
          throw new NotSupportedException("ImapMessage + ImapMessageFetchBodyOptions.DecodeContent is not supported");
        if (fetchRange.IsPartial)
          throw new NotSupportedException("ImapFetchRange + ImapMessageFetchBodyOptions.DecodeContent is not supported");

        contentTransferEncoding = MessagePart.ContentTransferEncoding;

        if (forceContentCharsetTo == null) {
          var contentTypeCharsetString = MessagePart.ContentTypeCharset;

          if (contentTypeCharsetString == null)
            contentCharset = DefaultCharset;
          else
            contentCharset = EncodingUtils.GetEncodingThrowException(contentTypeCharsetString, selectFallbackCharset);
        }
        else {
          contentCharset = forceContentCharsetTo;
        }
      }
    }

    private static Stream CreateDecodingStream(Stream stream, string contentTransferEncoding)
    {
      if (contentTransferEncoding == null)
        return stream;
      else
        return ContentTransferEncoding.CreateDecodingStream(stream, contentTransferEncoding);
    }

    internal TResult ReadMessageBody<TResult, TArg>(
      bool setBusy,
      Func<Stream, Encoding, ImapMessageAttributes, TArg, TResult> messageStreamConverter,
      TArg messageStreamConverterArgument,
      ImapMessageFetchBodyOptions options = default,
      EncodingSelectionCallback selectFallbackCharset = default,
      Encoding forceCharsetTo = default,
      string fetchSection = default,
      ImapFetchRange fetchRange = default,
      ImapFetchDataItems fetchDataItems = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = default)
    {
      Message.ThrowIfSequenceOrUidSetIsNotAvailable(out _);

      ValidateAndTranslateFetchOptions(fetchRange,
                                       options,
                                       selectFallbackCharset,
                                       forceCharsetTo,
                                       out var fetchPeek,
                                       out var contentTransferEncoding,
                                       out var contentCharset);

      if (setBusy)
        Message.Client.SetBusy();

      using (var stream = new FetchMessageBodyStream(Message,
                                                     fetchPeek,
                                                     fetchSection,
                                                     fetchRange,
                                                     cancellationToken,
                                                     progress,
                                                     unsetBusy: setBusy)) {
        var (_, messageAttributes) = stream.Prepare(fetchDataItems);

        cancellationToken.ThrowIfCancellationRequested();

        return messageStreamConverter(
          CreateDecodingStream(stream, contentTransferEncoding),
          contentCharset,
          messageAttributes,
          messageStreamConverterArgument
        );
      }
    }

    private protected Task<TResult> ReadMessageBodyAsync<TResult, TArg>(
      bool setBusy,
      Func<Stream, Encoding, ImapMessageAttributes, TArg, Task<TResult>> messageStreamConverter,
      TArg messageStreamConverterArgument,
      ImapMessageFetchBodyOptions options = default,
      EncodingSelectionCallback selectFallbackCharset = default,
      Encoding forceCharsetTo = default,
      string fetchSection = default,
      ImapFetchRange fetchRange = default,
      ImapFetchDataItems fetchDataItems = default,
      CancellationToken cancellationToken = default,
      IProgress<ImapMessageFetchProgressInfo> progress = default)
    {
      Message.ThrowIfSequenceOrUidSetIsNotAvailable(out _);

      ValidateAndTranslateFetchOptions(fetchRange,
                                       options,
                                       selectFallbackCharset,
                                       forceCharsetTo,
                                       out var fetchPeek,
                                       out var contentTransferEncoding,
                                       out var contentCharset);

      if (!setBusy)
        throw new NotSupportedException("runAsync && !setBusy is not supported");

      return Message.Client.RunExclusivelyAsync(async () => {
        using (var stream = new FetchMessageBodyStream(Message,
                                                       fetchPeek,
                                                       fetchSection,
                                                       fetchRange,
                                                       cancellationToken,
                                                       progress,
                                                       unsetBusy: true)) {
          var (_, messageAttributes) = stream.Prepare(fetchDataItems);

          cancellationToken.ThrowIfCancellationRequested();

          return await messageStreamConverter(
            CreateDecodingStream(stream, contentTransferEncoding),
            contentCharset,
            messageAttributes,
            messageStreamConverterArgument
          ).ConfigureAwait(false);
        }
      }, false);
    }
  }
}
