// 
// Copyright (c) 2019 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Collections;
using Smdn.Formats.Mime;
using Smdn.IO.Streams;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Text;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  public sealed class ImapMessageHeader : ImapMessageBodyBase {
    private const IEnumerable<string> allHeaderFields = null;

    private bool IsOfNonCapsulatedMessagePart => !(IsOfTopLevel || MessagePart.IsEncapsulatedMessagePart);

    internal ImapMessageHeader(ImapMessage message)
      : base(message)
    {
    }

    internal ImapMessageHeader(ImapMessagePart messagePart)
      : base(messagePart)
    {
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Message), Message),
        (nameof(PartSpecifier), PartSpecifier),
      });
    }

    private bool TranslateFetchSectionAndHeaderFields(
      IEnumerable<string> headerFields,
      bool excludePassedHeaderFields,
      out string fetchSectionAndHeaderFields)
    {
      fetchSectionAndHeaderFields = default;

      const string fetchAllHeaderFields = null;
      string joinedFetchHeaderFields;

      if (headerFields == allHeaderFields) {
        if (excludePassedHeaderFields)
          return false; // fetch no header fields
        else
          joinedFetchHeaderFields = fetchAllHeaderFields;
      }
      else {
        joinedFetchHeaderFields = string.Join(" ", headerFields);

        if (joinedFetchHeaderFields.Length == 0) { // headerFields is empty
          if (excludePassedHeaderFields)
            joinedFetchHeaderFields = fetchAllHeaderFields;
          else
            return false; // fetch no header fields
        }
      }

      // https://tools.ietf.org/html/rfc3501#section-6.4.5
      // 6.4.5.  FETCH Command
      //          The HEADER, HEADER.FIELDS, and HEADER.FIELDS.NOT part
      //          specifiers refer to the [RFC-2822] header of the message or of
      //          an encapsulated [MIME-IMT] MESSAGE/RFC822 message.
      //
      // 6.4.5.  FETCH Command
      //          The MIME part specifier refers to the [MIME-IMB] header for
      //          this part.
      //
      // 9.      Formal Syntax
      // section         = "[" [section-spec] "]"
      // section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
      //                   "TEXT"
      //                     ; top-level or MESSAGE/RFC822 part
      // section-part    = nz-number *("." nz-number)
      //                     ; body part nesting
      // section-spec    = section-msgtext / (section-part ["." section-text])
      // section-text    = section-msgtext / "MIME"
      //                     ; text other than actual body part (headers, etc.)
      if (IsOfNonCapsulatedMessagePart)
        joinedFetchHeaderFields = fetchAllHeaderFields; // ignore specification of headerFields if MessagePart is not message/rfc822 part

      if (joinedFetchHeaderFields == fetchAllHeaderFields) {
        if (IsOfTopLevel)
          fetchSectionAndHeaderFields = "HEADER";
        //else if (IsOfEncapsulatedMessagePart)
        //  fetchSectionAndHeaderFields = MessagePart.ParentPart.Specifier + ".HEADER";
        else
          fetchSectionAndHeaderFields = MessagePart.Specifier + ".MIME";
      }
      else {
        if (IsOfTopLevel)
          fetchSectionAndHeaderFields = string.Empty;
        else if (IsOfEncapsulatedMessagePart)
          fetchSectionAndHeaderFields = MessagePart.ParentPart.Specifier + ".";
        else
          throw new ImapProtocolViolationException("HEADER.FIELDS/HEADER.FIELDS.NOT can be specified only with top-level or message/rfc822 part");

        if (excludePassedHeaderFields)
          fetchSectionAndHeaderFields += "HEADER.FIELDS.NOT (" + joinedFetchHeaderFields + ")";
        else
          fetchSectionAndHeaderFields += "HEADER.FIELDS (" + joinedFetchHeaderFields + ")";
      }

      return true;
    }

    /*
     * ReadAs<TArg, TResult>(Func<Stream, TArg, TResult> converter)
     */
    private static TResult ReadAsConverter<TArg, TResult>(
      Stream headerStream,
      Encoding _0,
      ImapMessageAttributes _1,
      (Func<Stream, TArg, TResult> converter, TArg arg) args
    )
      => args.converter(headerStream, args.arg);

    public TResult ReadAs<TArg, TResult>(
      Func<Stream, TArg, TResult> converter,
      TArg arg,
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    )
    {
      var converterArgument = (converter ?? throw new ArgumentNullException(nameof(converter)), arg);

      if (!TranslateFetchSectionAndHeaderFields(headerFields: headerFields,
                                                excludePassedHeaderFields: excludePassedHeaderFields,
                                                out var fetchSectionAndHeaderFields))
        return ReadAsConverter(Stream.Null, null, null, converterArgument); // fetch nothing

      return ReadMessageBody(
        setBusy: true,
        messageStreamConverter: ReadAsConverter,
        messageStreamConverterArgument: converterArgument,
        fetchSection: fetchSectionAndHeaderFields,
        cancellationToken: cancellationToken
      );
    }

    public Task<TResult> ReadAsAsync<TArg, TResult>(
      Func<Stream, TArg, Task<TResult>> converter,
      TArg arg,
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    )
    {
      var converterArgument = (converter ?? throw new ArgumentNullException(nameof(converter)), arg);

      if (!TranslateFetchSectionAndHeaderFields(headerFields: headerFields,
                                                excludePassedHeaderFields: excludePassedHeaderFields,
                                                out var fetchSectionAndHeaderFields))
        return ReadAsConverter(Stream.Null, null, null, converterArgument); // fetch nothing

      return ReadMessageBodyAsync(
        setBusy: true,
        messageStreamConverter: ReadAsConverter,
        messageStreamConverterArgument: converterArgument,
        fetchSection: fetchSectionAndHeaderFields,
        cancellationToken: cancellationToken
      );
    }

    /*
     * ReadAs<TResult>(Converter<Stream, TResult> converter)
     */
    private static TResult ReadAsConverter<TResult>(Stream headerStream, Converter<Stream, TResult> converter)
      => converter(headerStream);

    public TResult ReadAs<TResult>(
      Converter<Stream, TResult> converter,
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAs(
      converter: ReadAsConverter,
      arg: converter ?? throw new ArgumentNullException(nameof(converter)),
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    public Task<TResult> ReadAsAsync<TResult>(
      Converter<Stream, Task<TResult>> converter,
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAsAsync(
      converter: ReadAsConverter,
      arg: converter ?? throw new ArgumentNullException(nameof(converter)),
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    /*
     * ReadAsString
     */
    private static string ReadAsStringConverter(Stream headerStream)
    {
      using (var reader = new StreamReader(headerStream, DefaultCharset, false)) {
        return reader.ReadToEnd();
      }
    }

    private static async Task<string> ReadAsStringAsyncConverter(Stream headerStream)
    {
      using (var reader = new StreamReader(headerStream, DefaultCharset, false)) {
        return await reader.ReadToEndAsync().ConfigureAwait(false);
      }
    }

    public string ReadAsString(
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAs(
      converter: ReadAsStringConverter,
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    public Task<string> ReadAsStringAsync(
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAsAsync(
      converter: ReadAsStringAsyncConverter,
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    /*
     * ReadAsNameValuePairs
     */
    private static IReadOnlyList<KeyValuePair<string, string>> ReadAsNameValuePairsConverter(
      Stream headerStream,
      (
        IEnumerable<string> headerFields,
        bool excludePassedHeaderFields,
        bool decodeHeaderFieldValues,
        EncodingSelectionCallback selectFallbackCharset,
        CancellationToken cancellationToken
      ) arg
    ) =>
    GetParsedHeaders(
      headerStream,
      arg.headerFields,
      arg.excludePassedHeaderFields,
      arg.decodeHeaderFieldValues,
      arg.selectFallbackCharset,
      arg.cancellationToken
    ).ToList();

    private static Task<IReadOnlyList<KeyValuePair<string, string>>> ReadAsNameValuePairsAsyncConverter(
      Stream headerStream,
      (
        IEnumerable<string> headerFields,
        bool excludePassedHeaderFields,
        bool decodeHeaderFieldValues,
        EncodingSelectionCallback selectFallbackCharset,
        CancellationToken cancellationToken
      ) arg
    ) => Task.FromResult(ReadAsNameValuePairsConverter(headerStream, arg)); // TODO: async

    private static IEnumerable<KeyValuePair<string, string>> GetParsedHeaders(
      Stream headerStream,
      IEnumerable<string> headerFields,
      bool excludePassedHeaderFields,
      bool decodeHeaderFieldValues,
      EncodingSelectionCallback selectFallbackCharset,
      CancellationToken cancellationToken
    )
    {
      foreach (var header in MimeUtils.ParseHeader(headerStream, true)) { // TODO: cancellationToken
        cancellationToken.ThrowIfCancellationRequested();

        if (headerFields != null) {
          // filter header fields on client side
          if (headerFields.Contains(header.Key, StringComparer.OrdinalIgnoreCase) == excludePassedHeaderFields)
            continue;
        }

        var value = header.Value.Trim();

        yield return KeyValuePair.Create(header.Key, decodeHeaderFieldValues ? MimeEncoding.Decode(value, selectFallbackCharset) : value);
      }
    }

    public IReadOnlyList<KeyValuePair<string, string>> ReadAsNameValuePairs(
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      bool decodeHeaderFieldValues = false,
      EncodingSelectionCallback selectFallbackCharset = null,
      CancellationToken cancellationToken = default
    ) =>
    ReadAs(
      converter: ReadAsNameValuePairsConverter,
      arg: (
        IsOfNonCapsulatedMessagePart ? headerFields : null,
        excludePassedHeaderFields,
        decodeHeaderFieldValues,
        selectFallbackCharset,
        cancellationToken
      ),
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    public Task<IReadOnlyList<KeyValuePair<string, string>>> ReadAsNameValuePairsAsync(
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      bool decodeHeaderFieldValues = false,
      EncodingSelectionCallback selectFallbackCharset = null,
      CancellationToken cancellationToken = default
    ) =>
    ReadAsAsync(
      converter: ReadAsNameValuePairsAsyncConverter,
      arg: (
        IsOfNonCapsulatedMessagePart ? headerFields : null,
        excludePassedHeaderFields,
        decodeHeaderFieldValues,
        selectFallbackCharset,
        cancellationToken
      ),
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    /*
     * Download()
     */
    private static Stream DownloadConverter(Stream bodyStream)
    {
      var stream = new ChunkedMemoryStream();

      CopyFetchMessageBodyStream(bodyStream, stream);

      stream.Position = 0L;

      return stream;
    }

    private static async Task<Stream> DownloadAsyncConverter(Stream bodyStream)
    {
      var stream = new ChunkedMemoryStream();

      await CopyFetchMessageBodyStreamAsync(bodyStream, stream).ConfigureAwait(false);

      stream.Position = 0L;

      return stream;
    }

    public Stream Download(
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAs(
      converter: DownloadConverter,
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    public Task<Stream> DownloadAsync(
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAsAsync(
      converter: DownloadAsyncConverter,
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    /*
     * WriteTo(Stream)
     */
    public void WriteTo(
      Stream stream,
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAs(
      converter: CopyFetchMessageBodyStream,
      arg: ThrowIfStreamIsNullOrNotWritable(stream),
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );

    public Task WriteToAsync(
      Stream stream,
      IEnumerable<string> headerFields = allHeaderFields,
      bool excludePassedHeaderFields = false,
      CancellationToken cancellationToken = default
    ) =>
    ReadAsAsync(
      converter: CopyFetchMessageBodyStreamAsync,
      arg: ThrowIfStreamIsNullOrNotWritable(stream),
      headerFields: headerFields,
      excludePassedHeaderFields: excludePassedHeaderFields,
      cancellationToken: cancellationToken
    );
  }
}
