//
// Copyright (c) 2008 smdn <smdn@smdn.jp>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

using Smdn.Collections;
using Smdn.Formats.Mime;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;
using Smdn.Net.Imap4.Urls;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  public sealed partial class ImapMessagePart : IImapUrl {
    public ImapClient Client {
      get { return message.Client; }
    }

    public ImapMessage Message {
      get { return message; }
    }

    public ImapMessageBody Body { get; }
    public ImapMessageHeader Header { get; }

    /*
     * BODYSTRUCTURE
     */
    public Uri Url {
      get { return structure.Url; }
    }

    public string Specifier {
      get { return structure.Section; }
    }

    public MimeType ContentType {
      get { return structure.MediaType; }
    }

    public bool IsMultiPart {
      get { return structure.IsMultiPart; }
    }

    public bool HasEncapsulatedMessage {
      get { return structure.HasEncapsulatedMessage; }
    }

    public bool IsEncapsulatedMessagePart {
      get {
        if (parent == null)
          return false;
        else
          return parent.HasEncapsulatedMessage;
      }
    }

    public ImapMessagePart ParentPart {
      get { return parent; }
    }

    public bool IsRootPart {
      get { return parent == null; }
    }

    public int NestLevel {
      get {
        if (parent == null)
          return 0;
        else
          return parent.NestLevel + 1;
      }
    }

    /*
     * BODYSTRUCTURE extensions
     */
    public string ContentDisposition {
      get {
        if (structure is IImapExtendedBodyStructure extendedStructure)
          return extendedStructure.Disposition?.Type;

        return null;
      }
    }

    public bool IsDispositionAttachment {
      get { return string.Equals(ContentDisposition, "attachment", StringComparison.OrdinalIgnoreCase); }
    }

    public bool IsDispositionInline {
      get { return string.Equals(ContentDisposition, "inline", StringComparison.OrdinalIgnoreCase); }
    }

    public IReadOnlyDictionary<string, string> ContentDispositionParameters {
      get {
        if (structure is IImapExtendedBodyStructure extendedStructure)
          return extendedStructure.Disposition?.Parameters ?? ReadOnlyDictionary<string, string>.Empty;

        return ReadOnlyDictionary<string, string>.Empty;
      }
    }

    public IReadOnlyList<string> ContentLanguages {
      get {
        if (structure is IImapExtendedBodyStructure extendedStructure)
          return extendedStructure.Languages ??
#if NET45
            ArrayExtensions.Empty<string>();
#else
            Array.Empty<string>();
#endif

#if NET45
        return ArrayExtensions.Empty<string>();
#else
        return Array.Empty<string>();
#endif
      }
    }

    public Uri ContentLocation {
      get {
        if (structure is IImapExtendedBodyStructure extendedStructure)
          return extendedStructure.Location;

        return null;
      }
    }

    /*
     * single part or message/rfc822
     */
    public IReadOnlyDictionary<string, string> ContentTypeParameters {
      get {
        if (structure.IsMultiPart) {
          var extendedMultipart = (structure as ImapExtendedMultiPartBodyStructure);

          if (extendedMultipart == null)
            return ReadOnlyDictionary<string, string>.Empty;
          else
            return extendedMultipart.Parameters;
        }
        else {
          return (structure as ImapSinglePartBodyStructure).Parameters;
        }
      }
    }

    public string ContentTypeCharset {
      get {
        if (ContentTypeParameters.TryGetValue("charset", out var charset))
          return charset;
        else
          return null;
      }
    }

    public string ContentID {
      get {
        if (structure.IsMultiPart)
          return null;
        else
          return (structure as ImapSinglePartBodyStructure).Id;
      }
    }

    public string ContentDescription {
      get {
        if (structure.IsMultiPart)
          return null;
        else
          return (structure as ImapSinglePartBodyStructure).Description;
      }
    }

    public string ContentTransferEncoding {
      get {
        if (structure.IsMultiPart)
          return null;
        else
          return (structure as ImapSinglePartBodyStructure).Encoding;
      }
    }

    public ContentTransferEncodingMethod ContentTransferEncodingMethod {
      get {
        var encoding = ContentTransferEncoding;

        if (string.IsNullOrEmpty(encoding))
          return ContentTransferEncodingMethod.Unknown;
        else
          return Smdn.Formats.Mime.ContentTransferEncoding.GetEncodingMethod(encoding);
      }
    }

    public string ContentMD5 {
      get {
        if (structure.IsMultiPart) {
          return null;
        }
        else if (structure.HasEncapsulatedMessage) {
          var extended = structure as ImapExtendedMessageRfc822BodyStructure;

          if (extended == null)
            return null;
          else
            return extended.MD5;
        }
        else {
          var extended = structure as ImapExtendedSinglePartBodyStructure;

          if (extended == null)
            return null;
          else
            return extended.MD5;
        }
      }
    }

    public long Length {
      get {
        if (structure.IsMultiPart)
          return 0L;
        else
          return (structure as ImapSinglePartBodyStructure).Size;
      }
    }

    /// <summary>
    /// gets the size of the body in text lines, or the size in text lines of the encapsulated message
    /// </summary>
    public long LineCount {
      get {
        if (structure.IsMultiPart)
          return 0L;
        else
          return (structure as ImapSinglePartBodyStructure).LineCount;
      }
    }

    private ImapMessagePart encapsulatedMessagePart = null;

    public ImapMessagePart EncapsulatedMessagePart {
      get {
        if (structure.HasEncapsulatedMessage) {
          return LazyInitializer.EnsureInitialized(ref encapsulatedMessagePart, delegate {
            return new ImapMessagePart(this, (structure as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure);
          });
        }
        else {
          return null;
        }
      }
    }

    public ImapEnvelope EncapsulatedMessageEnvelope {
      get {
        if (structure.HasEncapsulatedMessage)
          return (structure as ImapMessageRfc822BodyStructure).Envelope;
        else
          return null;
      }
    }

    /*
     * multipart
     */
    private IReadOnlyList<ImapMessagePart> nestedMessageParts = null;

    public IReadOnlyList<ImapMessagePart> NestedMessageParts {
      get {
        if (IsMultiPart) {
          return LazyInitializer.EnsureInitialized(ref nestedMessageParts, delegate {
            return (structure as ImapMultiPartBodyStructure).NestedPartStructures.ConvertAll(s => new ImapMessagePart(this, s));
          });
        }
        else {
#if NET45
          return ArrayExtensions.Empty<ImapMessagePart>();
#else
          return Array.Empty<ImapMessagePart>();
#endif
        }
      }
    }

    private readonly ImapMessage message;
    private readonly IImapBodyStructure structure;
    private readonly ImapMessagePart parent;

    internal ImapMessagePart(ImapMessage message)
    {
#if DEBUG
      if (message == null)
        throw new ArgumentNullException(nameof(message));
#endif

      this.message = message;
      this.structure = message.StaticAttributes.BodyStructure;
      this.parent = null;
      this.Body = new ImapMessageBody(message);
      this.Header = new ImapMessageHeader(message);
    }

    private ImapMessagePart(ImapMessagePart parent, IImapBodyStructure structure)
    {
#if DEBUG
      if (parent == null)
        throw new ArgumentNullException(nameof(parent));
      if (structure == null)
        throw new ArgumentNullException(nameof(structure));
#endif

      this.message = parent.message;
      this.structure = structure;
      this.parent = parent;
      this.Body = new ImapMessageBody(this);
      this.Header = new ImapMessageHeader(this);
    }

    /*
    * Traverse
    */
    public void Traverse(Action<ImapMessagePart> visitPart)
    {
      Traverse(false, visitPart);
    }

    public void Traverse(bool visitItself, Action<ImapMessagePart> visitPart)
    {
      Traverse(visitItself ? null : this, visitPart, null);
    }

    public void Traverse(Action<ImapMessagePart> visitInToPart,
                         Action<ImapMessagePart> visitOutFromPart)
    {
      Traverse(false, visitInToPart, visitOutFromPart);
    }

    public void Traverse(bool visitItself,
                         Action<ImapMessagePart> visitInToPart,
                         Action<ImapMessagePart> visitOutFromPart)
    {
      Traverse(visitItself ? null : this, visitInToPart, visitOutFromPart);
    }

    private void Traverse(ImapMessagePart rootPart,
                          Action<ImapMessagePart> visitInToPart,
                          Action<ImapMessagePart> visitOutFromPart)
    {
      if (visitInToPart != null && this != rootPart)
        visitInToPart(this);

      foreach (var nestedMessagePart in NestedMessageParts) {
        nestedMessagePart.Traverse(rootPart, visitInToPart, visitOutFromPart);
      }

      if (HasEncapsulatedMessage)
        EncapsulatedMessagePart.Traverse(rootPart, visitInToPart, visitOutFromPart);

      if (visitOutFromPart != null && this != rootPart)
        visitOutFromPart(this);
    }

    /*
     * Flatten()
     */
    public IReadOnlyList<ImapMessagePart> Flatten(bool containItself = false)
      => GetFlattenMessageParts(containItself);

    private List<ImapMessagePart> GetFlattenMessageParts(bool containItself)
    {
      var ret = new List<ImapMessagePart>();

      Traverse(containItself, delegate(ImapMessagePart part) {
        ret.Add(part);
      });

      return ret;
    }

    /*
     * Find()/FindAll()
     */
    public IReadOnlyList<ImapMessagePart> FindAll(Predicate<ImapMessagePart> match)
    {
      if (match == null)
        throw new ArgumentNullException(nameof(match));

      return GetFlattenMessageParts(false).FindAll(match);
    }

    public ImapMessagePart Find(Predicate<ImapMessagePart> match)
      => FindAll(match).FirstOrDefault();

    public static class Predicates {
      public static bool IsMultiPart(ImapMessagePart part)
      {
        return part.IsMultiPart;
      }

      public static bool HasEncapsulatedMessagePart(ImapMessagePart part)
      {
        return part.HasEncapsulatedMessage;
      }

      public static bool IsEncapsulatedMessagePart(ImapMessagePart part)
      {
        return part.IsEncapsulatedMessagePart;
      }

      public static bool IsTextPart(ImapMessagePart part)
      {
        return part.ContentType.TypeEqualsIgnoreCase("text");
      }

      public static bool IsPlainTextPart(ImapMessagePart part)
      {
        return part.ContentType.EqualsIgnoreCase(MimeType.TextPlain);
      }

      public static bool IsImagePart(ImapMessagePart part)
      {
        return part.ContentType.TypeEqualsIgnoreCase("image");
      }

      public static bool IsAttachmentPart(ImapMessagePart part)
      {
        return part.IsDispositionAttachment;
      }

      public static bool IsInlinePart(ImapMessagePart part)
      {
        return part.IsDispositionInline;
      }

      public static bool SeemsToBeAttachmentPart(ImapMessagePart part)
      {
        if (part.IsDispositionAttachment)
          return true;
        if (part.IsDispositionInline)
          return false;

        if (part.IsMultiPart)
          return false;
        if (part.IsRootPart)
          return false;

        if (MimeType.MessagePartial.EqualsIgnoreCase(part.ContentType))
          // partial message
          return false;

        if (MimeType.MultipartAlternative.EqualsIgnoreCase(part.parent.ContentType))
          // alternative content part
          return false;

        if (part.parent.IsRootPart &&
            MimeType.MultipartMixed.EqualsIgnoreCase(part.parent.ContentType) &&
            (part.ContentType.EqualsIgnoreCase("text/plain") || part.ContentType.EqualsIgnoreCase("text/html")))
          // text part under root part
          return false;

        return true;
      }
    }

    /*
     * utility methods
     */
    public string GetContentFileName()
    {
      return GetContentFileName(null, false);
    }

    public string GetContentFileName(EncodingSelectionCallback selectFallbackEncoding,
                                     bool throwIfMalformedOrUnsupported)
    {
      // try get filename from Content-Disposition filename parameter
      var ret = MimeHeaderUtils.GetDecodedParameterValue("filename",
                                                         ContentDispositionParameters,
                                                         selectFallbackEncoding,
                                                         throwIfMalformedOrUnsupported);

      // try get filename from Content-Type name parameter
      return ret ?? MimeHeaderUtils.GetDecodedParameterValue("name",
                                                             ContentTypeParameters,
                                                             selectFallbackEncoding,
                                                             throwIfMalformedOrUnsupported);
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      throw new NotSupportedException();
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessagePart: Authority='{0}', Mailbox='{1}', Uid={2}, UidValidity={3}, Specifier={4}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           Message.Mailbox.FullName,
                           Message.Uid,
                           Message.UidValidity,
                           Specifier);
    }
  }
}

