// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;

namespace Smdn.Net.Imap4.Client {
  internal class ConcreteMessageQuery : ImapMessageQuery {
    internal ConcreteMessageQuery(ImapOpenMailbox mailbox,
                                  ImapMessageFetchAttributeOptions fetchOptions,
                                  IImapNumberSet sequenceOrUidSet)
      : base(mailbox, fetchOptions, false)
    {
      this.sequenceOrUidSet = sequenceOrUidSet;
    }

    protected override IImapNumberSet GetSequenceOrUidSet()
    {
      return sequenceOrUidSet;
    }

    internal ImapMessage ToSingleMessage()
    {
      return Yield(ToSingleMessage);
    }

    internal Task<ImapMessage> ToSingleMessageAsync()
    {
      return YieldAsync(ToSingleMessage, CancellationToken.None);
    }

    private ImapMessage ToSingleMessage(IEnumerable<ImapMessage> messages)
    {
      var message = messages.FirstOrDefault();

      if (message == null)
        throw new ImapMessageNotFoundException(sequenceOrUidSet);

      return message;
    }

    private readonly IImapNumberSet sequenceOrUidSet;
  }

  public abstract class ImapMessageQuery : ImapMessageReference, IEnumerable<ImapMessage> {
    protected ImapMessageQuery(ImapOpenMailbox mailbox,
                               ImapMessageFetchAttributeOptions fetchOptions,
                               bool orderBySequenceNumber)
      : base(mailbox)
    {
      this.fetchOptions = fetchOptions;
      this.orderBySequenceNumber = orderBySequenceNumber;
    }

    private const int splitCount = 100;

    internal IEnumerable<ImapMessage> YieldMessages()
    {
      return YieldMessages(CancellationToken.None);
    }

    private IEnumerable<ImapMessage> YieldMessages(CancellationToken cancellationToken)
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      ThrowIfSequenceOrUidSetIsNotAvailable(out var sequenceOrUidSet);

      if (sequenceOrUidSet.IsEmpty)
        yield break;

      var fetchDataItems = ImapMessage.TranslateFetchOption(Mailbox,
                                                            fetchOptions,
                                                            out var fetchStaticAttr,
                                                            out var fetchDynamicAttr);

      foreach (var fetchSet in sequenceOrUidSet.SplitIntoEach(splitCount)) {
        cancellationToken.ThrowIfCancellationRequested();

        Mailbox.ProcessResult(Client.Session.Fetch(fetchSet,
                                                   fetchDataItems,
                                                   out ImapMessageAttributes[] messages),
                              null);

        if (orderBySequenceNumber) {
          Array.Sort(messages, ImapMessageCollection.CompareMessageAttributeSequence);

          foreach (var sequence in fetchSet) {
            var message = ImapMessageCollection.FindBySequence(messages, sequence);

            yield return Mailbox.ToMessageInfo(message, fetchStaticAttr, fetchDynamicAttr);
          }
        }
        else {
          foreach (var message in messages) {
            yield return Mailbox.ToMessageInfo(message, fetchStaticAttr, fetchDynamicAttr);
          }
        }
      }
    }

    public IEnumerator<ImapMessage> GetEnumerator()
    {
      return ToList().GetEnumerator();
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return ToList().GetEnumerator();
    }

    internal TResult Yield<TResult>(Converter<IEnumerable<ImapMessage>, TResult> converter)
    {
      return Client.RunExclusively(() => {
        return converter(YieldMessages(CancellationToken.None));
      });
    }

    internal Task<TResult> YieldAsync<TResult>(Converter<IEnumerable<ImapMessage>, TResult> converter,
                                               CancellationToken cancellationToken)
    {
      return Client.StartTaskExclusively(() => {
        return converter(YieldMessages(cancellationToken));
      }, cancellationToken);
    }

    public List<ImapMessage> ToList()
    {
      return Yield(Enumerable.ToList);
    }

    public ImapMessage[] ToArray()
    {
      return Yield(Enumerable.ToArray);
    }

    public Task<List<ImapMessage>> ToListAsync(CancellationToken cancellationToken = default)
    {
      return YieldAsync(Enumerable.ToList, cancellationToken);
    }

    public Task<ImapMessage[]> ToArrayAsync(CancellationToken cancellationToken = default)
    {
      return YieldAsync(Enumerable.ToArray, cancellationToken);
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessageQuery: Mailbox='{0}', UidValidity={1}}}",
                           Mailbox.FullName,
                           UidValidity);
    }

    private readonly ImapMessageFetchAttributeOptions fetchOptions;
    private readonly bool orderBySequenceNumber;
  }
}
