// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Net.Imap4.Protocol.Transmission;

namespace Smdn.Net.Imap4.Client {
  public abstract class ImapMessageReference {
    public ImapClient Client { get; }

    private ImapOpenMailbox mailbox;

    public ImapOpenMailbox Mailbox {
      get { ThrowIfNotAttachedToSelectedMailbox(); return mailbox; }
    }

    public bool IsAvailable {
      get { return mailbox != null && Client.OpeningMailbox == mailbox; }
    }

    public long UidValidity {
      get; protected set;
    }

    protected ImapMessageReference(ImapOpenMailbox mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException(nameof(mailbox));

      this.Client = mailbox.Client;
      this.mailbox = mailbox;
      this.UidValidity = mailbox.UidValidity; // save current UIDVALIDITY
    }

    protected abstract IImapNumberSet GetSequenceOrUidSet();

    protected void ThrowIfNotAttachedToSelectedMailbox()
    {
      if (!IsAvailable)
        throw new ImapUnavailableException();
    }

    internal protected void ThrowIfSequenceOrUidSetIsNotAvailable(out IImapNumberSet sequenceOrUidSet)
    {
      sequenceOrUidSet = ThrowIfSequenceOrUidSetIsNotAvailableCore(null);
    }

    private IImapNumberSet ThrowIfSequenceOrUidSetIsNotAvailableCore(IImapNumberSet specificSequenceOrUidSet)
    {
      ThrowIfNotAttachedToSelectedMailbox();

      var sequenceOrUidSet = specificSequenceOrUidSet ?? GetSequenceOrUidSet();

      mailbox.CheckUidValidity(mailbox.UidValidity, sequenceOrUidSet);

      return sequenceOrUidSet;
    }

    internal void DetachFromMailbox()
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      mailbox = null;
    }

    /*
     * operations
     */
    public void AddFlags(ImapMessageFlag flag,
                         params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItems.AddFlags(flag, flags));
    }

    public void AddKeywords(string keyword,
                            params string[] keywords)
    {
      Store(ImapStoreDataItems.AddFlags(keyword, keywords));
    }

    public void AddFlags(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      Store(ImapStoreDataItems.AddFlags(flagsAndKeywords));
    }

    public Task AddFlagsAsync(ImapMessageFlag flag,
                              params ImapMessageFlag[] flags)
    {
      return StoreAsync(ImapStoreDataItems.AddFlags(flag, flags));
    }

    public Task AddKeywordsAsync(string keyword,
                                 params string[] keywords)
    {
      return StoreAsync(ImapStoreDataItems.AddFlags(keyword, keywords));
    }

    public Task AddFlagsAsync(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      return StoreAsync(ImapStoreDataItems.AddFlags(flagsAndKeywords));
    }

    public void RemoveFlags(ImapMessageFlag flag,
                            params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItems.RemoveFlags(flag, flags));
    }

    public void RemoveKeywords(string keyword,
                               params string[] keywords)
    {
      Store(ImapStoreDataItems.RemoveFlags(keyword, keywords));
    }

    public void RemoveFlags(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      Store(ImapStoreDataItems.RemoveFlags(flagsAndKeywords));
    }

    public Task RemoveFlagsAsync(ImapMessageFlag flag,
                                 params ImapMessageFlag[] flags)
    {
      return StoreAsync(ImapStoreDataItems.RemoveFlags(flag, flags));
    }

    public Task RemoveKeywordsAsync(string keyword,
                                    params string[] keywords)
    {
      return StoreAsync(ImapStoreDataItems.RemoveFlags(keyword, keywords));
    }

    public Task RemoveFlagsAsync(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      return StoreAsync(ImapStoreDataItems.RemoveFlags(flagsAndKeywords));
    }

    public void SetFlags(ImapMessageFlag flag,
                          params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItems.ReplaceFlags(flag, flags));
    }

    public void SetKeywords(string keyword,
                            params string[] keywords)
    {
      Store(ImapStoreDataItems.ReplaceFlags(keyword, keywords));
    }

    public void SetFlags(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      Store(ImapStoreDataItems.ReplaceFlags(flagsAndKeywords));
    }

    public Task SetFlagsAsync(ImapMessageFlag flag,
                              params ImapMessageFlag[] flags)
    {
      return StoreAsync(ImapStoreDataItems.ReplaceFlags(flag, flags));
    }

    public Task SetKeywordsAsync(string keyword,
                                 params string[] keywords)
    {
      return StoreAsync(ImapStoreDataItems.ReplaceFlags(keyword, keywords));
    }

    public Task SetFlagsAsync(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      return StoreAsync(ImapStoreDataItems.ReplaceFlags(flagsAndKeywords));
    }

    public void Store(ImapStoreDataItems storeDataItems)
    {
      if (storeDataItems == null)
        throw new ArgumentNullException(nameof(storeDataItems));

      Client.RunExclusively(() => {
        StoreCore(storeDataItems);
      });
    }

    public Task StoreAsync(ImapStoreDataItems storeDataItems)
    {
      if (storeDataItems == null)
        throw new ArgumentNullException(nameof(storeDataItems));

      return Client.StartTaskExclusively(() => {
        StoreCore(storeDataItems);
      }, CancellationToken.None);
    }

    protected void StoreCore(ImapStoreDataItems storeDataItems, IImapNumberSet specificSequenceOrUidSet = null)
    {
      var sequenceOrUidSet = ThrowIfSequenceOrUidSetIsNotAvailableCore(specificSequenceOrUidSet);

      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      mailbox.ProcessResult(Client.Session.Store(sequenceOrUidSet, storeDataItems));
    }

    private static ImapStoreDataItems markAsDeletedStoreDataItems = ImapStoreDataItems.AddFlags(ImapMessageFlag.Deleted);

    public void MarkAsDeleted()
    {
      Store(markAsDeletedStoreDataItems);
    }

    public Task MarkAsDeletedAsync()
    {
      return StoreAsync(markAsDeletedStoreDataItems);
    }

    private static ImapStoreDataItems markAsSeenStoreDataItems = ImapStoreDataItems.AddFlags(ImapMessageFlag.Seen);

    public void MarkAsSeen()
    {
      Store(markAsSeenStoreDataItems);
    }

    public Task MarkAsSeenAsync()
    {
      return StoreAsync(markAsSeenStoreDataItems);
    }

    public void Delete()
    {
      Client.RunExclusively(DeleteCore);
    }

    public Task DeleteAsync()
    {
      return Client.StartTaskExclusively(DeleteCore, CancellationToken.None);
    }

    private void DeleteCore()
    {
      ThrowIfSequenceOrUidSetIsNotAvailable(out var sequenceOrUidSet);

      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      StoreCore(markAsDeletedStoreDataItems, sequenceOrUidSet);

      if (sequenceOrUidSet.IsUidSet && Client.IsCapabilityAvailable(ImapCapability.UidPlus))
        mailbox.ProcessResult(Client.Session.UidExpunge(sequenceOrUidSet));
      else
        mailbox.ProcessResult(Client.Session.Expunge());

      if (IsAvailable)
        mailbox.Detach(this);
    }

    public void CopyTo(ImapMailbox destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      Client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy, destinationMailbox));
    }

    public Task CopyToAsync(ImapMailbox destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      return Client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy, destinationMailbox),
                                         CancellationToken.None);
    }

    public void MoveTo(ImapMailbox destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      ThrowIfNotAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailbox.Mailbox))
        return; // do nothing

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      Client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move, destinationMailbox));
    }

    public Task MoveToAsync(ImapMailbox destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      ThrowIfNotAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailbox.Mailbox)) {
        // do nothing
        var tcs = new TaskCompletionSource<object>();

        tcs.SetResult(null);

        return tcs.Task;
      }

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      return Client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move, destinationMailbox),
                                         CancellationToken.None);
    }

    private void CopyToMoveToCore(ImapSession.CopyMoveCommand command,
                                  ImapMailbox destinationMailbox)
    {
      ThrowIfSequenceOrUidSetIsNotAvailable(out var sequenceOrUidSet);

      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      var move = command == ImapSession.CopyMoveCommand.Move;
      var markAsDelete = false;

      if (destinationMailbox.Client == this.Client) {
        // same session
        if (move && !Client.IsCapabilityAvailable(ImapCapability.Move)) {
          // use COPY+STORE instead of MOVE
          command = ImapSession.CopyMoveCommand.Copy;
          markAsDelete = true;
        }

        mailbox.ProcessResult(Client.Session.CopyMove(command, sequenceOrUidSet, destinationMailbox.Mailbox));
      }
      else {
        // different session
        if (move)
          markAsDelete = true;

        var appendMessage = this as ImapMessage;

        if (appendMessage == null) {
          var appendMessages = new ConcreteMessageQuery(mailbox,
                                                        ImapMessageFetchAttributeOptions.None,
                                                        sequenceOrUidSet);

          foreach (var message in appendMessages.YieldMessages()) { // not to SetBusy
            destinationMailbox.CopyMessageAcrossSession(message, false);
          }
        }
        else {
          destinationMailbox.CopyMessageAcrossSession(appendMessage, false);
        }
      }

      if (markAsDelete)
        StoreCore(markAsDeletedStoreDataItems, sequenceOrUidSet);

      if (move && IsAvailable)
        mailbox.Detach(this);
    }

    public void CopyTo(string destinationMailboxName)
    {
      CopyTo(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailbox CopyTo(string destinationMailboxName,
                              bool tryCreate)
    {
      ImapMailboxNames.GetValidated(destinationMailboxName, nameof(destinationMailboxName));

      return Client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy,
                                                          destinationMailboxName,
                                                          tryCreate));
    }

    public Task CopyToAsync(string destinationMailboxName)
    {
      return CopyToAsync(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public Task<ImapMailbox> CopyToAsync(string destinationMailboxName,
                                         bool tryCreate)
    {
      ImapMailboxNames.GetValidated(destinationMailboxName, nameof(destinationMailboxName));

      return Client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy,
                                                                destinationMailboxName,
                                                                tryCreate),
                                         CancellationToken.None);
    }

    public void MoveTo(string destinationMailboxName)
    {
      MoveTo(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailbox MoveTo(string destinationMailboxName,
                              bool tryCreate)
    {
      ThrowIfNotAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailboxName))
        return null; // do nothing

      ImapMailboxNames.GetValidated(destinationMailboxName, nameof(destinationMailboxName));

      return Client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move,
                                                          destinationMailboxName,
                                                          tryCreate));
    }

    public Task MoveToAsync(string destinationMailboxName)
    {
      return MoveToAsync(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public Task<ImapMailbox> MoveToAsync(string destinationMailboxName,
                                         bool tryCreate)
    {
      ThrowIfNotAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailboxName)) {
        // do nothing
        var tcs = new TaskCompletionSource<ImapMailbox>();

        tcs.SetResult(null);

        return tcs.Task;
      }

      ImapMailboxNames.GetValidated(destinationMailboxName, nameof(destinationMailboxName));

      return Client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move,
                                                                destinationMailboxName,
                                                                tryCreate),
                                         CancellationToken.None);
    }

    private ImapMailbox CopyToMoveToCore(ImapSession.CopyMoveCommand command,
                                         string destinationMailboxName,
                                         bool tryCreate)
    {
      ThrowIfSequenceOrUidSetIsNotAvailable(out var sequenceOrUidSet);

      if (sequenceOrUidSet.IsEmpty)
        return null; // do nothing

      var move = (command == ImapSession.CopyMoveCommand.Move);
      var markAsDelete = false;

      if (move && !Client.IsCapabilityAvailable(ImapCapability.Move)) {
        // use COPY+STORE instead of MOVE
        command = ImapSession.CopyMoveCommand.Copy;
        markAsDelete = true;
      }

      ImapMailboxStatus createdMailbox = null;

      if (tryCreate) {
        mailbox.ProcessResult(Client.Session.CopyMoveTryCreateMailbox(command,
                                                                      sequenceOrUidSet,
                                                                      destinationMailboxName,
                                                                      out createdMailbox));
      }
      else {
        mailbox.ProcessResult(Client.Session.CopyMove(command,
                                                      sequenceOrUidSet,
                                                      destinationMailboxName),
                              delegate(ImapTaggedStatusResponse status) {
          if (status.ResponseText.Code == ImapResponseCode.TryCreate)
            throw new ImapMailboxNotFoundException(destinationMailboxName);
          else
            return true;
        });
      }

      var createdMailboxInfo = (createdMailbox == null)
        ? null
        : Client.CreateFlagsRetrieved(createdMailbox);

      if (markAsDelete)
        StoreCore(markAsDeletedStoreDataItems, sequenceOrUidSet);

      if (move && IsAvailable)
        mailbox.Detach(this);

      return createdMailboxInfo;
    }
  }
}
