// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Client {
  public partial class ImapOpenMailbox : ImapMailbox, IDisposable {
    public override bool IsOpen {
      get { return Client.OpeningMailbox == this; }
    }

    public bool IsReadOnly {
      get { return Mailbox.ReadOnly; }
    }

    public long FirstUnseenMessageNumber {
      get { return Mailbox.FirstUnseen; }
    }

    public IReadOnlyStringSet<ImapMessageFlag> ApplicableFlags => Mailbox.ApplicableFlags;
    public IReadOnlyStringSet<ImapMessageFlag> PermanentFlags => Mailbox.PermanentFlags;

    public bool IsAllowedToCreateKeywords {
      get { return Mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords); }
    }

    public bool IsUidPersistent {
      get { return Mailbox.UidPersistent; }
    }

    private ImapMessageCollection messages = new ImapMessageCollection();

    internal ImapOpenMailbox(ImapClient client, ImapMailboxStatus mailbox)
      : base(client, mailbox)
    {
    }

    void IDisposable.Dispose()
    {
      if (Client.IsConnected && IsOpen)
        Client.CloseMailbox();
    }

    public override ImapOpenMailbox Open(bool asReadOnly = ImapClient.DefaultSelectReadOnly)
    {
      if (IsOpen)
        return this;
      else
        // messages.Clear();
        throw new NotImplementedException();
    }

    public override Task<ImapOpenMailbox> OpenAsync(bool asReadOnly = ImapClient.DefaultSelectReadOnly)
    {
      if (IsOpen) {
        var tcs = new TaskCompletionSource<ImapOpenMailbox>();

        tcs.SetResult(this);

        return tcs.Task;
      }
      else {
        // messages.Clear();
        throw new NotImplementedException();
      }
    }

    public void Close()
    {
      if (IsOpen)
        Client.CloseMailbox();
    }

    public Task CloseAsync()
    {
      if (IsOpen) {
        return Client.CloseMailboxAsync();
      }
      else {
        // do nothing
        var tcs = new TaskCompletionSource<object>();

        tcs.SetResult(null);

        return tcs.Task;
      }
    }

    internal void CheckSelected()
    {
      if (!IsOpen)
        throw new ImapUnavailableException();
    }

    internal void CheckUidValidity(long uidValidity, IImapNumberSet sequenceOrUidSet)
    {
      // TODO: impl
#if false
      if (!sequenceOrUidSet.IsUidSet)
        return;
      if (uidValidity != UidValidity)
        throw new ImapException("UIDVALIDITY value has been changed");
#endif
    }

    internal ImapMessage ToMessageInfo(ImapMessageAttributes messageAttributes, bool hasStaticAttr, bool hasDynamicAttr)
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      ImapMessage message;

      if (messages.TryGetByUid(messageAttributes.Uid, out message)) {
        if (message.Sequence != messageAttributes.Sequence)
            // XXX: this would be never occur?
          messages.ChangeSequenceNumber(message, messageAttributes.Sequence);
      }
      else {
        message = new ImapMessage(this, messageAttributes.Uid, messageAttributes.Sequence);

        messages.Add(message);
      }

      if (hasStaticAttr)
        message.StaticAttributes = messageAttributes.GetStaticAttributes();
      if (hasDynamicAttr)
        message.DynamicAttributes = messageAttributes.GetDynamicAttributes();

      return message;
    }

    internal void Detach(ImapMessageReference messageReference)
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      messageReference.DetachFromMailbox();

      if (messageReference is ImapMessage message)
        messages.Remove(message);
    }

    internal void DetachAllMailboxObjects()
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      foreach (var message in messages) {
        message.DetachFromMailbox();
      }

      messages.Clear();
    }

    internal protected override ImapCommandResult ProcessResult(ImapCommandResult result,
                                                                Func<ImapTaggedStatusResponse, bool> throwIfErrorResponse = null)
    {
      ProcessSizeAndStatusResponse(result.ReceivedResponses);

      return base.ProcessResult(result, throwIfErrorResponse);
    }

    private void ProcessSizeAndStatusResponse(IReadOnlyCollection<ImapResponse> receivedResponses)
    {
      var prevExistingMessageCount = Mailbox.ExistsMessage;
      var prevRecentMessageCount = Mailbox.RecentMessage;
      List<ImapMessage> deletedMessages = null;
      List<ImapMessage> statusChangedMessages = null;

      foreach (var data in receivedResponses.OfType<ImapServerDataResponse>()) {
        if (data.Type == ImapServerDataResponseType.Fetch) {
          // dynamic attribute of 'messages' will be updated if sequence number matched
          if (ImapServerDataResponseParser.FromFetch(messages, updateOnly: true, data, out var statusChangedMessageAttribute)) {
            if (messages.TryGetBySequence(statusChangedMessageAttribute.Sequence, out var statusChangedMessage)) {
              if (statusChangedMessages == null)
                statusChangedMessages = new List<ImapMessage>(capacity: receivedResponses.Count);

              statusChangedMessages.Add(statusChangedMessage);
            }
          }
        }
        else if (data.Type == ImapServerDataResponseType.Expunge) {
          var expunged = ImapServerDataResponseParser.FromExpunge(data);

          messages.Expunge(expunged, out var deletedMessage);

          if (deletedMessage != null) {
            if (deletedMessages == null)
              deletedMessages = new List<ImapMessage>(capacity: receivedResponses.Count);

            deletedMessages.Add(deletedMessage);
          }

          if (0L < Mailbox.ExistsMessage)
            Mailbox.ExistsMessage -= 1L;
        }
        else if (data.Type == ImapServerDataResponseType.Exists) {
          Mailbox.ExistsMessage = ImapServerDataResponseParser.FromExists(data);
        }
        else if (data.Type == ImapServerDataResponseType.Recent) {
          Mailbox.RecentMessage = ImapServerDataResponseParser.FromRecent(data);
        }
        else if (data.Type == ImapServerDataResponseType.Flags) {
          Mailbox.ApplicableFlags = ImapServerDataResponseParser.FromFlags(data);
        }
      }

      if (prevExistingMessageCount != Mailbox.ExistsMessage)
        Client.RaiseEvent(ImapClient.EventType.ExistingMessageCountChanged,
                          new ImapMailboxSizeChangedEventArgs(this, Mailbox.ExistsMessage, prevExistingMessageCount),
                          IsIdling);

      if (prevRecentMessageCount != Mailbox.RecentMessage)
        Client.RaiseEvent(ImapClient.EventType.RecentMessageCountChanged,
                          new ImapMailboxSizeChangedEventArgs(this, Mailbox.RecentMessage, prevRecentMessageCount),
                          IsIdling);

      if (statusChangedMessages != null)
        QueueMessageStatusChangedEvent(statusChangedMessages);

      if (deletedMessages != null)
        Client.RaiseEvent(ImapClient.EventType.MessageDeleted,
                          new ImapMessageStatusChangedEventArgs(deletedMessages),
                          IsIdling);
    }

    internal void QueueMessageStatusChangedEvent(IReadOnlyList<ImapMessage> statusChangedMessages)
    {
      if (statusChangedMessages == null || statusChangedMessages.Count == 0)
        return;

      Client.RaiseEvent(ImapClient.EventType.MessageStatusChanged,
                        new ImapMessageStatusChangedEventArgs(statusChangedMessages),
                        IsIdling);
    }

    /*
     * operations
     */
    internal protected override void RefreshCore(bool reretrieveMailboxFlags)
    {
      CheckSelected();

      if (reretrieveMailboxFlags)
        ReRetrieveMailboxFlags();

      ProcessResult(Client.Session.NoOp());
    }

    public void Expunge()
    {
      CheckSelected();

      Client.RunExclusively(() => ProcessResult(Client.Session.Expunge()));
    }

    public Task ExpungeAsync()
    {
      CheckSelected();

      return Client.StartTaskExclusively(() => ProcessResult(Client.Session.Expunge()),
                                         CancellationToken.None);
    }

    /*
     * MoveMessagesTo
     */
    public void MoveMessagesTo(ImapMailbox destinationMailbox)
    {
      MoveOrCopyMessagesToCore(false,
                               QueryMessages(ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void MoveMessagesTo(ImapMailbox destinationMailbox,
                               long uid,
                               params long[] uids)
    {
      MoveOrCopyMessagesToCore(false,
                               QueryMessages(ImapMessageFetchAttributeOptions.None,
                                             uid,
                                             uids),
                               destinationMailbox);
    }

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public void MoveMessagesTo(ImapSearchCriteria searchCriteria,
                               ImapMailbox destinationMailbox)
      => MoveMessagesTo(destinationMailbox, searchCriteria: searchCriteria, encoding: null);

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public void MoveMessagesTo(ImapSearchCriteria searchCriteria,
                               Encoding encoding,
                               ImapMailbox destinationMailbox)
      => MoveMessagesTo(destinationMailbox, searchCriteria: searchCriteria, encoding: encoding);

    public void MoveMessagesTo(ImapMailbox destinationMailbox,
                               ImapSearchCriteria searchCriteria,
                               Encoding encoding = null)
    {
      MoveOrCopyMessagesToCore(false,
                               QueryMessages(searchCriteria,
                                             encoding,
                                             ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public Task MoveMessagesToAsync(ImapMailbox destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(false,
                                           QueryMessages(ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    public Task MoveMessagesToAsync(ImapMailbox destinationMailbox,
                                    long uid,
                                    params long[] uids)
    {
      return MoveOrCopyMessagesToAsyncCore(false,
                                           QueryMessages(ImapMessageFetchAttributeOptions.None,
                                                         uid,
                                                         uids),
                                           destinationMailbox);
    }

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task MoveMessagesToAsync(ImapSearchCriteria searchCriteria,
                                    ImapMailbox destinationMailbox)
      => MoveMessagesToAsync(destinationMailbox, searchCriteria: searchCriteria, encoding: null);

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task MoveMessagesToAsync(ImapSearchCriteria searchCriteria,
                                    Encoding encoding,
                                    ImapMailbox destinationMailbox)
      => MoveMessagesToAsync(destinationMailbox, searchCriteria: searchCriteria, encoding: encoding);

    public Task MoveMessagesToAsync(ImapMailbox destinationMailbox,
                                    ImapSearchCriteria searchCriteria,
                                    Encoding encoding = null)
    {
      return MoveOrCopyMessagesToAsyncCore(false,
                                           QueryMessages(searchCriteria,
                                                         encoding,
                                                         ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    /*
     * CopyMessagesTo
     */
    public void CopyMessagesTo(ImapMailbox destinationMailbox)
    {
      MoveOrCopyMessagesToCore(true,
                               QueryMessages(ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public void CopyMessagesTo(ImapMailbox destinationMailbox,
                               long uid,
                               params long[] uids)
    {
      MoveOrCopyMessagesToCore(true,
                               QueryMessages(ImapMessageFetchAttributeOptions.None,
                                             uid,
                                             uids),
                               destinationMailbox);
    }

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public void CopyMessagesTo(ImapSearchCriteria searchCriteria,
                               ImapMailbox destinationMailbox)
      => CopyMessagesTo(destinationMailbox, searchCriteria: searchCriteria, encoding: null);

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public void CopyMessagesTo(ImapSearchCriteria searchCriteria,
                               Encoding encoding,
                               ImapMailbox destinationMailbox)
      => CopyMessagesTo(destinationMailbox, searchCriteria: searchCriteria, encoding: encoding);

    public void CopyMessagesTo(ImapMailbox destinationMailbox,
                               ImapSearchCriteria searchCriteria,
                               Encoding encoding = null)
    {
      MoveOrCopyMessagesToCore(true,
                               QueryMessages(searchCriteria,
                                             encoding,
                                             ImapMessageFetchAttributeOptions.None),
                               destinationMailbox);
    }

    public Task CopyMessagesToAsync(ImapMailbox destinationMailbox)
    {
      return MoveOrCopyMessagesToAsyncCore(true,
                                           QueryMessages(ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    public Task CopyMessagesToAsync(ImapMailbox destinationMailbox,
                                    long uid,
                                    params long[] uids)
    {
      return MoveOrCopyMessagesToAsyncCore(true,
                                           QueryMessages(ImapMessageFetchAttributeOptions.None,
                                                         uid,
                                                         uids),
                                           destinationMailbox);
    }

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task CopyMessagesToAsync(ImapSearchCriteria searchCriteria,
                                    ImapMailbox destinationMailbox)
      => CopyMessagesToAsync(destinationMailbox, searchCriteria: searchCriteria, encoding: null);

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task CopyMessagesToAsync(ImapSearchCriteria searchCriteria,
                                    Encoding encoding,
                                    ImapMailbox destinationMailbox)
      => CopyMessagesToAsync(destinationMailbox, searchCriteria: searchCriteria, encoding: encoding);

    public Task CopyMessagesToAsync(ImapMailbox destinationMailbox,
                                    ImapSearchCriteria searchCriteria,
                                    Encoding encoding = null)
    {
      return MoveOrCopyMessagesToAsyncCore(true,
                                           QueryMessages(searchCriteria,
                                                         encoding,
                                                         ImapMessageFetchAttributeOptions.None),
                                           destinationMailbox);
    }

    private void MoveOrCopyMessagesToCore(bool copy,
                                          ImapMessageQuery messageQuery,
                                          ImapMailbox destinationMailbox)
    {
      if (messageQuery == null)
        throw new ArgumentNullException(nameof(messageQuery));
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      if (copy)
        messageQuery.CopyTo(destinationMailbox);
      else
        messageQuery.MoveTo(destinationMailbox);
    }

    private Task MoveOrCopyMessagesToAsyncCore(bool copy,
                                               ImapMessageQuery messageQuery,
                                               ImapMailbox destinationMailbox)
    {
      if (messageQuery == null)
        throw new ArgumentNullException(nameof(messageQuery));
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      if (copy)
        return messageQuery.CopyToAsync(destinationMailbox);
      else
        return messageQuery.MoveToAsync(destinationMailbox);
    }
  }
}
