// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

using Smdn.Formats.Mime;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures {
  public static class ImapBodyStructureUtils {
    private static IEnumerable<IImapBodyStructure> EnumerateNestedOrEncapsulated(IImapBodyStructure bodyStructure)
    {
      if (bodyStructure is ImapMultiPartBodyStructure multiPart) {
        foreach (var nestedPartStructure in multiPart.NestedPartStructures)
          yield return nestedPartStructure;
      }
      if (bodyStructure is ImapMessageRfc822BodyStructure messageRfc822) {
        yield return messageRfc822.EncapsulatedMessageStructure;
      }
    }

    public static void Traverse(this IImapBodyStructure bodyStructure, Action<IImapBodyStructure> action)
    {
      if (bodyStructure == null)
        throw new ArgumentNullException(nameof(bodyStructure));
      if (action == null)
        throw new ArgumentNullException(nameof(action));

      foreach (var s in EnumerateNestedOrEncapsulated(bodyStructure)) {
        action(s);

        s.Traverse(action);
      }
    }

    public static IImapBodyStructure Find(this IImapBodyStructure bodyStructure, Predicate<IImapBodyStructure> match)
      => FindAll(bodyStructure, match).FirstOrDefault();

    public static IEnumerable<IImapBodyStructure> FindAll(this IImapBodyStructure bodyStructure, Predicate<IImapBodyStructure> match)
    {
      if (bodyStructure == null)
        throw new ArgumentNullException(nameof(bodyStructure));
      if (match == null)
        throw new ArgumentNullException(nameof(match));

      foreach (var s in EnumerateNestedOrEncapsulated(bodyStructure)) {
        if (match(s))
          yield return s;

        foreach (var ss in s.FindAll(match))
          yield return ss;
      }
    }

    public static IImapBodyStructure Find(this IImapBodyStructure bodyStructure, MimeType mediaType)
      => FindAll(bodyStructure, mediaType).FirstOrDefault();

    public static IEnumerable<IImapBodyStructure> FindAll(this IImapBodyStructure bodyStructure, MimeType mediaType)
    {
      if (mediaType == null)
        throw new ArgumentNullException(nameof(mediaType));

      return FindAll(bodyStructure, s => s.MediaType.EqualsIgnoreCase(mediaType));
    }

    public static IImapBodyStructure FindSection(this IImapBodyStructure bodyStructure, params int[] subSections)
    {
      var section = new StringBuilder(0x10);

      for (var i = 0; i < subSections.Length; i++) {
        if (subSections[i] < 1)
          throw new ArgumentOutOfRangeException(nameof(subSections),
                                                subSections[i],
                                                string.Format("contains zero or negative number (index: {0})", i));

        if (i != 0)
          section.Append('.');

        section.Append(subSections[i]);
      }

      return FindSection(bodyStructure, section.ToString());
    }

    public static IImapBodyStructure FindSection(this IImapBodyStructure bodyStructure, string section)
    {
      if (bodyStructure == null)
        throw new ArgumentNullException(nameof(bodyStructure));

      section = GetValidatedSectionSpecOfSectionPart(section);

      return Find(bodyStructure, s => string.Equals(s.Section, section, StringComparison.Ordinal));
    }

    public static IImapBodyStructure GetRootStructure(this IImapBodyStructure bodyStructure)
    {
      if (bodyStructure == null)
        throw new ArgumentNullException(nameof(bodyStructure));

      for (var s = bodyStructure;; ) {
        if (s.ParentStructure == null)
          return s;
        else
          s = s.ParentStructure;
      }
    }

    /*
     * section         = "[" [section-spec] "]"
     * section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
     *                   "TEXT"
     *                     ; top-level or MESSAGE/RFC822 part
     * section-part    = nz-number *("." nz-number)
     *                     ; body part nesting
     * section-spec    = section-msgtext / (section-part ["." section-text])
     * section-text    = section-msgtext / "MIME"
     *                     ; text other than actual body part (headers, etc.)
     */
    private static readonly Regex sectionPartRegex = new Regex(@"^[1-9][0-9]*(\.[1-9][0-9]*)*$", RegexOptions.Singleline | RegexOptions.Compiled | RegexOptions.CultureInvariant);

    public static string GetValidatedSectionSpecOfSectionPart(string sectionSpec)
    {
      if (string.IsNullOrEmpty(sectionSpec))
        return sectionSpec;
      else if (sectionPartRegex.IsMatch(sectionSpec))
        return sectionSpec;
      else
        throw new ArgumentException(string.Format("'{0}' is invalid section-part", sectionSpec), nameof(sectionSpec));
    }

    internal static void SetParentStructure(ImapMessageRfc822BodyStructure parent)
    {
      if (parent.EncapsulatedMessageStructure is ImapSinglePartBodyStructure singlePart)
        singlePart.ParentStructure = parent;
      else if (parent.EncapsulatedMessageStructure is ImapMultiPartBodyStructure multiPart)
        multiPart.ParentStructure = parent;
    }

    internal static void SetParentStructure(ImapMultiPartBodyStructure parent)
    {
      foreach (var nested in parent.NestedPartStructures) {
        if (nested is ImapSinglePartBodyStructure singlePart)
          singlePart.ParentStructure = parent;
        else if (nested is ImapMultiPartBodyStructure multiPart)
          multiPart.ParentStructure = parent;
      }
    }

    public static string GetContentFileName(this IImapBodyStructure bodyStructure,
                                            EncodingSelectionCallback selectFallbackEncoding = null,
                                            bool throwIfMalformedOrUnsupported = false)
    {
      if (bodyStructure == null)
        throw new ArgumentNullException(nameof(bodyStructure));

      // try get filename from Content-Disposition filename parameter
      if (bodyStructure is IImapExtendedBodyStructure extended && extended.Disposition != null)
        return MimeHeaderUtils.GetDecodedParameterValue("filename",
                                                        extended.Disposition.Parameters,
                                                        selectFallbackEncoding,
                                                        throwIfMalformedOrUnsupported);

      // try get filename from Content-Type name parameter
      if (bodyStructure is ImapSinglePartBodyStructure singlePart)
        return MimeHeaderUtils.GetDecodedParameterValue("name",
                                                        singlePart.Parameters,
                                                        selectFallbackEncoding,
                                                        throwIfMalformedOrUnsupported);

      return null;
    }
  }
}
