// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;
using Smdn.Net.Imap4.Urls;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes {
  public sealed class ImapMessageStaticAttributes :
    ImapMessageAttributesBase,
    IImapMessageStaticAttributes,
    IImapUrl
  {
    public long Uid { get; internal set; }
    public ImapEnvelope Envelope { get; internal set; }
    public IImapBodyStructure BodyStructure { get; internal set; }
    public long Rfc822Size { get; internal set; }
    public DateTimeOffset? InternalDate { get; internal set; }

    internal ImapMessageStaticAttributes(long sequence)
      : base(sequence)
    {
    }

    internal override ImapMessageDynamicAttributes GetDynamicAttributes() => null;
    internal override ImapMessageStaticAttributes GetStaticAttributes() => this;
    internal override ImapMessageBodyAttributes GetBodyAttributes() => null;

    /*
     * RFC 3516 - IMAP4 Binary Content Extension
     * http://tools.ietf.org/html/rfc3516
     */
    private Dictionary<string, long> sectionBinarySize;

    public IReadOnlyCollection<string> BinarySizeSpecifiers
#if NET45
      // KeyCollection does not implement IReadOnlyCollection<TKey>
      => new ReadOnlyBinarySizeSpecifierCollection(sectionBinarySize?.Keys);

    private class ReadOnlyBinarySizeSpecifierCollection : IReadOnlyCollection<string>, ICollection<string> {
      private readonly ICollection<string> specifiers;

      public ReadOnlyBinarySizeSpecifierCollection(ICollection<string> specifiers)
      {
        this.specifiers = specifiers ?? ArrayExtensions.Empty<string>();
      }

      public int Count => specifiers.Count;

      public IEnumerator<string> GetEnumerator() => specifiers.GetEnumerator();
      System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() => GetEnumerator();

      /* ICollection<T> */
      public bool IsReadOnly => true;
      public bool Contains(string item) => specifiers.Contains(item); // for Enumerable.Contains()
      public void CopyTo(string[] array, int arrayIndex) => specifiers.CopyTo(array, arrayIndex);

      public void Add(string item) => throw new NotSupportedException("read only");
      public void Clear() => throw new NotSupportedException("read only");
      public bool Remove(string item) => throw new NotSupportedException("read only");
    }
#else
      => (IReadOnlyCollection<string>)sectionBinarySize?.Keys ?? Array.Empty<string>();
#endif

    internal void SetSectionBinarySize(string section, long binarySize)
    {
      if (sectionBinarySize == null)
        sectionBinarySize = new Dictionary<string, long>(StringComparer.OrdinalIgnoreCase);

      sectionBinarySize[section] = binarySize;
    }

    public long GetFirstBinarySize()
    {
      if (sectionBinarySize == null)
        throw new InvalidOperationException("has no BINARY.SIZE attributes");

      return sectionBinarySize.Values.First();
    }

    public long GetBinarySizeOf(string section)
    {
      if (section == null)
        throw new ArgumentNullException(nameof(section));

      if (sectionBinarySize != null && sectionBinarySize.TryGetValue(section, out var binarySize))
        return binarySize;
      else
        throw new ArgumentException("no such section", nameof(section));
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Sequence), Sequence),
        (nameof(Uid), Uid),
        (nameof(BodyStructure), BodyStructure),
        (nameof(Envelope), Envelope),
        (nameof(InternalDate), InternalDate),
        (nameof(Rfc822Size), Rfc822Size),
        (nameof(BinarySizeSpecifiers), BinarySizeSpecifiers),
      });
    }

    /*
     * IImapUrl
     */
    private ImapUriBuilder uriBuilder;

    public Uri Url => ImapUriBuilder.BuildUri(uriBuilder);

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      uriBuilder = baseUrl.Clone();

      if (Uid == 0L) {
        uriBuilder.Uid = 0L;
        uriBuilder.SearchCriteria = GetSequenceNumberSearchCriteria();
      }
      else if (Uid != uriBuilder.Uid) {
        uriBuilder.Uid = Uid;
        uriBuilder.SearchCriteria = null;
      }

      if (BodyStructure != null)
        BodyStructure.SetBaseUrl(uriBuilder);
    }
  }
}
