// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
#if NETFRAMEWORK || NETSTANDARD2_0
using System.Net.Mail;
#endif
using System.Text;

using Smdn.Formats.Mime;
using Smdn.Text;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  // 7.4.2. FETCH Response
  //       ENVELOPE
  //         An address structure is a parenthesized list that describes an
  //         electronic mail address.  The fields of an address structure
  //         are in the following order: personal name, [SMTP]
  //         at-domain-list (source route), mailbox name, and host name.
  public sealed class ImapAddress {
    /*
    address         = "(" addr-name SP addr-adl SP addr-mailbox SP
                      addr-host ")"
    addr-adl        = nstring
                        ; Holds route from [RFC-2822] route-addr if
                        ; non-NIL
    addr-host       = nstring
                        ; NIL indicates [RFC-2822] group syntax.
                        ; Otherwise, holds [RFC-2822] domain name
    addr-mailbox    = nstring
                        ; NIL indicates end of [RFC-2822] group; if
                        ; non-NIL and addr-host is NIL, holds
                        ; [RFC-2822] group name.
                        ; Otherwise, holds [RFC-2822] local-part
                        ; after removing [RFC-2822] quoting
    addr-name       = nstring
                        ; If non-NIL, holds phrase from [RFC-2822]
                        ; mailbox after removing [RFC-2822] quoting
     */
    public string Name { get; }
    public string AtDomainList { get; }
    public string Mailbox { get; }
    public string Host { get; }

    [Obsolete("use AtDomainList instead")]
    public string AtDomailList => AtDomainList;

    internal ImapAddress(string name,
                         string atDomainList,
                         string mailbox,
                         string host)
    {
      Name = name;
      AtDomainList = atDomainList;
      Mailbox = mailbox;
      Host = host;
    }

    public string GetDecodedName(EncodingSelectionCallback selectFallbackEncoding = null,
                                 bool throwIfMalformedOrUnsupported = false)
      => GetDecodedName(selectFallbackEncoding,
                        throwIfMalformedOrUnsupported,
                        out _);

    private string GetDecodedName(EncodingSelectionCallback selectFallbackEncoding,
                                  bool throwIfMalformedOrUnsupported,
                                  out Encoding displayNameEncoding)
    {
      return MimeHeaderUtils.GetDecodedValue(Name,
                                             selectFallbackEncoding,
                                             throwIfMalformedOrUnsupported,
                                             out displayNameEncoding);
    }

#if NETFRAMEWORK || NETSTANDARD2_0
    public MailAddress ToMailAddress(EncodingSelectionCallback selectFallbackEncoding = null,
                                     bool throwIfMalformedOrUnsupported = false)
    {
      var address = string.Concat(Mailbox, "@", Host);
      var decodedDisplayName = GetDecodedName(selectFallbackEncoding,
                                              throwIfMalformedOrUnsupported,
                                              out var displayNameEncoding);

      if (decodedDisplayName == null)
        return new MailAddress(address);
      else
        return new MailAddress(address, decodedDisplayName, displayNameEncoding);
    }

    public static MailAddressCollection ToMailAddressCollection(IEnumerable<ImapAddress> addresses,
                                                                EncodingSelectionCallback selectFallbackEncoding = null,
                                                                bool throwIfMalformedOrUnsupported = false)
    {
      if (addresses == null)
        throw new ArgumentNullException(nameof(addresses));

      var addressCollection = new MailAddressCollection();

      foreach (var address in addresses) {
        addressCollection.Add(address.ToMailAddress(selectFallbackEncoding, throwIfMalformedOrUnsupported));
      }

      return addressCollection;
    }
#endif

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Name), Name),
        (nameof(AtDomainList), AtDomainList),
        (nameof(Mailbox), Mailbox),
        (nameof(Host), Host),
      });
    }
  }
}
