// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  // RFC 4315 Internet Message Access Protocol (IMAP) - UIDPLUS extension
  // http://tools.ietf.org/html/rfc4315
  // APPENDUID response code
#if SERIALIZATION
  [Serializable]
#endif
  public sealed class ImapAppendedUidSet : IImapNumberSet {
    public string Expression => AssignedUidSet.Expression;
    public bool IsUidSet => true;
    public bool IsEmpty => AssignedUidSet.IsEmpty;
    public bool IsSingle => AssignedUidSet.IsSingle;
    public long ConcreteCount => AssignedUidSet.ConcreteCount;

    /// <value>the UIDVALIDITY of the destination mailbox</value>
    public long UidValidity { get; }

    /// <value>the UID assigned to the appended message in the destination mailbox</value>
    public IImapNumberSet AssignedUidSet { get; }

    internal ImapAppendedUidSet(long uidValidity, IImapNumberSet assignedUidSet)
    {
      UidValidity = uidValidity;
      AssignedUidSet = assignedUidSet ?? throw new ArgumentNullException(nameof(assignedUidSet));
    }

#if SERIALIZATION
    private ImapAppendedUidSet(SerializationInfo info, StreamingContext context)
    {
      UidValidity = info.GetInt64(nameof(UidValidity));
      AssignedUidSet = (IImapNumberSet)info.GetValue(nameof(AssignedUidSet), typeof(IImapNumberSet));
    }

    public void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue(nameof(UidValidity), UidValidity);
      info.AddValue(nameof(AssignedUidSet), AssignedUidSet);
    }
#endif

    public IEnumerator<long> GetEnumerator() => AssignedUidSet.GetEnumerator();
    IEnumerator IEnumerable.GetEnumerator() => GetEnumerator();
    IImapNumberSet IImapNumberSet.CombineWith(IImapNumberSet other) => ImapNumberSet.Combine(this, other);
    public IEnumerable<IImapNumberSet> SplitIntoEach(int count) => AssignedUidSet.SplitIntoEach(count);

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(UidValidity), UidValidity),
        (nameof(AssignedUidSet), AssignedUidSet),
      });
    }
  }
}
