// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.Collections;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
#if SERIALIZATION
  [Serializable]
#endif
  public class ImapMatchedSequenceSet : IImapNumberSet {
    internal static ImapMatchedSequenceSet Empty(bool uid) => new ImapMatchedSequenceSet(ImapSequenceSet.Empty(asUidSet: uid));
    internal static ImapMatchedSequenceSet ToSavedResult(bool asUidSet, ImapMatchedSequenceSet result) => new SavedResultMatchedSequenceSet(asUidSet, result);

#if SERIALIZATION
    [Serializable]
#endif
    private class SavedResultMatchedSequenceSet : ImapMatchedSequenceSet {
      internal override IImapNumberSet NumberSet => throw new NotSupportedException("can't retrieve sequence or uid set from saved result");
      public override bool IsSavedResult => true;

      // RFC 5182 - IMAP Extension for Referencing the Last SEARCH Result
      // http://tools.ietf.org/html/rfc5182
      // 1. Introduction
      //    The client can use the "$" marker to reference the content of this
      //    internal variable.
      public override string Expression => "$";
      private readonly bool isUidSet;
      public override bool IsUidSet => isUidSet;
      public override bool IsEmpty => false;
      public override bool IsSingle => false;
      public override long ConcreteCount => -1L;

      public SavedResultMatchedSequenceSet(bool asUidSet, ImapMatchedSequenceSet savedResult)
      {
        this.isUidSet = asUidSet;
        this.HighestModSeq  = savedResult?.HighestModSeq;
        this.Tag            = savedResult?.Tag;
        this.Min            = savedResult?.Min;
        this.Max            = savedResult?.Max;
        this.Count          = savedResult?.Count;
      }

#if SERIALIZATION
      protected SavedResultMatchedSequenceSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        isUidSet = (bool)info.GetValue(nameof(IsUidSet), typeof(bool));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        info.AddValue(nameof(IsUidSet), IsUidSet);

        base.GetObjectData(info, context);
      }
#endif

      public override IEnumerator<long> GetEnumerator() => throw new NotSupportedException("can't enumerate saved result");
      public override IEnumerable<IImapNumberSet> SplitIntoEach(int count) => Singleton.CreateList(this);
    }

    /*
     * instance members
     */

    /*
     * RFC 5182 - IMAP Extension for Referencing the Last SEARCH Result
     * http://tools.ietf.org/html/rfc5182
     */
    public virtual bool IsSavedResult => false;

    private readonly IImapNumberSet numberSet;
    internal virtual IImapNumberSet NumberSet => numberSet;

    public virtual string Expression => NumberSet.Expression;
    public virtual bool IsUidSet => NumberSet.IsUidSet;
    public virtual bool IsEmpty => NumberSet.IsEmpty;
    public virtual bool IsSingle => NumberSet.IsSingle;
    public virtual long ConcreteCount => NumberSet.ConcreteCount;

    /*
     * RFC 4466 - Collected Extensions to IMAP4 ABNF
     * http://tools.ietf.org/html/rfc4466
     */
    //// <value>the highest mod-sequence for all messages being returned</value>
    [CLSCompliant(false)]
    public ulong? HighestModSeq { get; internal set; } = null;

    /*
     * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
     * http://tools.ietf.org/html/rfc4551
     */
    public string Tag { get; internal set; } = null;
    public long? Min { get; internal set; } = null;
    public long? Max { get; internal set; } = null;
    public long? Count { get; internal set; } = null;

    private protected ImapMatchedSequenceSet()
    {
    }

    //[CLSCompliant(false)]
    internal ImapMatchedSequenceSet(IImapNumberSet numberSet, ulong? highestModSeq = null)
    {
      this.numberSet = numberSet ?? throw new ArgumentNullException(nameof(numberSet));
      this.HighestModSeq = highestModSeq;
    }

#if SERIALIZATION
    protected ImapMatchedSequenceSet(SerializationInfo info, StreamingContext context)
    {
      numberSet = (IImapNumberSet)info.GetValue(nameof(NumberSet), typeof(IImapNumberSet));
      HighestModSeq = (ulong?)info.GetValue(nameof(HighestModSeq), typeof(ulong?));
      Tag = info.GetString(nameof(Tag));
      Min = (long?)info.GetValue(nameof(Min), typeof(long?));
      Max = (long?)info.GetValue(nameof(Max), typeof(long?));
      Count = (long?)info.GetValue(nameof(Count), typeof(long?));
    }

    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue(nameof(NumberSet), numberSet);
      info.AddValue(nameof(HighestModSeq), HighestModSeq);
      info.AddValue(nameof(Tag), Tag);
      info.AddValue(nameof(Min), Min);
      info.AddValue(nameof(Max), Max);
      info.AddValue(nameof(Count), Count);
    }
#endif

    public virtual IEnumerator<long> GetEnumerator() => NumberSet.GetEnumerator();
    IEnumerator IEnumerable.GetEnumerator() => GetEnumerator();
    IImapNumberSet IImapNumberSet.CombineWith(IImapNumberSet other) => ImapNumberSet.Combine(this, other);
    public virtual IEnumerable<IImapNumberSet> SplitIntoEach(int count) => NumberSet.SplitIntoEach(count);

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Expression), Expression),
        (nameof(IsUidSet), IsUidSet),
        (nameof(HighestModSeq), HighestModSeq),
        (nameof(Tag), Tag),
        (nameof(Min), Min),
        (nameof(Max), Max),
        (nameof(Count), Count),
      });
    }
  }
}
