// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  /*
   * RFC 5464 - The IMAP METADATA Extension
   * http://tools.ietf.org/html/rfc5464
   */
  public class ImapMetadata {
    /*
     * class members
     */
    public const string Separator = "/";
    public const char SeparatorChar = '/';
    private static readonly char[] separatorCharArray = { SeparatorChar };

    private static readonly char[] invalidEntryNameChars = {'*', '%', SeparatorChar};
    public static IReadOnlyList<char> InvalidEntryNameChars => invalidEntryNameChars;

    public static string[] SplitEntryName(string entryName)
    {
      if (entryName == null)
        throw new ArgumentNullException(nameof(entryName));
      if (!entryName.StartsWith(SeparatorChar))
        throw new ArgumentException($"{nameof(entryName)} must start with '/'", nameof(entryName));

      return entryName.Substring(1).Split(separatorCharArray);
    }

    public static string JoinEntryName(string hierarchicalEntryName, params string[] hierarchicalEntryNames)
      => JoinEntryName(GetValidatedHierarchicalEntryName(ParamArrayUtils.ToList(hierarchicalEntryName, hierarchicalEntryNames), nameof(hierarchicalEntryNames)));

    private static string JoinEntryName(IReadOnlyList<string> hierarchicalEntryName)
      => Separator + string.Join(Separator, hierarchicalEntryName);

    private static IReadOnlyList<string> GetValidatedHierarchicalEntryName(IReadOnlyList<string> hierarchicalEntryName, string paramName)
    {
      if (hierarchicalEntryName == null)
        throw new ArgumentNullException(paramName);
      if (hierarchicalEntryName.Count < 2)
        throw new ArgumentException($"{paramName} must have at least 2 components", paramName);

      foreach (var hierarchy in hierarchicalEntryName) {
        if (hierarchy == null)
          throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(paramName);
        if (0 <= hierarchy.IndexOfAny(invalidEntryNameChars))
          throw new ArgumentException($"{paramName} can't contain characters of {string.Join(",", invalidEntryNameChars.Select(c => "'" + c + "'"))}", paramName);
      }

      return hierarchicalEntryName;
    }

    [Obsolete("This feature will be deprecated in future release.")]
    public static ImapMetadata CreateQuoted(string entryName, string value)
      => new ImapMetadata(SplitEntryName(entryName),
                          ImapQuotedString.Create(value ?? throw new ArgumentNullException(nameof(value))));

    [Obsolete("This feature will be deprecated in future release.")]
    public static ImapMetadata CreateLiteral(string entryName, string value, Encoding charset)
      => new ImapMetadata(SplitEntryName(entryName),
                          ImapLiteralString.Create(value ?? throw new ArgumentNullException(nameof(value)),
                                                   charset ?? throw new ArgumentNullException(nameof(charset)),
                                                   ImapLiteralOptions.Literal8IfCapable));

    /// <summary>The <see cref="Value"/> property will be set to null to indicate that the entry is to be removed.</summary>
    internal static ImapMetadata CreateNil(string entryName) => new ImapMetadata(SplitEntryName(entryName), null);

#if false
    /// <param name="value">The value can be null to indicate that the entry is to be removed.</param>
    public static ImapMetadata CreateSharedVendorMetadata(ImapString @value, string hierarchicalEntryName, params string[] hierarchicalEntryNames)
      => CreateVendorMetadata(value, "shared", hierarchicalEntryName, hierarchicalEntryNames);

    /// <param name="value">The value can be null to indicate that the entry is to be removed.</param>
    public static ImapMetadata CreatePrivateVendorMetadata(ImapString @value, string hierarchicalEntryName, params string[] hierarchicalEntryNames)
      => CreateVendorMetadata(value, "private", hierarchicalEntryName, hierarchicalEntryNames);

    private static ImapMetadata CreateVendorMetadata(ImapString @value, string prefix, string hierarchicalEntryName, params string[] hierarchicalEntryNames)
    {
      var entryName = new string[3 + (hierarchicalEntryNames == null ? 0 : hierarchicalEntryNames.Length)];
      var index = 0;

      entryName[index++] = prefix;
      entryName[index++] = "vendor";
      entryName[index++] = hierarchicalEntryName;

      if (hierarchicalEntryNames != null) {
        foreach (var n in hierarchicalEntryNames) {
          entryName[index++] = n;
        }
      }

      if (@value == null)
        return new ImapMetadata(entryName);
      else
        return new ImapMetadata(entryName, @value);
    }
#endif

    /*
     * instance members
     */
    public IReadOnlyList<string> HierarchicalEntryName { get; }
    public string EntryName => JoinEntryName(HierarchicalEntryName);
    public bool IsShared => string.Equals(HierarchicalEntryName[0], "shared", StringComparison.OrdinalIgnoreCase);
    public bool IsPrivate => string.Equals(HierarchicalEntryName[0], "private", StringComparison.OrdinalIgnoreCase);

    public ImapString Value { get; }

    internal ImapMetadata(IReadOnlyList<string> hierarchicalEntryName, ImapString value)
    {
      this.HierarchicalEntryName = GetValidatedHierarchicalEntryName(hierarchicalEntryName, nameof(hierarchicalEntryName));
      this.Value = value;
    }
  }
}
