// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.Collections;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  // seq-number      = nz-number / "*"
  // seq-range       = seq-number ":" seq-number
  // sequence-set    = (seq-number / seq-range) *("," sequence-set)
#if SERIALIZATION
  [Serializable]
#endif
  internal abstract class ImapNumberSet : IImapNumberSet {
    internal static IImapNumberSet All(bool asUidSet) => new AllNumberSet(isUidSet: asUidSet);
    internal static IImapNumberSet Empty(bool asUidSet) => new EmptyNumberSet(isUidSet: asUidSet);
    internal static IImapNumberSet From(long from, bool asUidSet) => new FromNumberSet(isUidSet: asUidSet, from);
    internal static IImapNumberSet To(long to, bool asUidSet) => new ToNumberSet(isUidSet: asUidSet, to);
    internal static IImapNumberSet Range(long from, long to, bool asUidSet) => from == to ? Of(from, asUidSet: asUidSet) : new RangeNumberSet(isUidSet: asUidSet, from, to);
    internal static IImapNumberSet Of(long number, bool asUidSet) => new DiscreteNumberSet(isUidSet: asUidSet, Singleton.CreateList(number));
    internal static IImapNumberSet Of(bool asUidSet, long number, params long[] numbers) => new DiscreteNumberSet(isUidSet: asUidSet, ParamArrayUtils.ToList(number, numbers));
    internal static IImapNumberSet Of(IEnumerable<long> numbers, bool asUidSet = false) => new DiscreteNumberSet(isUidSet: asUidSet, new List<long>(numbers ?? throw new ArgumentNullException(nameof(numbers))));
    internal static IImapNumberSet Of(ArraySegment<long> numbers, bool asUidSet = false) => new DiscreteNumberSet(isUidSet: asUidSet, new List<long>(numbers));
#if false && NETSTANDARD2_1
    public static IImapSequenceSet Of(Span<long> numbers, bool asUidSet = false) => new NotImplementedException();
    public static IImapSequenceSet Of(Range numberRange, bool asUidSet = false) => new NotImplementedException();
#endif

    internal static IImapNumberSet FromListDirectly(IReadOnlyList<long> numbers, bool asUidSet) => new DiscreteNumberSet(isUidSet: asUidSet, numbers);

    internal static void ThrowIfEmpty(IImapNumberSet numberSet, string paramName)
    {
      if (numberSet == null)
        throw new ArgumentNullException(paramName, "must be a non-null sequence or uid set");
      if (numberSet.IsEmpty)
        throw new ArgumentException("must be a non-empty sequence or uid set", paramName);
    }

    private static bool IsSavedResult(IImapNumberSet numberSet)
    {
      if (numberSet is ImapMatchedSequenceSet matchedNumberSet)
        return matchedNumberSet.IsSavedResult;
      else
        return false;
    }

    internal static void ThrowIfEmptyOrNotUidSet(IImapNumberSet uidSet, string paramName)
    {
      if (uidSet == null)
        throw new ArgumentNullException(paramName, "must be a non-null uid set");
      if (uidSet.IsEmpty)
        throw new ArgumentException("must be a non-empty uid set", paramName);
      if (!uidSet.IsUidSet && !IsSavedResult(uidSet))
        throw new ArgumentException("must be a uid set", paramName);
    }

    internal static void ThrowIfEmptyOrNotSequenceSet(IImapNumberSet sequenceSet, string paramName)
    {
      if (sequenceSet == null)
        throw new ArgumentNullException(paramName, "must be a non-null sequence set");
      if (sequenceSet.IsEmpty)
        throw new ArgumentException("must be a non-empty sequence set", paramName);
      if (sequenceSet.IsUidSet && !IsSavedResult(sequenceSet))
        throw new ArgumentException("must be a sequence set", paramName);
    }

    public abstract string Expression { get; }
    public bool IsUidSet { get; }
    public abstract bool IsEmpty { get; }
    public abstract bool IsSingle { get; }
    public abstract long ConcreteCount { get; }
    /*
    public abstract long PossibleMin { get; }
    public abstract long PossibleMax { get; }
    */
    private protected ImapNumberSet(bool isUidSet)
    {
      IsUidSet = isUidSet;
    }

#if SERIALIZATION
    protected ImapNumberSet(SerializationInfo info, StreamingContext context)
    {
      IsUidSet = info.GetBoolean(nameof(IsUidSet));
    }

    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue(nameof(IsUidSet), IsUidSet);
    }
#endif

    public abstract IEnumerator<long> GetEnumerator();
    IEnumerator IEnumerable.GetEnumerator() => GetEnumerator();

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Expression), Expression),
        (nameof(IsUidSet), IsUidSet),
      });
    }

    public virtual IEnumerable<IImapNumberSet> SplitIntoEach(int count)
    {
      if (count <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(count), count);

      if (IsSingle) {
        yield return this;
        yield break;
      }

      var numbers = new long[count];
      var index = 0;
      var sequential = true;

      foreach (var number in this) {
        if (0 < index)
          sequential &= (numbers[index - 1] == number - 1L);

        numbers[index++] = number;

        if (count <= index) {
          if (sequential)
            yield return Range(numbers[0], number, asUidSet: IsUidSet);
          else
            yield return Of(numbers, asUidSet: IsUidSet);

          index = 0;
          sequential = true;
        }
      }

      if (0 < index) {
        if (sequential)
          yield return Range(numbers[0], numbers[index - 1], asUidSet: IsUidSet);
        else
          yield return Of(new ArraySegment<long>(numbers, 0, index), asUidSet: IsUidSet);
      }
    }

#region "create combined set"
    IImapNumberSet IImapNumberSet.CombineWith(IImapNumberSet other) => Combine(this, other);

    internal static IImapNumberSet Combine(IImapNumberSet x, IImapNumberSet y)
    {
      if (x is null)
        return y;
      if (y is null)
        return x;

      if (x.IsUidSet != y.IsUidSet)
        throw new ArgumentException("can't contain simultaneously both UID set and sequence set", nameof(x));

      if (x is ImapMatchedSequenceSet mx) {
        if (mx.IsSavedResult)
          throw new ArgumentException("can't combine with saved result", nameof(x));
        x = mx.NumberSet;
      }

      if (y is ImapMatchedSequenceSet my) {
        if (my.IsSavedResult)
          throw new ArgumentException("can't combine with saved result", nameof(y));
        y = my.NumberSet;
      }

      var cx = x as CombinedNumberSet;
      var cy = y as CombinedNumberSet;
      var capacity = (cx == null ? 1 : cx.Sets.Count) + (cy == null ? 1 : cy.Sets.Count);
      var allSets = new List<IImapNumberSet>(capacity);

      if (cx == null)
        allSets.Add(x);
      else
        allSets.AddRange(cx.Sets);

      if (cy == null)
        allSets.Add(y);
      else
        allSets.AddRange(cy.Sets);

      return GetRegularizedSet(x.IsUidSet, allSets, false);
    }

    internal static IImapNumberSet GetRegularizedSet(bool uid, ICollection<IImapNumberSet> sets, bool regularizeRangedSet)
    {
      if (regularizeRangedSet)
        throw new NotImplementedException();

      var filteredSets = new List<IImapNumberSet>(sets.Count);
#if REGULARIZE_RANGED_SET
      var fromMin = long.MaxValue;
      var toMax = 0L;
#endif

      foreach (var s in sets) {
        if (s is AllNumberSet)
          return new AllNumberSet(uid);

        if (s.IsEmpty)
          continue;

#if REGULARIZE_RANGED_SET
        var fromSet = s as FromSequenceSet;

        if (fromSet != null && fromSet.From < fromMin) {
          fromMin = fromSet.From;
          continue;
        }

        var toSet = s as ToSequenceSet;

        if (toSet != null && toMax < toSet.To) {
          toMax = toSet.To;
          continue;
        }
#endif

        filteredSets.Add(s);
      }

      // TODO: regularize 'from:to' and 'n1,n2,n3...'
#if REGULARIZE_RANGED_SET
      if (fromMin != long.MaxValue)
        filteredSets.Insert(0, ImapSequenceSet.CreateFromSet(uid, fromMin));
      if (toMax != 0)
        filteredSets.Add(ImapSequenceSet.CreateToSet(uid, toMax));
#endif

      return new CombinedNumberSet(uid, filteredSets);
    }
#endregion

#region "sequence set classes"
    // seq-number:*
#if SERIALIZATION
    [Serializable]
#endif
    private class FromNumberSet : ImapNumberSet {
      public long NumberFrom { get; }
      public override string Expression => string.Concat(NumberFrom, ":*");
      public override bool IsEmpty => false;
      public override bool IsSingle => false;
      public override long ConcreteCount => -1L;

      /*
      public override long PossibleMax => long.MaxValue;
      public override long PossibleMin => From;
      */

      public FromNumberSet(bool isUidSet, long from)
        : base(isUidSet)
      {
        NumberFrom = 0L < from ? from : throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(from), from);
      }

#if SERIALIZATION
      protected FromNumberSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        NumberFrom = info.GetInt64(nameof(NumberFrom));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue(nameof(NumberFrom), NumberFrom);
      }
#endif

      public override IEnumerator<long> GetEnumerator()
      {
        throw new NotSupportedException("can't enumerate 'n:*' set");
      }
    }

    // *:seq-number
#if SERIALIZATION
    [Serializable]
#endif
    private class ToNumberSet : ImapNumberSet {
      public long NumberTo { get; }
      public override string Expression => string.Concat("*:", NumberTo);
      public override bool IsEmpty => false;
      public override bool IsSingle => false;
      public override long ConcreteCount => NumberTo;

      /*
      public override long PossibleMax => To;
      public override long PossibleMin => 1L;
      */

      public ToNumberSet(bool isUidSet, long to)
        : base(isUidSet)
      {
        NumberTo = 0L < to ? to : throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(to), to);
      }

#if SERIALIZATION
      protected ToNumberSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        NumberTo = info.GetInt64(nameof(NumberTo));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue(nameof(NumberTo), NumberTo);
      }
#endif

      public override IEnumerator<long> GetEnumerator()
      {
        throw new NotSupportedException("can't enumerate '*:n' set");
      }
    }

    // seq-number:seq-number
#if SERIALIZATION
    [Serializable]
#endif
    private class RangeNumberSet : ImapNumberSet {
      public long NumberFrom { get; }
      public long NumberTo { get; }
      public override string Expression => string.Concat(NumberFrom, ":", NumberTo);
      public override bool IsEmpty => false;
      public override bool IsSingle => NumberFrom == NumberTo;
      public override long ConcreteCount => NumberTo - NumberFrom + 1L;

      /*
      public override long PossibleMax => To;
      public override long PossibleMin => From;
      */

      public RangeNumberSet(bool isUidSet, long from, long to)
        : base(isUidSet)
      {
        NumberFrom = 0L < from ? from : throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(from), from);
        NumberTo = from < to ? to : throw ExceptionUtils.CreateArgumentMustBeGreaterThan(nameof(from), nameof(to), to);
      }

#if SERIALIZATION
      protected RangeNumberSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        NumberFrom = info.GetInt64(nameof(NumberFrom));
        NumberTo = info.GetInt64(nameof(NumberTo));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue(nameof(NumberFrom), NumberFrom);
        info.AddValue(nameof(NumberTo), NumberTo);
      }
#endif

      public override IEnumerator<long> GetEnumerator()
      {
        for (var i = NumberFrom; i <= NumberTo; i++) {
          yield return i;
        }
      }
    }

    // *
#if SERIALIZATION
    [Serializable]
#endif
    private class AllNumberSet : ImapNumberSet {
      public override string Expression => "*";
      public override bool IsSingle => false;
      public override bool IsEmpty => false;
      public override long ConcreteCount => -1L;

      /*
      public override long PossibleMax => long.MaxValue;
      public override long PossibleMin => 1L;
      */

      public AllNumberSet(bool isUidSet)
        : base(isUidSet)
      {
      }

#if SERIALIZATION
      protected AllNumberSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
      }
#endif

      public override IEnumerator<long> GetEnumerator()
      {
        throw new NotSupportedException("can't enumerate '*' set");
      }
    }

    // ""
#if SERIALIZATION
    [Serializable]
#endif
    private class EmptyNumberSet : ImapNumberSet {
      public override string Expression => string.Empty;
      public override bool IsSingle => false;
      public override bool IsEmpty => true;
      public override long ConcreteCount => 0L;

      /*
      public override long PossibleMax => 0L;
      public override long PossibleMin => 0L;
      */

      public EmptyNumberSet(bool isUidSet)
        : base(isUidSet)
      {
      }

#if SERIALIZATION
      protected EmptyNumberSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
      }
#endif

      public override IEnumerator<long> GetEnumerator()
      {
        yield break;
      }
    }

    // seq-number *("," sequence-set)
#if SERIALIZATION
    [Serializable]
#endif
    private class DiscreteNumberSet : ImapNumberSet {
      public IReadOnlyList<long> Set { get; }
      public override string Expression => string.Join(",", Set);
      public override bool IsSingle => Set.Count == 1;
      public override bool IsEmpty => Set.Count == 0;
      public override long ConcreteCount => Set.Count;

      /*
      public override long PossibleMax => Set[Set.Length - 1]; // must be sorted
      public override long PossibleMin => Set[0]; // must be sorted
      */

      public DiscreteNumberSet(bool isUidSet, IReadOnlyList<long> @set)
        : base(isUidSet)
      {
        if (@set == null)
          throw new ArgumentNullException(nameof(set));

        for (var i = 0; i < @set.Count; i++) {
          if (@set[i] <= 0)
            throw new ArgumentOutOfRangeException(nameof(set),
                                                  @set[i],
                                                  string.Format("must be non-zero positive number (index: {0})", i));
        }

        Set = @set;
      }

#if SERIALIZATION
      protected DiscreteNumberSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        Set = (long[])info.GetValue(nameof(Set), typeof(long[]));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue(nameof(Set), Set.ToArray(), typeof(long[]));
      }
#endif

      public override IEnumerator<long> GetEnumerator()
      {
        foreach (var num in Set) {
          yield return num;
        }
      }
    }

    // (seq-number / seq-range) *("," sequence-set)
#if SERIALIZATION
    [Serializable]
#endif
    private class CombinedNumberSet : ImapNumberSet {
      private static string ToExpression(IImapNumberSet s) => s.Expression;
      private static bool IsEmptySet(IImapNumberSet s) => s.IsEmpty;
      private static bool IsSingleSet(IImapNumberSet s) => s.IsSingle;

      public IReadOnlyCollection<IImapNumberSet> Sets { get; }
      public override string Expression => string.Join(",", Sets.Select(ToExpression));
      public override bool IsEmpty => Sets.All(IsEmptySet);
      public override bool IsSingle => Sets.All(IsSingleSet);

      /*
      public override long PossibleMax {
        get
        {
          var max = 1L;

          foreach (var @set in Sets) {
            if (max < @set.PossibleMax)
              max = @set.PossibleMax;
          }

          return max;
        }
      }

      public override long PossibleMin {
        get
        {
          var min = long.MaxValue;

          foreach (var @set in Sets) {
            if (@set.PossibleMin < min)
              min = @set.PossibleMin;
          }

          return min;
        }
      }
      */

      public override long ConcreteCount {
        get {
          if (Sets.Count == 0)
            return -1;

          var c = 0L;

          foreach (var @set in Sets) {
            if (@set.ConcreteCount == -1)
              return -1;
            else
              c += @set.ConcreteCount;
          }

          return c;
        }
      }

      public CombinedNumberSet(bool isUidSet, IReadOnlyCollection<IImapNumberSet> sets)
        : base(isUidSet)
      {
        if (sets == null)
          throw new ArgumentNullException(nameof(sets));

        foreach (var s in sets) {
          if (s.IsUidSet != isUidSet)
            throw new ArgumentException("can't contain simultaneously both UID set and sequence set", nameof(sets));
        }

        Sets = sets;
      }

#if SERIALIZATION
      protected CombinedNumberSet(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
        Sets = (IImapNumberSet[])info.GetValue(nameof(Sets), typeof(IImapNumberSet[]));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        base.GetObjectData(info, context);

        info.AddValue(nameof(Sets), Sets.ToArray(), typeof(IImapNumberSet[]));
      }
#endif

      public override IEnumerator<long> GetEnumerator()
      {
        foreach (var s in Sets) {
          foreach (var num in s) {
            yield return num;
          }
        }
      }
    }
#endregion
  }
}
