// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Collections;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  // http://tools.ietf.org/html/rfc5256
  public sealed class ImapThreadList : IEnumerable<ImapThreadList> {
    /// <summary>sequence number or UID</summary>
    public long Number { get; }
    public bool IsRoot => Number == 0L;
    public IReadOnlyList<ImapThreadList> Children { get; }
    public bool IsUidList { get; }

    internal static ImapThreadList CreateRootedList(bool isUidList, IReadOnlyList<ImapThreadList> children)
      => new ImapThreadList(true, isUidList, 0, children);

    internal static ImapThreadList CreateNested(bool isUidList, long number, ImapThreadList child)
      => new ImapThreadList(false, isUidList, number, Singleton.CreateList(child ?? throw new ArgumentNullException(nameof(child))));

    internal static ImapThreadList CreateNested(bool isUidList, long number, IReadOnlyList<ImapThreadList> children)
      => new ImapThreadList(false, isUidList, number, children ?? throw new ArgumentNullException(nameof(children)));

    internal static ImapThreadList Create(bool isUidList, long number)
      => new ImapThreadList(false,
                            isUidList,
                            number,
#if NET45
                            ArrayExtensions.Empty<ImapThreadList>()
#else
                            Array.Empty<ImapThreadList>()
#endif
                            );

    private ImapThreadList(bool root,
                           bool isUidList,
                           long number,
                           IReadOnlyList<ImapThreadList> children)
    {
      if (root)
        number = 0L;
      else if (number <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(number), number);

      IsUidList = isUidList;
      Number = number;
      Children = children ??
#if NET45
        ArrayExtensions.Empty<ImapThreadList>();
#else
        Array.Empty<ImapThreadList>();
#endif
    }

    public IEnumerator<ImapThreadList> GetEnumerator() => (Children as IEnumerable<ImapThreadList>).GetEnumerator();
    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() => Children.GetEnumerator();

    public IImapNumberSet ToSequenceSet()
    {
      var sequenceOrUidSet = new List<long>();

      Traverse(t => sequenceOrUidSet.Add(t.Number));

      return ImapNumberSet.FromListDirectly(sequenceOrUidSet, asUidSet: IsUidList);
    }

    public void Traverse(Action<ImapThreadList> action)
    {
      if (action == null)
        throw new ArgumentNullException(nameof(action));

      foreach (var child in Children) {
        action(child);

        child.Traverse(action);
      }
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(IsUidList), IsUidList),
        (nameof(Number), Number),
        (nameof(Children), Children),
      });
    }
  }
}
