// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Threading;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Transmission;

namespace Smdn.Net.Imap4.Protocol.Session {
  public class ImapFetchMessageBodyStream : Stream {
    private const int defaultFetchBlockSize = 10 * 1024;
    public static readonly int DefaultFetchBlockSize = defaultFetchBlockSize;

    public override bool CanSeek => /*!IsClosed &&*/ false;
    public override bool CanRead => !IsClosed /*&& true*/;
    public override bool CanWrite => /*!IsClosed &&*/ false;
    public override bool CanTimeout => true;
    private bool IsClosed => Session == null;

    public override long Position {
      get { RejectDisposed(); return position; }
      set { throw ExceptionUtils.CreateNotSupportedSeekingStream(); }
    }

    public override long Length {
      get {
        RejectDisposed();

        if (length.HasValue)
          return length.Value;
        if (EntireMessage)
          throw new InvalidOperationException("can not refer Length before prepare");

        return indefiniteLength;
      }
    }

    public override int ReadTimeout {
      get { RejectDisposed(); return Session.ReceiveTimeout; }
      set { RejectDisposed(); Session.ReceiveTimeout = value; }
    }

    public override int WriteTimeout {
      get { RejectDisposed(); return Session.SendTimeout; }
      set { RejectDisposed(); Session.SendTimeout = value; }
    }

    protected ImapSession Session { get; private set; }
    protected CancellationToken CancellationToken { get; }
    internal protected long BlockSize { get; }
    private bool EntireMessage => string.IsNullOrEmpty(sectionSpecifier) && !range.IsPartial;

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  IImapNumberSet uid,
                                                  bool peek = false,
                                                  string sectionSpecifier = null,
                                                  ImapFetchRange range = default,
                                                  int blockSize = defaultFetchBlockSize,
                                                  CancellationToken cancellationToken = default)
    {
      this.Session = session ?? throw new ArgumentNullException(nameof(session));
      this.uid = ThrowIfNotSingleUid(uid, nameof(uid));
      this.peek = peek;
      this.sectionSpecifier = sectionSpecifier;
      this.range = range;
      this.BlockSize = 0 < blockSize /*&& blockSize <= ImapFetchRange.MaxValue*/ ? blockSize : throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(blockSize), blockSize);
      this.CancellationToken = cancellationToken;
      this.length = range.Length;

      /*static*/ IImapNumberSet ThrowIfNotSingleUid(IImapNumberSet number, string paramName)
      {
        ImapNumberSet.ThrowIfEmptyOrNotUidSet(number, paramName);

        if (!number.IsSingle)
          throw new ArgumentException("must be a single number", paramName);

        return number;
      }
    }

#if NETFRAMEWORK || NETSTANDARD2_0
    public override void Close() => Dispose(true);
#endif

    protected void DetachFromSession()
    {
      Session = null;
    }

    protected override void Dispose(bool disposing)
    {
      base.Dispose(disposing);

      bodyBuffer = null;
      Session = null;
    }

    public override void SetLength(long value) => throw ExceptionUtils.CreateNotSupportedSettingStreamLength();
    public override long Seek(long offset, SeekOrigin origin) => throw ExceptionUtils.CreateNotSupportedSeekingStream();

    public override void Flush()
    {
      // do nothing
    }

    public override void Write(byte[] buffer, int offset, int count) => throw ExceptionUtils.CreateNotSupportedWritingStream();

    public override int Read(byte[] buffer, int offset, int count)
    {
      RejectDisposed();

      CancellationToken.ThrowIfCancellationRequested();

      if (buffer == null)
        throw new ArgumentNullException(nameof(buffer));
      if (offset < 0)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(offset), offset);
      if (count < 0)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(count), count);
      if (buffer.Length - count < offset)
        throw ExceptionUtils.CreateArgumentAttemptToAccessBeyondEndOfArray(nameof(offset), buffer, offset, count);

      var initialPosition = position;

      for (;;) {
        if (count == 0)
          return (int)(position - initialPosition);

        if (bodyBufferRemain == 0) {
          if (hasFetchedFinalBlock)
            return (int)(position - initialPosition);

          FetchNextBlock<ImapMessageBodyAttributes>();
        }

        var bytesToRead = Math.Min(count, bodyBufferRemain);

        Buffer.BlockCopy(bodyBuffer, bodyBufferOffset, buffer, offset, bytesToRead);

        bodyBufferOffset += bytesToRead;
        bodyBufferRemain -= bytesToRead;
        offset           += bytesToRead;
        count            -= bytesToRead;
        position         += bytesToRead;
      }
    }

    // public override void CopyTo(Stream destination)
    public void ReadToEnd(Stream destination)
    {
      RejectDisposed();

      if (destination == null)
        throw new ArgumentNullException(nameof(destination));
      if (!destination.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream(nameof(destination));

      CopyTo(destination.Write);
    }

    public void ReadToEnd(BinaryWriter destination)
    {
      RejectDisposed();

      if (destination == null)
        throw new ArgumentNullException(nameof(destination));

      CopyTo(destination.Write);
    }

    private void CopyTo(Action<byte[], int, int> write)
    {
      // copy buffered data
      if (0 < bodyBufferRemain) {
        write(bodyBuffer, bodyBufferOffset, bodyBufferRemain);

        position += bodyBufferRemain;
        //bodyBufferRemain = 0;
      }

      // fetch and copy
      for (;;) {
        if (hasFetchedFinalBlock) {
          bodyBufferRemain = 0;
          return;
        }

        FetchNextBlock<ImapMessageBodyAttributes>();

        write(bodyBuffer, bodyBufferOffset, bodyBufferRemain);

        position += bodyBufferRemain;
        //bodyBufferRemain = 0;
      }
    }

    internal protected virtual (ImapCommandResult result, ImapMessageAttributes messageAttributes) Prepare(ImapFetchDataItems fetchDataItems)
    {
      var (result, attrs) = FetchNextBlock<ImapMessageFullAttributes>(
        extraFetchDataItems: fetchDataItems,
        throwIfFetchFailed: false,
        setLengthToRfc822Size: EntireMessage
      );

      return (result, attrs == null ? null : new ImapMessageAttributes(attrs));
    }

    internal protected virtual void ReportFetchedOctetCount(long fetchedOctetCount, long totalOctetCount)
    {
      // do nothing in this class
    }

    private　(ImapCommandResult, TMessageBody) FetchNextBlock<TMessageBody>(
      ImapFetchDataItems extraFetchDataItems = null,
      bool throwIfFetchFailed = true,
      bool setLengthToRfc822Size = false
    )
      where TMessageBody : ImapMessageBodyAttributes
    {
      CancellationToken.ThrowIfCancellationRequested();

      var endOfRange = length ?? long.MaxValue/*as infinite*/;

      if (position == endOfRange) {
        // reached to end of range
        bodyBufferOffset = 0;
        bodyBufferRemain = 0;
        hasFetchedFinalBlock = true;

        return (null, null);
      }

      var fetchDataItems = ImapFetchDataItems.BodyText(
        peek,
        section: sectionSpecifier,
        range: new ImapFetchRange(position + range.Start, Math.Min(endOfRange - position, BlockSize))
      );

      if (extraFetchDataItems != null)
        fetchDataItems = extraFetchDataItems + fetchDataItems;

      if (setLengthToRfc822Size && !fetchDataItems.ContainsOneOf(ImapFetchDataItems.Rfc822Size))
        fetchDataItems = ImapFetchDataItems.Rfc822Size + fetchDataItems;

      ImapCommandResult result;
      TMessageBody[] fetchedMessageBodies;

      try {
        result = Session.Fetch(uid, fetchDataItems, out fetchedMessageBodies);
      }
      catch (TimeoutException ex) {
        var timeoutException = CreateTimeoutException(ex);

        if (timeoutException == null)
          throw;

        throw timeoutException;
      }
      catch (Exception ex) {
        var unexpectedException = CreateUnexpectedException(ex);

        if (unexpectedException == null)
          throw;

        throw unexpectedException;
      }

      if (!result.IsSuccess) {
        if (throwIfFetchFailed)
          throw CreateFetchFailureException(result);

        return (result, null);
      }

      var fetchedBody = 0 < fetchedMessageBodies.Length ? fetchedMessageBodies[0] : throw CreateNoSuchMessageException(result);
      var len = fetchedBody.GetFirstBody(ref bodyBuffer);

      bodyBufferOffset = 0;
      bodyBufferRemain = len;

      hasFetchedFinalBlock = (len < BlockSize);

      if (setLengthToRfc822Size)
        length = (fetchedBody as IImapMessageStaticAttributes ?? throw new InvalidOperationException("can not refer static attributes")).Rfc822Size;

      var fetchedOctetCount = position + bodyBufferRemain;

      if (length.HasValue)
        hasFetchedFinalBlock |= (fetchedOctetCount == length.Value);

      ReportFetchedOctetCount(
        fetchedOctetCount: fetchedOctetCount,
        totalOctetCount: length ?? (hasFetchedFinalBlock ? fetchedOctetCount : indefiniteLength)
      );

      return (result, fetchedBody);
    }

    protected virtual Exception CreateNoSuchMessageException(ImapCommandResult result) => new ImapProtocolViolationException("no such message or expunged");
    protected virtual Exception CreateFetchFailureException(ImapCommandResult result) => new ImapProtocolViolationException(result.Description);
    protected virtual Exception CreateTimeoutException(TimeoutException baseException) => null;
    protected virtual Exception CreateUnexpectedException(Exception baseException) => null;

    private void RejectDisposed()
    {
      if (IsClosed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private readonly bool peek;
    private readonly IImapNumberSet uid;
    private readonly string sectionSpecifier;
    private readonly ImapFetchRange range;
    private const long indefiniteLength = -1;
    private long? length;
    private long position = 0L;
    private byte[] bodyBuffer;
    private int bodyBufferOffset = 0;
    private int bodyBufferRemain = 0;
    private bool hasFetchedFinalBlock = false;
  }
}
