// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Urls;
using Smdn.Net.MessageAccessProtocols.Strings;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.Session {
  public sealed class ImapMailboxStatus : IImapUrl {
    internal ImapSession Session { get; private set; }

    public Uri Url => ImapUriBuilder.BuildUri(UrlBuilder);
    internal ImapUriBuilder UrlBuilder { get; private set; }

    /*
     * basic properties
     */

    /// <value>decoded mailbox name</value>
    public string Name { get; private set; }
    public string HierarchyDelimiter { get; internal set; }
    public IReadOnlyStringSet<ImapMailboxFlag> Flags { get; internal set; } = ImapMailboxFlagSet.Empty;
    public IReadOnlyStringSet<ImapMessageFlag> ApplicableFlags { get; internal set; } = ImapMessageFlagSet.Empty;
    public IReadOnlyStringSet<ImapMessageFlag> PermanentFlags { get; internal set; } = ImapMessageFlagSet.Empty;
    public bool ReadOnly { get; internal set; }
    public long ExistsMessage { get; internal set; }
    public long RecentMessage { get; internal set; }
    public long UnseenMessage { get; internal set; }
    public long FirstUnseen { get; internal set; }
    public long UidNext { get; internal set; }

    public long UidValidity {
      get { return uidValidity; }
      internal set
      {
        uidValidity = value;

        if (UrlBuilder != null)
          UrlBuilder.UidValidity = value;
      }
    }

    /*
     * RFC 4315 Internet Message Access Protocol (IMAP) - UIDPLUS extension
     * http://tools.ietf.org/html/rfc4315
     */
    public bool UidPersistent { get; internal set; } = true;

    /*
     * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
     * http://tools.ietf.org/html/rfc4551
     */
    public bool ModificationSequences { get; internal set; }

    [CLSCompliant(false)]
    public ulong HighestModSeq { get; internal set; }

    /*
     * extended properties
     */
    public bool IsInbox => ImapMailboxNames.IsInbox(Name);

    public string LeafName {
      get {
        if (string.IsNullOrEmpty(HierarchyDelimiter))
          return Name;

        var index = Name.LastIndexOf(HierarchyDelimiter, StringComparison.Ordinal);

        if (index < 0)
          return Name;
        else
          return Name.Substring(index + 1);
      }
    }

    public string SuperiorName {
      get {
        if (string.IsNullOrEmpty(HierarchyDelimiter))
          return string.Empty;

        var index = Name.LastIndexOf(HierarchyDelimiter, StringComparison.Ordinal);

        if (index < 0)
          return string.Empty;
        else
          return Name.Substring(0, index);
      }
    }

    /*
     * RFC 5258 - Internet Message Access Protocol version 4 - LIST Command Extensions
     * http://tools.ietf.org/html/rfc5258
     * 3.4. Additional Requirements on LIST-EXTENDED Clients
     *    All clients that support this extension MUST treat an attribute with
     *    a stronger meaning as implying any attribute that can be inferred
     *    from it.  For example, the client must treat the presence of the
     *    \NoInferiors attribute as if the \HasNoChildren attribute was also
     *    sent by the server.
     */
    /// <remarks>returns true if the mailbox is flagged as \NoSelect or \NonExistent.</remarks>
    public bool IsUnselectable => Flags.Contains(ImapMailboxFlag.NoSelect) || Flags.Contains(ImapMailboxFlag.NonExistent);

    /// <remarks>returns true if the mailbox is flagged as \NoInferiors or \HasNoChildren.</remarks>
    public bool IsNonHierarchical => Flags.Contains(ImapMailboxFlag.NoInferiors) || Flags.Contains(ImapMailboxFlag.HasNoChildren);

    public bool NameEquals(ImapMailboxStatus other) => ImapMailboxNames.EqualityComparer.Equals(Name, other?.Name ?? throw new ArgumentNullException(nameof(other)));
    public bool NameEquals(string otherMailboxName) => ImapMailboxNames.EqualityComparer.Equals(Name, otherMailboxName);

    /// <remarks>name must be decoded</remarks>
    internal ImapMailboxStatus(string name)
    {
      if (name == null)
        throw new ArgumentNullException(nameof(name));

      SetName(name);

      this.Session = null;
    }

    internal void SetName(string newName)
    {
      Name = ImapMailboxNames.IsInbox(newName) ? ImapMailboxNames.Inbox : newName; // regularize character case of "INBOX"

      if (UrlBuilder != null)
        UrlBuilder.Mailbox = Name;
    }

    internal void AttachToSession(ImapSession session)
    {
      Session = session;

      (this as IImapUrl).SetBaseUrl(session.AuthorityBuilder);
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      UrlBuilder = baseUrl.Clone();
      UrlBuilder.Mailbox = Name;
      UrlBuilder.UidValidity = uidValidity;
    }

    internal void DetachFromSession()
    {
      Session = null;
    }

    internal void UpdateStatus(string statusAtt, ImapData value)
    {
      /*
      status-att      = "MESSAGES" / "RECENT" / "UIDNEXT" / "UIDVALIDITY" /
                        "UNSEEN"
      status-att-list =  status-att SP number *(SP status-att SP number)
      */

      /*
       * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
       * http://tools.ietf.org/html/rfc4551
       * 
       *    status-att-val      =/ "HIGHESTMODSEQ" SP mod-sequence-valzer
       *                           ;; extends non-terminal defined in [IMAPABNF].
       *                           ;; Value 0 denotes that the mailbox doesn't
       *                           ;; support persistent mod-sequences
       *                           ;; as described in Section 3.1.2
       */
      switch (statusAtt.ToUpperInvariant()) {
        case "MESSAGES":      ExistsMessage   = ImapDataParser.ToNumber(value); break;
        case "RECENT":        RecentMessage   = ImapDataParser.ToNumber(value); break;
        case "UNSEEN":        UnseenMessage   = ImapDataParser.ToNumber(value); break;
        case "UIDNEXT":       UidNext         = ImapDataParser.ToNumber(value); break;
        case "UIDVALIDITY":   UidValidity     = ImapDataParser.ToNumber(value); break;
        case "HIGHESTMODSEQ": HighestModSeq   = ImapDataParser.ToModSequenceValzer(value); break;
        default:
          //Trace.Info($"unsupported status-att: {statusAtt}");
          break;
      }
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Name), Name),
        (nameof(HierarchyDelimiter), HierarchyDelimiter),
        (nameof(Flags), Flags),
        (nameof(ReadOnly), ReadOnly),
        (nameof(ApplicableFlags), ApplicableFlags),
        (nameof(PermanentFlags), PermanentFlags),
        (nameof(ExistsMessage), ExistsMessage),
        (nameof(RecentMessage), RecentMessage),
        (nameof(UnseenMessage), UnseenMessage),
        (nameof(FirstUnseen), FirstUnseen),
        (nameof(UidNext), UidNext),
        (nameof(UidValidity), UidValidity),
        (nameof(UidPersistent), UidPersistent),
        (nameof(ModificationSequences), ModificationSequences),
        (nameof(HighestModSeq), HighestModSeq),
      });
    }

    private long uidValidity = 0L;
  }
}
