// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * transaction methods : authenticated state / APPEND
     */

    private enum AppendMode {
      Append,
      AppendBinary,
    }

    // messages  | binary | mailboxName       | tryCreate
    // --------------------------------------------------
    // single    | true   | string            | true
    // multiple  | false  | ImapMailboxStatus | false
    //
    // overloads of Append()
    //   Append      ({message|messages}, {mailboxName|ImapMailboxStatus});
    //   AppendBinary({message|messages}, {mailboxName|ImapMailboxStatus});
    // 
    //   AppendTryCreateMailbox      ({message|messages}, mailboxName, out ImapMailboxStatus);
    //   AppendBinaryTryCreateMailbox({message|messages}, mailboxName, out ImapMailboxStatus);

    /*
     * APPEND to mailbox name
     */
    public ImapCommandResult       Append(IImapAppendMessage message,               string mailboxName, out ImapAppendedUidSet appendedUids)
                                => Append(Singleton.CreateList(message),                   mailboxName, out                    appendedUids);

    public ImapCommandResult AppendBinary(IImapAppendMessage message,               string mailboxName, out ImapAppendedUidSet appendedUids)
                          => AppendBinary(Singleton.CreateList(message),                   mailboxName, out                    appendedUids);

    public ImapCommandResult       Append(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids)
    => AppendCore(AppendMode.      Append, tryCreate: false,              messages,        mailboxName, out                    appendedUids, out _);

    public ImapCommandResult AppendBinary(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids)
    => AppendCore(AppendMode.AppendBinary, tryCreate: false,              messages,        mailboxName, out                    appendedUids, out _);

    /*
     * APPEND to ImapMailboxStatus
     */
    public ImapCommandResult       Append(IImapAppendMessage message,               ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
                                => Append(Singleton.CreateList(message),                              mailbox, out                    appendedUids);

    public ImapCommandResult AppendBinary(IImapAppendMessage message,               ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
                          => AppendBinary(Singleton.CreateList(message),                              mailbox, out                    appendedUids);

    public ImapCommandResult       Append(IEnumerable<IImapAppendMessage> messages, ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
    => AppendCore(AppendMode.      Append, tryCreate: false,              messages, ValidateMailboxRelationship(mailbox).Name, out    appendedUids, out _);

    public ImapCommandResult AppendBinary(IEnumerable<IImapAppendMessage> messages, ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
    => AppendCore(AppendMode.AppendBinary, tryCreate: false,              messages, ValidateMailboxRelationship(mailbox).Name, out    appendedUids, out _);

    /*
     * APPEND with TRYCREATE
     */
    public ImapCommandResult       AppendTryCreateMailbox(IImapAppendMessage message,               string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
                                => AppendTryCreateMailbox(Singleton.CreateList(message),                   mailboxName, out                    appendedUids, out                   createdMailbox);

    public ImapCommandResult AppendBinaryTryCreateMailbox(IImapAppendMessage message,               string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
                          => AppendBinaryTryCreateMailbox(Singleton.CreateList(message),                   mailboxName, out                    appendedUids, out                   createdMailbox);

    public ImapCommandResult       AppendTryCreateMailbox(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
    => AppendCore(AppendMode.      Append, tryCreate: true,                               messages,        mailboxName, out                    appendedUids, out                   createdMailbox);

    public ImapCommandResult AppendBinaryTryCreateMailbox(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
    => AppendCore(AppendMode.AppendBinary, tryCreate: true,                               messages,        mailboxName, out                    appendedUids, out                   createdMailbox);


#if false
    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    [Obsolete] public ImapCommandResult Append(IImapAppendMessage message, ImapMailboxStatus mailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY extension.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinary(IImapAppendMessage message, ImapMailboxStatus mailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendMultiple(IEnumerable<IImapAppendMessage> messages, ImapMailboxStatus mailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY or MULTIAPPEND extension.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinaryMultiple(IEnumerable<IImapAppendMessage> messages, ImapMailboxStatus mailbox)
      => throw new NotImplementedException();



    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult Append(IImapAppendMessage message, ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinary(IImapAppendMessage message, ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendMultiple(IEnumerable<IImapAppendMessage> messages, ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY or MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinaryMultiple(IEnumerable<IImapAppendMessage> messages, ImapMailboxStatus mailbox, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();



    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    [Obsolete] public ImapCommandResult Append(IImapAppendMessage message, string mailboxName)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY extension.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinary(IImapAppendMessage message, string mailboxName)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY or MULTIAPPEND extension.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinaryMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName)
      => throw new NotImplementedException();



    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult Append(IImapAppendMessage message, string mailboxName, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinary(IImapAppendMessage message, string mailboxName, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY or MULTIAPPEND extension.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinaryMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids)
      => throw new NotImplementedException();



    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    [Obsolete] public ImapCommandResult Append(IImapAppendMessage message, string mailboxName, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinary(IImapAppendMessage message, string mailboxName, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY or MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinaryMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();



    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult Append(IImapAppendMessage message, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinary(IImapAppendMessage message, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();

    /// <summary>sends APPEND command</summary>
    /// <remarks>
    /// valid in authenticated state.
    /// this method will fail if server does not support BINARY or MULTIAPPEND extension.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="appendedUids"/> will be set if the server supports UIDPLUS extension and returns [APPENDUID] responce code, otherwise null.
    /// </remarks>
    [Obsolete] public ImapCommandResult AppendBinaryMultiple(IEnumerable<IImapAppendMessage> messages, string mailboxName, out ImapAppendedUidSet appendedUids, out ImapMailboxStatus createdMailbox)
      => throw new NotImplementedException();
#endif

    private ImapCommandResult AppendCore(AppendMode appendMode,
                                         bool tryCreate,
                                         IEnumerable<IImapAppendMessage> messages,
                                         string mailboxName,
                                         out ImapAppendedUidSet appendedUids,
                                         out ImapMailboxStatus createdMailbox)
    {
      RejectNonAuthenticatedState();

      if (messages == null)
        throw new ArgumentNullException(nameof(messages));

      var argMailboxName = ImapMailboxNameString.CreateMailboxName(mailboxName);

      appendedUids = null;
      createdMailbox = null;

      // append message
      var messagesToUpload = new List<ImapString>();
      var messageCount = 0;
      var literalOptions = ImapLiteralOptions.NonSynchronizingIfCapable;

      switch (appendMode) {
        case AppendMode.Append: literalOptions |= ImapLiteralOptions.Literal; break;
        case AppendMode.AppendBinary: literalOptions |= ImapLiteralOptions.Literal8; break;
        default: throw ExceptionUtils.CreateArgumentMustBeValidEnumValue(nameof(appendMode), appendMode);
      }

      foreach (var message in messages) {
        if (message == null)
          throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(nameof(messages));

        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         * 
         *    append          = "APPEND" SP mailbox 1*append-message
         *                      ;; only a single append-message may appear
         *                      ;; if MULTIAPPEND [MULTIAPPEND] capability
         *                      ;; is not present
         *    append-message  = append-opts SP append-data
         *    append-ext      = append-ext-name SP append-ext-value
         *                      ;; This non-terminal define extensions to
         *                      ;; to message metadata.
         *    append-ext-name = tagged-ext-label
         *    append-ext-value= tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    append-data     = literal / literal8 / append-data-ext
         *    append-data-ext = tagged-ext
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions,
         *                      ;; i.e., a mandatory label followed
         *                      ;; by parameters.
         *    append-opts     = [SP flag-list] [SP date-time] *(SP append-ext)
         *                      ;; message metadata
         */

        // flag-list
        if (message.Flags != null && 0 < message.Flags.Count)
          messagesToUpload.Add(ImapStringList.CreateParenthesized(message.Flags.Except(ImapMessageFlag.NonApplicableFlags).ToList()));

        // date-time
        if (message.InternalDate.HasValue)
          messagesToUpload.Add(ImapDateTimeFormat.ToDateTimeString(message.InternalDate.Value));

        // append-data
        messagesToUpload.Add(new AppendData(message, literalOptions));

        messageCount++;
      }

      if (messageCount == 0)
        throw new ArgumentException("at least 1 message must be specified", nameof(messages));

      ImapCommandResult failedResult = null;

      for (var i = 0; i < 2; i++) {
        var respTryCreate = false;

        using (var t = new AppendTransaction(connection, this, 1 < messageCount)) {
          // mailbox name
          t.RequestArguments["mailbox name"] = argMailboxName;

          // messages to upload
          t.RequestArguments["messages to upload"] = ImapStringList.Create(messagesToUpload);

          if (ProcessTransaction(t).IsSuccess) {
            appendedUids = t.TransactionResult;
            return t.Result;
          }
          else {
            if (ProcessMailboxRefferalResponse(t.Result.TaggedStatusResponse) || !tryCreate)
              return t.Result;
          }

          failedResult = t.Result;

          // 6.3.11. APPEND Command
          //       If the destination mailbox does not exist, a server MUST return an
          //       error, and MUST NOT automatically create the mailbox.  Unless it
          //       is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the APPEND
          //       if the CREATE is successful.
          respTryCreate = t.Result.TaggedStatusResponse.ResponseText.Code == ImapResponseCode.TryCreate;
        }

        // try create
        if (i == 0 && respTryCreate)
          if (!Create(mailboxName, out createdMailbox).IsSuccess)
            return failedResult;
      }

      return failedResult;
    }

    private class AppendData : ImapLiteralString {
      protected internal override long Length => message.Length;
      protected internal override Stream GetStream() => message.GetMessageStream();
      protected override string GetValueString() => "[MessageStream]";

      public AppendData(IImapAppendMessage message, ImapLiteralOptions options)
        : base(options)
      {
        this.message = message;
      }

      private readonly IImapAppendMessage message;
    }
  }
}
