// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * transaction methods : authenticated/selected state
     */

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, string resourceName, long resourceLimit)
      => SetQuota(quotaRoot, Singleton.CreateList(new ImapQuotaResourceLimit(resourceName, resourceLimit)), out _);

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, string resourceName, long resourceLimit, out ImapQuota changedQuota)
      => SetQuota(quotaRoot, Singleton.CreateList(new ImapQuotaResourceLimit(resourceName, resourceLimit)), out changedQuota);

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, ImapQuotaResourceLimit resourceLimit)
      => SetQuota(quotaRoot, Singleton.CreateList(resourceLimit), out _);

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, ImapQuotaResourceLimit resourceLimit, out ImapQuota changedQuota)
      => SetQuota(quotaRoot, Singleton.CreateList(resourceLimit), out changedQuota);

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, IReadOnlyCollection<ImapQuotaResourceLimit> resourceLimits)
      => SetQuota(quotaRoot, resourceLimits, out _);

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, IReadOnlyCollection<ImapQuotaResourceLimit> resourceLimits, out ImapQuota changedQuota)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));
      if (resourceLimits == null)
        throw new ArgumentNullException(nameof(resourceLimits));

      var listOfResourceLimits = new ImapString[resourceLimits.Count * 2];
      var index = 0;

      foreach (var resourceLimit in resourceLimits) {
        if (resourceLimit == null)
          throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(nameof(resourceLimits));

        listOfResourceLimits[index++] = resourceLimit.Name;
        listOfResourceLimits[index++] = resourceLimit.Limit.ToString();
      }

      RejectNonAuthenticatedState();

      changedQuota = null;

      using (var t = new SetQuotaTransaction(connection, this)) {
        // quota root
        t.RequestArguments["quota root"] = ImapQuotedString.Create(quotaRoot);
        // list of resource limits
        t.RequestArguments["list of resource limits"] = ImapStringList.CreateParenthesized(listOfResourceLimits);

        if (ProcessTransaction(t).IsSuccess)
          changedQuota = t.TransactionResult;

        return t.Result;
      }
    }

    /// <summary>sends GETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult GetQuota(string quotaRoot, out ImapQuota quota)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));

      RejectNonAuthenticatedState();

      quota = null;

      using (var t = new GetQuotaTransaction(connection, this)) {
        // quota root
        t.RequestArguments["quota root"] = ImapQuotedString.Create(quotaRoot);

        if (ProcessTransaction(t).IsSuccess)
          quota = t.TransactionResult;

        return t.Result;
      }
    }

    /// <summary>sends GETQUOTAROOT command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult GetQuotaRoot(ImapMailboxStatus mailbox, out IReadOnlyDictionary<string, IReadOnlyList<ImapQuota>> quotaRoots)
      =>GetQuotaRoot(ValidateMailboxRelationship(mailbox).Name, out quotaRoots);

    /// <summary>sends GETQUOTAROOT command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult GetQuotaRoot(string mailboxName, out IReadOnlyDictionary<string, IReadOnlyList<ImapQuota>> quotaRoots)
    {
      RejectNonAuthenticatedState();

      var argMailboxName = ImapMailboxNameString.CreateMailboxName(mailboxName);

      quotaRoots = null;

      using (var t = new GetQuotaRootTransaction(connection, this)) {
        // mailbox name
        t.RequestArguments["mailbox name"] = argMailboxName;

        if (ProcessTransaction(t).IsSuccess)
          quotaRoots = t.TransactionResult;

        return t.Result;
      }
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Namespace() => Namespace(out _);

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Namespace(out ImapNamespace namespaces)
    {
      RejectNonAuthenticatedState();

      namespaces = null;

      using (var t = new NamespaceTransaction(connection, this)) {
        if (ProcessTransaction(t).IsSuccess) {
          this.namespaces = t.TransactionResult;
          namespaces = t.TransactionResult;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult ComparatorGetActive() => ComparatorGetActive(out _);

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult ComparatorGetActive(out ImapCollationAlgorithm activeComparator)
    {
      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 4.7. COMPARATOR Command
       *    When issued with no arguments, it results in a
       *    COMPARATOR response indicating the currently active comparator.
       */
      activeComparator = null;

      using (var t = new ComparatorTransaction(connection, this)) {
        if (ProcessTransaction(t).IsSuccess) {
          this.activeComparator = t.TransactionResult.Item1;

          activeComparator = t.TransactionResult.Item1;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator(ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
      => Comparator(out _, out _, comparator, comparators);

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator(out IReadOnlyList<ImapCollationAlgorithm> matchingComparators, ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
      => Comparator(out _, out matchingComparators, comparator, comparators);

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator(out ImapCollationAlgorithm activeComparator, out IReadOnlyList<ImapCollationAlgorithm> matchingComparators, ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
    {
      RejectNonAuthenticatedState();

      if (comparator == null)
        throw new ArgumentNullException(nameof(comparator));

      var comparatorArguments = new ImapCollationAlgorithm[1 + (comparators == null ? 0 : comparators.Length)];

      comparatorArguments[0] = comparator;

      if (comparators != null) {
        for (var i = 0; i < comparators.Length; i++) {
          if (comparators[i] == null)
            throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(nameof(comparators));

          comparatorArguments[1 + i] = comparators[i];
        }
      }

      activeComparator = null;
      matchingComparators = null;

      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 4.7. COMPARATOR Command
       *    When issued with one or more comparator arguments, it changes the
       *    active comparator as directed.
       */
      using (var t = new ComparatorTransaction(connection, this)) {
        t.RequestArguments["comparator order arguments"] = ImapStringList.Create(comparatorArguments.Convert(comp => ImapQuotedString.Create(comp.Value)));

        if (ProcessTransaction(t).IsSuccess) {
          this.activeComparator = t.TransactionResult.Item1;

          activeComparator = t.TransactionResult.Item1;
          matchingComparators = t.TransactionResult.Item2;
        }

        return t.Result;
      }
    }

#region "GETMETADATA"
#region "server entries"
    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadataServer(string entrySpecifier, out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(null,
                                 Singleton.CreateList(entrySpecifier),
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadataServer(IReadOnlyCollection<string> entrySpecifiers, out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(null,
                                 entrySpecifiers,
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadataServer(string entrySpecifier, ImapGetMetadataOptions options, out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(null,
                                 Singleton.CreateList(entrySpecifier),
                                 options ?? throw new ArgumentNullException(nameof(options)),
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadataServer(IReadOnlyCollection<string> entrySpecifiers, ImapGetMetadataOptions options, out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(null,
                                 entrySpecifiers,
                                 options ?? throw new ArgumentNullException(nameof(options)),
                                 out metadata);
    }
#endregion

#region "mailbox entries"
    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailboxStatus mailbox,
                                         string entrySpecifier,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(ValidateMailboxRelationship(mailbox).Name,
                                 Singleton.CreateList(entrySpecifier),
                                 null,
                                 out metadata);
    }

    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailboxStatus mailbox,
                                         IReadOnlyCollection<string> entrySpecifiers,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(ValidateMailboxRelationship(mailbox).Name,
                                 entrySpecifiers,
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName,
                                         string entrySpecifier,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                                 Singleton.CreateList(entrySpecifier),
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName,
                                         IReadOnlyCollection<string> entrySpecifiers,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                                 entrySpecifiers,
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailboxStatus mailbox,
                                         string entrySpecifier,
                                         ImapGetMetadataOptions options,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(ValidateMailboxRelationship(mailbox).Name,
                                 Singleton.CreateList(entrySpecifier),
                                 options ?? throw new ArgumentNullException(nameof(options)),
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailboxStatus mailbox,
                                         IReadOnlyCollection<string> entrySpecifiers,
                                         ImapGetMetadataOptions options,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(ValidateMailboxRelationship(mailbox).Name,
                                 entrySpecifiers,
                                 options ?? throw new ArgumentNullException(nameof(options)),
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName,
                                         string entrySpecifier,
                                         ImapGetMetadataOptions options,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                                 Singleton.CreateList(entrySpecifier),
                                 options ?? throw new ArgumentNullException(nameof(options)),
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName,
                                         IReadOnlyCollection<string> entrySpecifiers,
                                         ImapGetMetadataOptions options,
                                         out IReadOnlyList<ImapMetadata> metadata)
    {
      return GetMetadataInternal(mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                                 entrySpecifiers,
                                 options ?? throw new ArgumentNullException(nameof(options)),
                                 out metadata);
    }
#endregion

    private ImapCommandResult GetMetadataInternal(string mailboxName,
                                                  IReadOnlyCollection<string> entrySpecifiers,
                                                  ImapGetMetadataOptions options,
                                                  out IReadOnlyList<ImapMetadata> metadata)
    {
      metadata = null;

      if (entrySpecifiers == null)
        throw new ArgumentNullException(nameof(entrySpecifiers));
      if (entrySpecifiers.Count == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyCollection(nameof(entrySpecifiers));

      if (handlesIncapableAsException)
        CheckServerCapabilityMetadata(mailboxName != null);

      var argMailboxName = mailboxName == null ? ImapQuotedString.Empty : ImapMailboxNameString.CreateMailboxName(mailboxName);

      using (var t = new GetMetadataTransaction(connection, this)) {
        // mailbox-name
        t.RequestArguments["mailbox-name"] = argMailboxName;

        // options
        if (options != null)
          t.RequestArguments["options"] = options;

        // entry-specifier
        if (entrySpecifiers.Count == 1)
          t.RequestArguments["entry-specifier"] = entrySpecifiers.First();
        else
          t.RequestArguments["entry-specifier"] = ImapStringList.CreateParenthesized(entrySpecifiers);

        if (ProcessTransaction(t).IsSuccess)
          metadata = t.TransactionResult;

        return t.Result;
      }
    }
#endregion

#region "SETMETADATA"
#region "server entries"
    /// <summary>sends SETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult SetMetadataServer(params ImapMetadata[] metadata)
    {
      return SetMetadataInternal(null,
                                 metadata);
    }

    /// <summary>sends SETMETADATA command and sets NIL value to remove entries</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult SetMetadataServerRemoval(string[] entrySpecifiers)
    {
      return SetMetadataInternal(null,
                                 entrySpecifiers.Convert(ImapMetadata.CreateNil));
    }
#endregion

#region "mailbox entries"
    /// <summary>sends SETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(ImapMailboxStatus mailbox, params ImapMetadata[] metadata)
    {
      return SetMetadataInternal(ValidateMailboxRelationship(mailbox).Name,
                                 metadata);
    }

    /// <summary>sends SETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(string mailboxName, params ImapMetadata[] metadata)
    {
      return SetMetadataInternal(mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                                 metadata);
    }

    /// <summary>sends SETMETADATA command and sets NIL value to remove entries</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadataRemoval(ImapMailboxStatus mailbox, params string[] entrySpecifiers)
    {
      return SetMetadataInternal(ValidateMailboxRelationship(mailbox).Name,
                                 entrySpecifiers.Convert(ImapMetadata.CreateNil));
    }

    /// <summary>sends SETMETADATA command and sets NIL value to remove entries</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadataRemoval(string mailboxName, params string[] entrySpecifiers)
    {
      return SetMetadataInternal(mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                                 entrySpecifiers.Convert(ImapMetadata.CreateNil));
    }
#endregion

    private ImapCommandResult SetMetadataInternal(string mailboxName,
                                                  ImapMetadata[] metadata)
    {
      if (metadata == null)
        throw new ArgumentNullException(nameof(metadata));

      if (handlesIncapableAsException)
        CheckServerCapabilityMetadata(mailboxName != null);

      var argMailboxName = mailboxName == null ? ImapQuotedString.Empty : ImapMailboxNameString.CreateMailboxName(mailboxName);

      using (var t = new SetMetadataTransaction(connection, this)) {
        // mailbox-name
        t.RequestArguments["mailbox-name"] = argMailboxName;

        // list of entry, values
        var listOfEntryAndValues = new ImapString[metadata.Length * 2];

        for (var i = 0; i < metadata.Length; i++) {
          var m = metadata[i];

          if (m == null)
            throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(nameof(metadata));

          listOfEntryAndValues[i * 2] = m.EntryName;
          listOfEntryAndValues[i * 2 + 1] = m.Value ?? ImapNilString.Nil;
        }

        t.RequestArguments["list of entry, values"] = ImapStringList.CreateParenthesized(listOfEntryAndValues);

        return ProcessTransaction(t);
      }
    }
#endregion

    private void CheckServerCapabilityMetadata(bool requireMailboxAnnotations)
    {
      /*
       * RFC 5464 - The IMAP METADATA Extension
       * http://tools.ietf.org/html/rfc5464
       * 1. Introduction and Overview
       *
       *    A server that supports both server and mailbox annotations indicates
       *    the presence of this extension by returning "METADATA" as one of the
       *    supported capabilities in the CAPABILITY command response.
       * 
       *    A server that supports only server annotations indicates the presence
       *    of this extension by returning "METADATA-SERVER" as one of the
       *    supported capabilities in the CAPABILITY command response.
       */
      if (serverCapabilities.Contains(ImapCapability.Metadata))
        return;
      else if (requireMailboxAnnotations)
        throw new ImapIncapableException(ImapCapability.Metadata);
      else if (!serverCapabilities.Contains(ImapCapability.MetadataServer))
        throw new ImapIncapableException(ImapCapability.MetadataServer);
    }
  }
}
