// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * transaction methods : selected state
     */

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Fetch(IImapNumberSet sequenceOrUidSet,
                                   ImapFetchDataItems messageDataItems)
    {
      return FetchInternal<ImapMessageNullAttributes>(sequenceOrUidSet, messageDataItems, null, null, null);
    }

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Fetch<TMessageAttributes>(IImapNumberSet sequenceOrUidSet,
                                                       ImapFetchDataItems messageDataItems,
                                                       out TMessageAttributes[] messages)
      where TMessageAttributes : ImapMessageAttributesBase
    {
      var messageAttributeCollection = MessageAttributeCollection<TMessageAttributes>.Create(sequenceOrUidSet?.ConcreteCount ?? throw new ArgumentNullException(nameof(sequenceOrUidSet)),
                                                                                             SelectedMailbox);

      var ret = FetchInternal(sequenceOrUidSet, messageDataItems, null, null, messageAttributeCollection);

      if (ret.IsSuccess)
        messages = messageAttributeCollection.ToArray();
      else
        messages = null;

      return ret;
    }

    /// <summary>sends FETCH command with CHANGEDSINCE FETCH modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapCommandResult FetchChangedSince<TMessageAttributes>(IImapNumberSet sequenceOrUidSet,
                                                                   ImapFetchDataItems messageDataItems,
                                                                   ulong modSeq,
                                                                   out TMessageAttributes[] messages)
      where TMessageAttributes : ImapMessageAttributesBase
    {
      var fetchModifiers = ImapStringList.CreateParenthesized(new[] { "CHANGEDSINCE", modSeq.ToString() });
      var messageAttributeCollection = MessageAttributeCollection<TMessageAttributes>.Create(sequenceOrUidSet?.ConcreteCount ?? throw new ArgumentNullException(nameof(sequenceOrUidSet)),
                                                                                             SelectedMailbox);

      var ret = FetchInternal(sequenceOrUidSet, messageDataItems, fetchModifiers, ImapCapability.CondStore, messageAttributeCollection);

      if (ret.IsSuccess)
        messages = messageAttributeCollection.ToArray();
      else
        messages = null;

      return ret;
    }

    private ImapCommandResult FetchInternal<TMessageAttributes>(IImapNumberSet sequenceOrUidSet,
                                                                ImapFetchDataItems messageDataItems,
                                                                ImapStringList fetchModifiers,
                                                                ImapCapability fetchModifiersCapabilityRequirement,
                                                                IImapMessageAttributeCollection<TMessageAttributes> messageAttributeCollection)
      where TMessageAttributes : ImapMessageAttributesBase
    {
      RejectNonSelectedState();

      ImapNumberSet.ThrowIfEmpty(sequenceOrUidSet, nameof(sequenceOrUidSet));

      if (messageDataItems == null)
        throw new ArgumentNullException(nameof(messageDataItems));

      using (var t = new FetchTransaction<TMessageAttributes>(connection,
                                                              this,
                                                              messageAttributeCollection,
                                                              sequenceOrUidSet.IsUidSet,
                                                              fetchModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    fetch           = "FETCH" SP sequence-set SP ("ALL" / "FULL" /
         *                      "FAST" / fetch-att /
         *                      "(" fetch-att *(SP fetch-att) ")")
         *                      [fetch-modifiers]
         *                      ;; modifies the original IMAP4 FETCH command to
         *                      ;; accept optional modifiers
         *    fetch-modifiers = SP "(" fetch-modifier *(SP fetch-modifier) ")"
         *    fetch-modifier  = fetch-modifier-name [ SP fetch-modif-params ]
         *    fetch-modif-params  = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    fetch-modifier-name = tagged-ext-label
         */
        // sequence set
        t.RequestArguments["sequence set"] = sequenceOrUidSet.Expression;

        // message data item names or macro
        t.RequestArguments["message data item names or macro"] = messageDataItems;

        // fetch-modifiers
        if (fetchModifiers != null)
          t.RequestArguments["fetch modifiers"] = fetchModifiers;

        return ProcessTransaction(t);
      }
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Store(IImapNumberSet sequenceOrUidSet,
                                   ImapStoreDataItems storeDataItems)
    {
      return StoreInternal(sequenceOrUidSet,
                           storeDataItems,
                           null,
                           null,
                           false,
                           out _);
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Store(IImapNumberSet sequenceOrUidSet,
                                   ImapStoreDataItems storeDataItems,
                                   out ImapMessageDynamicAttributes[] messageAttributes)
    {
      return StoreInternal(sequenceOrUidSet,
                           storeDataItems,
                           null,
                           null,
                           true,
                           out messageAttributes);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapCommandResult StoreUnchangedSince(IImapNumberSet sequenceOrUidSet,
                                                 ImapStoreDataItems storeDataItems,
                                                 ulong modSeq)
    {
      return StoreUnchangedSinceInternal(sequenceOrUidSet,
                                         storeDataItems,
                                         modSeq,
                                         false,
                                         out _,
                                         out _);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapCommandResult StoreUnchangedSince(IImapNumberSet sequenceOrUidSet,
                                                 ImapStoreDataItems storeDataItems,
                                                 ulong modSeq,
                                                 out ImapMessageDynamicAttributes[] messageAttributes)
    {
      return StoreUnchangedSinceInternal(sequenceOrUidSet,
                                         storeDataItems,
                                         modSeq,
                                         true,
                                         out messageAttributes,
                                         out _);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapCommandResult StoreUnchangedSince(IImapNumberSet sequenceOrUidSet,
                                                 ImapStoreDataItems storeDataItems,
                                                 ulong modSeq,
                                                 out IImapNumberSet failedMessageSet)
    {
      return StoreUnchangedSinceInternal(sequenceOrUidSet,
                                         storeDataItems,
                                         modSeq,
                                         false,
                                         out _,
                                         out failedMessageSet);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapCommandResult StoreUnchangedSince(IImapNumberSet sequenceOrUidSet,
                                                 ImapStoreDataItems storeDataItems,
                                                 ulong modSeq,
                                                 out ImapMessageDynamicAttributes[] messageAttributes,
                                                 out IImapNumberSet failedMessageSet)
    {
      return StoreUnchangedSinceInternal(sequenceOrUidSet,
                                         storeDataItems,
                                         modSeq,
                                         true,
                                         out messageAttributes,
                                         out failedMessageSet);
    }

    private ImapCommandResult StoreUnchangedSinceInternal(IImapNumberSet sequenceOrUidSet,
                                                          ImapStoreDataItems storeDataItems,
                                                          ulong modSeq,
                                                          bool retrieveMessageAttributes,
                                                          out ImapMessageDynamicAttributes[] messageAttributes,
                                                          out IImapNumberSet failedMessageSet)
    {
      failedMessageSet = null;
      messageAttributes = null;

      var storeModifiers = ImapStringList.CreateParenthesized(new[] { "UNCHANGEDSINCE", modSeq.ToString() });

      var ret = StoreInternal(sequenceOrUidSet,
                              storeDataItems,
                              storeModifiers,
                              ImapCapability.CondStore,
                              retrieveMessageAttributes,
                              out messageAttributes);

      if (ret.TaggedStatusResponse.ResponseText.Code == ImapResponseCode.Modified)
        failedMessageSet = ImapResponseTextParser.FromModified(ret.TaggedStatusResponse.ResponseText, sequenceOrUidSet.IsUidSet);

      return ret;
    }

    private ImapCommandResult StoreInternal(IImapNumberSet sequenceOrUidSet,
                                            ImapStoreDataItems storeDataItems,
                                            ImapStringList storeModifiers,
                                            ImapCapability storeModifiersCapabilityRequirement,
                                            bool retrieveMessageAttributes,
                                            out ImapMessageDynamicAttributes[] messageAttributes)
    {
      RejectNonSelectedState();

      ImapNumberSet.ThrowIfEmpty(sequenceOrUidSet, nameof(sequenceOrUidSet));

      if (storeDataItems == null)
        throw new ArgumentNullException(nameof(storeDataItems));

      messageAttributes = null;

      var messageAttributeCollection = retrieveMessageAttributes
        ? MessageAttributeCollection<ImapMessageDynamicAttributes>.Create(sequenceOrUidSet.ConcreteCount, SelectedMailbox)
        : null;

      using (var t = new StoreTransaction(connection,
                                          this,
                                          messageAttributeCollection,
                                          sequenceOrUidSet.IsUidSet,
                                          storeModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    store           = "STORE" SP sequence-set [store-modifiers]
         *                      SP store-att-flags
         *                      ;; extend [IMAP4] STORE command syntax
         *                      ;; to allow for optional store-modifiers
         *    store-modifiers =  SP "(" store-modifier *(SP store-modifier)
         *                        ")"
         *    store-modifier  = store-modifier-name [SP store-modif-params]
         *    store-modif-params = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    store-modifier-name = tagged-ext-label
         */
        // message set
        t.RequestArguments["message set"] = sequenceOrUidSet.Expression;

        // store-modifiers
        if (storeModifiers != null)
          t.RequestArguments["store modifiers"] = storeModifiers;

        // message data item name
        t.RequestArguments["message data item name"] = storeDataItems.ItemName;

        // value for message data item
        t.RequestArguments["value for message data item"] = storeDataItems;

        ProcessTransaction(t);

        messageAttributes = messageAttributeCollection?.ToArray();

        return t.Result;
      }
    }
  }
}
