// 
// Copyright (c) 2018 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * transaction methods : authenticated/selected state / IDLE
     */
    public ImapCommandResult Idle(Predicate<ImapResponse> keepIdlePredicate)
      => IdleInternal(Timeout.Infinite, keepIdlePredicate);

    public ImapCommandResult Idle(TimeSpan idleTimeout, Predicate<ImapResponse> keepIdlePredicate = null)
      => IdleInternal((int)idleTimeout.TotalMilliseconds, keepIdlePredicate);

    public ImapCommandResult Idle(int idleMillisecondsTimeout, Predicate<ImapResponse> keepIdlePredicate = null)
      => IdleInternal(idleMillisecondsTimeout, keepIdlePredicate);

    private ImapCommandResult IdleInternal(int idleMillisecondsTimeout, Predicate<ImapResponse> keepIdlePredicate)
    {
      if (idleMillisecondsTimeout < -1)
        throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(idleMillisecondsTimeout), idleMillisecondsTimeout);
      if (idleMillisecondsTimeout == Timeout.Infinite && keepIdlePredicate == null)
        throw new ArgumentException("non-infinite timeout or predicate must be specified");

      RejectNonAuthenticatedState();

      using (var cts = new CancellationTokenSource(idleMillisecondsTimeout)) {
        using (var t = new IdleTransaction(connection, this, keepIdlePredicate, cts.Token)) {
          return ProcessTransactionAsync(t).Result;
        }
      }
    }

    public Task<ImapCommandResult> IdleAsync(Predicate<ImapResponse> keepIdlePredicate)
      => IdleAsync(keepIdlePredicate ?? throw new ArgumentNullException(nameof(keepIdlePredicate)),
                   CancellationToken.None);

    public Task<ImapCommandResult> IdleAsync(CancellationToken stopIdleCancellationToken)
      => IdleAsync(null,
                   stopIdleCancellationToken);

    public Task<ImapCommandResult> IdleAsync(Predicate<ImapResponse> keepIdlePredicate, CancellationToken stopIdleCancellationToken)
    {
      RejectNonAuthenticatedState();

      // throw exception before starting task
      RejectTransactionProceeding();
      IsCapable(ImapCapability.Idle, throwIfIncapable: true);

      return Idle();

      async Task<ImapCommandResult> Idle()
      {
        using (var idleTransaction = new IdleTransaction(connection, this, keepIdlePredicate, stopIdleCancellationToken)) {
          var idleTask = ProcessTransactionAsync(idleTransaction);

          // wait for IDLE started or respond tagged error response
          idleTransaction.WaitForIdleStateChanged();

          return await idleTask;
        }
      }
    }
  }
}
