// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * transaction methods : authenticated state / LIST/LSUB etc.
     */

    /// <summary>sends LIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty mailbox name and an empty reference name.
    /// </remarks>
    public ImapCommandResult ListRoot(out ImapListedMailbox root)
      => ListRoot(string.Empty, out root);

    /// <summary>sends LIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty mailbox name.
    /// </remarks>
    public ImapCommandResult ListRoot(string referenceName, out ImapListedMailbox root)
    {
      RejectNonAuthenticatedState();

      var argReferenceName = ImapMailboxNameString.CreateReferenceName(referenceName);

      root = null;

      using (var t = new ListTransaction(connection, this)) {
        t.RequestArguments["reference name"] = argReferenceName;
        t.RequestArguments["mailbox name"] = ImapQuotedString.Empty;

        if (ProcessTransaction(t).IsSuccess) {
          hierarchyDelimiters[referenceName] = t.TransactionResult[0].HierarchyDelimiter;

          root = t.TransactionResult[0];
        }

        return t.Result;
      }
    }

    private static readonly string mailboxNameAll = "*";

    /// <summary>sends LIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapCommandResult ListAll(out IReadOnlyList<ImapMailboxStatus> mailboxes)
      => ListInternal(string.Empty, mailboxNameAll, out mailboxes);

    /// <summary>sends LSUB command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LSUB command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapCommandResult LsubAll(out IReadOnlyList<ImapMailboxStatus> mailboxes)
      => LsubInternal(string.Empty, mailboxNameAll, out mailboxes);

    /// <summary>sends RLIST command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends RLIST command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapCommandResult RListAll(out IReadOnlyList<ImapMailboxStatus> mailboxes)
      => RListInternal(string.Empty, mailboxNameAll, out mailboxes);

    /// <summary>sends RLSUB command</summary>
    /// <remarks>
    /// valid in authenticated state
    /// This method sends LIST command with an empty reference name and a wildcard "*" as mailbox name
    /// </remarks>
    public ImapCommandResult RLsubAll(out IReadOnlyList<ImapMailboxStatus> mailboxes)
      => RLsubInternal(string.Empty, mailboxNameAll, out mailboxes);



    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult List(string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return ListInternal(string.Empty,
                          mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                          out mailboxes);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult List(ImapMailboxStatus mailbox)
    {
      return ListInternal(string.Empty,
                          ValidateMailboxRelationship(mailbox).Name,
                          out _);
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult List(string referenceName, string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return ListInternal(referenceName ?? throw new ArgumentNullException(nameof(referenceName)),
                          mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                          out mailboxes);
    }

    private ImapCommandResult ListInternal(string referenceName,
                                           string mailboxName,
                                           out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      using (var t = new ListTransaction(connection, this)) {
        return ListLsubInternal(t, referenceName, ImapMailboxNameString.CreateListMailboxName(mailboxName), out mailboxes);
      }
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult Lsub(string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return LsubInternal(string.Empty,
                          mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                          out mailboxes);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult Lsub(ImapMailboxStatus mailbox)
    {
      return LsubInternal(string.Empty,
                          ValidateMailboxRelationship(mailbox).Name,
                          out _);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult Lsub(string referenceName, string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return LsubInternal(referenceName ?? throw new ArgumentNullException(nameof(referenceName)),
                          mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                          out mailboxes);
    }

    private ImapCommandResult LsubInternal(string referenceName,
                                           string mailboxName,
                                           out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      using (var t = new LsubTransaction(connection, this)) {
        return ListLsubInternal(t, referenceName, ImapMailboxNameString.CreateListMailboxName(mailboxName), out mailboxes);
      }
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult RList(string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return RListInternal(string.Empty,
                           mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                           out mailboxes);
    }

    /// <summary>sends RLIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult RList(string referenceName, string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return RListInternal(referenceName ?? throw new ArgumentNullException(nameof(referenceName)),
                           mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                           out mailboxes);
    }

    private ImapCommandResult RListInternal(string referenceName,
                                            string mailboxName,
                                            out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      using (var t = new RListTransaction(connection, this)) {
        return ListLsubInternal(t, referenceName, ImapMailboxNameString.CreateListMailboxName(mailboxName), out mailboxes);
      }
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult RLsub(string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return RLsubInternal(string.Empty,
                           mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                           out mailboxes);
    }

    /// <summary>sends RLSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult RLsub(string referenceName, string mailboxName, out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return RLsubInternal(referenceName ?? throw new ArgumentNullException(nameof(referenceName)),
                           mailboxName ?? throw new ArgumentNullException(nameof(mailboxName)),
                           out mailboxes);
    }

    private ImapCommandResult RLsubInternal(string referenceName,
                                            string mailboxName,
                                            out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      using (var t = new RLsubTransaction(connection, this)) {
        return ListLsubInternal(t, referenceName, ImapMailboxNameString.CreateListMailboxName(mailboxName), out mailboxes);
      }
    }

#region "extended LIST with no reference name"
#region "extended LIST with multiple mailbox name pattern"
    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(out IReadOnlyList<ImapMailboxStatus> mailboxes,
                                          string mailboxNamePattern,
                                          params string[] mailboxNamePatterns)
    {
      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, null, null, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(ImapListSelectionOptions selectionOptions,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes,
                                          string mailboxNamePattern,
                                          params string[] mailboxNamePatterns)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException(nameof(selectionOptions));

      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, selectionOptions, null, out mailboxes);
    }

      /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(ImapListReturnOptions returnOptions,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes,
                                          string mailboxNamePattern,
                                          params string[] mailboxNamePatterns)
    {
      if (returnOptions == null)
        throw new ArgumentNullException(nameof(returnOptions));

      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, null, returnOptions, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(ImapListSelectionOptions selectionOptions,
                                          ImapListReturnOptions returnOptions,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes,
                                          string mailboxNamePattern,
                                          params string[] mailboxNamePatterns)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException(nameof(selectionOptions));
      if (returnOptions == null)
        throw new ArgumentNullException(nameof(returnOptions));

      return ListExtendedInternalNoRefName(mailboxNamePattern, mailboxNamePatterns, selectionOptions, returnOptions, out mailboxes);
    }
#endregion

#region "extended LIST with single mailbox name pattern"
    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(string mailboxNamePattern,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return ListExtendedInternalNoRefName(mailboxNamePattern, null, null, null, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(string mailboxNamePattern,
                                          ImapListSelectionOptions selectionOptions,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException(nameof(selectionOptions));

      return ListExtendedInternalNoRefName(mailboxNamePattern, null, selectionOptions, null, out mailboxes);
    }

      /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(string mailboxNamePattern,
                                          ImapListReturnOptions returnOptions,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      if (returnOptions == null)
        throw new ArgumentNullException(nameof(returnOptions));

      return ListExtendedInternalNoRefName(mailboxNamePattern, null, null, returnOptions, out mailboxes);
    }

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(string mailboxNamePattern,
                                          ImapListSelectionOptions selectionOptions,
                                          ImapListReturnOptions returnOptions,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      if (selectionOptions == null)
        throw new ArgumentNullException(nameof(selectionOptions));
      if (returnOptions == null)
        throw new ArgumentNullException(nameof(returnOptions));

      return ListExtendedInternalNoRefName(mailboxNamePattern, null, selectionOptions, returnOptions, out mailboxes);
    }
#endregion

    private ImapCommandResult ListExtendedInternalNoRefName(string mailboxNamePattern,
                                                            string[] mailboxNamePatterns,
                                                            ImapListSelectionOptions selectionOptions,
                                                            ImapListReturnOptions returnOptions,
                                                            out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      return ListExtendedInternal(string.Empty,
                                  ParamArrayUtils.ToEnumerableNonNullable(nameof(mailboxNamePatterns), mailboxNamePattern, mailboxNamePatterns).ToList(),
                                  selectionOptions,
                                  returnOptions,
                                  out mailboxes);
    }
#endregion

    /// <summary>sends extended LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public ImapCommandResult ListExtended(string referenceName,
                                          IReadOnlyCollection<string> mailboxNamePatterns,
                                          ImapListSelectionOptions selectionOptions,
                                          ImapListReturnOptions returnOptions,
                                          out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      if (referenceName == null)
        throw new ArgumentNullException(nameof(referenceName));
      if (mailboxNamePatterns == null)
        throw new ArgumentNullException(nameof(mailboxNamePatterns));
      if (mailboxNamePatterns.Count == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyCollection(nameof(mailboxNamePatterns));
      if (selectionOptions == null)
        throw new ArgumentNullException(nameof(selectionOptions));
      if (returnOptions == null)
        throw new ArgumentNullException(nameof(returnOptions));

      return ListExtendedInternal(referenceName, mailboxNamePatterns, selectionOptions, returnOptions, out mailboxes);
    }

    private ImapCommandResult ListExtendedInternal(string referenceName,
                                                   IReadOnlyCollection<string> mailboxNamePatterns,
                                                   ImapListSelectionOptions selectionOptions,
                                                   ImapListReturnOptions returnOptions,
                                                   out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      using (var t = new ListExtendedTransaction(connection, this, mailboxManager)) {
        if (selectionOptions != null)
          t.RequestArguments["selection options"] = selectionOptions;

        if (returnOptions != null)
          t.RequestArguments["return options"] = returnOptions;

        var quotedMailboxNamePatterns = mailboxNamePatterns.ConvertAll(ImapMailboxNameString.CreateListMailboxName);
        
        if (quotedMailboxNamePatterns.Count == 1)
          ListLsubInternal(t, referenceName, quotedMailboxNamePatterns[0], out mailboxes);
        else
          ListLsubInternal(t, referenceName, ImapStringList.CreateParenthesized(quotedMailboxNamePatterns), out mailboxes);

        return t.Result;
      }
    }

    private ImapCommandResult ListLsubInternal(ListTransactionBase t,
                                               string referenceName,
                                               ImapString mailboxName,
                                               out IReadOnlyList<ImapMailboxStatus> mailboxes)
    {
      RejectNonAuthenticatedState();

      mailboxes = null;

      t.RequestArguments["reference name"] = ImapMailboxNameString.CreateReferenceName(referenceName);
      t.RequestArguments["mailbox name"] = mailboxName;

      if (ProcessTransaction(t).IsSuccess)
        mailboxes = t.TransactionResult.ConvertAll(mailboxManager.GetExistOrCreate);

      return t.Result;
    }
  }
}
