// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Linq;
using System.Net;
using System.Threading;

using Smdn.Net.Connection;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;
using Smdn.Net.Imap4.Urls;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * transaction methods : non-authenticated state
     */

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapCommandResult StartTls(UpgradeConnectionStreamCallback createAuthenticatedStream,
                                      bool reissueCapability = false)
    {
      RejectNonConnectedState();

      if (State != ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("already authenticated");

      if (createAuthenticatedStream == null)
        throw new ArgumentNullException(nameof(createAuthenticatedStream));

      ImapCommandResult result;

      using (var t = new StartTlsTransaction(connection, this, createAuthenticatedStream)) {
        if ((result = ProcessTransaction(t)).IsSuccess)
          // 6.2.1. STARTTLS Command
          //       Once [TLS] has been started, the client MUST discard cached
          //       information about server capabilities and SHOULD re-issue the
          //       CAPABILITY command.
          SetServerCapabilities(null);
        else
          return result;
      }

      if (reissueCapability)
        Capability();

      return result;
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapCommandResult Authenticate(ImapAuthenticationMechanism authenticationMechanism,
                                          ICredentialsByHost credentials,
                                          string username = null,
                                          bool reissueCapability = false,
                                          CancellationToken cancellationToken = default)
    {
      var ret = RejectNonConnectedOrGetAuthenticatedResult();

      if (ret != null)
        return ret;

      if (credentials == null)
        throw new ArgumentNullException(nameof(credentials));
      if (authenticationMechanism == null)
        throw new ArgumentNullException(nameof(authenticationMechanism));

      var credential = credentials.LookupCredential(connection, username, authenticationMechanism?.Value);

      if (credential == null)
        return ImapCommandResult.CreateFailure($"credential not found for {username};AUTH={authenticationMechanism.Value}@{connection.Host}:{connection.Port}");

      using (var saslMechanism = SaslClientMechanism.Create(authenticationMechanism.Value)) {
        // 6.2.2.  AUTHENTICATE Command
        //    The service name specified by this protocol's profile of [SASL] is
        //    "imap".
        saslMechanism.ServiceName = "imap";
        saslMechanism.Credential = credential;

        using (var t = new AuthenticateTransaction(connection,
                                                   this,
                                                   saslMechanism,
                                                   serverCapabilities.Contains(ImapCapability.SaslIR),
                                                   cancellationToken)) {
          t.RequestArguments["authentication mechanism name"] = authenticationMechanism; // capability will be checked

          return AuthenticateInternal(t,
                                      saslMechanism.Credential?.UserName,
                                      authenticationMechanism,
                                      reissueCapability);
        }
      }
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapCommandResult Authenticate(SaslClientMechanism specificAuthenticationMechanism,
                                          bool reissueCapability = false,
                                          CancellationToken cancellationToken = default)
    {
      var ret = RejectNonConnectedOrGetAuthenticatedResult();

      if (ret != null)
        return ret;

      if (specificAuthenticationMechanism == null)
        throw new ArgumentNullException(nameof(specificAuthenticationMechanism));

      using (var t = new AuthenticateTransaction(connection,
                                                 this,
                                                 specificAuthenticationMechanism,
                                                 serverCapabilities.Contains(ImapCapability.SaslIR),
                                                 cancellationToken)) {
        t.RequestArguments["authentication mechanism name"] = specificAuthenticationMechanism.Name; // capability will not be checked

        return AuthenticateInternal(t,
                                    specificAuthenticationMechanism.Credential?.UserName,
                                    ImapAuthenticationMechanism.GetKnownOrCreate(specificAuthenticationMechanism.Name),
                                    reissueCapability);
      }
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non-authenticated state</remarks>
    public ImapCommandResult Login(ICredentialsByHost credentials,
                                   string username = null,
                                   bool reissueCapability = false)
    {
      var ret = RejectNonConnectedOrGetAuthenticatedResult();

      if (ret != null)
        return ret;

      // RFC 2595 Using TLS with IMAP, POP3 and ACAP
      // http://tools.ietf.org/html/rfc2595
      // 3.2. IMAP LOGINDISABLED capability
      //    An IMAP client which complies with this specification MUST NOT issue
      //    the LOGIN command if this capability is present.
      if (serverCapabilities.Contains(ImapCapability.LoginDisabled))
        throw new ImapIncapableException("LOGIN is disabled");

      if (credentials == null)
        throw new ArgumentNullException(nameof(credentials));

      var credential = credentials.LookupCredential(connection, username, null);

      if (credential == null)
        return ImapCommandResult.CreateFailure($"credential not found for {username}@{connection.Host}:{connection.Port}");

      using (var t = new LoginTransaction(connection, this, credential)) {
        return AuthenticateInternal(t, credential.UserName, null, reissueCapability);
      }
    }

    private ImapCommandResult RejectNonConnectedOrGetAuthenticatedResult()
    {
      RejectNonConnectedState();

      if (State == ImapSessionState.NotAuthenticated)
        return null;
      else
        return ImapCommandResult.CreateSuccess("already authenticated");
    }

    private ImapCommandResult AuthenticateInternal(IImapTransaction t,
                                                   string username,
                                                   ImapAuthenticationMechanism authenticationMechanism,
                                                   bool reissueCapability)
    {
      ImapCommandResult result;

      try {
        result = ProcessTransaction(t);
      }
      catch (AuthenticateTransaction.SaslInitialExchangeFailedException ex) {
        return ImapCommandResult.CreateFailure(ex.Message);
      }

      var refferalResponseCode = result.GetStatusResponseOfCode(ImapResponseCode.Referral);

      if (refferalResponseCode != null) {
        // RFC 2221 IMAP4 Login Referrals
        // http://tools.ietf.org/html/rfc2221
        // 4.1. LOGIN and AUTHENTICATE Referrals
        //    An IMAP4 server MAY respond to a LOGIN or AUTHENTICATE command with a
        //    home server referral if it wishes to direct the user to another IMAP4
        //    server.
        var referToUri = ImapResponseTextParser.FromReferral(refferalResponseCode.ResponseText)[0];

        if (handlesReferralAsException) {
          throw new ImapLoginReferralException(string.Format("try another server: '{0}'", refferalResponseCode.ResponseText.Text),
                                               referToUri);
        }
        else {
          TraceInfo("login referral: '{0}'", refferalResponseCode.ResponseText.Text);

          if (result.IsSuccess)
            TraceInfo(string.Concat("  another server available at ", referToUri));
          else
            TraceInfo(string.Concat("  try to connect to ", referToUri));
        }
      }

      if (result.IsSuccess) {
        UpdateAuthority(username, authenticationMechanism);
        TransitStateTo(ImapSessionState.Authenticated);

        // 6.2.2. AUTHENTICATE Command
        //       A server MAY include a CAPABILITY response code in the tagged OK
        //       response of a successful AUTHENTICATE command in order to send
        //       capabilities automatically.

        // 6.2.3. LOGIN Command
        //       A server MAY include a CAPABILITY response code in the tagged OK
        //       response to a successful LOGIN command in order to send
        //       capabilities automatically.  It is unnecessary for a client to
        //       send a separate CAPABILITY command if it recognizes these
        //       automatic capabilities.
        var capabilityResponseCode = result.GetStatusResponseOfCode(ImapResponseCode.Capability);

        if (capabilityResponseCode == null) {
          var capability = result.ReceivedResponses.OfType<ImapServerDataResponse>().FirstOrDefault(data => data.Type == ImapServerDataResponseType.Capability);

          if (capability != null) {
            SetServerCapabilities(ImapServerDataResponseParser.FromCapability(capability));
            reissueCapability = false;
          }
        }
        else {
          SetServerCapabilities(ImapResponseTextParser.FromCapability(capabilityResponseCode.ResponseText));
          reissueCapability = false;
        }
      }
      else {
        return result;
      }

      if (reissueCapability)
        Capability();

      return result;
    }

    internal void UpdateAuthority(string username, ImapAuthenticationMechanism authType)
    {
      authority.Scheme    = connection.IsSecurePortConnection ? ImapUri.UriSchemeImaps : ImapUri.UriSchemeImap;
      authority.Host      = connection.Host;
      authority.Port      = connection.Port;
      authority.UserName  = username;
      authority.AuthType  = authType;

      TraceInfo(string.Concat("authority: ", authority));
    }
  }
}
