// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * transaction methods : selected state
     */

    /// <summary>sends CHECK command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Check()
    {
      RejectNonSelectedState();

      using (var t = new CheckTransaction(connection, this)) {
        return ProcessTransaction(t);
      }
    }

    /// <summary>sends CLOSE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Close()
    {
      RejectNonAuthenticatedState();

      if (State != ImapSessionState.Selected)
        return ImapCommandResult.CreateSuccess("mailbox has already closed or has not selected yet");

      using (var t = new CloseTransaction(connection, this)) {
        if (ProcessTransaction(t).IsSuccess) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Expunge()
      => ExpungeInternal(uidSet: null, retrieveExpungedMessages: false, out _);

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Expunge(out IReadOnlyList<long> expungedMessages)
      => ExpungeInternal(uidSet: null, retrieveExpungedMessages: true, out expungedMessages);

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult UidExpunge(IImapNumberSet uidSet)
    {
      ImapNumberSet.ThrowIfEmptyOrNotUidSet(uidSet, nameof(uidSet));

      return ExpungeInternal(uidSet, retrieveExpungedMessages: false, out _);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult UidExpunge(IImapNumberSet uidSet, out IReadOnlyList<long> expungedMessages)
    {
      ImapNumberSet.ThrowIfEmptyOrNotUidSet(uidSet, nameof(uidSet));

      return ExpungeInternal(uidSet, retrieveExpungedMessages: true, out expungedMessages);
    }

    private ImapCommandResult ExpungeInternal(IImapNumberSet uidSet,
                                              bool retrieveExpungedMessages,
                                              out IReadOnlyList<long> expungedMessages)
    {
      RejectNonSelectedState();

      expungedMessages = null;

      using (var t = new ExpungeTransaction(connection,
                                            this,
                                            retrieveExpungedMessages,
                                            (uidSet == null ? -1 : uidSet.ConcreteCount),
                                            (uidSet != null))) {
        // sequence set for UID EXPUNGE
        if (uidSet != null)
          t.RequestArguments["sequence set"] = uidSet.Expression;

        if (ProcessTransaction(t).IsSuccess)
          expungedMessages = t.TransactionResult;

        return t.Result;
      }
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Search(out ImapMatchedSequenceSet matched, ImapSearchCriteria searchingCriteria, Encoding charset = default)
    {
      using (var t = new SearchTransaction(connection, this, uid: false, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapCommandResult ESearch(out ImapMatchedSequenceSet matched, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, Encoding charset = default)
    {
      using (var t = new SearchTransaction(connection, this, uid: false, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult UidSearch(out ImapMatchedSequenceSet matched, ImapSearchCriteria searchingCriteria, Encoding charset = default)
    {
      using (var t = new SearchTransaction(connection, this, uid: true, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapCommandResult UidESearch(out ImapMatchedSequenceSet matched, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, Encoding charset = default)
    {
      using (var t = new SearchTransaction(connection, this, uid: true, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Sort(out ImapMatchedSequenceSet matched, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset = default)
    {
      using (var t = new SortTransaction(connection, this, uid: false, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapCommandResult ESort(out ImapMatchedSequenceSet matched, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, Encoding charset = default)
    {
      using (var t = new SortTransaction(connection, this, uid: false, ImapCapability.ESort)) {
        return ESearchESortInternal(t, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult UidSort(out ImapMatchedSequenceSet matched, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset = default)
    {
      using (var t = new SortTransaction(connection, this, uid: true, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends UID SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapCommandResult UidESort(out ImapMatchedSequenceSet matched, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, Encoding charset = default)
    {
      using (var t = new SortTransaction(connection, this, uid: true, ImapCapability.ESort)) {
        return ESearchESortInternal(t, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    private ImapCommandResult ESearchESortInternal(SearchTransactionBase t,
                                                   ImapSortCriteria sortingCriteria,
                                                   ImapSearchCriteria searchingCriteria,
                                                   Encoding charset,
                                                   ImapSearchResultOptions resultOptions,
                                                   out ImapMatchedSequenceSet matched)
    {
      if (resultOptions == null)
        throw new ArgumentNullException(nameof(resultOptions));

      /*
       * RFC 5182 - IMAP Extension for Referencing the Last SEARCH Result
       * http://tools.ietf.org/html/rfc5182
       */
      var requiredCapabilities = new ImapCapabilitySet((resultOptions as IImapExtension).RequiredCapabilities, asReadOnly: false);

      // ESEARCH or ESORT capability
      if (t is SearchTransaction) {
        requiredCapabilities.Add(ImapCapability.ESearch);
        requiredCapabilities.Remove(ImapCapability.ESort);
      }
      else if (t is SortTransaction) {
        requiredCapabilities.Add(ImapCapability.ESort);
        requiredCapabilities.Remove(ImapCapability.ESearch);
      }

      resultOptions = new ImapSearchResultOptions(requiredCapabilities, resultOptions);

      SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);

      var savedResult = resultOptions.ContainsOneOf(ImapSearchResultOptions.Save) &&
                        t.Result.IsSuccess &&
                        t.Result.GetStatusResponseOfCode(ImapResponseCode.NotSaved) == null;

      if (savedResult)
        matched = ImapMatchedSequenceSet.ToSavedResult(t.Uid, matched);

      return t.Result;
    }

    private ImapCommandResult SearchSortInternal(SearchTransactionBase t,
                                                 ImapSortCriteria sortingCriteria,
                                                 ImapSearchCriteria searchingCriteria,
                                                 Encoding charset,
                                                 ImapSearchResultOptions resultOptions,
                                                 out ImapMatchedSequenceSet matched)
    {
      if (t is SortTransaction && sortingCriteria == null)
        throw new ArgumentNullException(nameof(sortingCriteria));
      if (searchingCriteria == null)
        throw new ArgumentNullException(nameof(searchingCriteria));

      RejectNonSelectedState();

      matched = null;

      /*
       * 6.4.4. SEARCH Command
       *    US-ASCII MUST be supported; other [CHARSET]s MAY be supported.
       * 
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.SORT. SORT Command
       *    The US-ASCII and [UTF-8] charsets MUST be implemented.
       *    All other charsets are optional.
       */
      if (charset == null) {
        if (t is SortTransaction)
          charset = Encoding.UTF8;
        else
          charset = Encoding.ASCII;
      }

      var charsetSpecified = searchingCriteria.SetCharset(charset);

      /*
       * RFC 4466 - Collected Extensions to IMAP4 ABNF
       * http://tools.ietf.org/html/rfc4466
       *
       *    search          = "SEARCH" [search-return-opts]
       *                      SP search-program
       *    search-correlator  = SP "(" "TAG" SP tag-string ")"
       *    search-program     = ["CHARSET" SP charset SP]
       *                         search-key *(SP search-key)
       *                         ;; CHARSET argument to SEARCH MUST be
       *                         ;; registered with IANA.
       *    search-return-data = search-modifier-name SP search-return-value
       *                         ;; Note that not every SEARCH return option
       *                         ;; is required to have the corresponding
       *                         ;; ESEARCH return data.
       *    search-return-opts = SP "RETURN" SP "(" [search-return-opt
       *                         *(SP search-return-opt)] ")"
       *    search-return-opt = search-modifier-name [SP search-mod-params]
       *    search-return-value = tagged-ext-val
       *                         ;; Data for the returned search option.
       *                         ;; A single "nz-number"/"number" value
       *                         ;; can be returned as an atom (i.e., without
       *                         ;; quoting).  A sequence-set can be returned
       *                         ;; as an atom as well.
       *    search-modifier-name = tagged-ext-label
       *    search-mod-params = tagged-ext-val
       *                      ;; This non-terminal shows recommended syntax
       *                      ;; for future extensions.
       */

      /* 
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.SORT. SORT Command
       * The charset argument is mandatory (unlike SEARCH)
       */
      // charset specification
      if (t is SortTransaction || charsetSpecified)
        t.RequestArguments["charset specification"] = charset.WebName;

      // search-return-opt
      if (resultOptions != null)
        t.RequestArguments["result specifier"] = resultOptions;

      // searching criteria
      t.RequestArguments["searching criteria"] = searchingCriteria;

        // sort criteria
      if (sortingCriteria != null)
        t.RequestArguments["sort criteria"] = sortingCriteria;

      if (ProcessTransaction(t).IsSuccess)
        matched = t.TransactionResult;

      return t.Result;
    }

    /// <summary>sends THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Thread(out ImapThreadList threadList, ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset = default)
      => ThreadInternal(uid: false, threadingAlgorithm, searchingCriteria, charset, out threadList);

    /// <summary>sends UID THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult UidThread(out ImapThreadList threadList, ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset = default)
      => ThreadInternal(uid: true, threadingAlgorithm, searchingCriteria, charset, out threadList);

    private ImapCommandResult ThreadInternal(bool uid,
                                             ImapThreadingAlgorithm threadingAlgorithm,
                                             ImapSearchCriteria searchingCriteria,
                                             Encoding charset,
                                             out ImapThreadList threadList)
    {
      if (threadingAlgorithm == null)
        throw new ArgumentNullException(nameof(threadingAlgorithm));
      if (searchingCriteria == null)
        throw new ArgumentNullException(nameof(searchingCriteria));

      RejectNonSelectedState();

      threadList = null;

      using (var t = new ThreadTransaction(connection, this, uid)) {
        /*
         * http://tools.ietf.org/html/rfc5256
         * BASE.6.4.THREAD. THREAD Command
         *   The US-ASCII and [UTF-8] charsets MUST be implemented.
         */
        if (charset == null)
          charset = Encoding.UTF8;

        searchingCriteria.SetCharset(charset);

        // threading algorithm
        t.RequestArguments["threading algorithm"] = threadingAlgorithm;

        // charset specification
        t.RequestArguments["charset specification"] = charset.WebName;

        // searching criteria
        t.RequestArguments["searching criteria"] = searchingCriteria;

        if (ProcessTransaction(t).IsSuccess)
          threadList = t.TransactionResult;

        return t.Result;
      }
    }

    public enum CopyMoveCommand {
      Copy,
      Move,
    }

    #region "COPY"
    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Copy(IImapNumberSet sequenceOrUidSet, ImapMailboxStatus mailbox)
      => CopyMoveInternal(CopyMoveCommand.Copy, sequenceOrUidSet, ValidateMailboxRelationship(mailbox).Name, tryCreate: false, out _, out _);

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult Copy(IImapNumberSet sequenceOrUidSet, ImapMailboxStatus mailbox, out ImapCopiedUidSet copiedUids)
      => CopyMoveInternal(CopyMoveCommand.Copy, sequenceOrUidSet, ValidateMailboxRelationship(mailbox).Name, tryCreate: false, out copiedUids, out _);

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Copy(IImapNumberSet sequenceOrUidSet, string mailboxName)
      => CopyMoveInternal(CopyMoveCommand.Copy, sequenceOrUidSet, mailboxName, tryCreate: false, out _, out _);

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult Copy(IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids)
      => CopyMoveInternal(CopyMoveCommand.Copy, sequenceOrUidSet, mailboxName, tryCreate: false, out copiedUids, out _);

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>
    public ImapCommandResult CopyTryCreateMailbox(IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapMailboxStatus createdMailbox)
      => CopyMoveInternal(CopyMoveCommand.Copy, sequenceOrUidSet, mailboxName, tryCreate: true, out _, out createdMailbox);

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult CopyTryCreateMailbox(IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids, out ImapMailboxStatus createdMailbox)
      => CopyMoveInternal(CopyMoveCommand.Copy, sequenceOrUidSet, mailboxName, tryCreate: true, out copiedUids, out createdMailbox);
    #endregion

    #region "MOVE"
    /// <summary>sends MOVE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Move(IImapNumberSet sequenceOrUidSet, ImapMailboxStatus mailbox)
      => CopyMoveInternal(CopyMoveCommand.Move, sequenceOrUidSet, ValidateMailboxRelationship(mailbox).Name, tryCreate: false, out _, out _);

    /// <summary>sends MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult Move(IImapNumberSet sequenceOrUidSet, ImapMailboxStatus mailbox, out ImapCopiedUidSet copiedUids)
      => CopyMoveInternal(CopyMoveCommand.Move, sequenceOrUidSet, ValidateMailboxRelationship(mailbox).Name, tryCreate: false, out copiedUids, out _);

    /// <summary>sends MOVE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Move(IImapNumberSet sequenceOrUidSet, string mailboxName)
      => CopyMoveInternal(CopyMoveCommand.Move, sequenceOrUidSet, mailboxName, tryCreate: false, out _, out _);

    /// <summary>sends MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult Move(IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids)
      => CopyMoveInternal(CopyMoveCommand.Move, sequenceOrUidSet, mailboxName, tryCreate: false, out copiedUids, out _);

    /// <summary>sends MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>
    public ImapCommandResult MoveTryCreateMailbox(IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapMailboxStatus createdMailbox)
      => CopyMoveInternal(CopyMoveCommand.Move, sequenceOrUidSet, mailboxName, tryCreate: true, out _, out createdMailbox);

    /// <summary>sends MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult MoveTryCreateMailbox(IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids, out ImapMailboxStatus createdMailbox)
      => CopyMoveInternal(CopyMoveCommand.Move, sequenceOrUidSet, mailboxName, tryCreate: true, out copiedUids, out createdMailbox);
    #endregion

    #region "COPY/MOVE"
    /// <summary>sends COPY/MOVE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult CopyMove(CopyMoveCommand command, IImapNumberSet sequenceOrUidSet, ImapMailboxStatus mailbox)
      => CopyMoveInternal(command, sequenceOrUidSet, ValidateMailboxRelationship(mailbox).Name, tryCreate: false, out _, out _);

    /// <summary>sends COPY/MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult CopyMove(CopyMoveCommand command, IImapNumberSet sequenceOrUidSet, ImapMailboxStatus mailbox, out ImapCopiedUidSet copiedUids)
      => CopyMoveInternal(command, sequenceOrUidSet, ValidateMailboxRelationship(mailbox).Name, tryCreate: false, out copiedUids, out _);

    /// <summary>sends COPY/MOVE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult CopyMove(CopyMoveCommand command, IImapNumberSet sequenceOrUidSet, string mailboxName)
      => CopyMoveInternal(command, sequenceOrUidSet, mailboxName, tryCreate: false, out _, out _);

    /// <summary>sends COPY/MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult CopyMove(CopyMoveCommand command, IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids)
      => CopyMoveInternal(command, sequenceOrUidSet, mailboxName, tryCreate: false, out copiedUids, out _);

    /// <summary>sends COPY/MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// </remarks>
    public ImapCommandResult CopyMoveTryCreateMailbox(CopyMoveCommand command, IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapMailboxStatus createdMailbox)
      => CopyMoveInternal(command, sequenceOrUidSet, mailboxName, tryCreate: true, out _, out createdMailbox);

    /// <summary>sends COPY/MOVE command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapCommandResult CopyMoveTryCreateMailbox(CopyMoveCommand command, IImapNumberSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids, out ImapMailboxStatus createdMailbox)
      => CopyMoveInternal(command, sequenceOrUidSet, mailboxName, tryCreate: true, out copiedUids, out createdMailbox);
    #endregion

    private ImapCommandResult CopyMoveInternal(CopyMoveCommand command, IImapNumberSet sequenceOrUidSet, string mailboxName, bool tryCreate, out ImapCopiedUidSet copiedUids, out ImapMailboxStatus createdMailbox)
    {
      RejectNonSelectedState();

      ImapNumberSet.ThrowIfEmpty(sequenceOrUidSet, nameof(sequenceOrUidSet));

      var argMailboxName = ImapMailboxNameString.CreateMailboxName(mailboxName);

      copiedUids = null;
      createdMailbox = null;

      var commandTrueForCopy = default(bool);

      switch (command) {
        case CopyMoveCommand.Copy: commandTrueForCopy = true; break;
        case CopyMoveCommand.Move: commandTrueForCopy = false; break;
        default: throw ExceptionUtils.CreateArgumentMustBeValidEnumValue(nameof(command), command);
      }

      ImapCommandResult failedResult = null;

      for (var i = 0; i < 2; i++) {
        var respTryCreate = false;

        using (var t = new CopyMoveTransaction(connection, this, commandTrueForCopy, sequenceOrUidSet.IsUidSet)) {
          // sequence set
          t.RequestArguments["sequence set"] = sequenceOrUidSet.Expression;

          // mailbox name
          t.RequestArguments["mailbox name"] = argMailboxName;

          if (ProcessTransaction(t).IsSuccess) {
            copiedUids = t.TransactionResult;
            return t.Result;
          }
          else {
            if (ProcessMailboxRefferalResponse(t.Result.TaggedStatusResponse) || !tryCreate)
              return t.Result;
          }

          failedResult = t.Result;

          // 6.4.7. COPY Command
          //       If the destination mailbox does not exist, a server SHOULD return
          //       an error.  It SHOULD NOT automatically create the mailbox.  Unless
          //       it is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the COPY if
          //       the CREATE is successful.

          // RFC 6851 - Internet Message Access Protocol (IMAP) - MOVE Extension
          // http://tools.ietf.org/html/rfc6851
          // 3.3. Semantics of MOVE and UID MOVE
          //    Because of the similarity of MOVE to COPY, extensions that affect
          //    COPY affect MOVE in the same way.  Response codes such as TRYCREATE
          //    (see [RFC3501], Section 6.4.7), as well as those defined by
          //    extensions, are sent as appropriate.  See Section 4 for more
          //    information about how MOVE interacts with other IMAP extensions.
          respTryCreate = t.Result.TaggedStatusResponse.ResponseText.Code == ImapResponseCode.TryCreate;
        }

        // try create
        if (i == 0 && respTryCreate)
          if (!Create(mailboxName, out createdMailbox).IsSuccess)
            return failedResult;
      }

      return failedResult;
    }

    /// <summary>sends UNSELECT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapCommandResult Unselect()
    {
      RejectNonSelectedState();

      using (var t = new UnselectTransaction(connection, this)) {
        var r = ProcessTransaction(t);

        if (r.IsSuccess) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return r;
      }
    }

    private void ProcessUpdatedSizeAndStatusResponse(ImapCommandResult result)
    {
      if (selectedMailbox == null)
        return;

      // 7.2. Server Responses - Server and Mailbox Status
      // 7.2.6. FLAGS Response
      // 7.3. Server Responses - Mailbox Size
      // 7.3.1. EXISTS Response
      // 7.3.2. RECENT Response
      // 7.4. Server Responses - Message Status
      // 7.4.1. EXPUNGE Response
      // 7.4.2. FETCH Response
      foreach (var data in result.ReceivedResponses.OfType<ImapServerDataResponse>()) {
        if (data.Type == ImapServerDataResponseType.Flags) {
          //       The update from the FLAGS response MUST be recorded by the client.
          selectedMailbox.ApplicableFlags = ImapServerDataResponseParser.FromFlags(data);
        }
        else if (data.Type == ImapServerDataResponseType.Exists) {
          //       The update from the EXISTS response MUST be recorded by the
          //       client.
          selectedMailbox.ExistsMessage = ImapServerDataResponseParser.FromExists(data);
        }
        else if (data.Type == ImapServerDataResponseType.Recent) {
          //       The update from the RECENT response MUST be recorded by the
          //       client.
          selectedMailbox.RecentMessage = ImapServerDataResponseParser.FromRecent(data);
        }
        else if (data.Type == ImapServerDataResponseType.Expunge) {
          //       The update from the EXPUNGE response MUST be recorded by the
          //       client.
          if (0L < selectedMailbox.ExistsMessage)
            selectedMailbox.ExistsMessage -= 1L;
        }
      }
    }
  }
}
