// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Runtime.ExceptionServices;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;

using Trace = Smdn.Net.Imap4.Client.Trace;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    /*
     * methods for processing transaction
     */

    /// <remarks>processes transaction with no timeout.</remarks>
    private Task<ImapCommandResult> ProcessTransactionAsync(IImapTransaction transaction)
    {
      PreProcessTransaction(transaction);

      return ProcessAsync(transaction);

      async Task<ImapCommandResult> ProcessAsync(IImapTransaction t)
      {
        return PostProcessTransaction(await Task.Run(() => ProcessTransactionInternal(t)));
      }
    }

    private ImapCommandResult ProcessTransaction(IImapTransaction transaction)
    {
      var task = ProcessTransactionAsync(transaction);

      try {
        if (!task.Wait(transactionTimeout/* do not refer property TransactionTimeout to avoid throwing ObjectDisposedException */)) {
          CloseConnection();
          throw new TimeoutException(string.Format("transaction timeout ({0})", transaction.GetType().FullName));
        }
      }
      catch (AggregateException ex) {
        ExceptionDispatchInfo.Capture(ex.InnerException).Throw();
      }

      return task.Result;
    }

    private IImapTransaction ProcessTransactionInternal(IImapTransaction t)
    {
      lock (transactionLockObject) {
        Trace.LogRequest(t);

        TraceInfo(t is IdleTransaction, "idling");

        try {
          t.Process();
        }
        catch {
          CloseConnection();
          throw;
        }

        TraceInfo(t is IdleTransaction, "done");

        Trace.LogResponse(t);

        return t;
      }
    }

    private void PreProcessTransaction(IImapTransaction t)
    {
      // TODO: '5.5. Multiple Commands in Progress'
      //   The following are examples of valid non-waiting command sequences:
      //      FETCH + STORE + SEARCH + CHECK
      //      STORE + COPY + EXPUNGE
      RejectTransactionProceeding();

      // check and set literal options
      var isNonSyncLiteralCapable = serverCapabilities.Contains(ImapCapability.LiteralNonSync);
      var isLiteral8Capable = serverCapabilities.Contains(ImapCapability.Binary);
      var containsNonSyncLiteral = false;
      var containsLiteral8 = false;

      ImapString.Traverse(t.RequestArguments.Values, s => {
        if (s is ImapLiteralString literal) {
          var syncMode = literal.Options & ImapLiteralOptions.SynchronizationMode;
          var literalMode = literal.Options & ImapLiteralOptions.LiteralMode;

          switch (syncMode) {
            case ImapLiteralOptions.NonSynchronizingIfCapable:
              syncMode = isNonSyncLiteralCapable ? ImapLiteralOptions.NonSynchronizing : ImapLiteralOptions.Synchronizing;
              break;
            case ImapLiteralOptions.NonSynchronizing:
              containsNonSyncLiteral = true;
              break;
          }

          switch (literalMode) {
            case ImapLiteralOptions.Literal8IfCapable:
              literalMode = isLiteral8Capable ? ImapLiteralOptions.Literal8 : ImapLiteralOptions.Literal;
              break;
            case ImapLiteralOptions.Literal8:
              containsLiteral8 = true;
              break;
          }

          literal.Options = (syncMode | literalMode)
                            | (literal.Options & ~(ImapLiteralOptions.SynchronizationMode | ImapLiteralOptions.LiteralMode));
        }

        if (handlesIncapableAsException && s is IImapExtension argExtension)
          IsCapable(argExtension, throwIfIncapable: true);
      });

      if (!handlesIncapableAsException)
        return;

      // check capability
      if (!serverCapabilities.Contains(ImapCapability.Imap4Rev1) && !(t is CapabilityTransaction))
        throw new ImapIncapableException(ImapCapability.Imap4Rev1);

      if (containsNonSyncLiteral && !isNonSyncLiteralCapable)
        throw new ImapIncapableException(ImapCapability.LiteralNonSync);
      if (containsLiteral8 && !isLiteral8Capable)
        throw new ImapIncapableException(ImapCapability.Binary);

      if (t is IImapExtension trxExtension)
        IsCapable(trxExtension, throwIfIncapable: true);
    }

    private ImapCommandResult PostProcessTransaction(IImapTransaction t)
    {
      if (updateSelectedMailboxSizeAndStatus)
        ProcessUpdatedSizeAndStatusResponse(t.Result);

      if (t.Result.IsFinalResponseCondition(ImapResponseCondition.Bye)) {
        TransitStateTo(ImapSessionState.NotAuthenticated);
        CloseConnection();
      }

      return t.Result;
    }
  }
}
