// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.MessageAccessProtocols.Strings;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.Strings {
  // enum types:
  // * ImapStringEnum : ImapString
  //     => handles string constants
  //     ImapAuthenticationMechanism
  //       => handles authentication mechanisms
  //     ImapCapability
  //       => handles capability constants
  //     ImapCollationAlgorithm
  //       => handles COMPARATOR extension comparator order
  //     ImapCompressionMechanism
  //       => handles COMPRESS extension compression mechanism
  //     ImapMailboxFlag
  //       => handles mailbox flags
  //     ImapMessageFlag
  //       => handles system flags and keywords
  //     ImapThreadingAlgorithm
  //       => handles THREAD extension threading algorithm
  //     Protocol.ImapResponseCode
  //       => handles response codes
  //     Protocol.ImapDataResponseType
  //       => handles server response types
#if SERIALIZATION
  [Serializable]
#endif
  public abstract class ImapEnumValue : ImapString, IImapStringValue, IEquatable<string> {
    public string Value { get; }
    byte[] IImapStringValue.GetBytes() => OctetEncoding.SevenBits.GetBytes(Value); // TODO: optimize

    private protected ImapEnumValue(string value)
    {
      if (value == null)
        throw new ArgumentNullException(nameof(value));
      if (value.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(value));

      this.Value = value;
    }

#region "defined constants"
    private protected static ImapEnumValueSet<TEnumValue> CreateDefinedConstantsSet<TEnumValue>() where TEnumValue : ImapEnumValue
    {
      return new ImapEnumValueSet<TEnumValue>(TypedStringUtils.GetDefinedConstants<TEnumValue>(), asReadOnly: true);
    }
#endregion

#region "equatable"
    private protected static bool Equals<TEnumValue>(TEnumValue x, TEnumValue y) where TEnumValue : ImapEnumValue
      => ImapEnumValueEqualityComparer<TEnumValue>.Default.Equals(x, y);

    public bool Equals(string other) => ImapEnumValueEqualityComparer<ImapEnumValue>.Default.Equals(this, other);
    public override int GetHashCode() => ImapEnumValueEqualityComparer<ImapEnumValue>.Default.GetHashCode(this);

    public override bool Equals(object obj)
    {
      if (obj == null)
        return false;
      if (obj is string str)
        return Equals(str);
      if (obj is ImapEnumValue enumValue)
        return GetType() == enumValue.GetType() && ImapEnumValueEqualityComparer<ImapEnumValue>.Default.Equals(this, enumValue);

      return false;
    }
#endregion

    public override string ToString() => Value;
  }
}
