// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  /*
   * RFC 5255 - Internet Message Access Protocol Internationalization
   * http://tools.ietf.org/html/rfc5255
   */
  internal sealed class ComparatorTransaction :
    ImapTransactionBase<Tuple<ImapCollationAlgorithm, IReadOnlyList<ImapCollationAlgorithm>>>,
    IImapExtension
  {
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get { yield return ImapCapability.I18NLevel2; }
    }

    public ComparatorTransaction(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator)
      : base(connection, commandTagGenerator)
    {
    }

    /*
     * 4.7. COMPARATOR Command
     *    Arguments: Optional comparator order arguments.
     *    Response:  A possible COMPARATOR response (see Section 4.8).
     *    Result:    OK - Command completed
     *               NO - No matching comparator found
     *               BAD - Arguments invalid
     */
    protected override string GetCommand() => "COMPARATOR";

    protected override ImapString[] GetCommandArguments()
    {
      if (RequestArguments.TryGetValue("comparator order arguments", out var comparatorOrder))
        return new[] { comparatorOrder };
      else
        return null;
    }

    protected override void OnDataResponseReceived(ImapServerDataResponse data)
    {
      if (data.Type == ImapServerDataResponseType.Comparator)
        activeComparator = ImapServerDataResponseParser.FromComparator(data, out matchingComparators);

      base.OnDataResponseReceived(data);
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok)
        Finish(tagged, Tuple.Create(activeComparator, matchingComparators));
      else
        base.OnTaggedStatusResponseReceived(tagged);
    }

    private ImapCollationAlgorithm activeComparator;
    private IReadOnlyList<ImapCollationAlgorithm> matchingComparators;
  }
}
