// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Text;

using Trace = Smdn.Net.Imap4.Client.Trace;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  // Layers:
  //   Session.ImapSession
  // * Transaction.IImapTransaction
  //   Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  internal abstract class ImapTransactionBase<TTransactionResult> : ImapTransactionBase {
    public TTransactionResult TransactionResult { get; private set; } = default;

    protected ImapTransactionBase(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator)
      : base(connection, commandTagGenerator)
    {
    }

    protected void Finish(ImapTaggedStatusResponse finalStatusResponse, TTransactionResult transactionResult)
    {
      TransactionResult = transactionResult;

      Finish(finalStatusResponse);
    }
  }

  internal abstract class ImapTransactionBase : IImapTransaction, IDisposable
  {
    public ImapConnection Connection => connection;
    public IDictionary<string, ImapString> RequestArguments => requestArguments;
    public ImapCommandResult Result => result;

    protected bool IsFinished => (result != null) || (connection == null /* in case of being disposed or exception */ );

    private const int initialRequestArgumentCapacity = 4;
    private const int initialResponseCapacity = 4;

    protected ImapTransactionBase(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator)
    {
      this.connection = connection ?? throw new ArgumentNullException(nameof(connection));
      this.commandTagGenerator = commandTagGenerator ?? throw new ArgumentNullException(nameof(connection));
      this.requestArguments = new Dictionary<string, ImapString>(initialRequestArgumentCapacity, StringComparer.Ordinal);
      this.result = null;
      this.receivedResponses = new List<ImapResponse>(capacity: initialResponseCapacity);
    }

    public virtual void Dispose()
    {
      // nothing to do in this class
      connection = null;
    }

    private ImapConnection connection;
    private readonly IImapCommandTagGenerator commandTagGenerator;
    private Dictionary<string, ImapString> requestArguments;
    private ImapCommandResult result;

#region "processing transaction / state transition"
    protected abstract string GetCommand();
    protected virtual ImapString[] GetCommandArguments() => null;

    protected void SendContinuation(ImapString argument)
    {
#if DEBUG
      RejectFinished();
#endif

      commandContinuationContext = connection.SendCommand(ImapCommand.CreateContinuation(argument));
    }

    void IImapTransaction.Process()
    {
#if DEBUG
      if (connection == null)
        throw new ObjectDisposedException(GetType().FullName);

      RejectFinished();
#endif

      // construct and send command
      string commandTag = null;

      try {
        var commandString = GetCommand();
        var arguments = GetCommandArguments();

        if (commandString != null) {
          commandTag = commandTagGenerator.GetNext();

          var command = ImapCommand.Create(commandString,
                                           commandTag,
                                           arguments);

          commandContinuationContext = connection.SendCommand(command);
        }
      }
      catch (Exception ex) {
        connection = null;
        Trace.Log(ex);
        throw;
      }

      // receive response or send continuation
      try {
        for (;;) {
          ReceiveResponse(commandTag);

          if (IsFinished)
            return;
        }
      }
      catch (Exception ex) {
        connection = null;
        Trace.Log(ex);
        throw;
      }
    }

    private void ReceiveResponse(string commandTag)
    {
      ImapResponse response = null;

      try {
        for (;;) {
          response = connection.TryReceiveResponse();

          if (response != null) {
            receivedResponses.Add(response);
            break;
          }
        }
      }
      catch (ImapMalformedResponseException) {
        // protocol error, abort transaction
        throw;
      }

      try {
        switch (response) {
          case ImapTaggedStatusResponse tagged:
            OnStatusResponseReceived(tagged);

            if (string.Equals(tagged.Tag, commandTag, StringComparison.Ordinal))
              OnTaggedStatusResponseReceived(tagged);
            else
              // TODO: '5.5. Multiple Commands in Progress' or not matched tag
              throw new ImapUnexpectedResponseException($"command tag not matched (expected '{commandTag}' but was '{tagged.Tag}')", tagged);
            break;

          case ImapUntaggedStatusResponse untagged:
            OnStatusResponseReceived(untagged);
            OnUntaggedStatusResponseReceived(untagged);
            break;

          case ImapServerDataResponse data:
            OnDataResponseReceived(data);
            break;

          case ImapCommandContinuationRequest cont:
            OnCommandContinuationRequestReceived(cont);
            break;
        }
      }
      catch (ImapMalformedDataException ex) {
        if (response is ImapTaggedStatusResponse) {
          // error in response data, abort transaction
          throw;
        }
        else {
          // error in response data, continue processing
          Trace.Log(ex);
        }
      }
    }

    protected void Finish(ImapStatusResponse finalStatusResponse)
    {
#if DEBUG
      RejectFinished();
#endif

      result = new ImapCommandResult(finalStatusResponse as ImapTaggedStatusResponse,
                                     finalStatusResponse,
                                     receivedResponses);
    }

#if DEBUG
    protected Exception CreateRequestArgumentsException(string message)
    {
      return new InvalidOperationException(message);
    }

    private void RejectFinished()
    {
      if (IsFinished)
        throw new InvalidOperationException("transaction already finished");
    }
#endif

    private ImapConnection.CommandContinuationContext commandContinuationContext = null;
    private List<ImapResponse> receivedResponses;
#endregion

#region "response processing"
    protected virtual void OnDataResponseReceived(ImapServerDataResponse data)
    {
    }

    protected virtual void OnCommandContinuationRequestReceived(ImapCommandContinuationRequest continuationRequest)
    {
      if (commandContinuationContext == null)
        throw new ImapUnexpectedResponseException("unexpected command continuation or sending literal failed", continuationRequest);

      commandContinuationContext = connection.SendCommand(commandContinuationContext);
    }

    protected virtual void OnStatusResponseReceived(ImapStatusResponse status)
    {
      if (status.ResponseText.Code == ImapResponseCode.Alert)
        // ALERT
        Trace.Info(status.ResponseText.Text);
      else if (status.ResponseText.Code == ImapResponseCode.ClientBug)
        // CLIENTBUG
        Trace.Info(status.ResponseText.Text);
    }

    protected virtual void OnUntaggedStatusResponseReceived(ImapUntaggedStatusResponse untagged)
    {
      if (untagged.Condition == ImapResponseCondition.Bye)
        // The BYE response is always untagged
        Finish(untagged);
    }

    protected virtual void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      switch (tagged.Condition) {
        case ImapResponseCondition.Ok:
        case ImapResponseCondition.PreAuth:
        case ImapResponseCondition.No:
        case ImapResponseCondition.Bad:
          Finish(tagged);
          break;
        default: // BYE or undefined
          throw new InvalidOperationException($"tagged status response with invalid condition received (condition: {tagged.Condition})");
      }
    }
#endregion

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Connection), Connection),
        (nameof(RequestArguments), RequestArguments),
        (nameof(Result), Result),
      });
    }
  }
}
