// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal sealed class ListExtendedTransaction : ListTransactionBase, IImapExtension {
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get { yield return ImapCapability.ListExtended; }
    }

    public ListExtendedTransaction(ImapConnection connection,
                                   IImapCommandTagGenerator commandTagGenerator,
                                   ImapMailboxManager mailboxManager)
      : base(connection, commandTagGenerator)
    {
      this.mailboxManager = mailboxManager;
    }

    /*
     * http://tools.ietf.org/html/rfc5258
     * RFC 5258 - Internet Message Access Protocol version 4 - LIST Command Extensions
     * 3. Extended LIST Command
     *    The LIST command syntax is also extended in two additional ways: by
     *    adding a parenthesized list of command options between the command
     *    name and the reference name (LIST selection options) and an optional
     *    list of options at the end that control what kind of information
     *    should be returned (LIST return options).
     */
    protected override ImapServerDataResponseType ExpectedDataResponseType => ImapServerDataResponseType.List;

    protected override string GetCommand() => "LIST";

    protected override ImapString[] GetCommandArguments()
    {
#if DEBUG
      if (!RequestArguments.ContainsKey("reference name") ||
          !RequestArguments.ContainsKey("mailbox name"))
        throw CreateRequestArgumentsException("arguments 'reference name' and 'mailbox name' must be setted");
#endif

      var argsCount = 2;

      if (RequestArguments.TryGetValue("selection options", out var selectionOptions))
        argsCount += 1;
      if (RequestArguments.TryGetValue("return options", out var returnOptions))
        argsCount += 2;

      var args = new ImapString[argsCount];
      var index = 0;

      if (selectionOptions != null)
        args[index++] = selectionOptions;

      args[index++] = RequestArguments["reference name"];
      args[index++] = RequestArguments["mailbox name"];

      if (returnOptions != null) {
        args[index++] = "RETURN";
        args[index++] = RequestArguments["return options"];
      }

      return args;
    }

    protected override void OnDataResponseReceived(ImapServerDataResponse data)
    {
      /*
       * IMAP4 Extension for Returning STATUS Information in Extended LIST
       * http://tools.ietf.org/html/rfc5819
       */
      if (data.Type == ImapServerDataResponseType.Status)
        ImapServerDataResponseParser.FromStatus(data, mailboxManager);

      base.OnDataResponseReceived(data);
    }

    private readonly ImapMailboxManager mailboxManager;
  }
}
