// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal abstract class ListTransactionBase : ImapTransactionBase<IReadOnlyList<ImapListedMailbox>> {
    protected ListTransactionBase(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator)
      : base(connection, commandTagGenerator)
    {
    }

    /*
     * 6.3.8. LIST Command
     *    Arguments:  reference name
     *                mailbox name with possible wildcards
     *    Responses:  untagged responses: LIST
     *    Result:     OK - list completed
     *                NO - list failure: can't list that reference or name
     *                BAD - command unknown or arguments invalid
     *
     * 6.3.9. LSUB Command
     *    Arguments:  reference name
     *                mailbox name with possible wildcards
     *    Responses:  untagged responses: LSUB
     *    Result:     OK - lsub completed
     *                NO - lsub failure: can't list that reference or name
     *                BAD - command unknown or arguments invalid
     */

    /*
     * http://tools.ietf.org/html/rfc2193
     * RFC 2193 IMAP4 Mailbox Referrals
     * 
     * 5.1 RLIST command
     *   Arguments:  reference name
     *               mailbox name with possible wildcards
     *   Responses:  untagged responses: LIST
     *   Result:     OK - RLIST Completed
     *               NO - RLIST Failure
     *               BAD - command unknown or arguments invalid
     * 
     *   The RLIST command behaves identically to its LIST counterpart, except
     *   remote mailboxes are returned in addition to local mailboxes in the
     *   LIST responses.
     * 
     * 5.2 RLSUB Command
     *    Arguments:  reference name
     *                mailbox name with possible wildcards
     *    Responses:  untagged responses: LSUB
     *    Result:     OK - RLSUB Completed
     *                NO - RLSUB Failure
     *                BAD - command unknown or arguments invalid
     * 
     *    The RLSUB command behaves identically to its LSUB counterpart, except
     *    remote mailboxes are returned in addition to local mailboxes in the
     *    LSUB responses.
     */
    protected override ImapString[] GetCommandArguments()
    {
#if DEBUG
      if (!RequestArguments.ContainsKey("reference name") ||
          !RequestArguments.ContainsKey("mailbox name"))
        throw CreateRequestArgumentsException("arguments 'reference name' and 'mailbox name' must be setted");
#endif

      return new[] {
        RequestArguments["reference name"],
        RequestArguments["mailbox name"],
      };
    }

    protected abstract ImapServerDataResponseType ExpectedDataResponseType { get; }

    protected override void OnDataResponseReceived(ImapServerDataResponse data)
    {
      if (data.Type == ExpectedDataResponseType)
        mailboxLists.Add(ImapServerDataResponseParser.FromListLsub(data));

      base.OnDataResponseReceived(data);
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok)
        Finish(tagged, mailboxLists);
      else
        base.OnTaggedStatusResponseReceived(tagged);
    }

    private /*readonly*/ List<ImapListedMailbox> mailboxLists = new List<ImapListedMailbox>();
  }
}
