// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal sealed class SearchTransaction : SearchTransactionBase {
    public SearchTransaction(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator, bool uid, ImapCapability resultSpecifierCapabilityRequirement)
      : base(connection, commandTagGenerator, uid, resultSpecifierCapabilityRequirement)
    {
    }

    // RFC 4466 - Collected Extensions to IMAP4 ABNF
    // http://tools.ietf.org/html/rfc4466
    // 2.6. Extensions to SEARCH Command
    // 2.6.1. Extended SEARCH Command
    //    Arguments:  OPTIONAL result specifier
    //                OPTIONAL [CHARSET] specification
    //                searching criteria (one or more)
    //    Responses:  REQUIRED untagged response: SEARCH (*)
    //    Result:     OK - search completed
    //                NO - search error: cannot search that [CHARSET] or
    //                     criteria
    //                BAD - command unknown or arguments invalid
    protected override string GetCommand() => Uid ? "UID SEARCH" : "SEARCH";

    protected override ImapString[] GetCommandArguments()
    {
#if DEBUG
      if (!RequestArguments.ContainsKey("searching criteria"))
        throw CreateRequestArgumentsException("arguments 'searching criteria' must be setted");
#endif

      var argsCount = 1;

      if (RequestArguments.TryGetValue("result specifier", out var resultSpecifier))
        argsCount += 2;
      if (RequestArguments.TryGetValue("charset specification", out var charsetSpecification))
        argsCount += 2;

      var args = new ImapString[argsCount];
      var index = 0;

      if (resultSpecifier != null) {
        args[index++] = "RETURN";
        args[index++] = resultSpecifier;
      }

      if (charsetSpecification != null) {
        args[index++] = "CHARSET";
        args[index++] = charsetSpecification;
      }

      args[index++] = RequestArguments["searching criteria"];

      return args;
    }
  }
}
