// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal sealed class SortTransaction : SearchTransactionBase {
    protected override ImapCapability RequiredCapability {
      get { return base.RequiredCapability ?? ImapCapability.Sort; }
    }

    public SortTransaction(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator, bool uid, ImapCapability resultSpecifierCapabilityRequirement)
      : base(connection, commandTagGenerator, uid, resultSpecifierCapabilityRequirement)
    {
    }

    // http://tools.ietf.org/html/rfc5256
    // BASE.6.4.SORT. SORT Command
    //    Arguments:  sort program
    //                charset specification
    //                searching criteria (one or more)
    //    Data:       untagged responses: SORT
    //    Result:     OK - sort completed
    //                NO - sort error: can't sort that charset or
    //                     criteria
    //                BAD - command unknown or arguments invalid
    protected override string GetCommand() => Uid ? "UID SORT" : "SORT";

    protected override ImapString[] GetCommandArguments()
    {
#if DEBUG
      if (!RequestArguments.ContainsKey("searching criteria"))
        throw CreateRequestArgumentsException("arguments 'searching criteria' must be setted");

      // http://tools.ietf.org/html/rfc5256
      // BASE.6.4.SORT. SORT Command
      //   The charset argument is mandatory (unlike SEARCH)
      if (!RequestArguments.ContainsKey("charset specification") ||
          !RequestArguments.ContainsKey("sort criteria"))
        throw CreateRequestArgumentsException("arguments 'sort criteria' and 'charset specification' must be setted");
#endif

      if (RequestArguments.TryGetValue("result specifier", out var resultSpecifier))
        return new[] {
          "RETURN",
          resultSpecifier,
          RequestArguments["sort criteria"],
          RequestArguments["charset specification"],
          RequestArguments["searching criteria"],
        };
      else
        return new[] {
          RequestArguments["sort criteria"],
          RequestArguments["charset specification"],
          RequestArguments["searching criteria"],
        };
    }
  }
}
