// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  // Layers:
  //   Session.ImapSession
  //   Transaction.IImapTransaction
  // * Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  // handles '6. Client Commands'
  public sealed class ImapCommand {
    public static ImapCommand Create(string command, string tag, params ImapString[] arguments)
    {
      if (command == null)
        throw new ArgumentNullException(nameof(command));
      if (command.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyArray(nameof(command));
      if (tag == null)
        throw new ArgumentNullException(nameof(tag));
      if (tag.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyArray(nameof(tag));
      if (arguments != null && arguments.Any(a => a == null))
        throw ExceptionUtils.CreateAllItemsOfArgumentMustBeNonNull(nameof(arguments));

      return new ImapCommand(command, tag, arguments);
    }

    public static ImapCommand CreateContinuation(ImapString argument)
    {
      if (argument == null)
        throw new ArgumentNullException(nameof(argument));

      return new ImapCommand(null, null, new[] { argument });
    }

    /*
     * instance members
     */
    public string Tag { get; }
    public string Command { get; }
    public IReadOnlyList<ImapString> Arguments => arguments;

    private readonly ImapString[] arguments;

    private ImapCommand(string command, string tag, ImapString[] arguments)
    {
      Command = command;
      Tag = tag;
      this.arguments = arguments ??
#if NET45
        ArrayExtensions.Empty<ImapString>();
#else
        Array.Empty<ImapString>();
#endif
    }

    internal void ToArray(ref ImapString[] array, out int count)
    {
      count = (Command == null ? 0 : 2) + arguments.Length;

      if (array.Length < count)
        array = new ImapString[count]; // resize array

      if (Command == null) {
        Array.Copy(arguments, 0, array, 0, arguments.Length);
      }
      else {
        array[0] = Tag;
        array[1] = Command;

        Array.Copy(arguments, 0, array, 2, arguments.Length);
      }
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Command), Command),
        (nameof(Tag), Tag),
        (nameof(Arguments), Arguments),
      });
    }
  }
}
