// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.Transmission {
#if SERIALIZATION
  [Serializable]
#endif
  public class ImapCommandResult
#if SERIALIZATION
    : ISerializable
#endif
  {

#if SERIALIZATION
    [Serializable]
#endif
    private sealed class SuccessCommandResult : ImapCommandResult {
      public override bool IsConcludedByClientSide => true;
      public override bool IsSuccess => true;

      internal SuccessCommandResult(string description)
        : base(string.Concat("<", description, ">"))
      {
      }

#if SERIALIZATION
      internal SuccessCommandResult(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
      }
#endif
    }

#if SERIALIZATION
    [Serializable]
#endif
    private sealed class FailureCommandResult : ImapCommandResult {
      public override bool IsConcludedByClientSide => true;
      public override bool IsSuccess => false;

      internal FailureCommandResult(string description)
        : base(string.Concat("<", description, ">"))
      {
      }

#if SERIALIZATION
      internal FailureCommandResult(SerializationInfo info, StreamingContext context)
        : base(info, context)
      {
      }
#endif
    }

    internal static ImapCommandResult CreateSuccess(string description) => new SuccessCommandResult(description);
    internal static ImapCommandResult CreateFailure(string description) => new FailureCommandResult(description);

    /*
     * instance members
     */
    public virtual bool IsConcludedByClientSide => FinalStatusResponse == null;

    public virtual bool IsSuccess {
      get {
        if (FinalStatusResponse == null)
          return false;

        switch (FinalStatusResponse.Condition) {
          case ImapResponseCondition.Ok:
          case ImapResponseCondition.PreAuth:
            return true;

          default:
            return false;
        }
      }
    }

    public string Description { get; }
    public ImapTaggedStatusResponse TaggedStatusResponse { get; }
    public ImapStatusResponse FinalStatusResponse { get; }
    public IReadOnlyList<ImapResponse> ReceivedResponses { get; }

    private protected ImapCommandResult(string description)
    {
      Description = description;
      ReceivedResponses =
#if NET45
        ArrayExtensions.Empty<ImapResponse>();
#else
        Array.Empty<ImapResponse>();
#endif
    }

    internal ImapCommandResult(ImapTaggedStatusResponse taggedStatusResponse,
                               ImapStatusResponse finalStatusResponse,
                               IReadOnlyList<ImapResponse> receivedResponses)
    {
      Description = string.Concat("\"", finalStatusResponse.ResponseText.Text, "\"");
      TaggedStatusResponse = taggedStatusResponse;
      FinalStatusResponse = finalStatusResponse;
      ReceivedResponses = receivedResponses;
    }

#if SERIALIZATION
    internal protected ImapCommandResult(SerializationInfo info, StreamingContext context)
    {
      Description = info.GetString(nameof(Description));
      TaggedStatusResponse = (ImapTaggedStatusResponse)info.GetValue(nameof(TaggedStatusResponse), typeof(ImapTaggedStatusResponse));
      FinalStatusResponse = (ImapStatusResponse)info.GetValue(nameof(FinalStatusResponse), typeof(ImapStatusResponse));
      ReceivedResponses = (ImapResponse[])info.GetValue(nameof(ReceivedResponses), typeof(ImapResponse[]));
    }

    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue(nameof(Description), Description);
      info.AddValue(nameof(TaggedStatusResponse), TaggedStatusResponse);
      info.AddValue(nameof(FinalStatusResponse), FinalStatusResponse);
      info.AddValue(nameof(ReceivedResponses), ReceivedResponses.ToArray());
    }
#endif

    public bool IsFinalResponseCondition(ImapResponseCondition condition)
    {
      if (FinalStatusResponse == null)
        return false;

      return FinalStatusResponse.Condition == condition;
    }

    public ImapStatusResponse GetStatusResponseOfCode(ImapResponseCode code)
    {
      // XXX: resp-text of continue-req is ignored
      return ReceivedResponses.OfType<ImapStatusResponse>().FirstOrDefault(status => status.ResponseText.Code == code);
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(IsSuccess), IsSuccess),
        (nameof(Description), Description),
        (nameof(TaggedStatusResponse), TaggedStatusResponse),
        (nameof(FinalStatusResponse), FinalStatusResponse),
      });
    }
  }
}
