// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.IO.Streams;
using Smdn.IO.Streams.LineOriented;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  public abstract class ImapReceiver {
    private readonly LineOrientedBufferedStream stream;

    protected ImapReceiver(LineOrientedBufferedStream stream)
    {
      this.stream = stream ?? throw new ArgumentNullException(nameof(stream));
    }

    protected ByteString Receive()
    {
      var line = stream.ReadLine(true);

      if (line == null)
        throw new ImapConnectionException("stream closed");

      return ByteString.CreateImmutable(line);
    }

    protected sealed class ParsingContext {
      private const int initialCapacity = 32;
      private List<ImapData>[] nestedData = new List<ImapData>[1];

      internal ParsingContext()
      {
        nestedData[0] = new List<ImapData>(initialCapacity);
      }

      internal void Initialize()
      {
        for (var n = 0; n < nestedData.Length; n++) {
          nestedData[n].Clear();
        }

        nest = 0;
      }

      internal void AddData(ImapData data)
      {
        nestedData[nest].Add(data);
      }

      internal IReadOnlyList<ImapData> GetParsed()
      {
        return nestedData[0].ToArray();
      }

      /*
       * parenthesized strings
       */
      private int nest = 0;
      internal bool IsParenthesesClosed => nest == 0;

      internal void OpenParenthesis()
      {
        nest++;

        if (nestedData.Length <= nest) {
          Array.Resize(ref nestedData, nest + 1);

          nestedData[nest] = new List<ImapData>(initialCapacity);
        }
      }

      internal void CloseParenthesis()
      {
        if (nest == 0)
          throw new ImapMalformedDataException("unopened parenthesis");

        nestedData[nest - 1].Add(ImapData.CreateListData(nestedData[nest].ToArray()));
        nestedData[nest].Clear();

        nest--;
      }
    }

    private readonly ParsingContext parsingContext = new ParsingContext();

    private static readonly ByteString nil = ByteString.CreateImmutable("NIL");

    protected IReadOnlyList<ImapData> ParseData(ByteString text, int offset)
    {
      ParsingContext discard = null;

      return ParseData(text, offset, text.Length, ref discard);
    }

    protected IReadOnlyList<ImapData> ParseData(ByteString line, int offset, int indexOfEOL, ref ParsingContext context)
    {
      var index = offset;

      if (context == null) {
        parsingContext.Initialize();

        context = parsingContext;
      }
      else {
        if (index == indexOfEOL) {
          if (context.IsParenthesesClosed) {
            // end of data
            var ret = context.GetParsed();
            context = null;
            return ret;
          }
          else {
            throw new ImapMalformedDataException("unclosed parenthesis");
          }
        }
        else if (line[index] == Ascii.Octets.SP) {
          index++;
        }
      }

      for (;;) {
        if (line[index] == Ascii.Octets.SP)
          // skip SP
          index = line.IndexOfNot(Ascii.Octets.SP, index);

        switch (line[index]) {
          case ImapOctets.DQuote: {
            // quoted
            context.AddData(ImapData.CreateTextData(ParseQuotedString(line, indexOfEOL, ref index)));
            break;
          }

          case ImapOctets.OpenParen: {
            // parenthesized begin
            context.OpenParenthesis();
            index++; // '('
            break;
          }

          case ImapOctets.CloseParen: {
            // parenthesized end
            context.CloseParenthesis();
            index++; // ')'
            break;
          }

          case ImapOctets.OpenBrace:
          case ImapOctets.Tilde: {
            // literal/literal8
            var isLiteral = ParseLiteral((line[index] == ImapOctets.Tilde), line, indexOfEOL, ref index, out var parsed);

            context.AddData(parsed);

            if (isLiteral)
              return null; // incomplete
            else
              break;
          }

          default: {
            // non-quoted
            var text = ParseNonQuotedString(line, indexOfEOL, ref index);

            if (nil.EqualsIgnoreCase(text))
              context.AddData(ImapData.Nil);
            else
              context.AddData(ImapData.CreateTextData(text));

            break;
          }
        }

        if (index == indexOfEOL) {
          if (context.IsParenthesesClosed) {
            // end of data
            var ret = context.GetParsed();
            context = null;
            return ret;
          }
          else {
            return null; // incomplete
          }
        }
        else {
          if (line[index] == Ascii.Octets.SP)
            index++;
        }
      } // for
    }

    private bool ParseLiteral(bool literal8, ByteString line, int indexOfEOL, ref int index, out ImapData parsed)
    {
      // literal         = "{" number "}" CRLF *CHAR8
      //                ; Number represents the number of CHAR8s
      // CHAR8           = %x01-ff ; any OCTET except NUL, %x00
      // number          = 1*DIGIT
      //                     ; Unsigned 32-bit integer
      //                     ; (0 <= n < 4,294,967,296)
      if (literal8) {
        /*
         * RFC 3516 - IMAP4 Binary Content Extension
         * http://tools.ietf.org/html/rfc3516
         * 4.2. FETCH Command Extensions
         *    literal8       =   "~{" number "}" CRLF *OCTET
         *                       ; <number> represents the number of OCTETs
         *                       ; in the response string.
         */
        if (line[index + 1] != ImapOctets.OpenBrace) {
          // not literal8
          parsed = ImapData.CreateTextData(ParseNonQuotedString(line, indexOfEOL, ref index));
          return false;
        }

        index++; // '~'
      }

      index++; // '{'

      var number = 0L;

      for (;;) {
        if (index == indexOfEOL)
          throw new ImapMalformedDataException("malformed literal (unexpected EOL)");

        var o = line[index];

        if (0x30 <= o && o <= 0x39) {
          // number
          try {
            number = checked((number * 10L) + (o - 0x30 /* '0' */));
          }
          catch (OverflowException) {
            throw new ImapMalformedDataException("too lager literal (overflow)");
          }
        }
        else if (o == ImapOctets.CloseBrace) {
          break; // '}'
        }
        else {
          throw new ImapMalformedDataException("malformed literal (invalid number)");
        }

        index++;
      }

      if (++index != indexOfEOL)
        throw new ImapMalformedDataException("malformed literal (extra data before EOL)");

      /*
       * to avoid allocating buffers on Large Object Heap
       * 
       * http://msdn.microsoft.com/en-us/magazine/cc534993.aspx
       * http://tirania.org/blog/archive/2009/Nov-09.html
       */
      const long literalAllocationThreshold = 40960L; // 40kBytes

      if (number <= literalAllocationThreshold) {
        var literal = new byte[number];

        if (stream.Read(literal, 0, (int)number) != literal.Length)
          throw new ImapMalformedDataException("unexpected end of literal");

        parsed = ImapData.CreateTextData(ByteString.CreateImmutable(literal));
      }
      else {
        var literal = new ChunkedMemoryStream();

        if (stream.Read(literal, number) != number)
          throw new ImapMalformedDataException("unexpected end of literal");

        literal.Position = 0L;

        parsed = ImapData.CreateTextData(literal);
      }

      return true;
    }

    private ByteString ParseQuotedString(ByteString line, int indexOfEOL, ref int index)
    {
      // quoted          = DQUOTE *QUOTED-CHAR DQUOTE
      // QUOTED-CHAR     = <any TEXT-CHAR except quoted-specials> / 
      //                   "\" quoted-specials
      // TEXT-CHAR       = <any CHAR except CR and LF>
      // CHAR            = <any 7-bit US-ASCII character except NUL,
      //                   0x01 - 0x7f>
      // quoted-specials = DQUOTE / "\"
      // date-time       = DQUOTE date-day-fixed "-" date-month "-" date-year

      index++; // DQUOTE

      var bytes = line.Segment.Array;
      var offset = line.Segment.Offset + index;
      var offsetOfEOL = line.Segment.Offset + indexOfEOL;
      var dequoted = new byte[bytes.Length - 2 /* DQUOTE * 2 */];
      var len = 0;

      for (;;) {
        if (bytes[offset] == ImapOctets.BackSlash) {
          offset++; // skip quoted-specials
        }
        else {
          // non escaped chars
          if (bytes[offset] == ImapOctets.DQuote) {
            index = (offset - line.Segment.Offset) + 1/*DQUOTE*/;
            return ByteString.CreateImmutable(dequoted, 0, len);
          }
          else if (offset == offsetOfEOL) {
            throw new ImapMalformedDataException("malformed quoted string (quotation not closed)");
          }
        }

        dequoted[len++] = bytes[offset++]; // QUOTED-CHAR
      }
    }

    private ByteString ParseNonQuotedString(ByteString line, int indexOfEOL, ref int index)
    {
      // atom            = 1*ATOM-CHAR
      // ATOM-CHAR       = <any CHAR except atom-specials>
      // atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
      //                   quoted-specials / resp-specials
      // number          = 1*DIGIT
      // list-wildcards  = "%" / "*"
      // quoted-specials = DQUOTE / "\"
      // resp-specials   = "]"

      // section         = "[" [section-spec] "]"
      // section-msgtext = "HEADER" / "HEADER.FIELDS" [".NOT"] SP header-list /
      //                   "TEXT"
      //                     ; top-level or MESSAGE/RFC822 part
      // section-part    = nz-number *("." nz-number)
      //                     ; body part nesting
      // section-spec    = section-msgtext / (section-part ["." section-text])
      // section-text    = section-msgtext / "MIME"
      //                     ; text other than actual body part (headers, etc.)
      var bytes = line.Segment.Array;
      var offset = line.Segment.Offset + index;
      var start = index;

      for (;;) {
        if (index == indexOfEOL ||
            bytes[offset] == Ascii.Octets.SP ||
            bytes[offset] == ImapOctets.CloseParen) {
          return line.Substring(start, index - start);
        }
        else if (bytes[offset] == ImapOctets.OpenBracket) {
          // skip section
          for (;;) {
            index++;
            offset++;

            if (index == indexOfEOL)
              throw new ImapMalformedDataException("unclosed section bracket");
            else if (bytes[offset] == ImapOctets.CloseBracket)
              break;
          }
        }

        index++;
        offset++;
      }
    }
  }
}
