// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Text;

using Smdn.IO.Streams.LineOriented;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  public sealed class ImapResponseReceiver : ImapReceiver {
    private ParsingContext dataParsingContext;
    public bool IsResponseContinuing => dataParsingContext != null;
    public bool ReceiveResponseAsUTF8 { get; set; } = false;

    public ImapResponseReceiver(LineOrientedBufferedStream stream)
      : base(stream)
    {
    }

    public ImapResponse ReceiveResponse()
    {
      var line = Receive();

      try {
        return (dataParsingContext == null)
          ? ParseResponce(line, ref dataParsingContext)
          : ParseDataResponse(line, 0, ref dataParsingContext);
      }
      catch (ImapMalformedDataException ex) {
        dataParsingContext = null; // discard context
        throw new ImapMalformedResponseException(ex.Message, ex);
      }
      catch {
        dataParsingContext = null; // discard context
        throw;
      }
    }

    private readonly static ByteString continueReqMark = ByteString.CreateImmutable("+ ");

    private ImapResponse ParseResponce(ByteString line, ref ParsingContext parsingContext)
    {
      // response        = *(continue-req / response-data) response-done

      // continue-req    = "+" SP (resp-text / base64) CRLF
      if (line.StartsWith(continueReqMark)) {
        if (line[2] == ImapOctets.OpenBracket)
          // resp-text
          return new ImapCommandContinuationRequest(ParseRespText(line.Substring(2))); // remove leading "+" SP
        else
          // text (resp-text without resp-text-code) / base64
          return new ImapCommandContinuationRequest(line.Substring(2, line.Length - 4)); // remove leading "+" SP and trailing CRLF
      }

      // greeting        = "*" SP (resp-cond-auth / resp-cond-bye) CRLF
      // response-done   = response-tagged / response-fatal
      // response-data   = "*" SP (resp-cond-state / resp-cond-bye /
      //                   mailbox-data / message-data / capability-data) CRLF
      // response-fatal  = "*" SP resp-cond-bye CRLF
      //                     ; Server closes connection immediately
      // response-tagged = tag SP resp-cond-state CRLF

      // ("*" / tag) SP
      var tagSep = line.IndexOf(Ascii.Octets.SP);

      if (tagSep <= 0)
        // response-done and response-data must be ("*" / tag) SP
        throw new ImapMalformedResponseException("malformed response-done/response-data", line.ToString());

      var isTaggedResponse = !(1 == tagSep && line[0] == ImapOctets.Asterisk); // not ('*' SP)

      // ("OK" / "BAD" / "NO" / "BYE" / "PREAUTH" / text) SP
      var respCondSep = line.IndexOf(Ascii.Octets.SP, tagSep + 1);
      var respCondOrDataType = respCondSep == -1 ? null : line.Substring(tagSep + 1, respCondSep - tagSep - 1);

      if (respCondOrDataType == null) {
        if (isTaggedResponse)
          throw new ImapMalformedResponseException("malformed response-tagged", line.ToString());
        //else
        // '* SEARCH\r\n' (mailbox-data which contains no SP)
      }

      if (TryParseCondition(respCondOrDataType, out var condition)) {
        // resp-cond-auth / resp-cond-state / resp-cond-bye
        var responseText = ParseRespText(line.Substring(respCondSep + 1));

        if (isTaggedResponse)
          return new ImapTaggedStatusResponse(line.ToString(0, tagSep), condition, responseText);
        else
          return new ImapUntaggedStatusResponse(condition, responseText);
      }

      if (isTaggedResponse)
          throw new ImapMalformedResponseException($"undefined condition: {respCondOrDataType}", line.ToString());

      // mailbox-data / message-data / capability-data etc.
      return ParseDataResponse(line, tagSep + 1, ref parsingContext);
    }

    private static readonly ByteString respCondOk      = ByteString.CreateImmutable("OK");
    private static readonly ByteString respCondNo      = ByteString.CreateImmutable("NO");
    private static readonly ByteString respCondBad     = ByteString.CreateImmutable("BAD");
    private static readonly ByteString respCondBye     = ByteString.CreateImmutable("BYE");
    private static readonly ByteString respCondPreAuth = ByteString.CreateImmutable("PREAUTH");
    
    private bool TryParseCondition(ByteString cond, out ImapResponseCondition condition)
    {
      condition = ImapResponseCondition.Undefined;

      if (respCondOk.EqualsIgnoreCase(cond))
        condition = ImapResponseCondition.Ok;
      else if (respCondNo.EqualsIgnoreCase(cond))
        condition = ImapResponseCondition.No;
      else if (respCondBye.EqualsIgnoreCase(cond))
        condition = ImapResponseCondition.Bye;
      else if (respCondBad.EqualsIgnoreCase(cond))
        condition = ImapResponseCondition.Bad;
      else if (respCondPreAuth.EqualsIgnoreCase(cond))
        condition = ImapResponseCondition.PreAuth;

      return condition != ImapResponseCondition.Undefined;
    }

    private ImapResponseText ParseRespText(ByteString respText)
    {
      // resp-text       = ["[" resp-text-code "]" SP] text
      if (respText[0] != ImapOctets.OpenBracket)
        // resp-text which has no resp-text-code
        return new ImapResponseText(DecodeRespTextString(respText.Substring(0, respText.Length - 2))); // remove trailing CRLF

      // resp-text which has resp-text-code
      var respTextCodeEnd = respText.IndexOf(ImapOctets.CloseBracket);

      if (respTextCodeEnd == -1)
        throw new ImapMalformedResponseException("malformed resp-text-code", respText.ToString());

      var text = string.Empty;

      if (respTextCodeEnd < respText.Length) {
        var textOffset = respTextCodeEnd + 1;

        if (textOffset < respText.Length && respText[textOffset] == Ascii.Octets.SP)
          textOffset += 1;

        text = DecodeRespTextString(respText.Substring(textOffset, respText.Length - textOffset - 2 /*remove trailing CRLF */));
      }

      var respTextCode = respText.Substring(1, respTextCodeEnd - 1);
      var codeSep = respTextCode.IndexOf(Ascii.Octets.SP);

      if (codeSep == -1)
        // has no arguments; "READ-ONLY" / "READ-WRITE" / etc...
        return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(respTextCode),
#if NET45
                                    ArrayExtensions.Empty<ImapData>(),
#else
                                    Array.Empty<ImapData>(),
#endif
                                    text);

      // resp-text-code which has arguments or subcode
      //   "UIDNEXT" SP nz-number / "UIDVALIDITY" SP nz-number / etc...
      //   "METADATA" SP ("MAXSIZE" SP number / "TOOMANY" / "NOPRIVATE")
      var code = respTextCode.Substring(0, codeSep);
      var subCodeOrArguments = respTextCode.Substring(codeSep + 1);
      var subCodeOrArgumentList = ParseData(subCodeOrArguments, 0) ?? throw new ImapMalformedResponseException("malformed subcode or arguments of resp-text-code", subCodeOrArguments.ToString());

      return new ImapResponseText(ImapResponseCode.GetKnownOrCreate(code, subCodeOrArgumentList, out var arguments),
                                  arguments,
                                  text);

      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 3.5. Formal Syntax
       *     resp-text         = ["[" resp-text-code "]" SP ] UTF8-TEXT-CHAR
       *                         *(UTF8-TEXT-CHAR / "[")
       *         ; After the server is changed to a language other than
       *         ; i-default, this resp-text rule replaces the resp-text
       *         ; rule from [RFC3501].
       */
      string DecodeRespTextString(ByteString t) => ReceiveResponseAsUTF8 ? t.ToString(Encoding.UTF8) : t.ToString();
    }

    private ImapServerDataResponse ParseDataResponse(ByteString line, int offset, ref ParsingContext parsingContext)
    {
      // mailbox-data / message-data / capability-data etc.
      var parsedData = ParseData(line, offset, line.Length - 2 /*CRLF*/, ref parsingContext);

      if (parsedData == null)
        return null;
      else
        return ImapServerDataResponse.Create(parsedData);
    }
  }
}
