// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  // format converting methods for ImapServerDataResponse
  public static class ImapServerDataResponseParser {
    /*
     * 7.2. Server Responses - Server and Mailbox Status
     */

    // 7.2.1 CAPABILITY Response
    //   Contents:   capability listing
    //   Example:    S: * CAPABILITY IMAP4rev1 STARTTLS AUTH=GSSAPI XPIG-LATIN
    public static IReadOnlyStringSet<ImapCapability> FromCapability(ImapServerDataResponse data)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      capability-data = "CAPABILITY" *(SP capability) SP "IMAP4rev1"
                        *(SP capability)
                          ; Servers MUST implement the STARTTLS, AUTH=PLAIN,
                          ; and LOGINDISABLED capabilities
                          ; Servers which offer RFC 1730 compatibility MUST
                          ; list "IMAP4" as the first capability.
      */
      RejectMalformed(data, ImapServerDataResponseType.Capability, 0);

      return ImapDataParser.ToCapability(data.DataList);
    }

    // 7.2.2 LIST Response
    //   Contents:   name attributes
    //               hierarchy delimiter
    //               name
    //   Example:    S: * LIST (\Noselect) "/" ~/Mail/foo
    public static ImapListedMailbox FromList(ImapServerDataResponse data)
    {
      RejectMalformed(data, ImapServerDataResponseType.List, 0);

      return FromListLsub(data);
    }

    // 7.2.3 LSUB Response
    //   Contents:   name attributes
    //               hierarchy delimiter
    //               name
    //   Example:    S: * LSUB () "." #news.comp.mail.misc
    public static ImapListedMailbox FromLsub(ImapServerDataResponse data)
    {
      RejectMalformed(data, ImapServerDataResponseType.Lsub, 0);

      return FromListLsub(data);
    }

    public static ImapListedMailbox FromListLsub(ImapServerDataResponse data)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      mailbox-data    =  "FLAGS" SP flag-list / "LIST" SP mailbox-list /
                         "LSUB" SP mailbox-list / "SEARCH" *(SP nz-number) /
                         "STATUS" SP mailbox SP "(" [status-att-list] ")" /
                         number SP "EXISTS" / number SP "RECENT"
      */
      if (!(data.Type == ImapServerDataResponseType.List || data.Type == ImapServerDataResponseType.Lsub))
        throw new ImapMalformedDataException($"expected data type is {ImapServerDataResponseType.List} or {ImapServerDataResponseType.Lsub}, but was {data.Type}");

      return ImapDataParser.ToMailboxList(data.DataList);
    }

    // 7.2.4 STATUS Response
    //   Contents:   name
    //               status parenthesized list
    //   Example:    S: * STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292)
    public static (string mailbox, IReadOnlyDictionary<string, ImapData> statusAttributes) FromStatus(ImapServerDataResponse data)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      mailbox-data    =  "FLAGS" SP flag-list / "LIST" SP mailbox-list /
                         "LSUB" SP mailbox-list / "SEARCH" *(SP nz-number) /
                         "STATUS" SP mailbox SP "(" [status-att-list] ")" /
                         number SP "EXISTS" / number SP "RECENT"
      */
      RejectMalformed(data, ImapServerDataResponseType.Status, 2);

      var mailbox = ImapDataParser.ToMailbox(data.DataList[0]);
      var attributes = new Dictionary<string, ImapData>(StringComparer.OrdinalIgnoreCase);

      ImapDataParser.ToStatusAttList(data.DataList[1], (statusAtt, value) => attributes[statusAtt] = value);

      return (mailbox, attributes);
    }

    internal static void FromStatus(ImapServerDataResponse data, ImapMailboxManager mailboxManager)
    {
      RejectMalformed(data, ImapServerDataResponseType.Status, 2);

      var mailbox = mailboxManager.GetExistOrCreate(ImapDataParser.ToMailbox(data.DataList[0]));

      ImapDataParser.ToStatusAttList(data.DataList[1], mailbox.UpdateStatus);
    }

    // 7.2.5 SEARCH Response
    //   Contents:   zero or more numbers
    //   Example:    S: * SEARCH 2 3 6
    public static ImapMatchedSequenceSet FromSearch(ImapServerDataResponse data, bool uidSet)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      mailbox-data    =  "FLAGS" SP flag-list / "LIST" SP mailbox-list /
                         "LSUB" SP mailbox-list / "SEARCH" *(SP nz-number) /
                         "STATUS" SP mailbox SP "(" [status-att-list] ")" /
                         number SP "EXISTS" / number SP "RECENT"
      */
      RejectMalformed(data, ImapServerDataResponseType.Search, 0);

      return FromSearchOrSort(data, uidSet);
    }

    // http://tools.ietf.org/html/rfc5256
    // BASE.7.2.SORT. SORT Response
    //   Data:       zero or more numbers
    //   Example:    S: * SORT 2 3 6
    public static ImapMatchedSequenceSet FromSort(ImapServerDataResponse data, bool uidSet)
    {
      /*
      sort-data       = "SORT" *(SP nz-number)
      */
      RejectMalformed(data, ImapServerDataResponseType.Sort, 0);

      return FromSearchOrSort(data, uidSet);
    }

    private static ImapMatchedSequenceSet FromSearchOrSort(ImapServerDataResponse data, bool uidSet)
    {
      /*
       * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
       * http://tools.ietf.org/html/rfc4551
       * 
       * 3.5. Modified SEARCH Untagged Response
       *    Data:       zero or more numbers
       *                mod-sequence value (omitted if no match)
       * 
       *    mailbox-data        =/ "SEARCH" [1*(SP nz-number) SP
       *                           search-sort-mod-seq]
       *    search-sort-mod-seq = "(" "MODSEQ" SP mod-sequence-value ")"
       */
      if (data.DataList.Count == 0)
        return ImapMatchedSequenceSet.Empty(uidSet);

      var last = data.DataList.Count - 1;

      if (data.DataList[last].Format == ImapDataFormat.List) {
        if (data.DataList[last].List.Count < 2 || !ImapDataParser.IsTextEqualsToCaseInsensitive(data.DataList[last].List[0], "MODSEQ"))
          throw new ImapMalformedDataException("invalid format", data.DataList[last]);

        var numbers = new long[last];

        for (var i = 0; i < numbers.Length; i++) {
          numbers[i] = ImapDataParser.ToNonZeroNumber(data.DataList[i]);
        }

        return new ImapMatchedSequenceSet(ImapSequenceSet.Of(numbers, asUidSet: uidSet),
                                          ImapDataParser.ToModSequenceValue(data.DataList[last].List[1]));
      }
      else {
        IEnumerable<ImapData> numbers = data.DataList;

        if (data.DataList[last].GetTextLength() == 0) {
          // http://smdn.jp/misc/forum/libs/all.xhtml#entry170
          // XXX: workaround for Yahoo! IMAP; returns SEARCH response ends with extra SP
          //   '* SEARCH 2 5 6 '
          //   '* SEARCH '
          numbers = data.DataList.Take(last);
        }

        return new ImapMatchedSequenceSet(ImapSequenceSet.Of(numbers.Select(number => ImapDataParser.ToNonZeroNumber(number)), asUidSet: uidSet));
      }
    }

    // 7.2.6 FLAGS Response
    //   Contents:   flag parenthesized list
    //   Example:    S: * FLAGS (\Answered \Flagged \Deleted \Seen \Draft)
    public static IReadOnlyStringSet<ImapMessageFlag> FromFlags(ImapServerDataResponse data)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      mailbox-data    =  "FLAGS" SP flag-list / "LIST" SP mailbox-list /
                         "LSUB" SP mailbox-list / "SEARCH" *(SP nz-number) /
                         "STATUS" SP mailbox SP "(" [status-att-list] ")" /
                         number SP "EXISTS" / number SP "RECENT"
      */
      RejectMalformed(data, ImapServerDataResponseType.Flags, 1);

      return ImapDataParser.ToFlagList(data.DataList[0]);
    }

    /*
     * 7.3. Server Responses - Mailbox Size
     */

    // 7.3.1. EXISTS Response
    //   Contents:   none
    //   Example:    S: * 23 EXISTS
    public static long FromExists(ImapServerDataResponse data)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      mailbox-data    =  "FLAGS" SP flag-list / "LIST" SP mailbox-list /
                         "LSUB" SP mailbox-list / "SEARCH" *(SP nz-number) /
                         "STATUS" SP mailbox SP "(" [status-att-list] ")" /
                         number SP "EXISTS" / number SP "RECENT"
      */
      RejectMalformed(data, ImapServerDataResponseType.Exists, 0);

      return ImapDataParser.ToNumber(data.SizeStatusNumber);
    }

    // 7.3.2. RECENT Response
    //   Contents:   none
    //   Example:    S: * 5 RECENT
    public static long FromRecent(ImapServerDataResponse data)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      mailbox-data    =  "FLAGS" SP flag-list / "LIST" SP mailbox-list /
                         "LSUB" SP mailbox-list / "SEARCH" *(SP nz-number) /
                         "STATUS" SP mailbox SP "(" [status-att-list] ")" /
                         number SP "EXISTS" / number SP "RECENT"
      */
      RejectMalformed(data, ImapServerDataResponseType.Recent, 0);

      return ImapDataParser.ToNumber(data.SizeStatusNumber);
    }

    /*
     * 7.4. Server Responses - Message Status
     */

    // 7.4.1. EXPUNGE Response
    //   Contents:   none
    //   Example:    S: * 44 EXPUNGE
    public static long FromExpunge(ImapServerDataResponse data)
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      message-data    = nz-number SP ("EXPUNGE" / ("FETCH" SP msg-att))
      */
      RejectMalformed(data, ImapServerDataResponseType.Expunge, 0);

      return ImapDataParser.ToNonZeroNumber(data.SizeStatusNumber);
    }

    // 7.4.2. FETCH Response
    //   Contents:   message data
    //   Example:    S: * 23 FETCH (FLAGS (\Seen) RFC822.SIZE 44827)
    public static TMessageAttributes FromFetch<TMessageAttributes>(ImapServerDataResponse data)
      where TMessageAttributes : ImapMessageAttributesBase
    {
      RejectMalformed(data, ImapServerDataResponseType.Fetch, 1);

      TMessageAttributes attributes = null;

      ImapDataParser.ToFetchMessageData(ref attributes, data.SizeStatusNumber, data.DataList[0]);

      return attributes;
    }

    public static bool FromFetch<TMessageAttributes>(IImapMessageAttributeCollection<TMessageAttributes> messages,
                                                     bool updateOnly,
                                                     ImapServerDataResponse data,
                                                     out TMessageAttributes message)
      where TMessageAttributes : ImapMessageAttributesBase
    {
      /*
      response-data   = "*" SP (resp-cond-state / resp-cond-bye /
                        mailbox-data / message-data / capability-data) CRLF
      message-data    = nz-number SP ("EXPUNGE" / ("FETCH" SP msg-att))
      */
      RejectMalformed(data, ImapServerDataResponseType.Fetch, 1);

      return ImapDataParser.ToFetchMessageData(messages,
                                               updateOnly,
                                               data.SizeStatusNumber,
                                               data.DataList[0],
                                               out message);
    }

    /*
     * RFC 2087 - IMAP4 QUOTA extension
     * http://tools.ietf.org/html/rfc2087
     * 
     * 5.1. QUOTA Response
     *    Data:       quota root name
     *                list of resource names, usages, and limits
     * 
     *       This response occurs as a result of a GETQUOTA or GETQUOTAROOT
     *       command. The first string is the name of the quota root for which
     *       this quota applies.
     * 
     *       The name is followed by a S-expression format list of the resource
     *       usage and limits of the quota root.  The list contains zero or
     *       more triplets.  Each triplet conatins a resource name, the current
     *       usage of the resource, and the resource limit.
     * 
     *       Resources not named in the list are not limited in the quota root.
     *       Thus, an empty list means there are no administrative resource
     *       limits in the quota root.
     */
    public static ImapQuota FromQuota(ImapServerDataResponse data)
    {
      /* 
       *    quota_response  ::= "QUOTA" SP astring SP quota_list
       *    quota_list      ::= "(" #quota_resource ")"
       */
      RejectMalformed(data, ImapServerDataResponseType.Quota, 2);

      return new ImapQuota(ImapDataParser.ToAString(data.DataList[0]),
                           ImapDataParser.ToQuotaList(data.DataList[1]));
    }

    /*
     * RFC 2087 - IMAP4 QUOTA extension
     * http://tools.ietf.org/html/rfc2087
     * 
     * 5.2. QUOTAROOT Response
     *    Data:       mailbox name
     *                zero or more quota root names
     * 
     *       This response occurs as a result of a GETQUOTAROOT command.  The
     *       first string is the mailbox and the remaining strings are the
     *       names of the quota roots for the mailbox.
     */
    public static string FromQuotaRoot(ImapServerDataResponse data, out IReadOnlyList<string> quotaRootNames)
    {
      /* 
       *    quotaroot_response
       *                    ::= "QUOTAROOT" SP astring *(SP astring)
       */
      RejectMalformed(data, ImapServerDataResponseType.QuotaRoot, 1);

      var ret = ImapDataParser.ToMailbox(data.DataList[0]);

      if (data.DataList.Count == 1)
#if NET45
        quotaRootNames = ArrayExtensions.Empty<string>();
#else
        quotaRootNames = Array.Empty<string>();
#endif
      else
        quotaRootNames = data.DataList.Skip(1).Select(ImapDataParser.ToAString).ToList();

      return ret;
    }

    /*
     * RFC 5161 The IMAP ENABLE Extension
     * http://tools.ietf.org/html/rfc5161
     */

    // 3.2. The ENABLED Response
    //    Contents:   capability listing
    // 
    //    The ENABLED response occurs as a result of an ENABLE command.  The
    //    capability listing contains a space-separated listing of capability
    //    names that the server supports and that were successfully enabled.
    //    The ENABLED response may contain no capabilities, which means that no
    //    extensions listed by the client were successfully enabled.
    public static IReadOnlyStringSet<ImapCapability> FromEnabled(ImapServerDataResponse data)
    {
      /*
      response-data =/ "*" SP enable-data CRLF
      enable-data   = "ENABLED" *(SP capability)
      */
      RejectMalformed(data, ImapServerDataResponseType.Enabled, 0);

      return ImapDataParser.ToCapability(data.DataList);
    }

    /*
     * RFC 2971 IMAP4 ID extension
     * http://tools.ietf.org/html/rfc2971
     */

    // 3.2. ID Response
    //    Contents:   server parameter list
    // 
    //    In response to an ID command issued by the client, the server replies
    //    with a tagged response containing information on its implementation.
    //    The format is the same as the client list.
    public static IReadOnlyDictionary<string, string> FromId(ImapServerDataResponse data)
    {
      /*
      id_response ::= "ID" SPACE id_params_list
      */
      RejectMalformed(data, ImapServerDataResponseType.Id, 1);

      return ImapDataParser.ToIdParamsList(data.DataList[0]);
    }

    // http://tools.ietf.org/html/rfc5256
    // BASE.7.2.THREAD. THREAD Response
    //   Data:       zero or more threads
    //   Example:    S: * THREAD (2)(3 6 (4 23)(44 7 96))
    //   Example:    S: * THREAD ((3)(5))
    public static ImapThreadList FromThread(ImapServerDataResponse data, bool uidSet)
    {
      /*
      thread-data     = "THREAD" [SP 1*thread-list]
      */
      RejectMalformed(data, ImapServerDataResponseType.Thread, 0);

      return ImapDataParser.ToThreadList(uidSet, data.DataList);
    }

    /*
     * RFC 2342 IMAP4 Namespace
     * http://tools.ietf.org/html/rfc2342
     */

    // 5. NAMESPACE Command
    //   Response:  an untagged NAMESPACE response that contains the prefix
    //                and hierarchy delimiter to the server's Personal
    //                Namespace(s), Other Users' Namespace(s), and Shared
    //                Namespace(s) that the server wishes to expose. The
    //                response will contain a NIL for any namespace class
    //                that is not available. Namespace_Response_Extensions
    //                MAY be included in the response.
    //                Namespace_Response_Extensions which are not on the IETF
    //                standards track, MUST be prefixed with an "X-".
    //   Example 5.4:
    //   S: * NAMESPACE (("" "/")) (("~" "/")) (("#shared/" "/")
    //      ("#public/" "/")("#ftp/" "/")("#news." "."))
    public static ImapNamespace FromNamespace(ImapServerDataResponse data)
    {
      /*
       Namespace_Response = "*" SP "NAMESPACE" SP Namespace SP Namespace SP
          Namespace
      */
      RejectMalformed(data, ImapServerDataResponseType.Namespace, 3);

      return new ImapNamespace(ImapDataParser.ToNamespace(data.DataList[0]),
                               ImapDataParser.ToNamespace(data.DataList[1]),
                               ImapDataParser.ToNamespace(data.DataList[2]));
    }

    /*
     * RFC 5255 - Internet Message Access Protocol Internationalization
     * http://tools.ietf.org/html/rfc5255
     * 3.3. LANGUAGE Response
     *    Contents:  A list of one or more language tags.
     */
    public static IReadOnlyList<string> FromLanguage(ImapServerDataResponse data)
    {
      /*
       *     response-payload  =/ language-data
       *     language-data     = "LANGUAGE" SP "(" lang-tag-quoted *(SP
       *                       lang-tag-quoted) ")"
       */
      RejectMalformed(data, ImapServerDataResponseType.Language, 1);

      if (data.DataList[0].Format != ImapDataFormat.List)
        throw new ImapMalformedDataException("invalid format");

      return data.DataList[0].List.ConvertAll(ImapDataParser.ToLangTagQuoted);
    }

    /*
     * RFC 5255 - Internet Message Access Protocol Internationalization
     * http://tools.ietf.org/html/rfc5255
     * 4.8. COMPARATOR Response
     *   Contents:  The active comparator.  An optional list of available
     *              matching comparators
     */
    public static ImapCollationAlgorithm FromComparator(ImapServerDataResponse data, out IReadOnlyList<ImapCollationAlgorithm> matchingComparators)
    {
      /*
       * 
       *     comparator-data   = "COMPARATOR" SP comp-sel-quoted [SP "("
       *                         comp-id-quoted *(SP comp-id-quoted) ")"]
       *     comp-order-quoted = astring
       *         ; Once any literal wrapper or quoting is removed, this
       *         ; follows the collation-order rule from [RFC4790]
       *     comp-sel-quoted   = astring
       *         ; Once any literal wrapper or quoting is removed, this
       *         ; follows the collation-selected rule from [RFC4790]
       */
      RejectMalformed(data, ImapServerDataResponseType.Comparator, 1);

      var activeComparator = ImapDataParser.ToCompSelQuoted(data.DataList[0]);

      if (2 <= data.DataList.Count) {
        if (data.DataList[1].Format != ImapDataFormat.List)
          throw new ImapMalformedDataException("invalid format");

        matchingComparators = data.DataList[1].List.ConvertAll(ImapDataParser.ToCompOrderQuoted);
      }
      else {
#if NET45
        matchingComparators = ArrayExtensions.Empty<ImapCollationAlgorithm>();
#else
        matchingComparators = Array.Empty<ImapCollationAlgorithm>();
#endif
      }

      return activeComparator;
    }

    /*
     * RFC 4466 - Collected Extensions to IMAP4 ABNF
     * http://tools.ietf.org/html/rfc4466
     * 
     * 2.6.2. ESEARCH untagged response
     *    Contents:   one or more search-return-data pairs
     */
    public static ImapMatchedSequenceSet FromESearch(ImapServerDataResponse data)
    {
      /*
       *    esearch-response  = "ESEARCH" [search-correlator] [SP "UID"]
       *                         *(SP search-return-data)
       *                       ;; Note that SEARCH and ESEARCH responses
       *                       ;; SHOULD be mutually exclusive,
       *                       ;; i.e., only one of the response types
       *                       ;; should be
       *                       ;; returned as a result of a command.
       */
      RejectMalformed(data, ImapServerDataResponseType.ESearch, 0);

      var index = 0;
      var uid = false;
      string tag = null;

      if (index < data.DataList.Count && data.DataList[index].Format == ImapDataFormat.List) {
        // search-correlator
        tag = ImapDataParser.ToSearchCorrelator(data.DataList[index]);

        index++;
      }

      if (index < data.DataList.Count &&
          data.DataList[index].Format == ImapDataFormat.Text &&
          ImapDataParser.IsTextEqualsToCaseInsensitive(data.DataList[index], "UID")) {
        uid = true;

        index++;
      }

      // search-return-data
      return ImapDataParser.ToSearchReturnData(data.DataList, tag, uid, index);
    }

    /*
     * RFC 5464 - The IMAP METADATA Extension
     * http://tools.ietf.org/html/rfc5464
     */
    public static IReadOnlyList<string> FromMetadataEntryList(ImapServerDataResponse data, out string mailbox)
    {
      /*
       *       metadata-resp     = "METADATA" SP mailbox SP
       *                           (entry-values / entry-list)
       *                           ; empty string for mailbox implies
       *                           ; server annotation.
       */
      RejectMalformed(data, ImapServerDataResponseType.Metadata, 2);

      if (data.DataList[0].Format != ImapDataFormat.Text)
        throw new ImapMalformedDataException("invalid format");

      mailbox = ImapDataParser.ToMailbox(data.DataList[0]);

      return ImapDataParser.ToEntryList(data.DataList.Slice(1));
    }

    public static IReadOnlyList<ImapMetadata> FromMetadataEntryValues(ImapServerDataResponse data, out string mailbox)
    {
      /*
       *       metadata-resp     = "METADATA" SP mailbox SP
       *                           (entry-values / entry-list)
       *                           ; empty string for mailbox implies
       *                           ; server annotation.
       */
      RejectMalformed(data, ImapServerDataResponseType.Metadata, 2);

      if (data.DataList[0].Format != ImapDataFormat.Text)
        throw new ImapMalformedDataException("invalid format");

      mailbox = ImapDataParser.ToMailbox(data.DataList[0]);

      return ImapDataParser.ToEntryValues(data.DataList[1]);
    }

    private static void RejectMalformed(ImapServerDataResponse data, ImapServerDataResponseType expectedType, int expectedCount)
    {
      if (data.Type != expectedType)
        throw new ImapMalformedDataException(string.Format("expected data type is {0}, but was {1}", expectedType, data.Type));

      if (data.DataList.Count < expectedCount)
        throw new ImapMalformedDataException(string.Format("too few data counts; expected is {0} but was {1}", expectedCount, data.DataList.Count));

      if (data.SizeStatusNumber == null && ImapServerDataResponseType.SizeStatusTypes.Contains(expectedType))
        throw new ImapMalformedDataException(string.Format("size or message number missing"));
    }
  }
}
