// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Urls;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  // combinable data item types:
  //   ImapCombinableDataItem
  //     => handles 'data item names' etc.
  //   * ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //     ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapSortCriteria
  //       => handles 'sort criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //     ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.5. FETCH Command
  //   message data item names or macro
  public class ImapFetchDataItems : ImapCombinableDataItems, IImapExtension {
    // ALL
    //    Macro equivalent to: (FLAGS INTERNALDATE RFC822.SIZE ENVELOPE)
    public static readonly ImapFetchDataItems All
      = new ImapMacroFetchDataItems("ALL", ToList("FLAGS", "INTERNALDATE", "RFC822.SIZE", "ENVELOPE"));

    // FAST
    //    Macro equivalent to: (FLAGS INTERNALDATE RFC822.SIZE)
    public static readonly ImapFetchDataItems Fast
      = new ImapMacroFetchDataItems("FAST", ToList("FLAGS", "INTERNALDATE", "RFC822.SIZE"));

    // FULL
    //    Macro equivalent to: (FLAGS INTERNALDATE RFC822.SIZE ENVELOPE
    //    BODY)
    public static readonly ImapFetchDataItems Full
      = new ImapMacroFetchDataItems("FULL", ToList("FLAGS", "INTERNALDATE", "RFC822.SIZE", "ENVELOPE", "BODY"));

    internal static readonly ImapFetchDataItems Extensible
      = new ImapMacroFetchDataItems("FLAGS", ToList("INTERNALDATE", "RFC822.SIZE", "ENVELOPE", "BODYSTRUCTURE"));

    private sealed class ImapMacroFetchDataItems : ImapFetchDataItems {
      private protected override bool IsParenthesized => false;

      public IReadOnlyCollection<ImapString> EquivalentDataItemNames {
        get; private set;
      }

      public ImapMacroFetchDataItems(ImapString macro, IReadOnlyCollection<ImapString> equivalentDataItemNames)
        : base(Singleton.CreateList(macro))
      {
        this.EquivalentDataItemNames = equivalentDataItemNames;
      }
    }

    public static ImapFetchDataItems FromMacro(ImapFetchDataItemMacro macro)
    {
      switch (macro) {
        case ImapFetchDataItemMacro.All: return All;
        case ImapFetchDataItemMacro.Fast: return Fast;
        case ImapFetchDataItemMacro.Full: return Full;
        case ImapFetchDataItemMacro.Extensible: return Extensible;
        default: throw ExceptionUtils.CreateNotSupportedEnumValue(macro);
      }
    }

    // BODY
    //     Non-extensible form of BODYSTRUCTURE.
    public static readonly ImapFetchDataItems Body
      = new ImapFetchDataItems(ToList("BODY"));

    // BODY[<section>]<<partial>>
    //     The text of a particular body section.  The section
    //     specification is a set of zero or more part specifiers
    //     delimited by periods.  A part specifier is either a part number
    //     or one of the following: HEADER, HEADER.FIELDS,
    //     HEADER.FIELDS.NOT, MIME, and TEXT.  An empty section
    //     specification refers to the entire message, including the
    //     header.
    public static ImapFetchDataItems BodyText()
    {
      return BodyTextInternal(false, null, ImapFetchRange.All);
    }

    public static ImapFetchDataItems BodyText(string section)
    {
      return BodyTextInternal(false, section, ImapFetchRange.All);
    }

    public static ImapFetchDataItems BodyText(long partialStart, long partialLength)
    {
      return BodyTextInternal(false, null, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems BodyText(ImapFetchRange range)
    {
      return BodyTextInternal(false, null, range);
    }

    public static ImapFetchDataItems BodyText(string section, long partialStart, long partialLength)
    {
      return BodyTextInternal(false, section, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems BodyText(string section, ImapFetchRange range)
    {
      return BodyTextInternal(false, section, range);
    }

    // BODY.PEEK[<section>]<<partial>>
    //     An alternate form of BODY[<section>] that does not implicitly
    //     set the \Seen flag.
    public static ImapFetchDataItems BodyPeek()
    {
      return BodyTextInternal(true, null, ImapFetchRange.All);
    }

    public static ImapFetchDataItems BodyPeek(string section)
    {
      return BodyTextInternal(true, section, ImapFetchRange.All);
    }

    public static ImapFetchDataItems BodyPeek(long partialStart, long partialLength)
    {
      return BodyTextInternal(true, null, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems BodyPeek(ImapFetchRange range)
    {
      return BodyTextInternal(true, null, range);
    }

    public static ImapFetchDataItems BodyPeek(string section, long partialStart, long partialLength)
    {
      return BodyTextInternal(true, section, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems BodyPeek(string section, ImapFetchRange range)
    {
      return BodyTextInternal(true, section, range);
    }

    public static ImapFetchDataItems BodyText(bool peek)
    {
      return BodyTextInternal(peek, null, ImapFetchRange.All);
    }

    public static ImapFetchDataItems BodyText(bool peek, string section)
    {
      return BodyTextInternal(peek, section, ImapFetchRange.All);
    }

    public static ImapFetchDataItems BodyText(bool peek, long partialStart, long partialLength)
    {
      return BodyTextInternal(peek, null, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems BodyText(bool peek, ImapFetchRange range)
    {
      return BodyTextInternal(peek, null, range);
    }

    public static ImapFetchDataItems BodyText(bool peek, string section, long partialStart, long partialLength)
    {
      return BodyTextInternal(peek, section, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems BodyText(bool peek, string section, ImapFetchRange range)
    {
      return BodyTextInternal(peek, section, range);
    }

    // TODO: HEADER, HEADER.FIELDS, HEADER.FIELDS.NOT, MIME, and TEXT
    private static ImapFetchDataItems BodyTextInternal(bool peek, string section, ImapFetchRange range)
    {
      if (range.IsPartial) {
        try {
          return new ImapFetchDataItems(ToList(string.Concat((peek ? "BODY.PEEK[" : "BODY["), section, "]", range.ToString("f", null))));
        }
        catch (FormatException ex) {
          throw new ArgumentException(ex.Message, nameof(range));
        }
      }
      else {
        return new ImapFetchDataItems(ToList(string.Concat((peek ? "BODY.PEEK[" : "BODY["), section, "]")));
      }
    }

    // BODYSTRUCTURE
    //     The [MIME-IMB] body structure of the message.  This is computed
    //     by the server by parsing the [MIME-IMB] header fields in the
    //     [RFC-2822] header and [MIME-IMB] headers.
    public static readonly ImapFetchDataItems BodyStructure
      = new ImapFetchDataItems(ToList("BODYSTRUCTURE"));

    // ENVELOPE
    //     The envelope structure of the message.  This is computed by the
    //     server by parsing the [RFC-2822] header into the component
    //     parts, defaulting various fields as necessary.
    public static readonly ImapFetchDataItems Envelope
      = new ImapFetchDataItems(ToList("ENVELOPE"));

    // FLAGS
    //     The flags that are set for this message.
    public static readonly ImapFetchDataItems Flags
      = new ImapFetchDataItems(ToList("FLAGS"));

    // INTERNALDATE
    //     The internal date of the message.
    public static readonly ImapFetchDataItems InternalDate
      = new ImapFetchDataItems(ToList("INTERNALDATE"));

    // RFC822
    //     Functionally equivalent to BODY[], differing in the syntax of
    //     the resulting untagged FETCH data (RFC822 is returned).
    public static readonly ImapFetchDataItems Rfc822
      = new ImapFetchDataItems(ToList("RFC822"));

    // RFC822.HEADER
    //     Functionally equivalent to BODY.PEEK[HEADER], differing in the
    //     syntax of the resulting untagged FETCH data (RFC822.HEADER is
    //     returned).
    public static readonly ImapFetchDataItems Rfc822Header
      = new ImapFetchDataItems(ToList("RFC822.HEADER"));

    // RFC822.SIZE
    //     The [RFC-2822] size of the message.
    public static readonly ImapFetchDataItems Rfc822Size
      = new ImapFetchDataItems(ToList("RFC822.SIZE"));

    // RFC822.TEXT
    //     Functionally equivalent to BODY[TEXT], differing in the syntax
    //     of the resulting untagged FETCH data (RFC822.TEXT is returned).
    public static readonly ImapFetchDataItems Rfc822Text
      = new ImapFetchDataItems(ToList("RFC822.TEXT"));

    // UID
    //     The unique identifier for the message.
    public static readonly ImapFetchDataItems Uid
      = new ImapFetchDataItems(ToList("UID"));

    /*
     * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
     * http://tools.ietf.org/html/rfc4551
     */
    public static readonly ImapFetchDataItems ModSeq
      = new ImapFetchDataItems(new ImapCapabilitySet(new[] {ImapCapability.CondStore}, asReadOnly: true), ToList("MODSEQ"));

    /*
     * RFC 3516 - IMAP4 Binary Content Extension
     * http://tools.ietf.org/html/rfc3516
     * 4.2. FETCH Command Extensions
     */
    private static readonly IReadOnlyStringSet<ImapCapability> binaryCapabilitySet = new ImapCapabilitySet(new[] {ImapCapability.Binary}, asReadOnly: true);

    /*
     *       BINARY.SIZE<section-binary>
     *          Requests the decoded size of the section (i.e., the size to
     *          expect in response to the corresponding FETCH BINARY request).
     */ 
    public static ImapFetchDataItems BinarySize(string sectionBinary)
    {
      if (sectionBinary == null)
        throw new ArgumentNullException(nameof(sectionBinary));
      if (sectionBinary.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(sectionBinary));

      return new ImapFetchDataItems(binaryCapabilitySet, ToList(string.Concat("BINARY.SIZE[", sectionBinary, "]")));
    }

    /*
     *       BINARY.PEEK<section-binary>[<partial>]
     *          An alternate form of FETCH BINARY that does not implicitly set
     *          the \Seen flag.
     */
    public static ImapFetchDataItems BinaryPeek(string sectionBinary)
    {
      return BinaryInternal(true, sectionBinary, ImapFetchRange.All);
    }

    public static ImapFetchDataItems BinaryPeek(string sectionBinary, long partialStart, long partialLength)
    {
      return BinaryInternal(true, sectionBinary, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems BinaryPeek(string sectionBinary, ImapFetchRange range)
    {
      return BinaryInternal(true, sectionBinary, range);
    }

    /*
     *       BINARY<section-binary>[<partial>]
     *          Requests that the specified section be transmitted after
     *          performing CTE-related decoding.
     */
    public static ImapFetchDataItems Binary(string sectionBinary)
    {
      return BinaryInternal(false, sectionBinary, ImapFetchRange.All);
    }

    public static ImapFetchDataItems Binary(string sectionBinary, long partialStart, long partialLength)
    {
      return BinaryInternal(false, sectionBinary, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems Binary(string sectionBinary, ImapFetchRange range)
    {
      return BinaryInternal(false, sectionBinary, range);
    }

    public static ImapFetchDataItems Binary(bool peek, string sectionBinary)
    {
      return BinaryInternal(peek, sectionBinary, ImapFetchRange.All);
    }

    public static ImapFetchDataItems Binary(bool peek, string sectionBinary, long partialStart, long partialLength)
    {
      return BinaryInternal(peek, sectionBinary, new ImapFetchRange(partialStart, partialLength));
    }

    public static ImapFetchDataItems Binary(bool peek, string sectionBinary, ImapFetchRange range)
    {
      return BinaryInternal(peek, sectionBinary, range);
    }

    private static ImapFetchDataItems BinaryInternal(bool peek, string sectionBinary, ImapFetchRange range)
    {
      if (sectionBinary == null)
        throw new ArgumentNullException(nameof(sectionBinary));
      if (sectionBinary.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(sectionBinary));

      if (range.IsPartial) {
        try {
          return new ImapFetchDataItems(binaryCapabilitySet,
                                        ToList(string.Concat((peek ? "BINARY.PEEK[" : "BINARY["), sectionBinary, "]", range.ToString("f", null))));
        }
        catch (FormatException ex) {
          throw new ArgumentException(ex.Message, nameof(range));
        }
      }
      else {
        return new ImapFetchDataItems(binaryCapabilitySet,
                                      ToList(string.Concat((peek ? "BINARY.PEEK[" : "BINARY["), sectionBinary, "]")));
      }
    }

    /*
     * methods of combining
     */
    public ImapFetchDataItems CombineWith(ImapFetchDataItems dataItem)
    {
      return this + dataItem;
    }

    public static ImapFetchDataItems Combine(ImapFetchDataItems x, ImapFetchDataItems y)
    {
      return x + y;
    }

    public static ImapFetchDataItems operator +(ImapFetchDataItems x, ImapFetchDataItems y)
    {
      var requiredCapabilities = GetCombinedCapabilities(x, y);

      var itemsX = x is ImapMacroFetchDataItems mx ? mx.EquivalentDataItemNames : x.Items;
      var itemsY = y is ImapMacroFetchDataItems my ? my.EquivalentDataItemNames : y.Items;

      var items = new List<ImapString>(itemsX.Count + itemsY.Count);

      items.AddRange(itemsX);
      items.AddRange(itemsY);

      return new ImapFetchDataItems(requiredCapabilities, items);
    }

    /*
     * instance members
     */
    private readonly IReadOnlyStringSet<ImapCapability> requiredCapabilities;
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities => requiredCapabilities;

    private ImapFetchDataItems(IReadOnlyCollection<ImapString> dataItemNames)
      : this(ImapCapabilitySet.ReadOnlyEmpty, dataItemNames)
    {
    }

    private ImapFetchDataItems(IReadOnlyStringSet<ImapCapability> requiredCapabilities, IReadOnlyCollection<ImapString> dataItemNames)
      : base(dataItemNames)
    {
      this.requiredCapabilities = requiredCapabilities;
    }

    public static ImapFetchDataItems FromUri(Uri uri)
    {
      return BodyPeek(ImapStyleUriParser.GetSection(uri), ImapStyleUriParser.GetPartial(uri));
    }
  }
}
