// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  // enum types:
  //   ImapStringEnum : ImapString
  //     => handles string constants
  //     ImapAuthenticationMechanism
  //       => handles authentication mechanisms
  //     ImapCapability
  //       => handles capability constants
  //     ImapCompressionMechanism
  //       => handles COMPRESS extension compression mechanism
  //     ImapMailboxFlag
  //       => handles mailbox flags
  //   * ImapMessageFlag
  //       => handles system flags and keywords
  //     ImapThreadingAlgorithm
  //       => handles THREAD extension threading algorithm
  //     Protocol.ImapResponseCode
  //       => handles response codes
  //     Protocol.ImapDataResponseType
  //       => handles server response types

#if SERIALIZATION
  [Serializable]
#endif
  public sealed class ImapMessageFlag : ImapEnumValue, IImapExtension, IEquatable<ImapMessageFlag> {
    private static readonly ImapMessageFlagSet allFlags;
    public static IReadOnlyStringSet<ImapMessageFlag> AllFlags => allFlags;

    private static readonly ImapMessageFlagSet fetchFlags;
    public static IReadOnlyStringSet<ImapMessageFlag> FetchFlags => fetchFlags;

    private static readonly ImapMessageFlagSet permittedFlags;
    public static IReadOnlyStringSet<ImapMessageFlag> PermittedFlags => permittedFlags;

    private static readonly ImapMessageFlagSet nonApplicableFlags;
    public static IReadOnlyStringSet<ImapMessageFlag> NonApplicableFlags => nonApplicableFlags;

    /*
     * RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
     * http://tools.ietf.org/html/rfc3501
     */
    public static readonly ImapMessageFlag AllowedCreateKeywords = new ImapMessageFlag(PseudoValidator, @"\*", null);

    public static readonly ImapMessageFlag Seen     = new ImapMessageFlag(PseudoValidator, @"\Seen", null);
    public static readonly ImapMessageFlag Answered = new ImapMessageFlag(PseudoValidator, @"\Answered", null);
    public static readonly ImapMessageFlag Flagged  = new ImapMessageFlag(PseudoValidator, @"\Flagged", null);
    public static readonly ImapMessageFlag Deleted  = new ImapMessageFlag(PseudoValidator, @"\Deleted", null);
    public static readonly ImapMessageFlag Draft    = new ImapMessageFlag(PseudoValidator, @"\Draft", null);
    public static readonly ImapMessageFlag Recent   = new ImapMessageFlag(PseudoValidator, @"\Recent", null);

    static ImapMessageFlag()
    {
      allFlags = new ImapMessageFlagSet(CreateDefinedConstantsSet<ImapMessageFlag>());

      fetchFlags = new ImapMessageFlagSet(new[] {
        Answered,
        Flagged,
        Deleted,
        Seen,
        Draft,
        Recent,
      });

      permittedFlags = new ImapMessageFlagSet(new[] {
        Answered,
        Flagged,
        Deleted,
        Seen,
        Draft,
        AllowedCreateKeywords,
      });

      nonApplicableFlags = new ImapMessageFlagSet(new[] {
        Recent,
        AllowedCreateKeywords,
      });
    }

    internal static ImapMessageFlag GetKnownOrCreate(string flag)
    {
      if (allFlags.TryGetValue(flag, out var f))
        return f;

      /*
      if (flag.StartsWith(@"\"))
        Trace.Verbose("unknown system flag: {0}", flag);
      */

      return new ImapMessageFlag(flag);
    }

    public static bool IsValidKeyword(string keyword)
    {
      if (keyword == null)
        throw new ArgumentNullException(nameof(keyword));

      return IsValidKeyword(keyword, out _);
    }

    private static string PseudoValidator(string keyword) => keyword;

    /// <returns>validated keyword if valid</returns>
    public static string GetValidatedKeyword(string keyword)
    {
      if (keyword == null)
        throw new ArgumentNullException(nameof(keyword));

      if (IsValidKeyword(keyword, out var errorReason))
        return keyword;
      else
        throw new ArgumentException(errorReason, nameof(keyword));
    }

    private static bool IsValidKeyword(string keyword, out string errorReason)
    {
      /*
       * flag            = "\Answered" / "\Flagged" / "\Deleted" /
       *                   "\Seen" / "\Draft" / flag-keyword / flag-extension
       * flag-extension  = "\" atom
       * flag-keyword    = atom
       */
      errorReason = null;

      if (keyword.Length == 0) {
        errorReason = "keyword must be non-empty string";
        return false;
      }

      var index = ImapOctets.IndexOfNonAtomChar(keyword);

      if (0 <= index) {
        errorReason = string.Format("keyword '{0}' contains invalid charactor at {1}", keyword, index);
        return false;
      }

      return true;
    }

    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get
      {
        if (requiredCapability != null)
          yield return requiredCapability;
      }
    }

    public bool IsSystemFlag {
      get { return base.Value.StartsWith('\\'); }
    }

    public ImapMessageFlag(string keyword, ImapCapability requiredCapability = null)
      : this(GetValidatedKeyword, keyword, requiredCapability)
    {
    }

    private ImapMessageFlag(Func<string, string> validator, string keyword, ImapCapability requiredCapability)
      : base(validator(keyword))
    {
      this.requiredCapability = requiredCapability;
    }

    #region "equatable"
    public static bool operator ==(ImapMessageFlag x, ImapMessageFlag y) =>  Equals(x, y);
    public static bool operator !=(ImapMessageFlag x, ImapMessageFlag y) => !Equals(x, y);
    public bool Equals(ImapMessageFlag other) => Equals(this, other);
    public override bool Equals(object obj) => obj is ImapMessageFlag other ? Equals(this, other) : base.Equals(obj);
    public override int GetHashCode() => base.GetHashCode();
    #endregion

    private readonly ImapCapability requiredCapability;
  }
}
