// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  // combinable data item types:
  //   ImapCombinableDataItem
  //     => handles 'data item names' etc.
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //   * ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapSortCriteria
  //       => handles 'sort criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //     ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.4. SEARCH Command
  //   searching criteria
  public sealed partial class ImapSearchCriteria : ImapCombinableDataItems, IImapExtension {
    //  <sequence set>
    //     Messages with message sequence numbers corresponding to the
    //     specified message sequence number set.
    public static ImapSearchCriteria SequenceSet(IImapNumberSet sequenceSet)
    {
      ImapNumberSet.ThrowIfEmptyOrNotSequenceSet(sequenceSet, nameof(sequenceSet));

      return new ImapSearchCriteria(ToList(sequenceSet.Expression));
    }

    // UID <sequence set>
    //     Messages with unique identifiers corresponding to the specified
    //     unique identifier set.  Sequence set ranges are permitted.
    public static ImapSearchCriteria Uid(IImapNumberSet uidSet)
    {
      ImapNumberSet.ThrowIfEmptyOrNotUidSet(uidSet, nameof(uidSet));

      return new ImapSearchCriteria(ToList("UID", uidSet.Expression));
    }

    public static ImapSearchCriteria SequenceOrUidSet(IImapNumberSet sequenceOrUidSet)
    {
      ImapNumberSet.ThrowIfEmpty(sequenceOrUidSet, nameof(sequenceOrUidSet));

      if (sequenceOrUidSet.IsUidSet)
        return new ImapSearchCriteria(ToList("UID", sequenceOrUidSet.Expression));
      else
        return new ImapSearchCriteria(ToList(sequenceOrUidSet.Expression));
    }

    // ALL
    //     All messages in the mailbox; the default initial key for
    //     ANDing.
    public static readonly ImapSearchCriteria All
      = new ImapSearchCriteria(ToList("ALL"));

    // ANSWERED
    //     Messages with the \Answered flag set.
    public static readonly ImapSearchCriteria Answered
      = new ImapSearchCriteria(ToList("ANSWERED"));

    // BCC <string>
    //     Messages that contain the specified string in the envelope
    //     structure's BCC field.
    public static ImapSearchCriteria Bcc(string @value)
    {
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "BCC", new ImapTextLiteralString(@value) });
    }

    // BEFORE <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is earlier than the specified date.
    public static ImapSearchCriteria Before(DateTime date)
    {
      return new ImapSearchCriteria(ToList("BEFORE", ImapDateTimeFormat.ToDateString(date)));
    }

    // BODY <string>
    //     Messages that contain the specified string in the body of the
    //     message.
    public static ImapSearchCriteria Body(string @value)
    {
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "BODY", new ImapTextLiteralString(@value) });
    }

    // CC <string>
    //     Messages that contain the specified string in the envelope
    //     structure's CC field.
    public static ImapSearchCriteria Cc(string @value)
    {
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "CC", new ImapTextLiteralString(@value) });
    }

    // DELETED
    //     Messages with the \Deleted flag set.
    public static readonly ImapSearchCriteria Deleted
      = new ImapSearchCriteria(ToList("DELETED"));

    // DRAFT
    //     Messages with the \Draft flag set.
    public static readonly ImapSearchCriteria Draft
      = new ImapSearchCriteria(ToList("DRAFT"));

    // FLAGGED
    //     Messages with the \Flagged flag set.
    public static readonly ImapSearchCriteria Flagged
      = new ImapSearchCriteria(ToList("FLAGGED"));

    // FROM <string>
    //     Messages that contain the specified string in the envelope
    //     structure's FROM field.
    public static ImapSearchCriteria From(string @value)
    {
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "FROM", new ImapTextLiteralString(@value) });
    }

    // HEADER <field-name> <string>
    //     Messages that have a header with the specified field-name (as
    //     defined in [RFC-2822]) and that contains the specified string
    //     in the text of the header (what comes after the colon).  If the
    //     string to search is zero-length, this matches all messages that
    //     have a header line with the specified field-name regardless of
    //     the contents.
    public static ImapSearchCriteria Header(string fieldName, string @value)
    {
      if (fieldName == null)
        throw new ArgumentNullException(nameof(fieldName));
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "HEADER", ImapQuotedString.Create(fieldName), new ImapTextLiteralString(@value) });
    }

    // KEYWORD <flag>
    //     Messages with the specified keyword flag set.
    public static ImapSearchCriteria Keyword(ImapMessageFlag flag)
    {
      if (flag == null)
        throw new ArgumentNullException(nameof(flag));

      // TODO: optimize
      return new ImapSearchCriteria(new ImapString[] { "KEYWORD", flag });
    }

    public static ImapSearchCriteria Keyword(string keyword)
    {
      if (keyword == null)
        throw new ArgumentNullException(nameof(keyword));

      return new ImapSearchCriteria(ToList("KEYWORD", ImapMessageFlag.GetValidatedKeyword(keyword)));
    }

    // LARGER <n>
    //     Messages with an [RFC-2822] size larger than the specified
    //     number of octets.
    public static ImapSearchCriteria Larger(long size)
    {
      if (size < 0L)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(size), size);

      return new ImapSearchCriteria(ToList("LARGER", size.ToString()));
    }

    // NEW
    //     Messages that have the \Recent flag set but not the \Seen flag.
    //     This is functionally equivalent to "(RECENT UNSEEN)".
    public static readonly ImapSearchCriteria New
      = new ImapSearchCriteria(ToList("NEW"));

    // OLD
    //     Messages that do not have the \Recent flag set.  This is
    //     functionally equivalent to "NOT RECENT" (as opposed to "NOT
    //     NEW").
    public static readonly ImapSearchCriteria Old
      = new ImapSearchCriteria(ToList("OLD"));

    // ON <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is within the specified date.
    public static ImapSearchCriteria On(DateTime date)
    {
      return new ImapSearchCriteria(ToList("ON", ImapDateTimeFormat.ToDateString(date)));
    }

    // RECENT
    //     Messages that have the \Recent flag set.
    public static readonly ImapSearchCriteria Recent
      = new ImapSearchCriteria(ToList("RECENT"));

    // SEEN
    //     Messages that have the \Seen flag set.
    public static readonly ImapSearchCriteria Seen
      = new ImapSearchCriteria(ToList("SEEN"));

    // SENTBEFORE <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is earlier than the specified date.
    public static ImapSearchCriteria SentBefore(DateTime date)
    {
      return new ImapSearchCriteria(ToList("SENTBEFORE", ImapDateTimeFormat.ToDateString(date)));
    }

    // SENTON <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is within the specified date.
    public static ImapSearchCriteria SentOn(DateTime date)
    {
      return new ImapSearchCriteria(ToList("SENTON", ImapDateTimeFormat.ToDateString(date)));
    }

    // SENTSINCE <date>
    //     Messages whose [RFC-2822] Date: header (disregarding time and
    //     timezone) is within or later than the specified date.
    public static ImapSearchCriteria SentSince(DateTime date)
    {
      return new ImapSearchCriteria(ToList("SENTSINCE", ImapDateTimeFormat.ToDateString(date)));
    }

    // SINCE <date>
    //     Messages whose internal date (disregarding time and timezone)
    //     is within or later than the specified date.
    public static ImapSearchCriteria Since(DateTime date)
    {
      return new ImapSearchCriteria(ToList("SINCE", ImapDateTimeFormat.ToDateString(date)));
    }

    // SMALLER <n>
    //     Messages with an [RFC-2822] size smaller than the specified
    //     number of octets.
    public static ImapSearchCriteria Smaller(long size)
    {
      if (size < 0L)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(size), size);

      return new ImapSearchCriteria(ToList("SMALLER", size.ToString()));
    }

    // SUBJECT <string>
    //     Messages that contain the specified string in the envelope
    //     structure's SUBJECT field.
    public static ImapSearchCriteria Subject(string @value)
    {
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "SUBJECT", new ImapTextLiteralString(@value) });
    }

    // TEXT <string>
    //     Messages that contain the specified string in the header or
    //     body of the message.
    public static ImapSearchCriteria Text(string @value)
    {
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "TEXT", new ImapTextLiteralString(@value) });
    }

    // TO <string>
    //     Messages that contain the specified string in the envelope
    //     structure's TO field.
    public static ImapSearchCriteria To(string @value)
    {
      if (@value == null)
        throw new ArgumentNullException(nameof(value));

      return new ImapSearchCriteria(new ImapString[] { "TO", new ImapTextLiteralString(@value) });
    }

    // UNANSWERED
    //     Messages that do not have the \Answered flag set.
    public static readonly ImapSearchCriteria Unanswered
      = new ImapSearchCriteria(ToList("UNANSWERED"));

    // UNDELETED
    //     Messages that do not have the \Deleted flag set.
    public static readonly ImapSearchCriteria Undeleted
      = new ImapSearchCriteria(ToList("UNDELETED"));

    // UNDRAFT
    //     Messages that do not have the \Draft flag set.
    public static readonly ImapSearchCriteria Undraft
      = new ImapSearchCriteria(ToList("UNDRAFT"));

    // UNFLAGGED
    //     Messages that do not have the \Flagged flag set.
    public static readonly ImapSearchCriteria Unflagged
      = new ImapSearchCriteria(ToList("UNFLAGGED"));

    // UNKEYWORD <flag>
    //     Messages that do not have the specified keyword flag set.
    public static ImapSearchCriteria Unkeyword(ImapMessageFlag flag)
    {
      if (flag == null)
        throw new ArgumentNullException(nameof(flag));

      // TODO: optimize
      return new ImapSearchCriteria(new ImapString[] { "UNKEYWORD", flag });
    }

    public static ImapSearchCriteria Unkeyword(string keyword)
    {
      if (keyword == null)
        throw new ArgumentNullException(nameof(keyword));

      return new ImapSearchCriteria(ToList("UNKEYWORD", ImapMessageFlag.GetValidatedKeyword(keyword)));
    }

    // UNSEEN
    //     Messages that do not have the \Seen flag set.
    public static readonly ImapSearchCriteria Unseen
      = new ImapSearchCriteria(ToList("UNSEEN"));

    // NOT <search-key>
    //     Messages that do not match the specified search key.
    public ImapSearchCriteria Not()
    {
      return !this;
    }

    public static ImapSearchCriteria Not(ImapSearchCriteria c)
    {
      return !c;
    }

    public static ImapSearchCriteria operator! (ImapSearchCriteria c)
    {
      if (c == null)
        throw new ArgumentNullException(nameof(c));

      return new ImapSearchCriteria(c.requiredCapabilities,
                                    new ImapString[] { "NOT", ImapStringList.CreateParenthesized(c.Items) });
    }

    // OR <search-key1> <search-key2>
    //     Messages that match either search key.
    public ImapSearchCriteria Or(ImapSearchCriteria searchingCriteria)
    {
      return this | searchingCriteria;
    }

    public static ImapSearchCriteria Or(ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return x | y;
    }

    public static ImapSearchCriteria operator| (ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return new ImapSearchCriteria(GetCombinedCapabilities(x, y),
                                    new ImapString[] { "OR", ImapStringList.CreateParenthesized(x.Items), ImapStringList.CreateParenthesized(y.Items) });
    }

    // AND
    public ImapSearchCriteria And(ImapSearchCriteria searchingCriteria)
    {
      return this & searchingCriteria;
    }

    public static ImapSearchCriteria And(ImapSearchCriteria x, ImapSearchCriteria y)
    {
      return x & y;
    }

    public static ImapSearchCriteria operator& (ImapSearchCriteria x, ImapSearchCriteria y)
    {
      var (requiredCapabilities, items) = GetCombined(x, y);

      return new ImapSearchCriteria(requiredCapabilities, items);
    }

    /*
     * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
     * http://tools.ietf.org/html/rfc4551
     */

    // 3.4. MODSEQ Search Criterion in SEARCH
    //    The MODSEQ criterion for the SEARCH command allows a client to search
    //    for the metadata items that were modified since a specified moment.
    // 
    //    Syntax:  MODSEQ [<entry-name> <entry-type-req>] <mod-sequence-valzer>
    [CLSCompliant(false)]
    public static ImapSearchCriteria ModSeq(ulong modificationSequence)
    {
      return ModSeq(modificationSequence, null, null);
    }

    [CLSCompliant(false)]
    public static ImapSearchCriteria ModSeqSharedEntry(ulong modificationSequence, ImapMessageFlag flag)
    {
      if (flag == null)
        throw new ArgumentNullException(nameof(flag));

      return ModSeq(modificationSequence, flag, "shared");
    }

    [CLSCompliant(false)]
    public static ImapSearchCriteria ModSeqPrivateEntry(ulong modificationSequence, ImapMessageFlag flag)
    {
      if (flag == null)
        throw new ArgumentNullException(nameof(flag));

      return ModSeq(modificationSequence, flag, "priv");
    }

    [CLSCompliant(false)]
    public static ImapSearchCriteria ModSeqAllEntry(ulong modificationSequence, ImapMessageFlag flag)
    {
      if (flag == null)
        throw new ArgumentNullException(nameof(flag));

      return ModSeq(modificationSequence, flag, "all");
    }

    private static ImapSearchCriteria ModSeq(ulong modSequenceValzer, ImapMessageFlag entryFlagName, ImapString entryType)
    {
      if (entryType == null)
        return new ImapSearchCriteria(Singleton.CreateList(ImapCapability.CondStore),
                                      ToList("MODSEQ", modSequenceValzer.ToString()));
      else
        return new ImapSearchCriteria(Singleton.CreateList(ImapCapability.CondStore),
                                      new ImapString[] {
                                        "MODSEQ",
                                        ImapQuotedString.Create(string.Concat("/flags/", entryFlagName.Value)),
                                        entryType,
                                        modSequenceValzer.ToString()
                                      });
    }

    /*
     * RFC 5032 - WITHIN Search Extension to the IMAP Protocol
     * http://tools.ietf.org/html/rfc5032
     */

    // 3. Formal Syntax
    //    This document extends RFC 3501 [RFC3501] with two new search keys:
    //    OLDER <interval> and YOUNGER <interval>.
    // 
    //    search-key =/ ( "OLDER" / "YOUNGER" ) SP nz-number
    //                   ; search-key defined in RFC 3501
    public static ImapSearchCriteria Older(TimeSpan interval)
    {
      return Older((long)interval.TotalSeconds);
    }

    public static ImapSearchCriteria Older(long intervalSeconds)
    {
      if (intervalSeconds <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(intervalSeconds), intervalSeconds);

      return new ImapSearchCriteria(Singleton.CreateList(ImapCapability.Within), ToList("OLDER", intervalSeconds.ToString()));
    }

    public static ImapSearchCriteria Younger(TimeSpan interval)
    {
      return Younger((long)interval.TotalSeconds);
    }

    public static ImapSearchCriteria Younger(long intervalSeconds)
    {
      if (intervalSeconds <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(intervalSeconds), intervalSeconds);

      return new ImapSearchCriteria(Singleton.CreateList(ImapCapability.Within), ToList("YOUNGER", intervalSeconds.ToString()));
    }

    /*
     * RFC 5466 - IMAP4 Extension for Named Searches (Filters)
     * http://tools.ietf.org/html/rfc5466
     *
     * 4. Formal Syntax
     *    search-key            =/  "FILTER" SP filter-name
     *                          ;; New SEARCH criterion for referencing filters
     * 
     *    filter-name           =  1*<any ATOM-CHAR except "/">
     *                          ;; Note that filter-name disallows UTF-8 or
     *                          ;; the following characters: "(", ")", "{",
     *                          ;; " ", "%", "*", "]".  See definition of
     *                          ;; ATOM-CHAR [RFC3501].
     */
    public static ImapSearchCriteria Filter(string filterName)
    {
      return new ImapSearchCriteria(Singleton.CreateList(ImapCapability.Filters), ToList("FILTER", filterName));
    }

    /*
     * GetMessages by SEARCHing with X-GM-RAW
     * https://developers.google.com/gmail/imap_extensions
     */
    public static ImapSearchCriteria XGMRaw(string query)
    {
      return new ImapSearchCriteria(Singleton.CreateList(ImapCapability.GimapGmExt1), new ImapString[] { "X-GM-RAW", new ImapTextLiteralString(query) });
    }

    /*
     * instance members
     */
    private protected override bool IsParenthesized => false;

    private readonly IReadOnlyStringSet<ImapCapability> requiredCapabilities;
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities => requiredCapabilities;

    private ImapSearchCriteria(IReadOnlyCollection<ImapString> items)
      : this(ImapCapabilitySet.ReadOnlyEmpty, items)
    {
    }

    private ImapSearchCriteria(IReadOnlyCollection<ImapCapability> requiredCapabilities, IReadOnlyCollection<ImapString> items)
      : this(new ImapCapabilitySet(requiredCapabilities, asReadOnly: true), items)
    {
    }

    private ImapSearchCriteria(IReadOnlyStringSet<ImapCapability> requiredCapabilities, IReadOnlyCollection<ImapString> items)
      : base(items)
    {
      this.requiredCapabilities = requiredCapabilities;
    }

    internal bool SetCharset(Encoding charset)
    {
      var charsetSpecified = false;

      Traverse(this, s => {
        if (s is ImapTextLiteralString l) {
          l.Charset = charset;
          charsetSpecified = true;
        }
      });

      return charsetSpecified;
    }
  }
}
