// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  // combinable data item types:
  //   ImapCombinableDataItem
  //     => handles 'data item names' etc.
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //     ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapSortCriteria
  //       => handles 'sort criteria'
  //   * ImapStatusDataItem
  //       => handles 'status data item names'
  //     ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.3.10. STATUS Command
  //   status data item names
  public sealed class ImapStatusDataItems : ImapCombinableDataItems, IImapExtension {
    /// <summary>This field is equivalent to "MESSAGES RECENT UIDXNEXT UIDVALIDITY UNSEEN", contains no status data items defined in extensions.</summary>
    public static readonly ImapStatusDataItems StandardAll
      = new ImapStatusDataItems(ImapCapabilitySet.ReadOnlyEmpty, ToList("MESSAGES", "RECENT", "UIDNEXT", "UIDVALIDITY", "UNSEEN"));

    // MESSAGES
    public static readonly ImapStatusDataItems Messages
      = new ImapStatusDataItems(ImapCapabilitySet.ReadOnlyEmpty, ToList("MESSAGES"));

    // RECENT
    public static readonly ImapStatusDataItems Recent
      = new ImapStatusDataItems(ImapCapabilitySet.ReadOnlyEmpty, ToList("RECENT"));

    // UIDNEXT
    public static readonly ImapStatusDataItems UidNext
      = new ImapStatusDataItems(ImapCapabilitySet.ReadOnlyEmpty, ToList("UIDNEXT"));

    // UIDVALIDITY
    public static readonly ImapStatusDataItems UidValidity
      = new ImapStatusDataItems(ImapCapabilitySet.ReadOnlyEmpty, ToList("UIDVALIDITY"));

    // UNSEEN
    public static readonly ImapStatusDataItems Unseen
      = new ImapStatusDataItems(ImapCapabilitySet.ReadOnlyEmpty, ToList("UNSEEN"));

    /*
     * RFC 4551 - IMAP Extension for Conditional STORE Operation or Quick Flag Changes Resynchronization
     * http://tools.ietf.org/html/rfc4551
     */
    public static readonly ImapStatusDataItems HighestModSeq
      = new ImapStatusDataItems(new ImapCapabilitySet(new[] { ImapCapability.CondStore }, asReadOnly: true), ToList("HIGHESTMODSEQ"));

    /*
     * methods of combining
     */
    public ImapStatusDataItems CombineWith(ImapStatusDataItems dataItem)
    {
      return this + dataItem;
    }

    public static ImapStatusDataItems Combine(ImapStatusDataItems x, ImapStatusDataItems y)
    {
      return x + y;
    }

    public static ImapStatusDataItems operator+ (ImapStatusDataItems x, ImapStatusDataItems y)
    {
      var (requiredCapabilities, items) = GetCombined(x, y);

      return new ImapStatusDataItems(requiredCapabilities, items);
    }

    /*
     * instance members
     */
    private readonly IReadOnlyStringSet<ImapCapability> requiredCapabilities;
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities => requiredCapabilities;

    private ImapStatusDataItems(IReadOnlyStringSet<ImapCapability> requiredCapabilities, IReadOnlyCollection<ImapString> items)
      : base(items)
    {
      this.requiredCapabilities = requiredCapabilities;
    }
  }
}
