// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol {
  // combinable data item types:
  //   ImapCombinableDataItem
  //     => handles 'data item names' etc.
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //     ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapSortCriteria
  //       => handles 'sort criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //   * ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.6. STORE Command
  //   message data item name
  //   value for message data item
  public sealed class ImapStoreDataItems : ImapDataItems {
    private static readonly ImapString flagsReplace       = "FLAGS";
    private static readonly ImapString flagsReplaceSilent = "FLAGS.SILENT";
    private static readonly ImapString flagsAdd           = "+FLAGS";
    private static readonly ImapString flagsAddSilent     = "+FLAGS.SILENT";
    private static readonly ImapString flagsRemove        = "-FLAGS";
    private static readonly ImapString flagsRemoveSilent  = "-FLAGS.SILENT";

    // FLAGS <flag list>
    //     Replace the flags for the message (other than \Recent) with the
    //     argument.  The new value of the flags is returned as if a FETCH
    //     of those flags was done.
    public static ImapStoreDataItems ReplaceFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
      => Create(flagsReplace, Enumerable.Empty<string>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags));

    public static ImapStoreDataItems ReplaceFlags(string keyword, params string[] keywords)
      => Create(flagsReplace, ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords), Enumerable.Empty<ImapMessageFlag>());

    public static ImapStoreDataItems ReplaceFlags(IEnumerable<string> keywords, IEnumerable<ImapMessageFlag> flags)
      => Create(flagsReplace, keywords, flags);

    public static ImapStoreDataItems ReplaceFlags(IEnumerable<ImapMessageFlag> flags)
      => Create(flagsReplace, Enumerable.Empty<string>(), flags);

    // FLAGS.SILENT <flag list>
    //     Equivalent to FLAGS, but without returning a new value.
    public static ImapStoreDataItems ReplaceFlagsSilent(ImapMessageFlag flag, params ImapMessageFlag[] flags)
      => Create(flagsReplaceSilent, Enumerable.Empty<string>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags));

    public static ImapStoreDataItems ReplaceFlagsSilent(string keyword, params string[] keywords)
      => Create(flagsReplaceSilent, ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords), Enumerable.Empty<ImapMessageFlag>());

    public static ImapStoreDataItems ReplaceFlagsSilent(IEnumerable<string> keywords, IEnumerable<ImapMessageFlag> flags)
      => Create(flagsReplaceSilent, keywords, flags);

    public static ImapStoreDataItems ReplaceFlagsSilent(IEnumerable<ImapMessageFlag> flags)
      => Create(flagsReplaceSilent, Enumerable.Empty<string>(), flags);

    // +FLAGS <flag list>
    //     Add the argument to the flags for the message.  The new value
    //     of the flags is returned as if a FETCH of those flags was done.
    public static ImapStoreDataItems AddFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
      => Create(flagsAdd, Enumerable.Empty<string>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags));

    public static ImapStoreDataItems AddFlags(string keyword, params string[] keywords)
      => Create(flagsAdd, ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords), Enumerable.Empty<ImapMessageFlag>());

    public static ImapStoreDataItems AddFlags(IEnumerable<string> keywords, IEnumerable<ImapMessageFlag> flags)
      => Create(flagsAdd, keywords, flags);

    public static ImapStoreDataItems AddFlags(IEnumerable<ImapMessageFlag> flags)
      => Create(flagsAdd, Enumerable.Empty<string>(), flags);

    // +FLAGS.SILENT <flag list>
    //     Equivalent to +FLAGS, but without returning a new value.
    public static ImapStoreDataItems AddFlagsSilent(ImapMessageFlag flag, params ImapMessageFlag[] flags)
      => Create(flagsAddSilent, Enumerable.Empty<string>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags));

    public static ImapStoreDataItems AddFlagsSilent(string keyword, params string[] keywords)
      => Create(flagsAddSilent, ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords), Enumerable.Empty<ImapMessageFlag>());

    public static ImapStoreDataItems AddFlagsSilent(IEnumerable<string> keywords, IEnumerable<ImapMessageFlag> flags)
      => Create(flagsAddSilent, keywords, flags);

    public static ImapStoreDataItems AddFlagsSilent(IEnumerable<ImapMessageFlag> flags)
      => Create(flagsAddSilent, Enumerable.Empty<string>(), flags);

    // -FLAGS <flag list>
    //     Remove the argument from the flags for the message.  The new
    //     value of the flags is returned as if a FETCH of those flags was
    //     done.
    public static ImapStoreDataItems RemoveFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
      => Create(flagsRemove, Enumerable.Empty<string>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags));

    public static ImapStoreDataItems RemoveFlags(string keyword, params string[] keywords)
      => Create(flagsRemove, ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords), Enumerable.Empty<ImapMessageFlag>());

    public static ImapStoreDataItems RemoveFlags(IEnumerable<string> keywords, IEnumerable<ImapMessageFlag> flags)
      => Create(flagsRemove, keywords, flags);

    public static ImapStoreDataItems RemoveFlags(IEnumerable<ImapMessageFlag> flags)
      => Create(flagsRemove, Enumerable.Empty<string>(), flags);

    // -FLAGS.SILENT <flag list>
    //     Equivalent to -FLAGS, but without returning a new value.
    public static ImapStoreDataItems RemoveFlagsSilent(ImapMessageFlag flag, params ImapMessageFlag[] flags)
      => Create(flagsRemoveSilent, Enumerable.Empty<string>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags));

    public static ImapStoreDataItems RemoveFlagsSilent(string keyword, params string[] keywords)
      => Create(flagsRemoveSilent, ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords), Enumerable.Empty<ImapMessageFlag>());

    public static ImapStoreDataItems RemoveFlagsSilent(IEnumerable<string> keywords, IEnumerable<ImapMessageFlag> flags)
      => Create(flagsRemoveSilent, keywords, flags);

    public static ImapStoreDataItems RemoveFlagsSilent(IEnumerable<ImapMessageFlag> flags)
      => Create(flagsRemoveSilent, Enumerable.Empty<string>(), flags);

    // common construction method
    private static ImapStoreDataItems Create(ImapString itemName, IEnumerable<string> keywords, IEnumerable<ImapMessageFlag> flags)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      var items = new ImapMessageFlagSet(flags.Except(ImapMessageFlag.NonApplicableFlags), keywords);

      if (items.Count == 0)
        throw new ArgumentException("must contain at least one keyword or one applicable flag");

      return new ImapStoreDataItems(itemName, items);
    }

    /*
     * instance members
     */
    public ImapString ItemName { get; }

    private ImapStoreDataItems(ImapString itemName, IReadOnlyCollection<ImapString> items)
      : base(items)
    {
#if DEBUG
      if (itemName == null)
        throw new ArgumentNullException(nameof(itemName));
#endif
      ItemName = itemName;
    }

    public override string ToString()
    {
      return string.Concat(ItemName, " ", base.ToString());
    }
  }
}
