using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Threading;
using System.Text;
using System.Text.RegularExpressions;

namespace Smdn.Net.Imap4 {
  public class ImapPseudoServer : PseudoServer {
    private class Response : ResponseBase {
      public bool IsByeResponse { get; set; }
      public bool IsPushResponse { get; set; }
      public bool IsLiteralContinuationResponse { get; set; }
      public bool AcceptLiteral { get; set; }

      public Response(string expectedCommand,
                      string responseText,
                      Action actionBeforeRespond = null,
                      Action actionAfterRespond = null,
                      TimeSpan responseDelay = default)
        : base(expectedCommand,
               responseText,
               actionBeforeRespond: actionBeforeRespond,
               actionAfterRespond: actionAfterRespond,
               responseDelay: responseDelay)
      {
        IsByeResponse = false;
        IsPushResponse = false;
        IsLiteralContinuationResponse = false;
        AcceptLiteral = false;
      }
    }

    public ImapPseudoServer()
      : base()
    {
    }

    protected override void Initialize()
    {
      base.Initialize();

      responseTag = 0;

      SetGreetingResponse("* OK ready\r\n");
    }

    public static void Start(Action<ImapPseudoServer> action,
                             [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                             [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                             [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      Start(true, action, callerMemberName, sourceFilePath, sourceLinerNumber);
    }

    public static void Start(bool throwIfServerNotStopped,
                             Action<ImapPseudoServer> action,
                             [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                             [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                             [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start(true, throwIfServerNotStopped, action, callerMemberName, sourceFilePath, sourceLinerNumber);
      }
    }

    public static void StartNotAccepting(Action<ImapPseudoServer> action,
                                         [System.Runtime.CompilerServices.CallerMemberName] string callerMemberName = "",
                                         [System.Runtime.CompilerServices.CallerFilePath] string sourceFilePath = "",
                                         [System.Runtime.CompilerServices.CallerLineNumber] int sourceLinerNumber = 0)
    {
      using (var server = new ImapPseudoServer()) {
        server.Start(false, false, action, callerMemberName, sourceFilePath, sourceLinerNumber);
      }
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond = null,
                                    Action actionAfterRespond = null,
                                    bool isByeResponse = false)
    {
      if (greeting == null) {
        greetingResponse = null;
      }
      else {
        greetingResponse = new Response(expectedCommand: null,
                                        responseText: greeting,
                                        actionBeforeRespond: actionBeforeRespond,
                                        actionAfterRespond: actionAfterRespond);

        greetingResponse.IsByeResponse = isByeResponse;
      }
    }

    public void EnqueueTaggedResponse(string responseText)
      => EnqueueTaggedResponse(expectedCommand: null,
                               responseText: responseText);

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText,
                                      Action actionBeforeRespond = null,
                                      Action actionAfterRespond = null,
                                      bool isByeResponse = false,
                                      TimeSpan responseDelay = default)
    {
#if DEBUG
      if (!responseText.Contains("$tag"))
        throw new ArgumentException("is not tagged response", "responseText");
#endif

      responseText = responseText.Replace("$tag", (responseTag++).ToString("x4"));

      if (0x10000 == responseTag)
        responseTag = 0;

      var response = new Response(expectedCommand: expectedCommand,
                                  responseText: responseText,
                                  actionBeforeRespond: actionBeforeRespond,
                                  actionAfterRespond: actionAfterRespond,
                                  responseDelay: responseDelay);

      response.IsByeResponse = isByeResponse;

      InternalEnqueueResponse(response);
    }

    public void EnqueueUntaggedByeResponse(string responseText,
                                           bool asPushResponse = false)
    {
#if DEBUG
      if (!(responseText.StartsWith("* BYE") || responseText.Contains("\r\n* BYE")))
        throw new ArgumentException("is not BYE response", "responseText");
#endif

      var response = new Response(expectedCommand: null,
                                  responseText: responseText);

      response.IsByeResponse = true;
      response.IsPushResponse = asPushResponse;

      InternalEnqueueResponse(response);
    }

    public void EnqueueResponse(string responseText)
      => EnqueueResponse(expectedCommand: null,
                         responseText: responseText);

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond = null,
                                Action actionAfterRespond = null)
    {
      var response = new Response(expectedCommand: expectedCommand,
                                  responseText: responseText,
                                  actionBeforeRespond: actionBeforeRespond,
                                  actionAfterRespond: actionAfterRespond);

      InternalEnqueueResponse(response);
    }

    public void EnqueuePushResponse(string responseText,
                                    bool disconnectAfterSend = false)
    {
      var response = new Response(expectedCommand: null,
                                  responseText: responseText);

      response.IsPushResponse = true;
      response.IsByeResponse = disconnectAfterSend;

      InternalEnqueueResponse(response);
    }

    public void EnqueueLiteralContinuationResponse(string responseText,
                                                   bool acceptLiteral)
      => EnqueueLiteralContinuationResponse(expectedCommand: null,
                                            responseText: responseText,
                                            acceptLiteral: acceptLiteral);

    public void EnqueueLiteralContinuationResponse(string expectedCommand,
                                                   string responseText,
                                                   bool acceptLiteral)
    {
      if (responseText.Contains("$tag"))
        responseText = responseText.Replace("$tag", (responseTag++).ToString("x4"));

      var response = new Response(expectedCommand: expectedCommand,
                                  responseText: responseText);

      response.IsLiteralContinuationResponse = true;
      response.AcceptLiteral = acceptLiteral;

      InternalEnqueueResponse(response);
    }

    private void InternalEnqueueResponse(Response response)
    {
      lock (responseQueue)
        responseQueue.Enqueue(response);
    }

    // TODO: rename
    public string DequeueRequest()
    {
      return DequeueRequest(null);
    }

    public string DequeueRequest(Encoding encoding)
    {
      string ret;

      lock (commandQueue) {
        if (0 < commandQueue.Count)
          ret = commandQueue.Dequeue();
        else
          return null;
      }

      if (encoding == null)
        return ret;
      else
        return encoding.GetString(TransferCharset.GetBytes(ret));
    }

    protected override void ServerProc(TcpListener server, TcpClient client)
    {
      var commandStream = client.GetStream();

      if (greetingResponse != null) {
        Send(greetingResponse, commandStream);

        if (greetingResponse.IsByeResponse)
          return;
      }

      var commandReceiveBuffer = new byte[0x1000];
      var commandBuffer = new StringBuilder(0x1000);
      const long endOfLiteral = -1L;
      long literalOctetsToBeReceive = endOfLiteral;

      for (;;) {
        if (!IsConnected(client))
          return;

        Response pushResponse = null;

        lock (responseQueue) {
          if (0 < responseQueue.Count && responseQueue.Peek().IsPushResponse)
            pushResponse = responseQueue.Dequeue();
        }

        if (pushResponse != null) {
          Send(pushResponse, commandStream);

          if (pushResponse.IsByeResponse)
            return;
          else
            continue;
        }

        /*
         * wait for command sent
         */
        try {
          // XXX: NetworkStream.Read does not timeout on netfx
          if (Runtime.IsRunningOnNetFx && !commandStream.DataAvailable)
            continue;

          var commandLength = commandStream.Read(commandReceiveBuffer, 0, commandReceiveBuffer.Length);

          if (commandLength == 0 && !IsConnected(client))
            return;

          var received = TransferCharset.GetString(commandReceiveBuffer, 0, commandLength);

          WriteLog("C:", EscapeLogControlChars(received));

          commandBuffer.Append(received);

          literalOctetsToBeReceive -= commandLength;

          if (0 < literalOctetsToBeReceive) {
            continue; // continue to receive literal
          }
          else if (literalOctetsToBeReceive == 0L) {
            literalOctetsToBeReceive = endOfLiteral;
            continue; // receive next of literal
          }
        }
        catch (IOException ex) {
          if (!client.Connected)
            return; // disconnected

          var exSocket = ex.InnerException as SocketException;

          if (exSocket != null &&
              (exSocket.SocketErrorCode == SocketError.WouldBlock || exSocket.SocketErrorCode == SocketError.TimedOut)) {
            continue;
          }

          throw;
        }

        /*
         * process received command
         */
        var command = commandBuffer.ToString();

        var literalSizeMatches = literalSizeRegex.Matches(command);

        if (0 < literalSizeMatches.Count) {
          // command ends with literal
          var literalSizeMatch = literalSizeMatches[literalSizeMatches.Count - 1]; // last match

          literalOctetsToBeReceive = long.Parse(literalSizeMatch.Groups[2].Value);

          if ("+".Equals(literalSizeMatch.Groups[3].Value, StringComparison.Ordinal)) {
            // non-sync literal
            continue; // continue to receive literal
          }
          else {
            // sync literal
            Response literalContinuationResponse = null;

            lock (responseQueue) {
              if (0 < responseQueue.Count && responseQueue.Peek().IsLiteralContinuationResponse)
                literalContinuationResponse = responseQueue.Peek();
            }

            if (literalContinuationResponse != null) {
              if (literalContinuationResponse.AcceptLiteral) {
                Response r = null;

                lock (responseQueue)
                  r = responseQueue.Dequeue();

                Send(r, commandStream);

                continue; // continue to receive literal
              }
              else {
                literalOctetsToBeReceive = endOfLiteral;
              }
            }
            else {
              var continuationResponse = new Response(null, "+ continue\r\n", null, null);

              Send(continuationResponse, commandStream);

              continue; // continue to receive literal
            }
          }
        }

        if (!command.EndsWith("\r\n", StringComparison.Ordinal))
          // incomplete command
          continue;

        lock (commandQueue) {
          commandQueue.Enqueue(command);
        }

        /*
         * dequeue response
         */
        Response response = null;

        for (;;) {
          SetClientRequestReceived();

          lock (responseQueue) {
            if (0 < responseQueue.Count) {
              response = responseQueue.Dequeue();
              break;
            }
          }

          if (!IsConnected(client))
            return;

          Thread.Sleep(10);
        }

        /*
         * send response
         */
        Send(response, commandStream, command);

        commandBuffer.Length = 0;

        if (response.IsByeResponse)
          return;
      }
    }


    private readonly Queue<Response> responseQueue = new Queue<Response>();
    private readonly Queue<string> commandQueue = new Queue<string>();
    private Response greetingResponse;
    private int responseTag = 0;

    private static readonly Regex literalSizeRegex = new Regex(@"(~?)\{(\d+)(\+?)\}\r\n$", RegexOptions.Singleline | RegexOptions.Compiled | RegexOptions.CultureInvariant);
  }
}

