using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapClientEventsTests {
    [Test]
    public void TestExistingMessageCountChanged()
    {
      var selectResp =
        "* 10 EXISTS\r\n" +
        "* 3 RECENT\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(10L, mailbox.ExistingMessageCount);

        var raisedEventCount = 0;
        Exception caughtException = null;

        mailbox.Client.ExistingMessageCountChanged += (sender, e) => {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(mailbox, e.Mailbox);
            Assert.AreEqual(10L, e.PrevCount);
            Assert.AreEqual(12L, e.CurrentCount);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            raisedEventCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 12 EXISTS\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, raisedEventCount);
        Assert.AreEqual(12L, mailbox.ExistingMessageCount);
      });
    }

    [Test]
    public void TestExistMessageCountChanged()
    {
      var selectResp =
        "* 10 EXISTS\r\n" +
        "* 3 RECENT\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(10L, mailbox.ExistingMessageCount);

        var raisedObsoleteEventCount = 0;
        Exception caughtException = null;

#pragma warning disable 0618
        mailbox.Client.ExistMessageCountChanged += (sender, e) => {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(mailbox, e.Mailbox);
            Assert.AreEqual(10L, e.PrevCount);
            Assert.AreEqual(12L, e.CurrentCount);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            raisedObsoleteEventCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };
#pragma warning restore 0618

        var raisedEventCount = 0;

        mailbox.Client.ExistingMessageCountChanged += (sender, e) => raisedEventCount++;

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 12 EXISTS\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, raisedObsoleteEventCount);
        Assert.AreEqual(1, raisedEventCount);
        Assert.AreEqual(12L, mailbox.ExistingMessageCount);
      });
    }

    [Test]
    public void TestRecentMessageCountChanged()
    {
      var selectResp =
        "* 10 EXISTS\r\n" +
        "* 3 RECENT\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.RecentMessageCount);

        var raisedEventCount = 0;
        Exception caughtException = null;

        mailbox.Client.RecentMessageCountChanged += (sender, e) => {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(mailbox, e.Mailbox);
            Assert.AreEqual(3L, e.PrevCount);
            Assert.AreEqual(4L, e.CurrentCount);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            raisedEventCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 4 RECENT\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, raisedEventCount);
        Assert.AreEqual(4L, mailbox.RecentMessageCount);
      });
    }

    [Test]
    public void TestMessageStatusChanged_Refreshing_Raise()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* 2 FETCH (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* 3 FETCH (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.IsTrue(messages[0].IsSeen);
        Assert.IsFalse(messages[0].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsFalse(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[2].IsSeen);
        Assert.IsFalse(messages[2].IsMarkedAsDeleted);

        var raisedEventCount = 0;
        Exception caughtException = null;

        mailbox.Client.MessageStatusChanged += (sender, e) => {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(2, e.Messages.Count);

            Assert.AreEqual(1L, e.Messages[0].Sequence);
            Assert.IsTrue(e.Messages[0].IsSeen);
            Assert.IsTrue(e.Messages[0].IsMarkedAsDeleted);

            Assert.AreEqual(3L, e.Messages[1].Sequence);
            Assert.IsFalse(e.Messages[1].IsSeen);
            Assert.IsFalse(e.Messages[1].IsMarkedAsDeleted);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            raisedEventCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 1 FETCH (FLAGS (\\Seen \\Deleted))\r\n" +
                                     "* 3 FETCH (FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, raisedEventCount);

        Assert.IsTrue(messages[0].IsSeen);
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsFalse(messages[1].IsMarkedAsDeleted);
        Assert.IsFalse(messages[2].IsSeen);
        Assert.IsFalse(messages[2].IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestMessageStatusChanged_Refreshing_Raise_WithMessageDeletion()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* 2 FETCH (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* 3 FETCH (UID 3 FLAGS (\\Seen))\r\n" +
                                     "* 4 FETCH (UID 4 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.IsTrue(messages[0].IsSeen);
        Assert.IsFalse(messages[0].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsFalse(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[2].IsSeen);
        Assert.IsFalse(messages[2].IsMarkedAsDeleted);
        Assert.IsTrue(messages[3].IsSeen);
        Assert.IsFalse(messages[3].IsMarkedAsDeleted);

        var eventMessageStatusChangedRaisedCount = 0;
        var eventMessageDeletedRaisedCount = 0;
        Exception caughtExceptionInMessageStatusChanged = null;
        Exception caughtExceptionInMessageDeleted = null;

        mailbox.Client.MessageStatusChanged += (sender, e) => {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(1, e.Messages.Count);

            Assert.AreSame(messages[2], e.Messages[0]);
            Assert.AreEqual(2L, e.Messages[0].Sequence); // 3 -> (2 EXPUNGE) -> 2
            Assert.AreEqual(3L, e.Messages[0].Uid);
            Assert.IsFalse(e.Messages[0].IsSeen);
            Assert.IsFalse(e.Messages[0].IsMarkedAsDeleted);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            eventMessageStatusChangedRaisedCount++;
          }
          catch (Exception ex) {
            caughtExceptionInMessageStatusChanged = ex;
          }
        };

        mailbox.Client.MessageDeleted += (sender, e) => {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(1, e.Messages.Count);

            Assert.AreSame(messages[1], e.Messages[0]);
            Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, e.Messages[0].Sequence);
            Assert.AreEqual(2L, e.Messages[0].Uid);
            Assert.IsTrue(e.Messages[0].IsDeleted);

            eventMessageDeletedRaisedCount++;
          }
          catch (Exception ex) {
            caughtExceptionInMessageDeleted = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 3 FETCH (FLAGS ())\r\n" +
                                     "* 2 EXPUNGE\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtExceptionInMessageStatusChanged);
        Assert.IsNull(caughtExceptionInMessageDeleted);

        Assert.AreEqual(1, eventMessageStatusChangedRaisedCount);
        Assert.AreEqual(1, eventMessageDeletedRaisedCount);

        Assert.AreEqual(1L, messages[0].Sequence); // 1 -> (2 EXPUNGE) -> 1
        Assert.IsTrue(messages[0].IsSeen);
        Assert.IsFalse(messages[0].IsMarkedAsDeleted);

        Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, messages[1].Sequence); // 2 -> (2 EXPUNGE) -> 0(vanished)
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.AreEqual(2L, messages[2].Sequence); // 3 -> (2 EXPUNGE) -> 2
        Assert.IsFalse(messages[2].IsSeen);
        Assert.IsFalse(messages[2].IsMarkedAsDeleted);

        Assert.AreEqual(3L, messages[3].Sequence); // 4 -> (2 EXPUNGE) -> 3
        Assert.IsTrue(messages[3].IsSeen);
        Assert.IsFalse(messages[3].IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestMessageStatusChanged_RefreshingMessage_Raise_NonManagingMessages()
    {
      MessageStatusChanged_RefreshingMessage(false);
    }

    [Test]
    public void TestMessageStatusChanged_RefreshingMessage_Raise_ManagingMessages()
    {
      MessageStatusChanged_RefreshingMessage(true);
    }

    private void MessageStatusChanged_RefreshingMessage(bool dynamicAttributeFetched)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(1L, mailbox.ExistingMessageCount);

        ImapMessage message;

        if (dynamicAttributeFetched) {
          server.EnqueueTaggedResponse("FETCH 1 (UID FLAGS)\r\n",
                                       "* 1 FETCH (UID 1 FLAGS ())\r\n" +
                                       "$tag OK done\r\n");

          message = mailbox.GetMessageBySequence(1L, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

          Assert.IsFalse(message.IsSeen);
        }
        else {
          server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                       "* 1 FETCH (UID 1)\r\n" +
                                       "$tag OK done\r\n");

          message = mailbox.GetMessageBySequence(1L, ImapMessageFetchAttributeOptions.None); // fetch no attributes (except UID)
        }

        var raisedEventCount = 0;
        IReadOnlyList<ImapMessage> statusChangedMessages = null;

        mailbox.Client.MessageStatusChanged += (sender, e) => {
          raisedEventCount++;
          statusChangedMessages = e.Messages;
        };

        server.EnqueueTaggedResponse("FETCH 1 (FLAGS)\r\n",
                                     "* 1 FETCH (FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        message.Refresh();

        Assert.IsTrue(message.IsSeen);

        Assert.AreEqual(1, raisedEventCount); // event must be raised
        Assert.IsNotNull(statusChangedMessages);
        Assert.AreEqual(1, statusChangedMessages.Count);
        Assert.AreSame(message, statusChangedMessages[0]);
      });
    }

    [Test]
    public void TestMessageStatusChanged_FetchingBody_NotRaise_NonManagingMessages()
    {
      MessageStatusChanged_FetchingBody(false);
    }

    [Test]
    public void TestMessageStatusChanged_FetchingBody_NotRaise_ManagingMessages()
    {
      MessageStatusChanged_FetchingBody(true);
    }

    private void MessageStatusChanged_FetchingBody(bool dynamicAttributeFetched)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      var messageBody = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

test message

3 uid fetch 6 (body[])
* 6 FETCH (UID 6 BODY[] {1476}
hogehogehoge
)
3 OK Fetch completed.
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(1L, mailbox.ExistingMessageCount);

        ImapMessage message;

        if (dynamicAttributeFetched) {
          server.EnqueueTaggedResponse("FETCH 1 (UID FLAGS)\r\n",
                                       "* 1 FETCH (UID 1 FLAGS ())\r\n" +
                                       "$tag OK done\r\n");

          message = mailbox.GetMessageBySequence(1L, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

          Assert.IsFalse(message.IsSeen);
        }
        else {
          server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                       "* 1 FETCH (UID 1)\r\n" +
                                       "$tag OK done\r\n");

          message = mailbox.GetMessageBySequence(1L, ImapMessageFetchAttributeOptions.None); // fetch no attributes (except UID)
        }

        var raisedEventCount = 0;
        IReadOnlyList<ImapMessage> statusChangedMessages = null;

        mailbox.Client.MessageStatusChanged += (sender, e) => {
          raisedEventCount++;
          statusChangedMessages = e.Messages;
        };

        var octets = OctetEncoding.EightBits.GetByteCount(messageBody);

        server.EnqueueTaggedResponse("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n",
                                     $"* {message.Uid} FETCH (RFC822.SIZE {octets} BODY[] {{{octets}}}\r\n{messageBody})\r\n" +
                                     "$tag OK done\r\n");

        message.Body.WriteTo(stream: Stream.Null, options: ImapMessageFetchBodyOptions.Default);

        Assert.AreEqual(0, raisedEventCount); // event must not be raised
        Assert.IsNull(statusChangedMessages);
      });
    }

    [Test]
    public void TestMessageStatusChanged_FetchingBody_NotRaise_ByResponseWithDynamicAttributes_NonManagingMessages()
    {
      MessageStatusChanged_FetchingBody_ResponseWithDynamicAttributes(false);
    }

    [Test]
    public void TestMessageStatusChanged_FetchingBody_Raise_ByResponseWithDynamicAttributes_ManagingMessages()
    {
      MessageStatusChanged_FetchingBody_ResponseWithDynamicAttributes(true);
    }

    private void MessageStatusChanged_FetchingBody_ResponseWithDynamicAttributes(bool dynamicAttributeFetched)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      var messageBody = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

test message

3 uid fetch 6 (body[])
* 6 FETCH (UID 6 BODY[] {1476}
hogehogehoge
)
3 OK Fetch completed.
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(1L, mailbox.ExistingMessageCount);

        ImapMessage message;

        if (dynamicAttributeFetched) {
          server.EnqueueTaggedResponse("FETCH 1 (UID FLAGS)\r\n",
                                       "* 1 FETCH (UID 1 FLAGS ())\r\n" +
                                       "$tag OK done\r\n");

          message = mailbox.GetMessageBySequence(1L, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

          Assert.IsFalse(message.IsSeen);
        }
        else {
          server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                       "* 1 FETCH (UID 1)\r\n" +
                                       "$tag OK done\r\n");

          message = mailbox.GetMessageBySequence(1L, ImapMessageFetchAttributeOptions.None); // fetch no attributes (except UID)
        }

        var raisedEventCount = 0;
        IReadOnlyList<ImapMessage> statusChangedMessages = null;

        mailbox.Client.MessageStatusChanged += (sender, e) => {
          raisedEventCount++;
          statusChangedMessages = e.Messages;
        };

        var octets = OctetEncoding.EightBits.GetByteCount(messageBody);

        server.EnqueueTaggedResponse("UID FETCH 1 (RFC822.SIZE BODY[]<0.10240>)\r\n",
                                     $"* {message.Uid} FETCH (RFC822.SIZE {octets} FLAGS (\\Seen) BODY[] {{{octets}}}\r\n{messageBody})\r\n" +
                                     "$tag OK done\r\n");

        message.Body.WriteTo(stream: Stream.Null, options: ImapMessageFetchBodyOptions.SetSeen);

        if (dynamicAttributeFetched) {
          Assert.AreEqual(1, raisedEventCount); // event must be raised
          Assert.AreEqual(1, statusChangedMessages.Count);
          Assert.AreSame(message, statusChangedMessages[0]);
          Assert.IsTrue(message.IsSeen);
        }
        else {
          Assert.AreEqual(0, raisedEventCount); // event must not be raised
        }
      });
    }

    [Test]
    public void TestMessageStatusChanged_Refreshing_NotRaise_ByMessagesOutOfManaging()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        var raisedEventCount = 0;

        mailbox.Client.MessageStatusChanged += (sender, e) => raisedEventCount++;

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 1 FETCH (FLAGS (\\Seen \\Deleted))\r\n" +
                                     "* 3 FETCH (FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.AreEqual(0, raisedEventCount);
      });
    }

    [Test]
    public void TestMessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithNoAttributes()
    {
      MessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithNoAttributes(false);
    }

    [Test]
    public void TestMessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithNoAttributes_ManagingMessages()
    {
      MessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithNoAttributes(true);
    }

    private void MessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithNoAttributes(bool dynamicAttributeFetched)
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        IReadOnlyList<ImapMessage> messages = null;

        if (dynamicAttributeFetched) {
          // FETCH
          server.EnqueueTaggedResponse("UID FETCH 1:3 (UID FLAGS)\r\n",
                                       "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                       "* 2 FETCH (UID 2 FLAGS (\\Seen))\r\n" +
                                       "* 3 FETCH (UID 3 FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");

          messages = mailbox.GetMessages(new[] { 1L, 2L, 3L }, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

          server.DequeueRequest();
        }

        var raisedEventCount = 0;

        mailbox.Client.MessageStatusChanged += (sender, e) => raisedEventCount++;

        // FETCH
        server.EnqueueTaggedResponse("UID FETCH 1:3 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "* 2 FETCH (UID 2)\r\n" +
                                     "* 3 FETCH (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        messages = mailbox.GetMessages(new[] { 1L, 2L, 3L }, ImapMessageFetchAttributeOptions.None); // fetch no attributes (except UID)

        server.DequeueRequest();

        Assert.AreEqual(0, raisedEventCount); // event must not be raised
      });
    }

    [Test]
    public void TestMessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithStaticAttributes()
    {
      MessageStatusChanged_YieldingMessages_ByYieldingWithStaticAttributes(false);
    }

    [Test]
    public void TestMessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithStaticAttributes_ManagingMessages()
    {
      MessageStatusChanged_YieldingMessages_ByYieldingWithStaticAttributes(true);
    }

    private void MessageStatusChanged_YieldingMessages_ByYieldingWithStaticAttributes(bool dynamicAttributeFetched)
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        IReadOnlyList<ImapMessage> messages = null;

        if (dynamicAttributeFetched) {
          // FETCH
          server.EnqueueTaggedResponse("UID FETCH 1:3 (UID FLAGS)\r\n",
                                       "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                       "* 2 FETCH (UID 2 FLAGS (\\Seen))\r\n" +
                                       "* 3 FETCH (UID 3 FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");

          messages = mailbox.GetMessages(new[] { 1L, 2L, 3L }, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

          server.DequeueRequest();
        }

        var raisedEventCount = 0;

        mailbox.Client.MessageStatusChanged += (sender, e) => raisedEventCount++;

        // FETCH
        var bodystructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL)";
        var envelope = "(" +
                       "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
                       "\"test message\" " +
                       "((\"from address\" NIL \"from\" \"mail.example.com\")) " +
                       "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " +
                       "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " +
                       "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                       "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                       "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                       "\"<in.reply.to@mail.example.com>\" " +
                       "\"<message.id@mail.example.com>\" " +
                       ")";
        var internaldate = "\"23-Mar-2010 09:46:52 +0900\"";

        server.EnqueueTaggedResponse("UID FETCH 1:3 (UID BODYSTRUCTURE ENVELOPE INTERNALDATE RFC822.SIZE)\r\n",
                                     $"* 1 FETCH (UID 1 BODYSTRUCTURE {bodystructure} ENVELOPE {envelope} INTERNALDATE {internaldate} RFC822.SIZE 1024)\r\n" +
                                     $"* 2 FETCH (UID 2 BODYSTRUCTURE {bodystructure} ENVELOPE {envelope} INTERNALDATE {internaldate} RFC822.SIZE 1024)\r\n" +
                                     $"* 3 FETCH (UID 3 BODYSTRUCTURE {bodystructure} ENVELOPE {envelope} INTERNALDATE {internaldate} RFC822.SIZE 1024)\r\n" +
                                     "$tag OK done\r\n");

        messages = mailbox.GetMessages(new[] { 1L, 2L, 3L }, ImapMessageFetchAttributeOptions.StaticAttributes); // fetch only static attributes

        server.DequeueRequest();

        Assert.AreEqual(0, raisedEventCount); // event must not be raised
      });
    }

    [Test]
    public void TestMessageStatusChanged_YieldingMessages_NotRaise_ByYieldingWithDynamicAttributes()
    {
      MessageStatusChanged_YieldingMessages_WithDynamicAttributes(false);
    }

    [Test]
    public void TestMessageStatusChanged_YieldingMessages_Raise_ByYieldingWithDynamicAttributes_ManagingMessages()
    {
      MessageStatusChanged_YieldingMessages_WithDynamicAttributes(true);
    }

    private void MessageStatusChanged_YieldingMessages_WithDynamicAttributes(bool yieldAgain)
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        var raisedEventCount = 0;
        IReadOnlyList<ImapMessage> statusChangedMessages = null;

        mailbox.Client.MessageStatusChanged += (sender, e) => {
          raisedEventCount++;
          statusChangedMessages = e.Messages;
        };

        // FETCH
        server.EnqueueTaggedResponse("UID FETCH 1:3 (UID FLAGS)\r\n",
                                     "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* 2 FETCH (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* 3 FETCH (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(new[] { 1L, 2L, 3L }, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

        server.DequeueRequest();

        Assert.AreEqual(0, raisedEventCount); // event must not be raised
        Assert.IsNull(statusChangedMessages);

        if (yieldAgain) {
          // FETCH
          server.EnqueueTaggedResponse("UID FETCH 1:3 (UID FLAGS)\r\n",
                                       "* 1 FETCH (UID 1 FLAGS (\\Seen $label1))\r\n" +
                                       "* 2 FETCH (UID 2 FLAGS (\\Seen $label2))\r\n" +
                                       "* 3 FETCH (UID 3 FLAGS (\\Seen $label3))\r\n" +
                                       "$tag OK done\r\n");

          messages = mailbox.GetMessages(new[] { 1L, 2L, 3L }, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

          server.DequeueRequest();

          Assert.AreEqual(1, raisedEventCount); // event must be raised
          Assert.IsNotNull(statusChangedMessages);
          Assert.AreEqual(3, statusChangedMessages.Count);
          CollectionAssert.Contains(statusChangedMessages, messages[0]);
          CollectionAssert.Contains(statusChangedMessages, messages[1]);
          CollectionAssert.Contains(statusChangedMessages, messages[2]);
        }
      });
    }

    [Test]
    public void TestMessageStatusChanged_FetchingAttributes_NotRaise_ByFetchingStaticAttributes()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        var raisedEventCount = 0;

        mailbox.Client.MessageStatusChanged += (sender, e) => raisedEventCount++;

        // FETCH
        server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS)\r\n",
                                     "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes); // fetch only dymamic attributes (except UID)

        server.DequeueRequest();

        Assert.AreEqual(0, raisedEventCount); // event must not be raised
        Assert.IsTrue(message.IsSeen);

        // run EnsureStaticAttributesFetched / FETCH (msg-att-static) will be issued
        var bodystructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL)";
        var envelope = "(" +
                       "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
                       "\"test message\" " +
                       "((\"from address\" NIL \"from\" \"mail.example.com\")) " +
                       "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " +
                       "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " +
                       "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                       "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                       "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                       "\"<in.reply.to@mail.example.com>\" " +
                       "\"<message.id@mail.example.com>\" " +
                       ")";
        var internaldate = "\"23-Mar-2010 09:46:52 +0900\"";

        server.EnqueueTaggedResponse("UID FETCH 1 (BODYSTRUCTURE ENVELOPE INTERNALDATE RFC822.SIZE)\r\n",
                                     $"* 1 FETCH (UID 1 BODYSTRUCTURE {bodystructure} ENVELOPE {envelope} INTERNALDATE {internaldate} RFC822.SIZE 1024)\r\n" +
                                     "$tag OK done\r\n");

        Assert.AreEqual(1024L, message.Length);

        server.DequeueRequest();

        Assert.AreEqual(0, raisedEventCount); // event must not be raised
      });
    }

    [Test]
    public void TestMessageStatusChanged_FetchingAttributes_Raise_ByFetchingDynamicAttributes()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        var raisedEventCount = 0;
        IReadOnlyList<ImapMessage> statusChangedMessages = null;

        mailbox.Client.MessageStatusChanged += (sender, e) => {
          raisedEventCount++;
          statusChangedMessages = e.Messages;
        };

        // FETCH
        server.EnqueueTaggedResponse("UID FETCH 1 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.None); // fetch no attributes (except UID)

        server.DequeueRequest();

        Assert.AreEqual(0, raisedEventCount); // event must not be raised

        // run EnsureDynamicAttributesFetched / FETCH (msg-att-dynamic) will be issued
        server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                     "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(message.IsSeen);

        server.DequeueRequest();

        Assert.AreEqual(1, raisedEventCount); // event must be raised
        Assert.AreEqual(1, statusChangedMessages.Count);
        Assert.AreSame(message, statusChangedMessages[0]);
      });
    }

    [Test]
    public void TestMessageDeleted()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* 2 FETCH (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* 3 FETCH (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsFalse(messages[0].IsDeleted);
        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsFalse(messages[0].IsDeleted);
        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsFalse(messages[0].IsDeleted);

        var raisedEventCount = 0;
        Exception caughtException = null;

        mailbox.Client.MessageDeleted += (sender, e) => {
          try {
            Assert.IsFalse(mailbox.Client.IsBusy);

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(2, e.Messages.Count);

            Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, e.Messages[0].Sequence);
            Assert.AreEqual(3L, e.Messages[0].Uid);
            Assert.IsTrue(e.Messages[0].IsDeleted);

            Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, e.Messages[1].Sequence);
            Assert.AreEqual(1L, e.Messages[1].Uid);
            Assert.IsTrue(e.Messages[1].IsDeleted);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => mailbox.Client.Refresh(), "new operation from event handler");

            raisedEventCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 3 EXPUNGE\r\n" +
                                     "* 1 EXPUNGE\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, raisedEventCount);

        Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsTrue(messages[0].IsDeleted);
        Assert.AreEqual(1L, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsFalse(messages[1].IsDeleted);
        Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestAlertReceived_WithUntaggedOkResponse()
    {
      EventAlertReceived("* OK [ALERT] operation successed\r\n" +
                         "$tag OK done\r\n",
                         "operation successed",
                         ImapResponseCondition.Ok);
    }

    [Test]
    public void TestAlertReceived_WithUntaggedNoResponse()
    {
      EventAlertReceived("* NO [ALERT] operation failed\r\n" +
                         "$tag OK done\r\n",
                         "operation failed",
                         ImapResponseCondition.No);
    }

    [Test]
    public void TestAlertReceived_WithTaggedOkResponse()
    {
      EventAlertReceived("$tag OK [ALERT] done\r\n",
                         "done",
                         ImapResponseCondition.Ok);
    }

    private void EventAlertReceived(string noopResponse,
                                    string expectedAlert,
                                    ImapResponseCondition expectedCondition)
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        var raisedEventCount = 0;
        Exception caughtException = null;

        client.AlertReceived += (sender, e) => {
          try {
            Assert.IsFalse(client.IsBusy);

            Assert.AreSame(client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(expectedAlert, e.Alert);
            Assert.AreEqual(expectedCondition, e.Condition);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            Assert.DoesNotThrow(() => client.Refresh(), "new operation from event handler");

            raisedEventCount++;
          }
          catch (Exception ex) {
            caughtException = ex;
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     noopResponse);

        client.Refresh();

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, raisedEventCount);
      });
    }

    [Test]
    public void TestEvents_WhileIdling()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistingMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* 1 FETCH (UID 1 FLAGS (\\Seen))\r\n" +
                                     "* 2 FETCH (UID 2 FLAGS (\\Seen))\r\n" +
                                     "* 3 FETCH (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes);

        var raisedEventCountMessageDeleted = 0;
        var raisedEventCountRecentMessageCountChanged = 0;
        Exception caughtExceptionInMessageDeleted = null;
        Exception caughtExceptionInRecentMessageCountChanged = null;

        mailbox.Client.MessageDeleted += (sender, e) => {
          try {
            Assert.IsTrue(mailbox.Client.IsBusy, "client.IsBusy");
            Assert.IsTrue(mailbox.Client.IsIdling, "client.IsIdling");
            Assert.IsTrue(mailbox.IsIdling, "mailbox.IsIdling");

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);

            Assert.IsNotNull(e.Messages);
            Assert.AreEqual(1, e.Messages.Count);
            Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, e.Messages[0].Sequence);
            Assert.AreEqual(2L, e.Messages[0].Uid);
            Assert.IsTrue(e.Messages[0].IsDeleted);

            Assert.AreEqual(1L, messages[0].Sequence);
            Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, messages[1].Sequence);
            Assert.AreEqual(2L, messages[2].Sequence);

            Assert.Throws<InvalidOperationException>(() => mailbox.Refresh());

            raisedEventCountMessageDeleted++;
          }
          catch (Exception ex) {
            caughtExceptionInMessageDeleted = ex;
          }
        };

        mailbox.Client.RecentMessageCountChanged += (sender, e) => {
          try {
            Assert.IsTrue(mailbox.Client.IsBusy, "client.IsBusy");
            Assert.IsTrue(mailbox.Client.IsIdling, "client.IsIdling");
            Assert.IsTrue(mailbox.IsIdling, "mailbox.IsIdling");

            Assert.AreSame(mailbox.Client, sender);

            Assert.IsNotNull(e);
            Assert.AreEqual(mailbox, e.Mailbox);
            Assert.AreEqual(0L, e.PrevCount);
            Assert.AreEqual(1L, e.CurrentCount);

            Assert.Throws<InvalidOperationException>(() => mailbox.Refresh());

            raisedEventCountRecentMessageCountChanged++;
          }
          catch (Exception ex) {
            caughtExceptionInRecentMessageCountChanged = ex;
          }
        };

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        server.EnqueueResponse("IDLE\r\n",
                               "+ idling\r\n");

        mailbox.StartIdle();

        server.EnqueuePushResponse("* 4 EXISTS\r\n");
        server.EnqueuePushResponse("* 1 RECENT\r\n");
        server.EnqueuePushResponse("* 2 EXPUNGE\r\n");

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");

        Assert.IsNull(caughtExceptionInMessageDeleted);
        Assert.IsNull(caughtExceptionInRecentMessageCountChanged);

        Assert.AreEqual(1, raisedEventCountMessageDeleted, "MessageDeleted raised count");
        Assert.AreEqual(1, raisedEventCountRecentMessageCountChanged, "RecentMessageCountChanged raised count");
      });
    }

    [Test]
    public void TestEvents_IgnoreExceptionsFromEventHandler()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        var raisedEventCount = 0;

        client.AlertReceived += (sender, e) => {
          raisedEventCount++;

          throw new InvalidOperationException("this exception will be ignored");
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] done\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(1, raisedEventCount);
      });
    }

    [Test]
    public void TestEvents_RaisedInEachOperation()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        var receivedAlerts = new List<string>();

        client.AlertReceived += (sender, e) => receivedAlerts.Add(e.Alert);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] 1st\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(1, receivedAlerts.Count);
        Assert.AreEqual("1st", receivedAlerts[0]);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] 2nd\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(2, receivedAlerts.Count);
        Assert.AreEqual("2nd", receivedAlerts[1]);
      });
    }

    [Test]
    public void TestEvents_NoEventsRaised()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        var raisedEventCount = 0;

        client.AlertReceived += (sender, e) => raisedEventCount++;

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        Assert.AreEqual(0, raisedEventCount);
      });
    }

#if NETFRAMEWORK || NETSTANDARD2_0
    private class PseudoSynchronizingObject : ISynchronizeInvoke {
      public bool InvokeRequired {
        get; private set;
      }

      private bool expectSynchronousCall;

      public PseudoSynchronizingObject(bool invokeRequired, bool expectSynchronousCall)
      {
        this.InvokeRequired = invokeRequired;
        this.expectSynchronousCall = expectSynchronousCall;
      }

      public IAsyncResult BeginInvoke(Delegate method, object[] args)
      {
        if (expectSynchronousCall)
          throw new InvalidOperationException("synchronous call expected");

        ThreadPool.QueueUserWorkItem(delegate {
          method.DynamicInvoke(args);
        });

        return null;
      }

      public object EndInvoke(IAsyncResult result)
      {
        throw new NotImplementedException();
      }

      public object Invoke(Delegate method, object[] args)
      {
        if (!expectSynchronousCall)
          throw new InvalidOperationException("asynchronous call expected");

        return method.DynamicInvoke(args);
      }
    }

    [Test]
    public void TestEvents_RaiseEventWithSynchronizingObject_Null()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        const bool expectSynchronousCall = true;

        client.SynchronizingObject = null;

        HandleEventWithSynchronizingObject(server, client, expectSynchronousCall);
      });
    }

    [Test]
    public void TestEvents_RaiseEventWithSynchronizingObject_InvokeRequired()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        const bool expectSynchronousCall = false;

        client.SynchronizingObject = new PseudoSynchronizingObject(true, expectSynchronousCall);

        HandleEventWithSynchronizingObject(server, client, expectSynchronousCall);
      });
    }

    [Test]
    public void TestEvents_RaiseEventWithSynchronizingObject_InvokeNotRequired()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        const bool expectSynchronousCall = true;

        client.SynchronizingObject = new PseudoSynchronizingObject(false, expectSynchronousCall);

        HandleEventWithSynchronizingObject(server, client, expectSynchronousCall);
      });
    }

    private void HandleEventWithSynchronizingObject(ImapPseudoServer server, ImapClient client, bool expectSynchronousCall)
    {
      using (var waitHandle = new ManualResetEvent(false)) {
        var raisedEventCount = 0;
        var eventHandlerThreadID = -1;
        Exception caughtException = null;

        client.AlertReceived += (sender, e) => {
          eventHandlerThreadID = Thread.CurrentThread.ManagedThreadId;

          try {
            try {
              Assert.AreSame(client, sender);

              Assert.IsNotNull(e);
              Assert.AreEqual("done",e.Alert);

              raisedEventCount++;
            }
            catch (Exception ex) {
              caughtException = ex;
            }
          }
          finally {
            waitHandle.Set();
          }
        };

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK [ALERT] done\r\n");

        Assert.DoesNotThrow(() => client.Refresh());

        if (!waitHandle.WaitOne(1000)) // wait for event handler finished
          Assert.Fail("event handler not called or not finished");

        Assert.IsNull(caughtException);

        Assert.AreEqual(1, raisedEventCount);

        if (expectSynchronousCall)
          Assert.AreEqual(Thread.CurrentThread.ManagedThreadId,
                          eventHandlerThreadID);
        else
          Assert.AreNotEqual(Thread.CurrentThread.ManagedThreadId,
                             eventHandlerThreadID);
      }
    }
#endif
  }
}

