using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageBodyTests : ImapMessageBodyBaseTestsBase {
    [SetUp]
    public void SetUp()
    {
      TestUtils.Encodings.RegisterCodePagesEncodingProvider();
    }

    private void TestArgumentNull(string expectedParamName, Action<ImapMessageBody> testAction)
    {
      TestCases.WithMessage((server, message) => {
        var ex = Assert.Throws<ArgumentNullException>(() => testAction(message.Body));

        Assert.AreEqual(expectedParamName, ex.ParamName);
      });
    }

    [Test]
    public void TestReadMessageBody_Peek_OptionsDefault_OfMessage()
    {
      var initialFlags = new[] {"$label1"};

      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), initialFlags, testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Peek_OptionsDefault(message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Peek_OptionsDefault_OfMessagePart()
    {
      var initialFlags = new[] {"$label1"};

      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), initialFlags: initialFlags, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Peek_OptionsDefault(messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Peek_OptionsDefault(ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      Assert.AreEqual(1, body.Message.Flags.Count);
      Assert.IsFalse(body.Message.Flags.Contains(ImapMessageFlag.Seen));
      Assert.IsTrue(body.Message.Flags.Contains("$label1"));
      Assert.IsFalse(body.Message.IsSeen);

      messageBodyGenerator.EmitNextResponse(server);

      body.Read(options: ImapMessageFetchBodyOptions.Default, read: (Stream stream) => {
        StringAssert.EndsWith($"UID FETCH 1 ({(body.IsOfTopLevel ? "RFC822.SIZE " : "")}BODY.PEEK[{body.PartSpecifier}]<0.10240>)\r\n", server.DequeueRequest());

        Assert.AreEqual(1, body.Message.Flags.Count);
        Assert.IsFalse(body.Message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(body.Message.Flags.Contains("$label1"));
        Assert.IsFalse(body.Message.IsSeen);
      });
    }

    [Test]
    public void TestReadMessageBody_Peek_OptionsSetSeen_OfMessage()
    {
      var initialFlags = new[] {"$label1"};

      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1, newFlags: new[] { "$label1", "\\Seen" }), initialFlags, testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Peek_OptionsSetSeen(message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Peek_OptionsSetSeen_OfMessagePart()
    {
      var initialFlags = new[] {"$label1"};

      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1, newFlags: new[] { "$label1", "\\Seen" }), initialFlags: initialFlags, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Peek_OptionsSetSeen(messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Peek_OptionsSetSeen(ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      Assert.AreEqual(1, body.Message.Flags.Count);
      Assert.IsFalse(body.Message.Flags.Contains(ImapMessageFlag.Seen));
      Assert.IsTrue(body.Message.Flags.Contains("$label1"));
      Assert.IsFalse(body.Message.IsSeen);

      messageBodyGenerator.EmitNextResponse(server);

      body.Read(options: ImapMessageFetchBodyOptions.SetSeen, read: (Stream stream) => {
        StringAssert.EndsWith($"UID FETCH 1 ({(body.IsOfTopLevel ? "RFC822.SIZE " : "")}BODY[{body.PartSpecifier}]<0.10240>)\r\n", server.DequeueRequest());

        Assert.AreEqual(2, body.Message.Flags.Count);
        Assert.IsTrue(body.Message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(body.Message.Flags.Contains("$label1"));
        Assert.IsTrue(body.Message.IsSeen);
      });
    }

    [Test]
    public void TestReadMessageBody_Peek_OptionsSetSeen_IsAlreadySeen_OfMessage()
    {
      var initialFlags = new[] {"\\Seen", "$label1"};

      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), initialFlags, testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Peek_OptionsSetSeen_IsAlreadySeen(message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Peek_OptionsSetSeen_IsAlreadySeen_OfMessagePart()
    {
      var initialFlags = new[] {"\\Seen", "$label1"};

      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), initialFlags: initialFlags, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Peek_OptionsSetSeen_IsAlreadySeen(messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Peek_OptionsSetSeen_IsAlreadySeen(ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      var initialFlags = new[] {"\\Seen", "$label1"};

      Assert.AreEqual(2, body.Message.Flags.Count);
      Assert.IsTrue(body.Message.Flags.Contains(ImapMessageFlag.Seen));
      Assert.IsTrue(body.Message.Flags.Contains("$label1"));
      Assert.IsTrue(body.Message.IsSeen);

      messageBodyGenerator.EmitNextResponse(server);

      body.Read(options: ImapMessageFetchBodyOptions.SetSeen, read: (Stream stream) => {
        StringAssert.EndsWith($"UID FETCH 1 ({(body.IsOfTopLevel ? "RFC822.SIZE " : "")}BODY[{body.PartSpecifier}]<0.10240>)\r\n", server.DequeueRequest());

        Assert.AreEqual(2, body.Message.Flags.Count);
        Assert.IsTrue(body.Message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(body.Message.Flags.Contains("$label1"));
        Assert.IsTrue(body.Message.IsSeen);
      });
    }

    [Test]
    public void TestReadMessageBody_Partial_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Partial(message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Partial_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Partial(messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Partial(ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      var fetchRange = new ImapFetchRange(128, 64);

      messageBodyGenerator.EmitNextResponse(server);

      body.Read(fetchRange: fetchRange, read: stream => {
        StringAssert.EndsWith($"UID FETCH 1 (BODY.PEEK[{body.PartSpecifier}]<128.64>)\r\n", server.DequeueRequest());

        Assert.IsNotNull(stream);

        Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(stream.Length));
      });
    }

    [Test]
    public void TestReadMessageBody_Partial_All_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Partial_All(message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Partial_All_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Partial_All(messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Partial_All(ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      var fetchRange = ImapFetchRange.All;

      messageBodyGenerator.EmitNextResponse(server);

      body.Read(fetchRange: fetchRange, read: stream => {
        StringAssert.EndsWith($"UID FETCH 1 ({(body.IsOfTopLevel ? "RFC822.SIZE " : "")}BODY.PEEK[{body.PartSpecifier}]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsNotNull(stream);

        Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(stream.Length));
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDefault_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.UTF8, "body", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        ReadMessageBody_ContentCharset_OptionsDefault(message.Body);
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDefault_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.UTF8, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        ReadMessageBody_ContentCharset_OptionsDefault(messagePart.Body);
      });
    }

    private void ReadMessageBody_ContentCharset_OptionsDefault(ImapMessageBody body)
    {
      body.ReadAs(options: ImapMessageFetchBodyOptions.Default, arg: 0, converter: (bodyStream, contentEncoding, _) => {
        Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
        Assert.AreEqual(ImapMessageBodyBase.DefaultCharset, contentEncoding, nameof(contentEncoding));

        return _;
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, "body", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        ReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset(message.Body);
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        ReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset(messagePart.Body);
      });
    }

    private void ReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset(ImapMessageBody body)
    {
      var forceCharsetTo = Encoding.UTF8;

      body.ReadAs(options: ImapMessageFetchBodyOptions.Default, forceCharsetTo: forceCharsetTo, arg: 0, converter: (bodyStream, contentEncoding, _) => {
        Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
        Assert.AreEqual(forceCharsetTo, contentEncoding, nameof(contentEncoding));

        return _;
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset_Null_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, "body", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        ReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset_Null(message.Body);
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset_Null_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        ReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset_Null(messagePart.Body);
      });
    }

    private void ReadMessageBody_ContentCharset_OptionsDefault_ForceContentCharset_Null(ImapMessageBody body)
    {
      body.ReadAs(options: ImapMessageFetchBodyOptions.Default, forceCharsetTo: null, arg: 0, converter: (bodyStream, contentEncoding, _) => {
        Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
        Assert.AreEqual(ImapMessageBodyBase.DefaultCharset, contentEncoding, nameof(contentEncoding));

        return _;
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ForceContentCharset()
    {
      var forceCharsetTo = Encoding.UTF8;

      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, forceCharsetTo: forceCharsetTo, arg: 0, converter: (bodyStream, contentEncoding, _) => {
          Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
          Assert.AreEqual(forceCharsetTo, contentEncoding, nameof(contentEncoding));

          return _;
        });
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ForceContentCharset_Null_HasContentCharset()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, forceCharsetTo: null, arg: 0, converter: (bodyStream, contentEncoding, _) => {
          Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
          Assert.AreEqual(Encoding.ASCII, contentEncoding, nameof(contentEncoding));

          return _;
        });
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ForceContentCharset_Null_HasNoContentCharset()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: null, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, forceCharsetTo: null, arg: 0, converter: (bodyStream, contentEncoding, _) => {
          Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
          Assert.AreEqual(ImapMessageBodyBase.DefaultCharset, contentEncoding, nameof(contentEncoding));

          return _;
        });
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ContentTypeCharset()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, arg: 0, converter: (bodyStream, contentEncoding, _) => {
          Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
          Assert.AreEqual(Encoding.ASCII, contentEncoding, nameof(contentEncoding));

          return _;
        });
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ContentTypeCharset_HasNoContentCharset()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: null, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, arg: 0, converter: (bodyStream, contentEncoding, _) => {
          Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
          Assert.AreEqual(ImapMessageBodyBase.DefaultCharset, contentEncoding, nameof(contentEncoding));

          return _;
        });
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ContentTypeCharset_UnknownCharset()
    {
      const string pseudoCharsetName = "X-UNKNOWN";

      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, contentTypeCharsetName: pseudoCharsetName, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        // messageBodyGenerator.EmitNextResponse(server);

        var ex = Assert.Throws<EncodingNotSupportedException>(() => {
          messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, arg: 0, converter: (bodyStream, contentEncoding, _) => {
            Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
            Assert.AreEqual(ImapMessageBodyBase.DefaultCharset, contentEncoding, nameof(contentEncoding));

            return _;
          });
        });

        Assert.AreEqual(pseudoCharsetName, ex.EncodingName);
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ContentTypeCharset_Fallback()
    {
      const string pseudoCharsetName = "X-UNKNOWN";
      var fallbackCharset = Encoding.UTF8;

      var selectFallbackCharset = new EncodingSelectionCallback(name => {
        Assert.AreEqual(pseudoCharsetName, name, nameof(name));

        return fallbackCharset;
      });

      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, contentTypeCharsetName: pseudoCharsetName, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, selectFallbackCharset: selectFallbackCharset, arg: 0, converter: (bodyStream, contentEncoding, _) => {
          Assert.IsNotNull(contentEncoding, nameof(contentEncoding));
          Assert.AreEqual(fallbackCharset, contentEncoding, nameof(contentEncoding));

          return _;
        });
      });
    }

    [Test]
    public void TestReadMessageBody_ContentCharset_OptionsDecodeContent_ContentTypeCharset_Fallback_NotSupportedException()
    {
      const string pseudoCharsetName = "X-UNKNOWN";

      var selectFallbackCharset = new EncodingSelectionCallback(name => {
        Assert.AreEqual(pseudoCharsetName, name, nameof(name));
        return null;
      });

      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, contentTypeCharsetName: pseudoCharsetName, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        // messageBodyGenerator.EmitNextResponse(server);

        var ex = Assert.Throws<EncodingNotSupportedException>(() => {
          messagePart.Body.ReadAs(options: ImapMessageFetchBodyOptions.DecodeContent, selectFallbackCharset: selectFallbackCharset, arg: 0, converter: (bodyStream, contentEncoding, _) => {
            Assert.Fail("converter called");
            return _;
          });
        });

        Assert.AreEqual(pseudoCharsetName, ex.EncodingName);
      });
    }

    [Test]
    public void TestReadMessageBody_DecodeContent_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, "body", testAction: (server, message, messageBodyGenerator) => {
        // messageBodyGenerator.EmitNextResponse(server);
        var ex = Assert.Throws<NotSupportedException>(() => {
          message.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, read: (Stream stream) => Assert.Fail("converter called"));
        });
      });
    }

    [Test]
    public void TestReadMessageBody_DecodeContent_Partial_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, "body", testAction: (server, message, messageBodyGenerator) => {
        // messageBodyGenerator.EmitNextResponse(server);
        var ex = Assert.Throws<NotSupportedException>(() => {
          message.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, fetchRange: new ImapFetchRange(0L, 1L), read: (Stream stream) => Assert.Fail("converter called"));
        });
      });
    }

    [Test]
    public void TestReadMessageBody_DecodeContent_Partial_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        // messageBodyGenerator.EmitNextResponse(server);
        var ex = Assert.Throws<NotSupportedException>(() => {
          messagePart.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, fetchRange: new ImapFetchRange(0L, 1L), read: (Stream stream) => Assert.Fail("converter called"));
        });
      });
    }

    [Test]
    public void TestReadMessageBody_DecodeContent_ContentTransferEncoding_Unknown()
    {
      WithMessagePartSingleTextPlain(contentTypeCharset: Encoding.ASCII, contentTransferEncoding: "X-UNKNOWN", messagePartBody: "body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        var ex = Assert.Throws<NotSupportedException>(() => {
          messagePart.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, read: (Stream stream) => Assert.Fail("converter called"));
        });

        StringAssert.Contains("X-UNKNOWN", ex.Message);
      });
    }

    private static void AssertFetchedMessageReader(ImapMessagePart messagePart, StreamReader reader)
    {
      Assert.IsTrue(messagePart.Client.IsBusy, "IsBusy; stream still opening");

      Assert.DoesNotThrow(reader.Dispose);

      Assert.IsFalse(messagePart.Client.IsBusy, "IsBusy; stream closed");

      Assert.DoesNotThrow(reader.Dispose);
      Assert.Throws<ObjectDisposedException>(() => reader.ReadLine());
    }

    [Test]
    public void TestReadMessageBody_DecodeContent_ContentTransferEncoding_7bit()
    {
      WithMessagePartSingleTextPlain(contentTransferEncoding: "7bit", contentTypeCharset: Encoding.ASCII, messagePartBody: "7bit body", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, read: (StreamReader reader) => {
          Assert.AreEqual("7bit body", reader.ReadToEnd());

          AssertFetchedMessageReader(messagePart, reader);
        });
      });
    }

    [Test]
    public void TestReadMessageBody_DecodeContent_ContentTransferEncoding_Base64()
    {
      WithMessagePartSingleTextPlain(contentTransferEncoding: "base64", contentTypeCharset: Encoding.UTF8, messagePartBody: @"YmFzZTY0IGVuY29kZWQg5pys5paH", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, read: (StreamReader reader) => {
          Assert.AreEqual("base64 encoded 本文", reader.ReadToEnd());

          AssertFetchedMessageReader(messagePart, reader);
        });
      });
    }

    [Test]
    public void TestReadMessageBody_DecodeContent_ContentTransferEncoding_QuotedPrintable()
    {
      WithMessagePartSingleTextPlain(contentTransferEncoding: "quoted-printable", contentTypeCharset: Encoding.UTF8, messagePartBody: @"quoted-printable =E6=9C=AC=E6=96=87", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        messagePart.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, read: (StreamReader reader) => {
          Assert.AreEqual("quoted-printable 本文", reader.ReadToEnd());

          AssertFetchedMessageReader(messagePart, reader);
        });
      });
    }

    private class FetchProgress : IProgress<ImapMessageFetchProgressInfo> {
      private readonly Action<ImapMessageFetchProgressInfo> report;

      public FetchProgress(Action<ImapMessageFetchProgressInfo> report)
      {
        this.report = report;
      }

      public void Report(ImapMessageFetchProgressInfo value)
      {
        report(value);
      }
    }

    [TestCase(1)]
    [TestCase(2)]
    [TestCase(3)]
    [TestCase(20)]
    public void TestReadMessageBody_Progress_OfMessage(int blockCount)
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: blockCount), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Progress(message, message.Body, server, messageBodyGenerator);
      });
    }

    [TestCase(1)]
    [TestCase(2)]
    [TestCase(3)]
    [TestCase(20)]
    public void TestReadMessageBody_Progress_OfMessagePart(int blockCount)
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: blockCount), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Progress(messagePart.Message, messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Progress(ImapMessage message, ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      var cts = new CancellationTokenSource();
      var progressReportedCount = 0;
      var progressReportedFinish = false;
      var progressLastPercentage = int.MinValue;
      var progress = new FetchProgress(p => {
        progressReportedCount++;

        if (p.FetchedOctetCount == messageBodyGenerator.TotalOctets)
          progressReportedFinish = true;

        Assert.AreEqual(cts.Token, p.CancellationToken, "CancellationToken #{0}", progressReportedCount);
        Assert.AreEqual(messageBodyGenerator.TotalOctets, p.TotalOctetCount, "Total #{0}", progressReportedCount);

        Assert.GreaterOrEqual(p.ProgressPercentage, 0, "0 <= ProgressPercentage #{0}", progressReportedCount);
        Assert.LessOrEqual(p.ProgressPercentage, 100, "ProgressPercentage <= 100 #{0}", progressReportedCount);
        Assert.AreEqual(p.ProgressPercentage, 100 * p.FetchedOctetCount / p.TotalOctetCount);
        Assert.Greater(p.ProgressPercentage, progressLastPercentage, "progressLastPercentage < ProgressPercentage #{0}", progressReportedCount);

        progressLastPercentage = p.ProgressPercentage;
      });

      for (var c = 0; c < messageBodyGenerator.BlockCount; c++) {
        messageBodyGenerator.EmitNextResponse(server);
      }

      Assert.DoesNotThrow(() => {
        message.Body.Read(progress: progress, cancellationToken: cts.Token, read: (Stream stream) => {
          stream.CopyTo(Stream.Null);
        });
      });

      Assert.AreEqual(messageBodyGenerator.BlockCount, progressReportedCount, "reported count");
      Assert.IsTrue(progressReportedFinish, "reported finish");
    }

    [TestCase(1)]
    [TestCase(2)]
    [TestCase(3)]
    [TestCase(20)]
    public void TestReadMessageBody_Progress_IndefiniteLength(int blockCount)
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: blockCount, respondRfc822Size: false), testAction: (server, message, messageBodyGenerator) => {
        var fetchRange = new ImapFetchRange(start: 1);
        var progressReportedFinish = false;
        var progressReportedCount = 0;
        var progress = new FetchProgress(p => {
          progressReportedCount++;

          Assert.Greater(p.FetchedOctetCount, 0);

          if (p.FetchedOctetCount == messageBodyGenerator.TotalOctets)
            progressReportedFinish = true;

          if (progressReportedFinish) {
            Assert.AreEqual(100, p.ProgressPercentage);
          }
          else {
            Assert.Less(p.TotalOctetCount, 0);
            Assert.AreEqual(0, p.ProgressPercentage);
          }
        });

        for (var c = 0; c < messageBodyGenerator.BlockCount; c++) {
          messageBodyGenerator.EmitNextResponse(server);
        }

        Assert.DoesNotThrow(() => {
          message.Body.Read(fetchRange: fetchRange, progress: progress, read: stream => stream.CopyTo(Stream.Null));
        });

        Assert.AreEqual(messageBodyGenerator.BlockCount, progressReportedCount, "reported count");
        Assert.IsTrue(progressReportedFinish, "reported finish");
      });
    }

    [Test]
    public void TestReadMessageBody_Progress_ExceptionThrown_ByReportOnPrepare()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 2), testAction: (server, message, messageBodyGenerator) => {
        var progressReportedCount = 0;
        var progress = new FetchProgress(p => {
          progressReportedCount++;
          throw new NotSupportedException();
        });

        messageBodyGenerator.EmitNextResponse(server);

        Assert.Throws<NotSupportedException>(() => {
          message.Body.Read(progress: progress, read: (Stream stream) => Assert.Fail("converter called"));
        });

        Assert.AreEqual(1, progressReportedCount);
      });
    }

    [Test]
    public void TestReadMessageBody_Progress_ExceptionThrown_ByReportOnFillBuffer()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 2), testAction: (server, message, messageBodyGenerator) => {
        var progressReportedCount = 0;
        var progress = new FetchProgress(p => {
          progressReportedCount++;

          if (progressReportedCount == 2)
            throw new NotSupportedException();
        });

        messageBodyGenerator.EmitNextResponse(server);

        Assert.Throws<NotSupportedException>(() => {
          messageBodyGenerator.EmitNextResponse(server);

          message.Body.Read(progress: progress, read: (Stream stream) => stream.CopyTo(Stream.Null));
        });

        Assert.AreEqual(2, progressReportedCount);
      });
    }

    /*
     * Read(), ReadAs(), etc.
     */
    private class ReadProgress : IProgress<ImapMessageFetchProgressInfo> {
      private readonly int expectedReportCount;
      private int actualReportCount;
      private readonly CancellationToken expectedCancellationToken;
      private CancellationToken actualCancellationToken;

      public ReadProgress(int expectedReportCount, CancellationToken expectedCancellationToken)
      {
        this.expectedReportCount = expectedReportCount;
        this.expectedCancellationToken = expectedCancellationToken;
      }

      public void Report(ImapMessageFetchProgressInfo value)
      {
        Assert.AreNotEqual(0, value.FetchedOctetCount);
        Assert.DoesNotThrow(() => Assert.IsNotNull(value.ProgressPercentage));

        actualReportCount++;
        actualCancellationToken = value.CancellationToken;
      }

      public void AssertProgress()
      {
        Assert.AreEqual(expectedReportCount, actualReportCount, nameof(actualReportCount));
        Assert.AreEqual(expectedCancellationToken, actualCancellationToken, nameof(actualCancellationToken));
      }
    }

    private void ReadCommon(
      Action<
        ImapMessageBody,
        ImapMessageFetchBodyOptions,
        ImapFetchRange,
        IProgress<ImapMessageFetchProgressInfo>,
        CancellationToken,
        MessageBodyGenerator
      > testAction,
      bool rangedFetch = true)
    {
      WithMessageSingleTextPlain(null, MessageBodyGenerator.Create(1L, 3, newFlags: new[] { "\\Seen" }, respondRfc822Size: !rangedFetch), testAction: (server, message, messageBodyGenerator) => {
        using (var cts = new CancellationTokenSource()) {
          var options = ImapMessageFetchBodyOptions.SetSeen;
          var progress = new ReadProgress(messageBodyGenerator.BlockCount, cts.Token);
          var fetchRange = new ImapFetchRange(0L, messageBodyGenerator.TotalOctets);

          messageBodyGenerator.EmitAllResponse(server);

          testAction(message.Body, options, fetchRange, progress, cts.Token, messageBodyGenerator);

          Assert.IsFalse(message.Client.IsBusy);
          Assert.IsTrue(message.IsSeen);

          progress.AssertProgress();
        }

        for (int b = 0, offset = 0; b < messageBodyGenerator.BlockCount; b++) {
          var length = b == 0 ? MessageBodyGenerator.DefaultBlockSize : messageBodyGenerator.GetBlockLengthOf(b);
          var fetchRfc822Size = !rangedFetch && b == 0;

          StringAssert.EndsWith($"UID FETCH 1 ({(fetchRfc822Size ? "RFC822.SIZE ": "")}BODY[]<{offset}.{length}>)\r\n", server.DequeueRequest());

          offset += length;
        }
      });
    }

    [Test]
    public void TestRead_Stream()
      => Read_Stream(runAsync: false);

    [Test]
    public void TestReadAsync_Stream()
      => Read_Stream(runAsync: true);

    private void Read_Stream(bool runAsync)
    {
      ReadCommon((body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        var result = new MemoryStream();

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            await body.ReadAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange,
              read: (Stream stream) => stream.CopyToAsync(result)
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            body.Read(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange,
              read: (Stream stream) => stream.CopyTo(result)
            );
          });
        }

        using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
          FileAssert.AreEqual(messageBodyStream, result);
        }
      });
    }

    [Test]
    public void TestRead_Stream_ReadNull()
      => TestArgumentNull("read", body => body.Read(read: (Action<Stream>)null));

    [Test]
    public void TestReadAsync_Stream_ReadNull()
      => TestArgumentNull("read", body => body.ReadAsync(read: (Func<Stream, Task>)null));

    [Test]
    public void TestRead_StreamReader()
      => Read_StreamReader(runAsync: false);

    [Test]
    public void TestReadAsync_StreamReader()
      => Read_StreamReader(runAsync: true);

    private void Read_StreamReader(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        string result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            await body.ReadAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              forceCharsetTo: Encoding.UTF8,
              read: async (StreamReader reader) => {
                Assert.AreSame(Encoding.UTF8, reader.CurrentEncoding);
                result = await reader.ReadToEndAsync();
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            body.Read(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              forceCharsetTo: Encoding.UTF8,
              read: (StreamReader reader) => {
                Assert.AreSame(Encoding.UTF8, reader.CurrentEncoding);
                result = reader.ReadToEnd();
              }
            );
          });
        }

        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result);
      });
    }

    [Test]
    public void TestRead_StreamReader_ReadNull()
      => TestArgumentNull("read", body => body.Read(read: (Action<StreamReader>)null));

    [Test]
    public void TestReadAsync_StreamReader_ReadNull()
      => TestArgumentNull("read", body => body.ReadAsync(read: (Func<StreamReader, Task>)null));

    [Test]
    public void TestReadAs_Stream_Converter()
      => ReadAs_Stream_Converter(runAsync: false);

    [Test]
    public void TestReadAsAsync_Stream_Converter()
      => ReadAs_Stream_Converter(runAsync: true);

    private void ReadAs_Stream_Converter(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        string result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              converter: async (Stream stream) => {
                return await new StreamReader(stream, OctetEncoding.SevenBits).ReadToEndAsync();
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAs(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              converter: (Stream stream) => {
                return new StreamReader(stream, OctetEncoding.SevenBits).ReadToEnd();
              }
            );
          });
        }

        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result);
      });
    }

    [Test]
    public void TestReadAs_Stream_Converter_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAs((Converter<Stream, int>)null));

    [Test]
    public void TestReadAsAsync_Stream_Converter_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAsAsync((Converter<Stream, Task<int>>)null));

    [Test]
    public void TestReadAs_Stream_ConverterFunc()
      => ReadAs_Stream_ConverterFunc(runAsync: false);

    [Test]
    public void TestReadAsAsync_Stream_ConverterFunc()
      => ReadAs_Stream_ConverterFunc(runAsync: true);

    private void ReadAs_Stream_ConverterFunc(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        (ImapMessageBody arg, string ret) result = default;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              arg: body,
              converter: async (Stream stream, ImapMessageBody arg) => {
                return (arg, await new StreamReader(stream, OctetEncoding.SevenBits).ReadToEndAsync());
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAs(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              arg: body,
              converter: (Stream stream, ImapMessageBody arg) => {
                return (arg, new StreamReader(stream, OctetEncoding.SevenBits).ReadToEnd());
              }
            );
          });
        }

        Assert.AreSame(body, result.arg);
        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result.ret);
      });
    }

    [Test]
    public void TestReadAs_Stream_ConverterFunc_Arg()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          message.Body.ReadAs(arg: 1, converter: (Stream stream, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return 0;
          });
        });
      });
    }

    [Test]
    public void TestReadAsAsync_Stream_ConverterFunc_Arg()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrowAsync(async () => {
          await message.Body.ReadAsAsync(arg: 1, converter: (Stream stream, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return Task.FromResult(0);
          });
        });
      });
    }

    [Test]
    public void TestReadAs_StreamReader_Converter()
      => ReadAs_StreamReader_Converter(runAsync: false);

    [Test]
    public void TestReadAsAsync_StreamReader_Converter()
      => ReadAs_StreamReader_Converter(runAsync: true);

    private void ReadAs_StreamReader_Converter(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        string result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              converter: async (StreamReader reader) => {
                return await reader.ReadToEndAsync();
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAs(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              converter: (StreamReader reader) => {
                return reader.ReadToEnd();
              }
            );
          });
        }

        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result);
      });
    }

    [Test]
    public void TestReadAs_StreamReader_Converter_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAs((Converter<StreamReader, int>)null));

    [Test]
    public void TestReadAsAsync_StreamReader_Converter_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAsAsync((Converter<StreamReader, Task<int>>)null));

    [Test]
    public void TestReadAs_StreamReader_ConverterFunc_Arg()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          message.Body.ReadAs(arg: 1, converter: (StreamReader reader, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return 0;
          });
        });
      });
    }

    [Test]
    public void TestReadAs_StreamReader_ConverterFunc()
      => ReadAs_StreamReader_ConverterFunc(runAsync: false);

    [Test]
    public void TestReadAsAsync_StreamReader_ConverterFunc()
      => ReadAs_StreamReader_ConverterFunc(runAsync: true);

    private void ReadAs_StreamReader_ConverterFunc(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        (ImapMessageBody arg, string ret) result = default;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              arg: body,
              converter: async (StreamReader reader, ImapMessageBody arg) => {
                return (arg, await reader.ReadToEndAsync());
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAs(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              arg: body,
              converter: (StreamReader reader, ImapMessageBody arg) => {
                return (arg, reader.ReadToEnd());
              }
            );
          });
        }

        Assert.AreSame(body, result.arg);
        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result.ret);
      });
    }

    [Test]
    public void TestReadAsAsync_StreamReader_ConverterFunc_Arg()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrowAsync(async () => {
          await message.Body.ReadAsAsync(arg: 1, converter: (StreamReader reader, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return Task.FromResult(0);
          });
        });
      });
    }

    [Test]
    public void TestReadAs_Stream_ConverterFunc_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAs(converter: (Func<Stream, int, int>)null, arg: default));

    [Test]
    public void TestReadAsAsync_Stream_ConverterFunc_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAsAsync(converter: (Func<Stream, int, Task<int>>)null, arg: default));

    [Test]
    public void TestReadAs_StreamReader_ConverterFunc_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAs(converter: (Func<StreamReader, int, int>)null, arg: default));

    [Test]
    public void TestReadAsAsync_StreamReader_ConverterFunc_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAsAsync(converter: (Func<StreamReader, int, Task<int>>)null, arg: default));

    [Test]
    public void TestReadAs_Stream_ConverterFuncWithEncoding()
      => ReadAs_Stream_ConverterFuncWithEncoding(runAsync: false);

    [Test]
    public void TestReadAsAsync_Stream_ConverterFuncWithEncoding()
      => ReadAs_Stream_ConverterFuncWithEncoding(runAsync: true);

    private void ReadAs_Stream_ConverterFuncWithEncoding(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        (ImapMessageBody arg, Encoding encoding, string ret) result = default;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              arg: body,
              converter: async (Stream stream, Encoding encoding, ImapMessageBody arg) => {
                return (arg, encoding, await new StreamReader(stream, encoding).ReadToEndAsync());
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAs(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              arg: body,
              converter: (Stream stream, Encoding encoding, ImapMessageBody arg) => {
                return (arg, encoding, new StreamReader(stream, encoding).ReadToEnd());
              }
            );
          });
        }

        Assert.AreSame(body, result.arg);
        Assert.AreEqual(result.encoding, ImapMessageBodyBase.DefaultCharset);
        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result.ret);
      });
    }

    [Test]
    public void TestReadAs_Stream_ConverterFuncWithEncoding_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAs(converter: (Func<Stream, Encoding, int, int>)null, arg: default));

    [Test]
    public void TestReadAsAsync_Stream_ConverterFuncWithEncoding_ConverterNull()
      => TestArgumentNull("converter", body => body.ReadAsAsync(converter: (Func<Stream, Encoding, int, Task<int>>)null, arg: default));

    [Test]
    public void TestReadAs_Stream_ConverterFuncWithEncoding_Arg()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          message.Body.ReadAs(arg: 1, converter: (Stream stream, Encoding encoding, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return 0;
          });
        });
      });
    }

    [Test]
    public void TestReadAsAsync_Stream_ConverterFuncWithEncoding_Arg()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrowAsync(async () => {
          await message.Body.ReadAsAsync(arg: 1, converter: (Stream stream, Encoding encoding, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return Task.FromResult(0);
          });
        });
      });
    }

    [Test]
    public void TestReadAsByteArray()
      => ReadAsByteArray(runAsync: false);

    [Test]
    public void TestReadAsByteArrayAsync()
      => ReadAsByteArray(runAsync: true);

    private void ReadAsByteArray(bool runAsync)
    {
      ReadCommon((body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        byte[] result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsByteArrayAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAsByteArray(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }

        using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
          FileAssert.AreEqual(messageBodyStream, new MemoryStream(result, writable: false));
        }
      });
    }

    [Test]
    public void TestReadAsString()
      => ReadAsString(runAsync: false);

    [Test]
    public void TestReadAsStringAsync()
      => ReadAsString(runAsync: true);

    private void ReadAsString(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        string result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsStringAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              forceCharsetTo: Encoding.UTF8
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAsString(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              forceCharsetTo: Encoding.UTF8
            );
          });
        }

        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result);
      });
    }

    [Test]
    public void TestReadAsStringLines()
      => ReadAsStringLines(runAsync: false);

    [Test]
    public void TestReadAsStringLinesAsync()
      => ReadAsStringLines(runAsync: true);

    private void ReadAsStringLines(bool runAsync)
    {
      ReadCommon(rangedFetch: false, testAction: (body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        IReadOnlyList<string> result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.ReadAsStringLinesAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              forceCharsetTo: Encoding.UTF8
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.ReadAsStringLines(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              forceCharsetTo: Encoding.UTF8
            );
          });
        }

        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), string.Join("\r\n", result));
      });
    }

    [Test]
    public void TestDownload()
      => Download(runAsync: false);

    [Test]
    public void TestDownloadAsync()
      => Download(runAsync: true);

    private void Download(bool runAsync)
    {
      ReadCommon((body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        Stream result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await body.DownloadAsync(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = body.Download(
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }

        using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
          FileAssert.AreEqual(messageBodyStream, result);
        }
      });
    }

    [Test]
    public void TestWriteTo_Stream()
      => WriteTo_Stream(runAsync: false);

    [Test]
    public void TestWriteToAsync_Stream()
      => WriteTo_Stream(runAsync: true);

    private void WriteTo_Stream(bool runAsync)
    {
      ReadCommon((body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        var destination = new MemoryStream();

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            await body.WriteToAsync(
              stream: destination,
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            body.WriteTo(
              stream: destination,
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }

        using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
          FileAssert.AreEqual(messageBodyStream, destination);
        }
      });
    }

    [Test]
    public void TestWriteTo_Stream_DestinationNull()
      => TestArgumentNull("stream", body => body.WriteTo(stream: null));

    [Test]
    public void TestWriteToAsync_Stream_DestinationNull()
      => TestArgumentNull("stream", body => body.WriteToAsync(stream: null));

    [Test]
    public void TestWriteTo_Stream_DestinationNotWritable()
      => WriteTo_Stream_DestinationNotWritable("stream", (body, s) => body.WriteTo(stream: s));

    [Test]
    public void TestWriteToAsync_Stream_DestinationNotWritable()
      => WriteTo_Stream_DestinationNotWritable("stream", (body, s) => body.WriteToAsync(stream: s));

    private void WriteTo_Stream_DestinationNotWritable(string expectedParamName, Action<ImapMessageBody, Stream> testAction)
    {
      TestCases.WithMessage((server, message) => {
        using (var s = new MemoryStream(new byte[0], writable: false)) {
          var ex = Assert.Throws<ArgumentException>(() => testAction(message.Body, s));

          Assert.AreEqual(expectedParamName, ex.ParamName);
        }
      });
    }

    [Test]
    public void TestWriteTo_BinaryWriter()
      => WriteTo_BinaryWriter(runAsync: false);

    [Test]
    public void TestWriteToAsync_BinaryWriter()
      => WriteTo_BinaryWriter(runAsync: true);

    private void WriteTo_BinaryWriter(bool runAsync)
    {
      ReadCommon((body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
        var result = new MemoryStream();
        var destination = new BinaryWriter(result);

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            await body.WriteToAsync(
              writer: destination,
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            body.WriteTo(
              writer: destination,
              options: options,
              progress: progress,
              cancellationToken: cancellationToken,
              fetchRange: fetchRange
            );
          });
        }

        using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
          FileAssert.AreEqual(messageBodyStream, result);
        }
      });
    }

    [Test]
    public void TestWriteTo_BinaryWriter_DestinationNull()
      => TestArgumentNull("writer", body => body.WriteTo(writer: null));

    [Test]
    public void TestWriteToAsync_BinaryWriter_DestinationNull()
      => TestArgumentNull("writer", body => body.WriteToAsync(writer: null));

    [Test]
    public void TestSave()
      => Save(runAsync: false);

    [Test]
    public void TestSaveAsync()
      => Save(runAsync: true);

    private void Save(bool runAsync)
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        ReadCommon((body, options, fetchRange, progress, cancellationToken, messageBodyGenerator) => {
          if (runAsync) {
            Assert.DoesNotThrowAsync(async () => {
              await body.SaveAsync(
                path: outfile,
                options: options,
                progress: progress,
                cancellationToken: cancellationToken,
                fetchRange: fetchRange
              );
            });
          }
          else {
            Assert.DoesNotThrow(() => {
              body.Save(
                path: outfile,
                options: options,
                progress: progress,
                cancellationToken: cancellationToken,
                fetchRange: fetchRange
              );
            });
          }

          Assert.IsTrue(File.Exists(outfile));

          using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
            using (var actualSavedMessageBodyStream = File.OpenRead(outfile)) {
              FileAssert.AreEqual(messageBodyStream, actualSavedMessageBodyStream);
            }
          }
        });
      });
    }

    [Test]
    public void TestSave_PathNull()
      => TestArgumentNull("path", body => body.Save(path: null));

    [Test]
    public void TestSaveAsync_PathNull()
      => TestArgumentNull("path", body => body.SaveAsync(path: null));

    [Test]
    public void TestSave_PathEmpty()
      => Save_PathEmpty("path", body => body.Save(path: string.Empty));

    [Test]
    public void TestSaveAsync_PathEmpty()
      => Save_PathEmpty("path", body => body.SaveAsync(path: string.Empty));

    private void Save_PathEmpty(string expectedParamName, Action<ImapMessageBody> testAction)
    {
      TestCases.WithMessage((server, message) => {
        var ex = Assert.Throws<ArgumentException>(() => testAction(message.Body));

        Assert.AreEqual(expectedParamName, ex.ParamName);
      });
    }

    [Test]
    public void TestSave_FileMustNotBeCreatedOnUnavailableException_MessageVanished()
      => Save_FileMustNotBeCreatedOnUnavailableException_MessageVanished((body, path) => body.Save(path));

    [Test]
    public void TestSaveAsync_FileMustNotBeCreatedOnUnavailableException_MessageVanished()
      => Save_FileMustNotBeCreatedOnUnavailableException_MessageVanished((body, path) => body.SaveAsync(path));

    private void Save_FileMustNotBeCreatedOnUnavailableException_MessageVanished(Action<ImapMessageBody, string> testAction)
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        TestCases.WithMessageVanished((server, message) => {
          Assert.Throws<ImapUnavailableException>(() => testAction(message.Body, outfile));

          Assert.IsFalse(File.Exists(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestSave_FileMustNotBeCreatedOnUnavailableException_MailboxClosed()
      => Save_FileMustNotBeCreatedOnUnavailableException_MailboxClosed((body, path) => body.Save(path));

    [Test]
    public void TestSaveAsync_FileMustNotBeCreatedOnUnavailableException_MailboxClosed()
      => Save_FileMustNotBeCreatedOnUnavailableException_MailboxClosed((body, path) => body.SaveAsync(path));

    private void Save_FileMustNotBeCreatedOnUnavailableException_MailboxClosed(Action<ImapMessageBody, string> testAction)
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        TestCases.WithMessageMailboxClosed((server, message) => {
          Assert.Throws<ImapUnavailableException>(() => testAction(message.Body, outfile));

          Assert.IsFalse(File.Exists(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestSave_FileMustNotBeCreatedOnMessageVanishedException()
      => Save_FileMustNotBeCreatedOnMessageVanishedException(runAsync: false);

    [Test]
    public void TestSaveAsync_FileMustNotBeCreatedOnMessageVanishedException()
      => Save_FileMustNotBeCreatedOnMessageVanishedException(runAsync: true);

    private void Save_FileMustNotBeCreatedOnMessageVanishedException(bool runAsync)
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        TestCases.WithMessage((server, message) => {
          // FETCH (message vanished)
          server.EnqueueTaggedResponse("$tag OK done\r\n");

          var ex = runAsync
            ? Assert.ThrowsAsync<ImapMessageVanishedException>(async () => await message.Body.SaveAsync(outfile))
            : Assert.Throws<ImapMessageVanishedException>(() => message.Body.Save(outfile));

          Assert.AreSame(message, ex.VanishedMessage, nameof(ex.VanishedMessage));

          Assert.IsFalse(File.Exists(outfile), "file existence");

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestSave_OverwriteExsintingFile()
      => Save_OverwriteExsintingFile(runAsync: false);

    [Test]
    public void TestSaveAsync_OverwriteExsintingFile()
      => Save_OverwriteExsintingFile(runAsync: true);

    private void Save_OverwriteExsintingFile(bool runAsync)
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");
      const int existingFileLength = 10 * 1024;

      TestUtils.UsingFile(outfile, () => {
        File.WriteAllText(outfile, new string('x', existingFileLength));

        WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
          Assert.Less(messageBodyGenerator.TotalOctets, new FileInfo(outfile).Length, "less than existing file length");

          messageBodyGenerator.EmitNextResponse(server);

          if (runAsync)
            Assert.DoesNotThrowAsync(async () => await message.Body.SaveAsync(outfile));
          else
            Assert.DoesNotThrow(() => message.Body.Save(outfile));

          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsTrue(File.Exists(outfile), "file existence");
          Assert.AreEqual(messageBodyGenerator.TotalOctets, new FileInfo(outfile).Length, "file length");
          Assert.AreEqual(OctetEncoding.SevenBits.GetBytes(messageBodyGenerator.GetMessageBody()),
                          File.ReadAllBytes(outfile));
        });
      });
    }
  }
}
