using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageHeaderTests : ImapMessageBodyBaseTestsBase {
    [SetUp]
    public void SetUp()
    {
      TestUtils.Encodings.RegisterCodePagesEncodingProvider();
    }

    private void TestArgumentNull(string expectedParamName, Action<ImapMessageHeader> testAction)
    {
      TestCases.WithMessage((server, message) => {
        var ex = Assert.Throws<ArgumentNullException>(() => testAction(message.Header));

        Assert.AreEqual(expectedParamName, ex.ParamName);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_All_OfMessage()
    {
      WithMessageSingleTextPlain(null, "name: value", testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_All(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_All_OfMessagePartEncapsulated()
    {
      WithMessagePartEncapsulated(messagePartBody: "name: value", testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_All(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_All_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_All(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadMessageHeader_HeaderFields_All(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      messageBodyGenerator.EmitNextResponse(server);

      header.ReadAs(headerFields: null, converter: headerStream => {
        Assert.IsNotNull(headerStream);
        //Assert.Throws<NotSupportedException>(() => Assert.AreNotEqual(0, headerStream.Length));

        return headerStream;
      });

      var request = server.DequeueRequest();

      StringAssert.Contains("BODY.PEEK[", request);

      if (header.IsOfTopLevel)
        StringAssert.Contains("HEADER]<0.", request);
      else
        StringAssert.Contains($"{header.PartSpecifier}.MIME]<0.", request);
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_Empty_Exclude_OfMessage()
    {
      WithMessageSingleTextPlain(null, "name: value", testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_Empty_Exclude(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_Empty_Exclude_OfMessagePartEncapsulated()
    {
      WithMessagePartEncapsulated(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_Empty_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_Empty_Exclude_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_Empty_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadMessageHeader_HeaderFields_Empty_Exclude(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      messageBodyGenerator.EmitNextResponse(server);

      header.ReadAs(headerFields: Enumerable.Empty<string>(), excludePassedHeaderFields: true, converter: headerStream => {
        Assert.IsNotNull(headerStream);
        //Assert.Throws<NotSupportedException>(() => Assert.AreNotEqual(0, headerStream.Length));

        return headerStream;
      });

      var request = server.DequeueRequest();

      StringAssert.Contains("BODY.PEEK[", request);

      if (header.IsOfTopLevel)
        StringAssert.Contains("HEADER]<0.", request);
      else
        StringAssert.Contains($"{header.PartSpecifier}.MIME]<0.", request);
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_Empty_OfMessage()
    {
      WithMessageSingleTextPlain(null, "name: value", testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_Empty(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_Empty_OfMessagePartEncapsulated()
    {
      WithMessagePartEncapsulated(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_Empty(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_Empty_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_Empty(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadMessageHeader_HeaderFields_Empty(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      //messageBodyGenerator.EmitNextResponse(server);

      // no-op
      Assert.AreEqual(string.Empty, header.ReadAsString(headerFields: Enumerable.Empty<string>()));

      Assert.DoesNotThrowAsync(async () => {
        Assert.AreEqual(string.Empty, await header.ReadAsStringAsync(headerFields: Enumerable.Empty<string>()));
      });

      Assert.IsNull(server.DequeueRequest());
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_All_Exclude_OfMessage()
    {
      WithMessageSingleTextPlain(null, "name: value", testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_All_Exclude(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_All_Exclude_OfMessagePartEncapsulated()
    {
      WithMessagePartEncapsulated(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_All_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_All_Exclude_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_All_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadMessageHeader_HeaderFields_All_Exclude(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      //messageBodyGenerator.EmitNextResponse(server);

      // no-op
      Assert.AreEqual(string.Empty, header.ReadAsString(headerFields: null, excludePassedHeaderFields: true));

      Assert.DoesNotThrowAsync(async () => {
        Assert.AreEqual(string.Empty, await header.ReadAsStringAsync(headerFields: null, excludePassedHeaderFields: true));
      });

      Assert.IsNull(server.DequeueRequest());
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NonEmpty_OfMessage()
    {
      WithMessageSingleTextPlain(null, "name: value" , testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NonEmpty(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NonEmpty_OfMessagePartEncapsulated()
    {
      WithMessagePartEncapsulated(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NonEmpty(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NonEmpty_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NonEmpty(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadMessageHeader_HeaderFields_NonEmpty(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      messageBodyGenerator.EmitNextResponse(server);

      header.ReadAs(headerFields: new[] { "name", "X-HEADER" }, converter: headerStream => {
        Assert.IsNotNull(headerStream);
        //Assert.Throws<NotSupportedException>(() => Assert.AreNotEqual(0, headerStream.Length));

        return headerStream;
      });

      var request = server.DequeueRequest();

      StringAssert.Contains("BODY.PEEK[", request);

      if (header.IsOfTopLevel || header.IsOfEncapsulatedMessagePart)
        StringAssert.Contains("HEADER.FIELDS (name X-HEADER)]<0.", request);
      else
        StringAssert.Contains($"{header.PartSpecifier}.MIME]<0.", request);
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NonEmpty_Exclude_OfMessage()
    {
      WithMessageSingleTextPlain(null, "name: value", testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NonEmpty_Exclude(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NonEmpty_Exclude_OfMessagePartEncapsulated()
    {
      WithMessagePartEncapsulated(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NonEmpty_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NonEmpty_Exclude_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(messagePartBody: "name: value", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NonEmpty_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadMessageHeader_HeaderFields_NonEmpty_Exclude(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      messageBodyGenerator.EmitNextResponse(server);

      header.ReadAs(headerFields: new[] { "name", "X-HEADER" }, excludePassedHeaderFields: true, converter: headerStream => {
        Assert.IsNotNull(headerStream);
        //Assert.Throws<NotSupportedException>(() => Assert.AreNotEqual(0, headerStream.Length));

        return headerStream;
      });

      var request = server.DequeueRequest();

      StringAssert.Contains("BODY.PEEK[", request);

      if (header.IsOfTopLevel || header.IsOfEncapsulatedMessagePart)
        StringAssert.Contains("HEADER.FIELDS.NOT (name X-HEADER)]<0.", request);
      else
        StringAssert.Contains($"{header.PartSpecifier}.MIME]<0.", request);
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NoMatchedHeaderFields_OfMessage()
    {
      WithMessageSingleTextPlain(null, string.Empty, testAction: (server, message, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NoMatchedHeaderFields(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NoMatchedHeaderFields_OfMessagePartEncapsulated()
    {
      WithMessagePartEncapsulated(messagePartBody: string.Empty, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NoMatchedHeaderFields(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadMessageHeader_HeaderFields_NoMatchedHeaderFields_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(messagePartBody: string.Empty, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageHeader_HeaderFields_NoMatchedHeaderFields(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadMessageHeader_HeaderFields_NoMatchedHeaderFields(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      messageBodyGenerator.EmitNextResponse(server);

      Assert.AreEqual(string.Empty, header.ReadAsString());

      var request = server.DequeueRequest();

      StringAssert.Contains("BODY.PEEK[", request);

      if (header.IsOfTopLevel)
        StringAssert.Contains("HEADER]<0.", request);
      else
        StringAssert.Contains($"{header.PartSpecifier}.MIME]<0.", request);
    }

    /*
     * ReadAsNameValuePairs
     */
    [Test]
    public void TestReadAsNameValuePairs_OfMessage()
    {
      const string header = @"content-type: text/plain
Received: from 3.example.com
received: from 2.example.com
RECEIVED: from 1.example.com
X-Extension: foo";

      WithMessageSingleTextPlain(null, header, testAction: (server, message, messageBodyGenerator) => {
        ReadAsNameValuePairs(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadAsNameValuePairs_OfMessagePartEncapsulated()
    {
      const string header = @"content-type: text/plain
Received: from 3.example.com
received: from 2.example.com
RECEIVED: from 1.example.com
X-Extension: foo";

      WithMessagePartEncapsulated(messagePartBody: header, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadAsNameValuePairs(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadAsNameValuePairs_OfMessagePart()
    {
      const string header = @"content-type: text/plain
From: from@example.com (must be excluded by client)
To: to@example.com (must be excluded by client)
Received: from 3.example.com
received: from 2.example.com
RECEIVED: from 1.example.com
X-Extension: foo";

      WithMessagePartSingleTextPlain(messagePartBody: header, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadAsNameValuePairs(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadAsNameValuePairs(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      messageBodyGenerator.EmitNextResponse(server);

      var headerFields = header.ReadAsNameValuePairs(headerFields: new[] { "Content-Type", "Received", "X-Extension" });

      Assert.IsNotNull(headerFields);
      Assert.AreEqual(5, headerFields.Count);

      var headerContentType = headerFields.First(p => string.Equals(p.Key, "Content-Type", StringComparison.OrdinalIgnoreCase));

      Assert.AreEqual("text/plain", headerContentType.Value);

      var headerXExtension = headerFields.First(p => string.Equals(p.Key, "X-Extension", StringComparison.OrdinalIgnoreCase));

      Assert.AreEqual("foo", headerXExtension.Value);

      var headersReceived = headerFields.Where(p => string.Equals(p.Key, "Received", StringComparison.OrdinalIgnoreCase)).ToArray();

      Assert.AreEqual(3, headersReceived.Length);
      Assert.AreEqual("from 3.example.com", headersReceived[0].Value);
      Assert.AreEqual("from 2.example.com", headersReceived[1].Value);
      Assert.AreEqual("from 1.example.com", headersReceived[2].Value);

      var request = server.DequeueRequest();

      StringAssert.Contains("BODY.PEEK[", request);

      if (header.IsOfTopLevel || header.IsOfEncapsulatedMessagePart)
        StringAssert.Contains("HEADER.FIELDS (Content-Type Received X-Extension)]<0.", request);
      else
        StringAssert.Contains($"{header.PartSpecifier}.MIME]<0.", request);
    }

    [Test]
    public void TestReadAsNameValuePairs_Async_Exclude_OfMessage()
    {
      const string header = @"content-type: text/plain
FROM: from@example.com
To: to@example.com";

      WithMessageSingleTextPlain(null, header, testAction: (server, message, messageBodyGenerator) => {
        ReadAsNameValuePairs_Async_Exclude(server, messageBodyGenerator, message.Header);
      });
    }

    [Test]
    public void TestReadAsNameValuePairs_Async_Exclude_OfMessagePartEncapsulated()
    {
      const string header = @"content-type: text/plain
FROM: from@example.com
To: to@example.com";

      WithMessagePartEncapsulated(messagePartBody: header, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadAsNameValuePairs_Async_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    [Test]
    public void TestReadAsNameValuePairs_Async_Exclude_OfMessagePart()
    {
      const string header = @"content-type: text/plain
FROM: from@example.com
To: to@example.com
Received: from 3.example.com (must be excluded by client)
received: from 2.example.com (must be excluded by client)
RECEIVED: from 1.example.com (must be excluded by client)
X-Extension: foo (must be excluded by client)";

      WithMessagePartSingleTextPlain(messagePartBody: header, testAction: (server, messagePart, messageBodyGenerator) => {
        ReadAsNameValuePairs_Async_Exclude(server, messageBodyGenerator, messagePart.Header);
      });
    }

    private void ReadAsNameValuePairs_Async_Exclude(ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator, ImapMessageHeader header)
    {
      messageBodyGenerator.EmitNextResponse(server);

      IReadOnlyList<KeyValuePair<string, string>> excludedHeaderFields = null;

      Assert.DoesNotThrowAsync(async () => {
        excludedHeaderFields = await header.ReadAsNameValuePairsAsync(headerFields: new[] { "Received", "X-Extension" }, excludePassedHeaderFields: true);
      });

      Assert.IsNotNull(excludedHeaderFields);
      Assert.AreEqual(3, excludedHeaderFields.Count);

      StringAssert.AreEqualIgnoringCase("Content-Type", excludedHeaderFields[0].Key);
      Assert.AreEqual("text/plain", excludedHeaderFields[0].Value);

      StringAssert.AreEqualIgnoringCase("From", excludedHeaderFields[1].Key);
      Assert.AreEqual("from@example.com", excludedHeaderFields[1].Value);

      StringAssert.AreEqualIgnoringCase("To", excludedHeaderFields[2].Key);
      Assert.AreEqual("to@example.com", excludedHeaderFields[2].Value);

      var request = server.DequeueRequest();

      StringAssert.Contains("BODY.PEEK[", request);

      if (header.IsOfTopLevel || header.IsOfEncapsulatedMessagePart)
        StringAssert.Contains("HEADER.FIELDS.NOT (Received X-Extension)]<0.", request);
      else
        StringAssert.Contains($"{header.PartSpecifier}.MIME]<0.", request);
    }

    [Test]
    public void TestReadAsNameValuePairs_NoMatchedHeaderFields()
    {
      WithMessageSingleTextPlain(null, string.Empty, testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        var headerFields = message.Header.ReadAsNameValuePairs(headerFields: new[] { "X-NonExistent" });

        Assert.IsNotNull(headerFields);
        CollectionAssert.IsEmpty(headerFields);
      });
    }

    [TestCase("Subject: =?utf-8?q?=E3=82=BF=E3=82=A4=E3=83=88=E3=83=AB?=")]
    [TestCase("Subject: =?utf-8?b?44K/44Kk44OI44Or?=")]
    public void TestReadAsNameValuePairs_Decode(string subjectHeaderField)
    {
      WithMessageSingleTextPlain(null, subjectHeaderField, testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.AreEqual("タイトル",
                        message.Header.ReadAsNameValuePairs(new[] { "Subject" }, decodeHeaderFieldValues: true).First().Value);
      });
    }

    [Test]
    public void TestReadAsNameValuePairs_Decode_CharsetUnknown()
    {
      WithMessageSingleTextPlain(null, "Subject: =?x-unknown?b?44K/44Kk44OI44Or?=", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        var ex = Assert.Throws<EncodingNotSupportedException>(() => message.Header.ReadAsNameValuePairs(new[] { "Subject" }, decodeHeaderFieldValues: true));

        Assert.AreEqual(ex.EncodingName, "x-unknown");
      });
    }

    [Test]
    public void TestReadAsNameValuePairs_Decode_SelectFallbackCharset()
    {
      WithMessageSingleTextPlain(null, "Subject: =?x-unknown?b?44K/44Kk44OI44Or?=", testAction: (server, message, messageBodyGenerator) => {
        var fallbackCharsetSelected = false;
        var selectFallbackCharset = new EncodingSelectionCallback(name => {
          fallbackCharsetSelected = true;
          Assert.AreEqual(name, "x-unknown");
          return Encoding.UTF8;
        });

        messageBodyGenerator.EmitNextResponse(server);

        Assert.AreEqual("タイトル",
                        message.Header.ReadAsNameValuePairs(new[] { "Subject" }, decodeHeaderFieldValues: true, selectFallbackCharset: selectFallbackCharset).First().Value);

        Assert.IsTrue(fallbackCharsetSelected);
      });
    }

    /*
     * ReadAs(), etc.
     */
    private void ReadCommon(
      Action<
        ImapMessageHeader,
        IEnumerable<string>,
        bool,
        CancellationToken,
        MessageBodyGenerator
      > testAction
    )
    {
      WithMessageSingleTextPlain(null, "name: value\r\n", testAction: (server, message, messageBodyGenerator) => {
        var headerFields = new[] { "exclude" };
        var excludePassedHeaderFields = true;

        messageBodyGenerator.EmitNextResponse(server);

        using (var cts = new CancellationTokenSource()) {
          testAction(message.Header, headerFields, excludePassedHeaderFields, cts.Token, messageBodyGenerator);
        }

        Assert.IsFalse(message.Client.IsBusy);

        StringAssert.EndsWith("UID FETCH 1 (BODY.PEEK[HEADER.FIELDS.NOT (exclude)]<0.10240>)\r\n", server.DequeueRequest());
      });
    }

    [Test]
    public void TestReadAs_Converter()
      => ReadAs_Converter(runAsync: false);

    [Test]
    public void TestReadAsAsync_Converter()
      => ReadAs_Converter(runAsync: true);

    private void ReadAs_Converter(bool runAsync)
    {
      ReadCommon((header, headerFields, excludePassedHeaderFields, cancellationToken, messageBodyGenerator) => {
        string result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await header.ReadAsAsync(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken,
              converter: async (Stream stream) => {
                return await new StreamReader(stream, OctetEncoding.SevenBits).ReadToEndAsync();
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = header.ReadAs(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken,
              converter: (Stream stream) => {
                return new StreamReader(stream, OctetEncoding.SevenBits).ReadToEnd();
              }
            );
          });
        }

        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result);
      });
    }

    [Test]
    public void TestReadAs_Converter_ConverterNull()
      => TestArgumentNull("converter", header => header.ReadAs((Converter<Stream, int>)null));

    [Test]
    public void TestReadAsAsync_Converter_ConverterNull()
      => TestArgumentNull("converter", header => header.ReadAsAsync((Converter<Stream, Task<int>>)null));

    [Test]
    public void TestReadAs_ConverterFunc()
      => ReadAs_ConverterFunc(runAsync: false);

    [Test]
    public void TestReadAsAsync_ConverterFunc()
      => ReadAs_ConverterFunc(runAsync: true);

    private void ReadAs_ConverterFunc(bool runAsync)
    {
      ReadCommon((header, headerFields, excludePassedHeaderFields, cancellationToken, messageBodyGenerator) => {
        (ImapMessageHeader arg, string ret) result = default;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await header.ReadAsAsync(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken,
              arg: header,
              converter: async (Stream stream, ImapMessageHeader arg) => {
                return (arg, await new StreamReader(stream, OctetEncoding.SevenBits).ReadToEndAsync());
              }
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = header.ReadAs(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken,
              arg: header,
              converter: (Stream stream, ImapMessageHeader arg) => {
                return (arg, new StreamReader(stream, OctetEncoding.SevenBits).ReadToEnd());
              }
            );
          });
        }

        Assert.AreSame(header, result.arg);
        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result.ret);
      });
    }

    [Test]
    public void TestReadAs_ConverterFunc_ConverterNull()
      => TestArgumentNull("converter", header => header.ReadAs(converter: (Func<Stream, int, int>)null, arg: default));

    [Test]
    public void TestReadAsAsync_ConverterFunc_ConverterNull()
      => TestArgumentNull("converter", header => header.ReadAsAsync(converter: (Func<Stream, int, Task<int>>)null, arg: default));

    [Test]
    public void TestReadAs_ConverterFunc_Arg()
    {
      WithMessageSingleTextPlain(null, "name: value", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          message.Header.ReadAs(arg: 1, converter: (Stream stream, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return 0;
          });
        });
      });
    }

    [Test]
    public void TestReadAsAsync_ConverterFunc_Arg()
    {
      WithMessageSingleTextPlain(null, "name: value", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrowAsync(async () => {
          await message.Header.ReadAsAsync(arg: 1, converter: (Stream stream, int arg) => {
            Assert.AreEqual(arg, 1, nameof(arg));
            return Task.FromResult(0);
          });
        });
      });
    }

    [Test]
    public void TestReadAsString()
      => ReadAsString(runAsync: false);

    [Test]
    public void TestReadAsStringAsync()
      => ReadAsString(runAsync: true);

    private void ReadAsString(bool runAsync)
    {
      ReadCommon((header, headerFields, excludePassedHeaderFields, cancellationToken, messageBodyGenerator) => {
        string result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await header.ReadAsStringAsync(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = header.ReadAsString(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken
            );
          });
        }

        Assert.AreEqual(messageBodyGenerator.GetMessageBody(), result);
      });
    }

    [Test]
    public void TestDownload()
      => Download(runAsync: false);

    [Test]
    public void TestDownloadAsync()
      => Download(runAsync: true);

    private void Download(bool runAsync)
    {
      ReadCommon((header, headerFields, excludePassedHeaderFields, cancellationToken, messageBodyGenerator) => {
        Stream result = null;

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            result = await header.DownloadAsync(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            result = header.Download(
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken
            );
          });
        }

        using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
          FileAssert.AreEqual(messageBodyStream, result);
        }
      });
    }

    [Test]
    public void TestWriteTo_Stream()
      => WriteTo_Stream(runAsync: false);

    [Test]
    public void TestWriteToAsync_Stream()
      => WriteTo_Stream(runAsync: true);

    private void WriteTo_Stream(bool runAsync)
    {
      ReadCommon((header, headerFields, excludePassedHeaderFields, cancellationToken, messageBodyGenerator) => {
        var destination = new MemoryStream();

        if (runAsync) {
          Assert.DoesNotThrowAsync(async () => {
            await header.WriteToAsync(
              stream: destination,
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken
            );
          });
        }
        else {
          Assert.DoesNotThrow(() => {
            header.WriteTo(
              stream: destination,
              headerFields: headerFields,
              excludePassedHeaderFields: excludePassedHeaderFields,
              cancellationToken: cancellationToken
            );
          });
        }

        using (var messageBodyStream = messageBodyGenerator.GetMessageBodyStream()) {
          FileAssert.AreEqual(messageBodyStream, destination);
        }
      });
    }

    [Test]
    public void TestWriteTo_Stream_DestinationNull()
      => TestArgumentNull("stream", header => header.WriteTo(stream: null));

    [Test]
    public void TestWriteToAsync_Stream_DestinationNull()
      => TestArgumentNull("stream", header => header.WriteToAsync(stream: null));

    [Test]
    public void TestWriteTo_Stream_DestinationNotWritable()
      => WriteTo_Stream_DestinationNotWritable("stream", (header, s) => header.WriteTo(stream: s));

    [Test]
    public void TestWriteToAsync_Stream_DestinationNotWritable()
      => WriteTo_Stream_DestinationNotWritable("stream", (header, s) => header.WriteToAsync(stream: s));

    private void WriteTo_Stream_DestinationNotWritable(string expectedParamName, Action<ImapMessageHeader, Stream> testAction)
    {
      TestCases.WithMessage((server, message) => {
        using (var s = new MemoryStream(new byte[0], writable: false)) {
          var ex = Assert.Throws<ArgumentException>(() => testAction(message.Header, s));

          Assert.AreEqual(expectedParamName, ex.ParamName);
        }
      });
    }
  }
}
