using System;
using System.Collections.Generic;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenMailboxOperationsFetchingTests {
    [Test]
    public void TestQueryMessages()
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        var messages = mailbox.QueryMessages();

        Assert.AreEqual(5L, mailbox.ExistingMessageCount);

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* 5 EXPUNGE\r\n" +
                                       "* 4 EXISTS\r\n" + 
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 1:4 (UID)\r\n",
                                       "* 1 FETCH (UID 3)\r\n" +
                                       "* 2 FETCH (UID 5)\r\n" +
                                       "* 3 FETCH (UID 7)\r\n" +
                                       "* 4 FETCH (UID 9)\r\n" +
                                       "$tag OK done\r\n");
        });

        var enumeratedMessages = new ImapMessage[4];

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(messages.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.AreEqual(4L, mailbox.ExistingMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");
          Assert.AreEqual($"{mailbox.Url}/;UID=3", enumerator.Current.Url.AbsoluteUri, "#0 url");

          enumeratedMessages[0] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(5L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");
          Assert.AreEqual($"{mailbox.Url}/;UID=5", enumerator.Current.Url.AbsoluteUri, "#1 url");

          enumeratedMessages[1] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(3L, enumerator.Current.Sequence, "#2 seq");
          Assert.AreEqual($"{mailbox.Url}/;UID=7", enumerator.Current.Url.AbsoluteUri, "#2 url");

          enumeratedMessages[2] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#3 non-null");
          Assert.AreEqual(9L, enumerator.Current.Uid, "#3 uid");
          Assert.AreEqual(4L, enumerator.Current.Sequence, "#3 seq");

          enumeratedMessages[3] = enumerator.Current;

          Assert.IsFalse(enumerator.MoveNext());
        }

        /*
         * re-enumerate
         */
        Assert.AreEqual(4L, mailbox.ExistingMessageCount);

        responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* 2 EXPUNGE\r\n" +
                                       "* 1 EXPUNGE\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 1:2 (UID)\r\n",
                                       "* 1 FETCH (UID 7)\r\n" +
                                       "* 2 FETCH (UID 9)\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(messages.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.AreEqual(2L, mailbox.ExistingMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");
          Assert.AreEqual($"{mailbox.Url}/;UID=7", enumerator.Current.Url.AbsoluteUri, "#0 url");

          Assert.AreSame(enumeratedMessages[2], enumerator.Current, "returns same instance #0");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(9L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");
          Assert.AreEqual($"{mailbox.Url}/;UID=9", enumerator.Current.Url.AbsoluteUri, "#1 url");

          Assert.AreSame(enumeratedMessages[3], enumerator.Current, "returns same instance #1");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages()
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 5 EXPUNGE\r\n" +
                                     "* 4 EXISTS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:4 (UID)\r\n",
                                     "* 1 FETCH (UID 3)\r\n" +
                                     "* 2 FETCH (UID 5)\r\n" +
                                     "* 3 FETCH (UID 7)\r\n" +
                                     "* 4 FETCH (UID 9)\r\n" +
                                     "$tag OK done\r\n");

        Assert.AreEqual(5L, mailbox.ExistingMessageCount);

        var messages = mailbox.GetMessages();

        Assert.IsFalse(mailbox.Client.IsBusy);
        Assert.AreEqual(4L, mailbox.ExistingMessageCount);

        Assert.AreEqual(4, messages.Count);

        Assert.IsNotNull(messages[0], "#0 non-null");
        Assert.AreEqual(3L, messages[0].Uid, "#0 uid");
        Assert.AreEqual(1L, messages[0].Sequence, "#0 seq");
        Assert.AreEqual($"{mailbox.Url}/;UID=3", messages[0].Url.AbsoluteUri, "#0 url");

        Assert.IsNotNull(messages[1], "#1 non-null");
        Assert.AreEqual(5L, messages[1].Uid, "#1 uid");
        Assert.AreEqual(2L, messages[1].Sequence, "#1 seq");
        Assert.AreEqual($"{mailbox.Url}/;UID=5", messages[1].Url.AbsoluteUri, "#1 url");

        Assert.IsNotNull(messages[2], "#2 non-null");
        Assert.AreEqual(7L, messages[2].Uid, "#2 uid");
        Assert.AreEqual(3L, messages[2].Sequence, "#2 seq");
        Assert.AreEqual($"{mailbox.Url}/;UID=7", messages[2].Url.AbsoluteUri, "#2 url");

        Assert.IsNotNull(messages[3], "#3 non-null");
        Assert.AreEqual(9L, messages[3].Uid, "#3 uid");
        Assert.AreEqual(4L, messages[3].Sequence, "#3 seq");
        Assert.AreEqual($"{mailbox.Url}/;UID=9", messages[3].Url.AbsoluteUri, "#3 url");
      });
    }

    [Test]
    public void TestGetMessages_NoMessageExists()
    {
      var selectResp =
        "* 2 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 1 EXPUNGE\r\n" +
                                     "* 1 EXPUNGE\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0L, mailbox.ExistingMessageCount);

        Assert.AreEqual(0, messages.Count);
      });
    }

    [Test]
    public void TestGetMessages_MalformedResponseShouldBeIgnored()
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "* 2 FETCH (boom!)\r\n" +
                                     "* 3 FETCH (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(2, messages.Count);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.AreEqual(3L, messages[1].Uid);
      });
    }

    [Test]
    public void TestGetMessageByUid()
    {
      GetMessageByUidOrSequence(true, mailbox => mailbox.GetMessageByUid(3L));
    }

    [Test]
    public void TestGetMessageBySequence()
    {
      GetMessageByUidOrSequence(false, mailbox => mailbox.GetMessageBySequence(1L));
    }

    [Test]
    public void TestGetMessageByUidAsync()
    {
      GetMessageByUidOrSequence(true, mailbox => mailbox.GetMessageByUidAsync(3L).Result);
    }

    [Test]
    public void TestGetMessageBySequenceAsync()
    {
      GetMessageByUidOrSequence(false, mailbox => mailbox.GetMessageBySequenceAsync(1L).Result);
    }

    private void GetMessageByUidOrSequence(bool uid, Func<ImapOpenMailbox, ImapMessage> testFunc)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse(uid ? "UID FETCH 3 (UID)\r\n" : "FETCH 1 (UID)\r\n",
                                     "* 1 FETCH (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var message = testFunc(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(message);
        Assert.AreEqual(3L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(23L, message.UidValidity);
        Assert.AreSame(mailbox, message.Mailbox);

        server.EnqueueTaggedResponse(uid ? "UID FETCH 3 (UID)\r\n" : "FETCH 1 (UID)\r\n",
                                     "* 1 FETCH (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var messageSecond = testFunc(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(message, messageSecond, "returns same instance");
      });
    }

    [Test]
    public void TestGetMessageByUid_NotFound()
    {
      GetMessageByUid_NotFound(mailbox => mailbox.GetMessageByUid(3L));
    }

    [Test]
    public void TestGetMessageByUidAsync_NotFound()
    {
      GetMessageByUid_NotFound(mailbox => mailbox.GetMessageByUidAsync(3L).Result);
    }

    private void GetMessageByUid_NotFound(Func<ImapOpenMailbox, ImapMessage> testFunc)
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("UID FETCH 3 (UID)\r\n",
                                     "$tag OK done\r\n");

        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapMessageNotFoundException>(() => testFunc(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.SequenceOrUidSet);
        Assert.AreEqual("3", ex.SequenceOrUidSet.Expression);

        TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate(ImapMessageNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.SequenceOrUidSet);
          Assert.AreEqual("3", deserialized.SequenceOrUidSet.Expression);
        });
      });
    }

    [Test]
    public void TestGetMessageByUid_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageByUid(0L));
      });
    }

    [Test]
    public void TestGetMessageByUidAsync_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageByUidAsync(0L));
      });
    }

    [Test]
    public void TestGetMessageBySequence_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequence(0L));
      });
    }

    [Test]
    public void TestGetMessageBySequenceAsync_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequenceAsync(0L));
      });
    }

    [Test]
    public void TestGetMessageBySequence_GreaterThanExistCount_1()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequence(1L));
      });
    }

    [Test]
    public void TestGetMessageBySequence_GreaterThanExistCount_2()
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequence(2L));
      });
    }

    [Test]
    public void TestGetMessageBySequence_NoResponse()
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                     "$tag NO such message\r\n");

        Assert.Throws<ImapErrorResponseException>(() => mailbox.GetMessageBySequence(1L));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetMessageBySequence_MalformedResponseShouldBeIgnored()
                 => GetMessageByUidOrSequence_MalformedResponseShouldBeIgnored(false, mailbox => mailbox.GetMessageBySequence(1L));

    [Test]
    public void TestGetMessageBySequenceAsync_MalformedResponseShouldBeIgnored()
                 => GetMessageByUidOrSequence_MalformedResponseShouldBeIgnored(false, mailbox => mailbox.GetMessageBySequenceAsync(1L).Result);

    [Test]
    public void TestGetMessageByUid_MalformedResponseShouldBeIgnored()
                 => GetMessageByUidOrSequence_MalformedResponseShouldBeIgnored(true, mailbox => mailbox.GetMessageByUid(1L));

    [Test]
    public void TestGetMessageByUidAsync_MalformedResponseShouldBeIgnored()
                 => GetMessageByUidOrSequence_MalformedResponseShouldBeIgnored(true, mailbox => mailbox.GetMessageByUidAsync(1L).Result);

    private void GetMessageByUidOrSequence_MalformedResponseShouldBeIgnored(bool uid, Func<ImapOpenMailbox, ImapMessage> testFunc)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        server.EnqueueTaggedResponse(uid ? "UID FETCH 1 (UID)\r\n" : "FETCH 1 (UID)\r\n",
                                     "* 1 FETCH (boom!)\r\n" +
                                     "$tag OK done\r\n");

        TestUtils.Assert.ThrowsOrAggregates<ImapMessageNotFoundException>(() => testFunc(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetMessageFirstUnseen()
    {
      GetMessageFirstUnseen(mailbox => mailbox.GetMessageFirstUnseen());
    }

    [Test]
    public void TestGetMessageFirstUnseenAsync()
    {
      GetMessageFirstUnseen(mailbox => mailbox.GetMessageFirstUnseenAsync().Result);
    }

    private void GetMessageFirstUnseen(Func<ImapOpenMailbox, ImapMessage> testFunc)
    {
      var selectResp =
        "* 18 EXISTS\r\n" +
        "* OK [UNSEEN 17]\r\n" +
        "* OK [UIDVALIDITY 3857529045]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        Assert.AreEqual(17L, mailbox.FirstUnseenMessageNumber);

        server.EnqueueTaggedResponse("FETCH 17 (UID)\r\n",
                                     "* 17 FETCH (UID 20)\r\n" +
                                     "$tag OK done\r\n");

        var message = testFunc(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(message);
        Assert.AreEqual(20L, message.Uid);
        Assert.AreEqual(17L, message.Sequence);
        Assert.AreEqual($"{mailbox.Url}/;UID=20", message.Url.AbsoluteUri);
        Assert.AreEqual(3857529045L, message.UidValidity);
        Assert.AreSame(mailbox, message.Mailbox);
      });
    }

    [Test]
    public void TestGetMessageFirstUnseen_FirstUnseenMissing()
    {
      GetMessageFirstUnseen_FirstUnseenMissing(mailbox => mailbox.GetMessageFirstUnseen());
    }

    [Test]
    public void TestGetMessageFirstUnseenAsync_FirstUnseenMissing()
    {
      GetMessageFirstUnseen_FirstUnseenMissing(mailbox => mailbox.GetMessageFirstUnseenAsync().Result);
    }

    private void GetMessageFirstUnseen_FirstUnseenMissing(Func<ImapOpenMailbox, ImapMessage> testFunc)
    {
      var selectResp =
        "* 18 EXISTS\r\n" +
        "* OK [UIDVALIDITY 3857529045]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.AreEqual(0L, mailbox.FirstUnseenMessageNumber);

        Assert.IsNull(testFunc(mailbox));
      });
    }

    [Test]
    public void TestGetMessages_ByUids()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (server, mailbox) => {
        server.EnqueueTaggedResponse("FETCH 1,3,5 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "* 3 FETCH (UID 3)\r\n" +
                                     "* 5 FETCH (UID 5)\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(1L, 3L, 5L);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(3, messages.Count);

        Assert.IsNotNull(messages[0], "#0 non-null");
        Assert.AreEqual(1L, messages[0].Uid, "#0 uid");
        Assert.AreEqual(1L, messages[0].Sequence, "#0 seq");
        Assert.AreEqual($"{mailbox.Url}/;UID=1", messages[0].Url.AbsoluteUri, "#0 url");

        Assert.IsNotNull(messages[1], "#1 non-null");
        Assert.AreEqual(3L, messages[1].Uid, "#1 uid");
        Assert.AreEqual(3L, messages[1].Sequence, "#1 seq");
        Assert.AreEqual($"{mailbox.Url}/;UID=3", messages[1].Url.AbsoluteUri, "#1 url");

        Assert.IsNotNull(messages[2], "#2 non-null");
        Assert.AreEqual(5L, messages[2].Uid, "#2 uid");
        Assert.AreEqual(5L, messages[2].Sequence, "#2 seq");
        Assert.AreEqual($"{mailbox.Url}/;UID=5", messages[2].Url.AbsoluteUri, "#2 url");
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
          "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, GetMessagesWithSearchCriteria);
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresDisabled()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Searchres);

        GetMessagesWithSearchCriteria(server, mailbox);
      });
    }

    private void GetMessagesWithSearchCriteria(ImapPseudoServer server, ImapOpenMailbox mailbox)
    {
      server.EnqueueTaggedResponse("SEARCH RECENT\r\n",
                                   "* SEARCH 1 2 3\r\n" +
                                   "$tag OK done\r\n");
      server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                   "* 1 FETCH (UID 3)\r\n" +
                                   "* 2 FETCH (UID 5)\r\n" +
                                   "* 3 FETCH (UID 7)\r\n" +
                                   "$tag OK done\r\n");

      var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

      Assert.IsFalse(mailbox.Client.IsBusy);

      Assert.IsNotNull(messages[0], "#0 non-null");
      Assert.AreEqual(3L, messages[0].Uid, "#0 uid");
      Assert.AreEqual(1L, messages[0].Sequence, "#0 seq");
      Assert.AreEqual($"{mailbox.Url}/;UID=3", messages[0].Url.AbsoluteUri, "#0 url");

      Assert.IsNotNull(messages[1], "#1 non-null");
      Assert.AreEqual(5L, messages[1].Uid, "#1 uid");
      Assert.AreEqual(2L, messages[1].Sequence, "#1 seq");
      Assert.AreEqual($"{mailbox.Url}/;UID=5", messages[1].Url.AbsoluteUri, "#1 url");

      Assert.IsNotNull(messages[2], "#2 non-null");
      Assert.AreEqual(7L, messages[2].Uid, "#2 uid");
      Assert.AreEqual(3L, messages[2].Sequence, "#2 seq");
      Assert.AreEqual($"{mailbox.Url}/;UID=7", messages[2].Url.AbsoluteUri, "#2 url");
    }

    [Test]
    public void TestQueryMessages_WithSearchCriteria_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        var messages = mailbox.QueryMessages(ImapSearchCriteria.Subject("test"), Encoding.UTF8);

        server.EnqueueTaggedResponse("SEARCH CHARSET utf-8 SUBJECT {4}\r\ntest\r\n",
                                     "$tag NO [BADCHARSET (US-ASCII Shift_JIS)] charset not supported\r\n");

        var ex = Assert.Throws<EncodingNotSupportedException>(() => messages.GetEnumerator());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("utf-8", ex.EncodingName);
        StringAssert.Contains("US-ASCII", ex.Message);
        StringAssert.Contains("Shift_JIS", ex.Message);
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH CHARSET utf-8 SUBJECT {4}\r\ntest\r\n",
                                     "$tag NO [BADCHARSET (US-ASCII Shift_JIS)] charset not supported\r\n");

        var ex = Assert.Throws<EncodingNotSupportedException>(() => {
          mailbox.GetMessages(ImapSearchCriteria.Subject("test"), Encoding.UTF8);
        });

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("utf-8", ex.EncodingName);
        StringAssert.Contains("US-ASCII", ex.Message);
        StringAssert.Contains("Shift_JIS", ex.Message);
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) RECENT\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH $ (UID)\r\n",
                                     "* 1 FETCH (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(1, messages.Count);

        Assert.IsNotNull(messages[0], "#0 non-null");
        Assert.AreEqual(3L, messages[0].Uid, "#0 uid");
        Assert.AreEqual(1L, messages[0].Sequence, "#0 seq");
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable_NoNotSaved()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) RECENT\r\n",
                                     "$tag NO [NOTSAVED] not saved\r\n");
        server.EnqueueTaggedResponse("SEARCH RECENT\r\n",
                                     "* SEARCH 1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                     "* 1 FETCH (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(1, messages.Count);

        Assert.IsNotNull(messages[0], "#0 non-null");
        Assert.AreEqual(3L, messages[0].Uid, "#0 uid");
        Assert.AreEqual(1L, messages[0].Sequence, "#0 seq");
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) CHARSET shift_jis SUBJECT {4}\r\ntest\r\n",
                                     "$tag NO [BADCHARSET] charset not supported\r\n");

        var ex = Assert.Throws<EncodingNotSupportedException>(() => {
          mailbox.GetMessages(ImapSearchCriteria.Subject("test"), TestUtils.Encodings.ShiftJis);
        });

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("shift_jis", ex.EncodingName);
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_NothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH RECENT\r\n",
                                     "* SEARCH\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsEmpty(messages);
      });
    }

    [Test]
    public void TestGetMessagesUnseen()
    {
      GetMessagesWithPresetSearchCriteria("UNSEEN", mailbox => mailbox.GetMessagesUnseen());
    }

    [Test]
    public void TestGetMessagesRecent()
    {
      GetMessagesWithPresetSearchCriteria("RECENT", mailbox => mailbox.GetMessagesRecent());
    }

    [Test]
    public void TestGetMessagesNew()
    {
      GetMessagesWithPresetSearchCriteria("NEW", mailbox => mailbox.GetMessagesNew());
    }

    private void GetMessagesWithPresetSearchCriteria(string expectedSearchCriteria,
                                                     Func<ImapOpenMailbox, IReadOnlyList<ImapMessage>> getMessages)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH " + expectedSearchCriteria + "\r\n",
                                     "* SEARCH 1 2 3\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* 1 FETCH (UID 3)\r\n" +
                                     "* 2 FETCH (UID 5)\r\n" +
                                     "* 3 FETCH (UID 7)\r\n" +
                                     "$tag OK done\r\n");

        var messages = getMessages(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(3, messages.Count);
      });
    }

    [Test]
    public void TestGetSortedMessages()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.Sort}, "INBOX", selectResp, (server, mailbox) => {
        server.EnqueueTaggedResponse("SORT (ARRIVAL) utf-8 RECENT\r\n",
                                     "* SORT 2 4 1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 2,4,1 (UID)\r\n",
                                     "* 1 FETCH (UID 2)\r\n" +
                                     "* 2 FETCH (UID 4)\r\n" +
                                     "* 4 FETCH (UID 8)\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(3, messages.Count);

        Assert.IsNotNull(messages[0], "#0 non-null");
        Assert.AreEqual(2L, messages[0].Sequence, "#0 seq");
        Assert.AreEqual(4L, messages[0].Uid, "#0 uid");
        Assert.AreEqual($"{mailbox.Url}/;UID=4", messages[0].Url.AbsoluteUri, "#0 url");

        Assert.IsNotNull(messages[1], "#1 non-null");
        Assert.AreEqual(4L, messages[1].Sequence, "#1 seq");
        Assert.AreEqual(8L, messages[1].Uid, "#1 uid");
        Assert.AreEqual($"{mailbox.Url}/;UID=8", messages[1].Url.AbsoluteUri, "#1 url");

        Assert.IsNotNull(messages[2], "#2 non-null");
        Assert.AreEqual(1L, messages[2].Sequence, "#2 seq");
        Assert.AreEqual(2L, messages[2].Uid, "#2 uid");
        Assert.AreEqual($"{mailbox.Url}/;UID=2", messages[2].Url.AbsoluteUri, "#2 url");
      });
    }

    [Test]
    public void TestGetSortedMessages_SortIncapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetSortedMessages(ImapSortCriteria.Subject, ImapSearchCriteria.Recent));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetSortedMessages_SortCriteriaIncapable_SortDisplay()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.Sort}, "INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetSortedMessages(ImapSortCriteria.DisplayTo, ImapSearchCriteria.Recent));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetSortedMessages_NothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.Sort},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SORT (ARRIVAL) utf-8 RECENT\r\n",
                                     "* SORT\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsEmpty(messages);
      });
    }

    [Test]
    public void TestGetThreadedMessages()
    {
      GetThreadedMessages((mailbox, algorithm) => mailbox.GetThreadedMessages(algorithm));
    }

    [Test]
    public void TestGetThreadedMessagesAsync()
    {
      GetThreadedMessages((mailbox, algorithm) => mailbox.GetThreadedMessagesAsync(algorithm).Result);
    }

    private void GetThreadedMessages(Func<ImapOpenMailbox, ImapThreadingAlgorithm, ImapMessageThread> testFunc)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("THREAD REFERENCES utf-8 ALL\r\n",
                                     "* THREAD (2)(3 6 (4 23)(44))\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 2,3,6,4,23,44 (UID)\r\n",
                                     "* 2 FETCH (UID 2)\r\n" +
                                     "* 3 FETCH (UID 3)\r\n" +
                                     "* 6 FETCH (UID 6)\r\n" +
                                     "* 4 FETCH (UID 4)\r\n" +
                                     "* 23 FETCH (UID 23)\r\n" +
                                     "* 44 FETCH (UID 44)\r\n" +
                                     "$tag OK done\r\n");

        var thread = testFunc(mailbox, ImapThreadingAlgorithm.References);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(thread);
        Assert.IsTrue(thread.IsRoot);
        Assert.IsNull(thread.Parent);
        Assert.AreEqual(0, thread.Depth);
        Assert.Throws<NotSupportedException>(() => Assert.IsNull(thread.Message));

        /*
          -- 2
          -- 3
             `-- 6
                 |-- 4
                 |   `-- 23
                 |
                 `-- 44
        */
        var expectedMessageSequences = new[] {
          2L, 3L, 6L, 4L, 23L, 44L,
        };
        var expectedDepths = new[] {
          1L, 1L, 2L, 3L, 4L, 3L,
        };
        var expectedChildrenCounts = new[] {
          0L, 1L, 2L, 1L, 0L, 0L,
        };
        var traversedMessageCount = 0;

        thread.Traverse(delegate(ImapMessageThread t) {
          Assert.IsFalse(t.IsRoot, "IsRoot #{0}", traversedMessageCount);
          Assert.IsNotNull(t.Parent, "Parent #{0}", traversedMessageCount);
          Assert.AreEqual(expectedDepths[traversedMessageCount], t.Depth, "Depth #{0}", traversedMessageCount);
          Assert.IsNotNull(t.Children, "Children #{0}", traversedMessageCount);
          Assert.AreEqual(expectedChildrenCounts[traversedMessageCount], t.Children.Count, "Children.Count #{0}", traversedMessageCount);
          Assert.AreEqual(expectedMessageSequences[traversedMessageCount], t.Message.Sequence, "Message.Sequence #{0}", traversedMessageCount);

          traversedMessageCount++;
        });

        Assert.AreEqual(6, traversedMessageCount);
      });
    }

    [Test]
    public void TestGetThreadedMessages_ThreadingAlgorithmIncapable_References()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessages(ImapThreadingAlgorithm.References));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetThreadedMessagesAsync_ThreadingAlgorithmIncapable_References()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessagesAsync(ImapThreadingAlgorithm.References));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetThreadedMessages_ThreadingAlgorithmIncapable_OrderedSubject()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessages(ImapThreadingAlgorithm.OrderedSubject));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetThreadedMessages_NothingMatched()
    {
      GetThreadedMessages_NothingMatched((mailbox, algorithm) => mailbox.GetThreadedMessages(algorithm));
    }

    [Test]
    public void TestGetThreadedMessagesAsync_NothingMatched()
    {
      GetThreadedMessages_NothingMatched((mailbox, algorithm) => mailbox.GetThreadedMessagesAsync(algorithm).Result);
    }

    private void GetThreadedMessages_NothingMatched(Func<ImapOpenMailbox, ImapThreadingAlgorithm, ImapMessageThread> testFunc)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("THREAD REFERENCES utf-8 ALL\r\n",
                                     "* THREAD\r\n" +
                                     "$tag OK done\r\n");

        var thread = testFunc(mailbox, ImapThreadingAlgorithm.References);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(thread);
        Assert.IsTrue(thread.IsRoot);
        Assert.IsNull(thread.Parent);
        Assert.AreEqual(0, thread.Depth);
        Assert.Throws<NotSupportedException>(() => Assert.IsNull(thread.Message));
        Assert.IsEmpty(thread.Children);

        thread.Traverse(delegate(ImapMessageThread t) {
          Assert.Fail("children traversed");
        });
      });
    }

    [Test]
    public void TestGetThreadedMessages_NoBadCharset()
    {
      GetThreadedMessages_NoBadCharset((mailbox, algorithm, criteria, encoding) => mailbox.GetThreadedMessages(algorithm, criteria, encoding));
    }

    [Test]
    public void TestGetThreadedMessagesAsync_NoBadCharset()
    {
      GetThreadedMessages_NoBadCharset((mailbox, algorithm, criteria, encoding) => mailbox.GetThreadedMessagesAsync(algorithm, criteria, encoding).Result);
    }

    private void GetThreadedMessages_NoBadCharset(Func<ImapOpenMailbox, ImapThreadingAlgorithm, ImapSearchCriteria, Encoding, ImapMessageThread> testFunc)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("THREAD REFERENCES shift_jis SUBJECT {4}\r\ntest\r\n",
                                     "$tag NO [BADCHARSET] charset not supported\r\n");

        var ex = TestUtils.Assert.ThrowsOrAggregates<EncodingNotSupportedException>(delegate {
          testFunc(mailbox,
                   ImapThreadingAlgorithm.References,
                   ImapSearchCriteria.Subject("test"),
                   TestUtils.Encodings.ShiftJis);
        });

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("shift_jis", ex.EncodingName);
      });
    }

    [Test]
    public void TestQuerySortedMessages_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.Sort},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        var messages = mailbox.QuerySortedMessages(ImapSortCriteria.Date,
                                                   ImapSearchCriteria.Subject("test"),
                                                   TestUtils.Encodings.ShiftJis);

        server.EnqueueTaggedResponse("SORT (DATE) shift_jis SUBJECT {4}\r\ntest\r\n",
                                     "$tag NO [BADCHARSET] charset not supported\r\n");

        var ex = Assert.Throws<EncodingNotSupportedException>(() => messages.GetEnumerator());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("shift_jis", ex.EncodingName);
      });
    }

    [Test]
    public void TestGetSortedMessages_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.Sort},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SORT (DATE) shift_jis SUBJECT {4}\r\ntest\r\n",
                                     "$tag NO [BADCHARSET] charset not supported\r\n");

        var ex = Assert.Throws<EncodingNotSupportedException>(() => {
          mailbox.GetSortedMessages(ImapSortCriteria.Date,
                                    ImapSearchCriteria.Subject("test"),
                                    TestUtils.Encodings.ShiftJis);
        });

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("shift_jis", ex.EncodingName);
      });
    }

    [Test]
    public void TestGetMessagesGmailSearch()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.GimapGmExt1},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH CHARSET ",
                                     "* SEARCH 1 2 3\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* 1 FETCH (UID 3)\r\n" +
                                     "* 2 FETCH (UID 5)\r\n" +
                                     "* 3 FETCH (UID 7)\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessagesGmailSearch("from:太郎 has:attachment");

        Assert.IsFalse(mailbox.Client.IsBusy);

        StringAssert.EndsWith("SEARCH CHARSET utf-8 X-GM-RAW {26}\r\n" + 
                              "from:太郎 has:attachment\r\n",
                              server.DequeueRequest(Encoding.UTF8));

         Assert.AreEqual(3, messages.Count);
      });
    }

    [Test]
    public void TestGetMessagesGmailSearch_GimapGmExt1Incapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
          "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        var ex = Assert.Throws<ImapIncapableException>(() => mailbox.GetMessagesGmailSearch("from:太郎 has:attachment"));

        Assert.IsFalse(mailbox.Client.IsBusy);

        CollectionAssert.Contains(ex.RequiredCapabilities, ImapCapability.GimapGmExt1);
      });
    }
  }
}

