using System;
using System.Collections.Generic;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures {
  [TestFixture]
  public class ImapBodyDispositionTests {
    [Test]
    public void TestAttachment()
    {
      var bodyStructure = new SinglePartMessage(
        MimeType.TextPlain,
        extended: true,
        disposition: new MessageContentDispsition("attachment", new Dictionary<string, string> { { "filename", "test.txt" } })
      ).ToBodyStructure();

      var disp = (bodyStructure as ImapExtendedSinglePartBodyStructure).Disposition;

      Assert.AreEqual("attachment", disp.Type);
      Assert.IsTrue(disp.IsAttachment);
      Assert.IsNotNull(disp.Parameters);
      Assert.AreEqual(1, disp.Parameters.Count);
      Assert.AreEqual("test.txt", disp.FileName);
    }

    [Test]
    public void TestInline()
    {
      var bodyStructure = new SinglePartMessage(
        MimeType.TextPlain,
        extended: true,
        disposition: new MessageContentDispsition("inline", null)
      ).ToBodyStructure();

      var disp = (bodyStructure as ImapExtendedSinglePartBodyStructure).Disposition;

      Assert.AreEqual("inline", disp.Type);
      Assert.IsTrue(disp.IsInline);
      Assert.IsNotNull(disp.Parameters);
      Assert.AreEqual(0, disp.Parameters.Count);
      Assert.IsNull(disp.FileName);
    }

    [TestCase("attachment")]
    [TestCase("Attachment")]
    [TestCase("ATTACHMENT")]
    public void TestIsAttachment_CaseInsensitive(string type)
    {
      var bodyStructure = new SinglePartMessage(
        MimeType.TextPlain,
        extended: true,
        disposition: new MessageContentDispsition(type)
      ).ToBodyStructure();

      var disp = (bodyStructure as ImapExtendedSinglePartBodyStructure).Disposition;

      Assert.IsTrue(disp.IsAttachment);
      Assert.IsFalse(disp.IsInline);
    }

    [TestCase("inline")]
    [TestCase("Inline")]
    [TestCase("INLINE")]
    public void TestIsInline_CaseInsensitive(string type)
    {
      var bodyStructure = new SinglePartMessage(
        MimeType.TextPlain,
        extended: true,
        disposition: new MessageContentDispsition(type)
      ).ToBodyStructure();

      var disp = (bodyStructure as ImapExtendedSinglePartBodyStructure).Disposition;

      Assert.IsTrue(disp.IsInline);
      Assert.IsFalse(disp.IsAttachment);
    }

    [TestCase("filename")]
    [TestCase("FileName")]
    [TestCase("FILENAME")]
    public void TestParameters_CaseInsensitive(string headerField)
    {
      var bodyStructure = new SinglePartMessage(
        MimeType.TextPlain,
        extended: true,
        disposition: new MessageContentDispsition("attachment", new Dictionary<string, string> { { headerField, "test.txt" } })
      ).ToBodyStructure();

      var disp = (bodyStructure as ImapExtendedSinglePartBodyStructure).Disposition;

      Assert.IsNotNull(disp.Parameters);
      Assert.AreEqual(1, disp.Parameters.Count);
      Assert.IsTrue(disp.Parameters.ContainsKey("filename"));
      Assert.IsTrue(disp.Parameters.ContainsKey("FileName"));
      Assert.IsTrue(disp.Parameters.ContainsKey("FILENAME"));
      Assert.AreEqual("test.txt", disp.FileName);
    }

    [Test]
    public void TestToString()
    {
      var bodyStructure = new SinglePartMessage(
        MimeType.TextPlain,
        extended: true,
        disposition: new MessageContentDispsition("attachment", new Dictionary<string, string> { { "filename", "test.txt" } })
      ).ToBodyStructure();

      var disp = (bodyStructure as ImapExtendedSinglePartBodyStructure).Disposition;

      Assert.DoesNotThrow(() => disp.ToString());
    }
  }
}
