using System;
using System.Collections.Generic;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  [TestFixture]
  public class ImapAddressTests {
    private ImapAddress CreateAddress(Address address)
    {
      return CreateAddressList(address)[0];
    }

    private IReadOnlyList<ImapAddress> CreateAddressList(params Address[] addresses)
    {
      var envelope = new MessageEnvelope(from: addresses);
      var bodyStructure = new Rfc822Message(
        new SinglePartMessage(
          MimeType.TextPlain
        ),
        encapsulatedMessageEnvelope: envelope
      ).ToBodyStructure() as ImapMessageRfc822BodyStructure;

      return bodyStructure.Envelope.From;
    }

    [Test]
    public void TestGetDecodedName()
    {
      if (!TestUtils.Encodings.IsEncodingAvailable("iso-2022-jp"))
        Assert.Ignore("encoding iso-2022-jp not available");

      var address = CreateAddress(new Address("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com"));

      Assert.AreEqual("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", address.Name);
      Assert.AreEqual("テスト", address.GetDecodedName());
    }

    [Test]
    public void TestGetDecodedName_InvalidMimeEncoding()
    {
      if (!TestUtils.Encodings.IsEncodingAvailable("iso-2022-jp"))
        Assert.Ignore("encoding iso-2022-jp not available");

      var address = CreateAddress(new Address("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com"));

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.Name);
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.GetDecodedName());
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.GetDecodedName(null ,false));

      Assert.Throws<FormatException>(() => address.GetDecodedName(null, true));
    }

    [Test]
    public void TestGetDecodedName_InvalidCharset()
    {
      var address = CreateAddress(new Address("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com"));

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.Name);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.GetDecodedName());
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.GetDecodedName(null ,false));

      Assert.Throws<EncodingNotSupportedException>(() => address.GetDecodedName(null, true));

      Assert.AreEqual("テスト", address.GetDecodedName(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return TestUtils.Encodings.Jis;
      }, true));
    }

#if NETFRAMEWORK || NETSTANDARD2_0
    [Test]
    public void TestToMailAddress_WithDisplayName()
    {
      var address = CreateAddress(new Address("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com")).ToMailAddress();

      Assert.AreEqual("テスト", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);

      address = CreateAddress(new Address("test", null, "test", "example.com")).ToMailAddress();

      Assert.AreEqual("test", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);
    }

    [Test]
    public void TestToMailAddress_WithoutDisplayName()
    {
      var address = CreateAddress(new Address(null, null, "test", "example.com")).ToMailAddress();

      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);
    }

    [Test]
    public void TestToMailAddress_InvalidMimeEncoding()
    {
      var a = CreateAddress(new Address("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com"));

      var address = a.ToMailAddress();

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);

      Assert.DoesNotThrow(() => a.ToMailAddress(null, false));
      Assert.Throws<FormatException>(() => a.ToMailAddress(null, true));
    }

    [Test]
    public void TestToMailAddress_InvalidCharset()
    {
      var a = CreateAddress(new Address("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com"));
      var address = a.ToMailAddress();

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);

      Assert.DoesNotThrow(() => a.ToMailAddress(null, false));
      Assert.Throws<EncodingNotSupportedException>(() => a.ToMailAddress(null, true));

      Assert.AreEqual("テスト", a.ToMailAddress(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return TestUtils.Encodings.Jis;
      }, true).DisplayName);
    }

    [Test]
    public void TestToMailAddressCollection()
    {
      var col = ImapAddress.ToMailAddressCollection(
        CreateAddressList(
          new Address("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.net"),
          new Address("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com"),
          new Address(null, null, "test", "example.com")
        )
      );

      Assert.IsNotNull(col);
      Assert.AreEqual(3, col.Count);

      Assert.AreEqual("テスト", col[0].DisplayName);
      Assert.AreEqual("test@example.net", col[0].Address);

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", col[1].DisplayName);
      Assert.AreEqual("test@example.com", col[1].Address);

      Assert.IsEmpty(col[2].DisplayName);
      Assert.AreEqual("test@example.com", col[2].Address);
    }

    [Test]
    public void TestToMailAddressCollection_InvalidCharset()
    {
      var addresses = CreateAddressList(
        new Address("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com")
      );

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=",
                      ImapAddress.ToMailAddressCollection(addresses)[0].DisplayName);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=",
                      ImapAddress.ToMailAddressCollection(addresses, null, false)[0].DisplayName);

      Assert.Throws<EncodingNotSupportedException>(() => ImapAddress.ToMailAddressCollection(addresses, null, true));

      Assert.AreEqual("テスト", ImapAddress.ToMailAddressCollection(addresses, delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return TestUtils.Encodings.Jis;
      }, true)[0].DisplayName);
    }

    [Test]
    public void TestToMailAddressCollection_InvalidMimeEncoding()
    {
      var addresses = CreateAddressList(
        new Address("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com")
      );

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=",
                      ImapAddress.ToMailAddressCollection(addresses)[0].DisplayName);
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=",
                      ImapAddress.ToMailAddressCollection(addresses, null, false)[0].DisplayName);

      Assert.Throws<FormatException>(() => ImapAddress.ToMailAddressCollection(addresses, null, true));
    }

    [Test]
    public void TestToMailAddressCollection_Null()
    {
      Assert.Throws<ArgumentNullException>(() => ImapAddress.ToMailAddressCollection(null));
    }
#endif

    [Test]
    public void TestToString()
    {
      var address = CreateAddress(new Address("test", "test", "test", "example.com"));

      Assert.DoesNotThrow(() => address.ToString());
    }
  }
}
