using System;
using System.Text;
using NUnit.Framework;

using Smdn.Text.Encodings;

using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  [TestFixture]
  public class ImapEnvelopeTests {
    private ImapEnvelope CreateEnvelope(MessageEnvelope envelope)
    {
      var bodyStructure = new Rfc822Message(
        new SinglePartMessage(
          MimeType.TextPlain
        ),
        encapsulatedMessageEnvelope: envelope
      ).ToBodyStructure() as ImapMessageRfc822BodyStructure;

      return bodyStructure.Envelope;
    }

    [Test]
    public void TestAddressListProperties_NonNull()
    {
      var envelope = CreateEnvelope(new MessageEnvelope());

      CollectionAssert.IsEmpty(envelope.From, "From");
      CollectionAssert.IsEmpty(envelope.Sender, "Sender");
      CollectionAssert.IsEmpty(envelope.ReplyTo, "ReplyTo");
      CollectionAssert.IsEmpty(envelope.To, "To");
      CollectionAssert.IsEmpty(envelope.Cc, "Cc");
      CollectionAssert.IsEmpty(envelope.Bcc, "Bcc");
    }

    [TestCase("Fri, 15 Mar 2002 12:32:23 +0900")]
    [TestCase("Fri, 15 Mar 2002 12:32:23 +0900 (JST)")]
    [TestCase("Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002 12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)")]
    [TestCase("Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002\r\n12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)")]
    public void TestGetDecodedDate(string input)
    {
      var expected = new DateTimeOffset(2002, 3, 15, 12, 32, 23, new TimeSpan(+9, 0, 0));
      var envelope = CreateEnvelope(new MessageEnvelope(date: input));

      Assert.AreEqual(expected, envelope.GetDecodedDate(), input);
    }

    [TestCase("")]
    [TestCase(null)]
    public void TestGetDecodedDate_NilOrEmpty(string input)
    {
      var envelope = CreateEnvelope(new MessageEnvelope(date: input));

      Assert.IsNull(envelope.GetDecodedDate(), "#1");
      Assert.IsNull(envelope.GetDecodedDate(true), "#2");
    }

    [Test]
    public void TestGetDecodedDate_Malformed()
    {
      var envelope = CreateEnvelope(new MessageEnvelope(date: "malformed date and time"));

      Assert.IsNull(envelope.GetDecodedDate());
      Assert.IsNull(envelope.GetDecodedDate(false));
      Assert.Throws<FormatException>(() => envelope.GetDecodedDate(true));
    }

    [Test]
    public void TestGetDecodedSubject()
    {
      var envelope1 = CreateEnvelope(new MessageEnvelope(subject: "not encoded"));

      Assert.AreEqual("not encoded", envelope1.GetDecodedSubject());
      Assert.AreEqual("not encoded", envelope1.Subject);

      var envelope2 = CreateEnvelope(new MessageEnvelope(subject: "=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?="));

      Assert.AreEqual("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", envelope2.Subject);
      Assert.AreEqual("テスト", envelope2.GetDecodedSubject());
    }

    [TestCase("")]
    [TestCase(null)]
    public void TestGetDecodedSubject_NilOrEmpty(string input)
    {
      var envelope = CreateEnvelope(new MessageEnvelope(subject: input));

      Assert.AreEqual(input, envelope.Subject);
      Assert.IsNull(envelope.GetDecodedSubject());
    }

    [Test]
    public void TestGetDecodedSubject_InvalidMimeEncoding()
    {
      var envelope = CreateEnvelope(new MessageEnvelope(subject: "=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?="));

      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.Subject);
      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject());
      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject(null, false));

      Assert.Throws<FormatException>(() => envelope.GetDecodedSubject(null, true));
    }

    [Test]
    public void TestGetDecodedSubject_InvalidCharset()
    {
      var envelope = CreateEnvelope(new MessageEnvelope(subject: "=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?="));

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.Subject);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject());
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject(null ,false));

      Assert.Throws<EncodingNotSupportedException>(() => envelope.GetDecodedSubject(null, true));

      Assert.AreEqual("テスト", envelope.GetDecodedSubject(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return TestUtils.Encodings.Jis;
      }, true));
    }

    [Test]
    public void TestToString()
    {
      var addresses = new[] {
        new Address("test1", "test1", "test1", "test1"),
        new Address("test2", "test2", "test2", "test2"),
      };
      var envelope = CreateEnvelope(new MessageEnvelope(
        date: "date",
        subject: "subject",
        from: addresses,
        sender: addresses,
        replyTo: addresses,
        to: addresses,
        cc: addresses,
        bcc: addresses,
        inReplyTo: "in-reply-to",
        messageId: "message-id"
      ));

      Assert.DoesNotThrow(() => envelope.ToString());
    }
  }
}
