using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  [TestFixture]
  public class ImapNumberSetTests {
    private void AssertSerialization(IImapNumberSet s)
    {
      TestUtils.Assert.IsSerializableBinaryFormat(s, d => {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual(s.Expression, d.Expression);
        Assert.AreEqual(s.IsUidSet, d.IsUidSet);
        Assert.AreEqual(s.IsEmpty, d.IsEmpty);
        Assert.AreEqual(s.IsSingle, d.IsSingle);
        Assert.AreEqual(s.ConcreteCount, d.ConcreteCount);
      });
    }

    private void AssertGetEnumeratorNotSupported(IImapNumberSet s)
    {
      Assert.Throws<NotSupportedException>(() => {
        foreach (var n in s) {
          Assert.Fail("enumerated");
        }
      });
    }

    [Test]
    public void TestAll()
    {
      Test(ImapUidSet.All(), true);
      Test(ImapSequenceSet.All(), false);
      Test(ImapSequenceSet.All(asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual("*", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.Less(s.ConcreteCount, 0L);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        AssertGetEnumeratorNotSupported(s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestEmpty()
    {
      Test(ImapUidSet.Empty(), true);
      Test(ImapSequenceSet.Empty(), false);
      Test(ImapSequenceSet.Empty(asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual(string.Empty, s.Expression);
        Assert.IsTrue(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(0L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.IsEmpty(s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestFrom()
    {
      Test(ImapUidSet.From(72), true);
      Test(ImapSequenceSet.From(72), false);
      Test(ImapSequenceSet.From(72, asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual("72:*", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.Less(s.ConcreteCount, 0L);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        AssertGetEnumeratorNotSupported(s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestTo()
    {
      Test(ImapUidSet.To(25), true);
      Test(ImapSequenceSet.To(25), false);
      Test(ImapSequenceSet.To(25, asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual("*:25", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(25L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        AssertGetEnumeratorNotSupported(s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestRange()
    {
      Test(ImapUidSet.Range(15, 18), true);
      Test(ImapSequenceSet.Range(15, 18), false);
      Test(ImapSequenceSet.Range(15, 18, asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual("15:18", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(4L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 15, 16, 17, 18 }, s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestRange_Single()
    {
      Test(ImapUidSet.Range(15, 15), true);
      Test(ImapSequenceSet.Range(15, 15), false);
      Test(ImapSequenceSet.Range(15, 15, asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual("15", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsTrue(s.IsSingle);
        Assert.AreEqual(1L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 15 }, s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestOf_Single()
    {
      Test(ImapUidSet.Of(611), true);
      Test(ImapSequenceSet.Range(611, 611), false);
      Test(ImapSequenceSet.Range(611, 611, asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual("611", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsTrue(s.IsSingle);
        Assert.AreEqual(1L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 611 }, s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestOf_Empty()
    {
      Test(ImapUidSet.Of(Enumerable.Empty<long>()), true);
      Test(ImapSequenceSet.Of(Enumerable.Empty<long>()), false);
      Test(ImapSequenceSet.Of(Enumerable.Empty<long>(), asUidSet: true), true);

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual(string.Empty, s.Expression);
        Assert.IsTrue(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(0L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[0], s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestOf_Array()
    {
      long[] baseArray;

      baseArray = new long[] { 2, 3, 4 }; Test(ImapUidSet.Of(1, baseArray), baseArray, true);
      baseArray = new long[] { 2, 3, 4 }; Test(ImapSequenceSet.Of(1, baseArray), baseArray, false);
      baseArray = new long[] { 2, 3, 4 }; Test(ImapSequenceSet.Of(asUidSet: true, 1, baseArray), baseArray, true);

      void Test(IImapNumberSet s, long[] _baseArray, bool isUidSet)
      {
        Array.Clear(_baseArray, 0, _baseArray.Length);

        Assert.AreEqual("1,2,3,4", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(4L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 1, 2, 3, 4 }, s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestOf_ArraySegment()
    {
      long[] baseArray;

      baseArray = new long[] { 1, 2, 3, 4, 5 }; Test(ImapUidSet.Of(new ArraySegment<long>(baseArray, 1, 3)), baseArray, true);
      baseArray = new long[] { 1, 2, 3, 4, 5 }; Test(ImapSequenceSet.Of(new ArraySegment<long>(baseArray, 1, 3)), baseArray, false);
      baseArray = new long[] { 1, 2, 3, 4, 5 }; Test(ImapSequenceSet.Of(new ArraySegment<long>(baseArray, 1, 3), asUidSet: true), baseArray, true);

      void Test(IImapNumberSet s, long[] _baseArray, bool isUidSet)
      {
        Array.Clear(_baseArray, 0, _baseArray.Length);

        Assert.AreEqual("2,3,4", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(3L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 2, 3, 4 }, s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestOf_IEnumerable_List()
    {
      List<long> baseList;

      baseList = new List<long> { 1, 2, 3, 4 }; Test(ImapUidSet.Of(baseList), baseList, true);
      baseList = new List<long> { 1, 2, 3, 4 }; Test(ImapSequenceSet.Of(baseList), baseList, false);
      baseList = new List<long> { 1, 2, 3, 4 }; Test(ImapSequenceSet.Of(baseList, asUidSet: true), baseList, true);

      void Test(IImapNumberSet s, List<long> _baseList, bool isUidSet)
      {
        _baseList.Clear();

        Assert.AreEqual("1,2,3,4", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(4L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 1, 2, 3, 4 }, s);

        AssertSerialization(s);
      }
    }

    [Test]
    public void TestOf_IEnumerable_Iterator()
    {
      Test(ImapUidSet.Of(Iterate()), true);
      Test(ImapSequenceSet.Of(Iterate()), false);
      Test(ImapSequenceSet.Of(Iterate(), asUidSet: true), true);

      IEnumerable<long> Iterate()
      {
        yield return 1L;
        yield return 2L;
        yield return 3L;
        yield return 4L;
      }

      void Test(IImapNumberSet s, bool isUidSet)
      {
        Assert.AreEqual("1,2,3,4", s.Expression);
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(4L, s.ConcreteCount);
        Assert.AreEqual(isUidSet, s.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 1, 2, 3, 4 }, s);

        AssertSerialization(s);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombine(bool uid)
    {
      var set1 = ImapSequenceSet.Range(1, 3, asUidSet: uid);
      var set2 = ImapSequenceSet.Of(asUidSet: uid, 5, 6, 8, 9);

      var c = set1.CombineWith(set2);

      Assert.AreEqual("1:3,5,6,8,9", c.Expression);
      Assert.AreEqual(uid, c.IsUidSet);
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(7L, c.ConcreteCount);

      AssertSerialization(c);

      c = set2.CombineWith(set1);

      Assert.AreEqual("5,6,8,9,1:3", c.Expression);
      Assert.AreEqual(uid, c.IsUidSet);
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(7L, c.ConcreteCount);

      AssertSerialization(c);
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombine_WithTo(bool uid)
    {
      var set1 = ImapSequenceSet.To(7, asUidSet: uid);
      var set2 = ImapSequenceSet.Of(uid, 5, 6, 8, 9);

      var c = set1.CombineWith(set2);

      //TODO: Assert.AreEqual("*:7,8,9", c.Expression);
      Assert.AreEqual("*:7,5,6,8,9", c.Expression);
      Assert.AreEqual(uid, c.IsUidSet);
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      //TODO: Assert.AreEqual(9L, c.ConcreteCount);
      Assert.AreEqual(11L, c.ConcreteCount);

      AssertSerialization(c);
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombine_WithFrom(bool uid)
    {
      var set1 = ImapSequenceSet.From(7, asUidSet: uid);
      var set2 = ImapSequenceSet.Of(asUidSet: uid, 5, 6, 8, 9);

      var c = set1.CombineWith(set2);

      //TODO: Assert.AreEqual("5,6,7:*", c.Expression);
      Assert.AreEqual("7:*,5,6,8,9", c.Expression);
      Assert.AreEqual(uid, c.IsUidSet);
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.Less(c.ConcreteCount, 0L);

      AssertSerialization(c);
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombine_NonEmpty_WithEmpty(bool uid)
    {
      var set1 = ImapSequenceSet.Range(2, 5, asUidSet: uid);
      var set2 = ImapSequenceSet.Empty(asUidSet: uid);

      Assert.IsFalse(set1.IsEmpty);
      Assert.IsFalse(set1.IsSingle);
      Assert.IsTrue(set2.IsEmpty);
      Assert.IsFalse(set2.IsSingle);

      foreach (var c in new[] {
        set1.CombineWith(set2),
        set2.CombineWith(set1),
      }) {
        Assert.AreEqual("2:5", c.Expression);
        Assert.AreEqual(uid, c.IsUidSet);
        Assert.IsFalse(c.IsEmpty);
        Assert.IsFalse(c.IsSingle);
        Assert.AreEqual(4L, c.ConcreteCount);

        AssertSerialization(c);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombine_Single_WithEmpty(bool uid)
    {
      var set1 = ImapSequenceSet.Of(3, asUidSet: uid);
      var set2 = ImapSequenceSet.Empty(asUidSet: uid);

      Assert.IsFalse(set1.IsEmpty);
      Assert.IsTrue(set1.IsSingle);
      Assert.IsTrue(set2.IsEmpty);
      Assert.IsFalse(set2.IsSingle);

      foreach (var c in new[] {
        set1.CombineWith(set2),
        set2.CombineWith(set1),
      }) {
        Assert.AreEqual("3", c.Expression);
        Assert.AreEqual(uid, c.IsUidSet);
        Assert.IsFalse(c.IsEmpty);
        Assert.IsTrue(c.IsSingle);
        Assert.AreEqual(1L, c.ConcreteCount);

        AssertSerialization(c);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombine_WithAll(bool uid)
    {
      var set1 = ImapSequenceSet.Range(1, 3, asUidSet: uid);
      var set2 = ImapSequenceSet.All(asUidSet: uid);

      foreach (var c in new[] {
        set1.CombineWith(set2),
        set2.CombineWith(set1),
      }) {
        Assert.AreEqual("*", c.Expression);
        Assert.AreEqual(uid, c.IsUidSet);
        Assert.IsFalse(c.IsEmpty);
        Assert.IsFalse(c.IsSingle);
        Assert.Less(c.ConcreteCount, 0L);

        AssertSerialization(c);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombine_WithCombinedSet(bool uid)
    {
      var set1 = ImapSequenceSet.Of(asUidSet: uid, 1, 2, 3).CombineWith(ImapSequenceSet.Range(5, 7, asUidSet: uid));
      var set2 = ImapSequenceSet.Range(9, 10, asUidSet: uid).CombineWith(ImapSequenceSet.Of(asUidSet: uid, 15, 16, 18));

      var c = set1.CombineWith(set2);

      Assert.AreEqual("1,2,3,5:7,9:10,15,16,18", c.Expression);
      Assert.AreEqual(uid, c.IsUidSet);
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(11, c.ConcreteCount);

      AssertSerialization(c);
    }

    [Test]
    public void TestCombine_WithSequenceSetAndUidSet()
    {
      foreach (var test in new[] {
        new {Set1 = ImapSequenceSet.All(),         Set2 = ImapUidSet.All()},
        new {Set1 = ImapSequenceSet.All(),         Set2 = ImapUidSet.From(2)},
        new {Set1 = ImapSequenceSet.From(3),       Set2 = ImapUidSet.To(5)},
        new {Set1 = ImapSequenceSet.Range(1, 3),   Set2 = ImapUidSet.Range(5, 7)},
        new {Set1 = ImapSequenceSet.To(10),        Set2 = ImapUidSet.Of(3)},
      }) {
        Assert.Throws<ArgumentException>(() => test.Set1.CombineWith(test.Set2), "{0} {1}", test.Set1, test.Set2);
        Assert.Throws<ArgumentException>(() => test.Set2.CombineWith(test.Set1), "{0} {1}", test.Set1, test.Set2);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombined_GetEnumerator(bool uid)
    {
      var set1 = ImapSequenceSet.Range(1, 3, asUidSet: uid);
      var set2 = ImapSequenceSet.Of(asUidSet: uid, 5, 6, 8, 9);

      CollectionAssert.AreEqual(new long[] { 1L, 2L, 3L, 5L, 6L, 8L, 9L }, set1.CombineWith(set2));
      CollectionAssert.AreEqual(new long[] { 5L, 6L, 8L, 9L, 1L, 2L, 3L }, set2.CombineWith(set1));
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestCombined_WithFromSet_GetEnumerator(bool uid)
    {
      var set1 = ImapSequenceSet.From(10, asUidSet: uid);
      var set2 = ImapSequenceSet.Of(asUidSet: uid, 5, 6, 8, 9);

      AssertGetEnumeratorNotSupported(set1.CombineWith(set2));
      // TODO: AssertGetEnumeratorNotSupported(set2.CombineWith(set1));
    }

    [Test]
    public void TestSplitIntoEach_Sequential1()
    {
      var range = ImapSequenceSet.Range(1L, 345L);
      var expectedExpression = new[] {
        "1:100",
        "101:200",
        "201:300",
        "301:345",
      };
      var expectedConcreteCount = new[] {
        100L,
        100L,
        100L,
        45L,
      };
      var index = 0;

      Assert.AreEqual(345L, range.ConcreteCount);

      foreach (var splitted in range.SplitIntoEach(100)) {
        Assert.AreEqual(expectedExpression[index], splitted.Expression);

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.Assert.IsSerializableBinaryFormat(splitted, d => {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedExpression[index], d.Expression);
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(4, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential2()
    {
      var range = ImapSequenceSet.Range(1L, 400L);
      var expectedExpression = new[] {
        "1:100",
        "101:200",
        "201:300",
        "301:400",
      };
      var expectedConcreteCount = new[] {
        100L,
        100L,
        100L,
        100L,
      };
      var index = 0;

      Assert.AreEqual(400L, range.ConcreteCount);

      foreach (var splitted in range.SplitIntoEach(100)) {
        Assert.AreEqual(expectedExpression[index], splitted.Expression);

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.Assert.IsSerializableBinaryFormat(splitted, d => {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedExpression[index], d.Expression);
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(4, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential3()
    {
      var range = ImapSequenceSet.Range(1L, 100L);
      var expectedExpression = new[] {
        "1:100",
      };
      var expectedConcreteCount = new[] {
        100L,
      };
      var index = 0;

      Assert.AreEqual(100L, range.ConcreteCount);

      foreach (var splitted in range.SplitIntoEach(100)) {
        Assert.AreEqual(expectedExpression[index], splitted.Expression);

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.Assert.IsSerializableBinaryFormat(splitted, d => {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedExpression[index], d.Expression);
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(1, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential4()
    {
      var range = ImapSequenceSet.Range(3L, 12L);
      var expectedExpression = new[] {
        "3:6",
        "7:10",
        "11:12",
      };
      var expectedConcreteCount = new[] {
        4L,
        4L,
        2L,
      };
      var index = 0;

      foreach (var splitted in range.SplitIntoEach(4)) {
        Assert.AreEqual(expectedExpression[index], splitted.Expression);

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.Assert.IsSerializableBinaryFormat(splitted, d => {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedExpression[index], d.Expression);
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(3, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential5()
    {
      var set1 = ImapUidSet.Range(3L, 8L);
      var set2 = ImapUidSet.Range(9L, 15L);
      var combined = set1.CombineWith(set2);
      var expectedExpression = new[] {
        "3:7",
        "8:12",
        "13:15",
      };
      var expectedConcreteCount = new[] {
        5L,
        5L,
        3L,
      };
      var index = 0;

      foreach (var splitted in combined.SplitIntoEach(5)) {
        Assert.AreEqual(expectedExpression[index], splitted.Expression);

        Assert.IsTrue(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.Assert.IsSerializableBinaryFormat(splitted, d => {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedExpression[index], d.Expression);
          Assert.IsTrue(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(3, index);
    }

    [Test]
    public void TestSplitIntoEach_NonSequential1()
    {
      var set1 = ImapSequenceSet.Range(7L, 10L);
      var set2 = ImapSequenceSet.Of(1L, 3L, 5L);
      var set3 = ImapSequenceSet.Range(20L, 25L);
      var set4 = ImapSequenceSet.Of(30L);
      var combined = set1.CombineWith(set2).CombineWith(set3).CombineWith(set4);
      var expectedExpression = new[] {
        "7:9",
        "10,1,3",
        "5,20,21",
        "22:24",
        "25,30",
      };
      var expectedConcreteCount = new[] {
        3L,
        3L,
        3L,
        3L,
        2L,
      };
      var index = 0;

      foreach (var splitted in combined.SplitIntoEach(3)) {
        Assert.AreEqual(expectedExpression[index], splitted.Expression);

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.Assert.IsSerializableBinaryFormat(splitted, d => {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedExpression[index], d.Expression);
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(5, index);
    }

    [Test]
    public void TestSplitIntoEach_NonSequential2()
    {
      var set1 = ImapSequenceSet.Range(7L, 10L);
      var set2 = ImapSequenceSet.Of(1L, 3L, 5L);
      var set3 = ImapSequenceSet.Range(20L, 25L);
      var set4 = ImapSequenceSet.Of(30L);
      var combined = set1.CombineWith(set2).CombineWith(set3).CombineWith(set4);
      var expectedExpression = new[] {
        "7,8,9,10,1",
        "3,5,20,21,22",
        "23,24,25,30",
      };
      var expectedConcreteCount = new[] {
        5L,
        5L,
        4L,
      };
      var index = 0;

      foreach (var splitted in combined.SplitIntoEach(5)) {
        Assert.AreEqual(expectedExpression[index], splitted.Expression);

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.Assert.IsSerializableBinaryFormat(splitted, d => {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedExpression[index], d.Expression);
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(3, index);
    }

    [Test]
    public void TestSplitIntoEach_Single()
    {
      var set = ImapSequenceSet.Of(1L);
      var index = 0;

      foreach (var splitted in set.SplitIntoEach(5)) {
        Assert.AreSame(splitted, set);

        index++;
      }

      Assert.AreEqual(1, index);
    }
  }
}
