using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Connection;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionTests {
    private void AssertSessionNotConnected(ImapSession session)
    {
      Assert.AreEqual(ImapSessionState.NotConnected, session.State);
      Assert.IsFalse(session.IsTransactionProceeding);

      Assert.Throws<InvalidOperationException>(() => { Assert.Fail("IsSecureConnection = {0}", session.IsSecureConnection); });

      Assert.Throws<InvalidOperationException>(() => session.NoOp());
      Assert.AreEqual(ImapSessionState.NotConnected, session.State);

      Assert.Throws<InvalidOperationException>(() => session.Login(new NetworkCredential("user", "pass")));
      Assert.AreEqual(ImapSessionState.NotConnected, session.State);

      Assert.Throws<InvalidOperationException>(() => session.ListRoot(out _));
      Assert.AreEqual(ImapSessionState.NotConnected, session.State);

      Assert.Throws<InvalidOperationException>(() => session.Check());
      Assert.AreEqual(ImapSessionState.NotConnected, session.State);

      Assert.DoesNotThrow(() => { session.ReceiveTimeout = 1; });
      Assert.DoesNotThrow(() => { Assert.AreEqual(1, session.ReceiveTimeout); });
      Assert.DoesNotThrow(() => { session.SendTimeout = 1; });
      Assert.DoesNotThrow(() => { Assert.AreEqual(1, session.SendTimeout); });
    }

    [Test]
    public void TestConstruct()
    {
      using (var session = new ImapSession()) {
        AssertSessionNotConnected(session);
      }
    }

    internal static void AssertSessionDisposed(ImapSession session, ImapPseudoServer server, bool disposedByTransactionTimeout = false)
    {
      Assert.AreEqual(ImapSessionState.Disconnected, session.State);

      if (!disposedByTransactionTimeout)
        Assert.IsFalse(session.IsTransactionProceeding);

      // attempt to send command
      Assert.Throws<ObjectDisposedException>(() => session.NoOp());

      // attempt to reconnect
      if (server != null) {
        Assert.Throws<ObjectDisposedException>(() => session.ConnectAsync(server.Host, server.Port));
        Assert.AreEqual(ImapSessionState.Disconnected, session.State);

        if (!disposedByTransactionTimeout)
          Assert.IsFalse(session.IsTransactionProceeding);
      }

      // attempt to re-disconnect
      Assert.Throws<ObjectDisposedException>(() => session.Disconnect());

      // attempt to re-dispose
      Assert.DoesNotThrow(() => session.Dispose());

      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("SendTimeout = {0}", session.SendTimeout); });
      Assert.Throws<ObjectDisposedException>(() => { session.SendTimeout = 0; });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("ReceiveTimeout = {0}", session.ReceiveTimeout); });
      Assert.Throws<ObjectDisposedException>(() => { session.ReceiveTimeout = 0; });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("TransactionTimeout = {0}", session.TransactionTimeout); });
      Assert.Throws<ObjectDisposedException>(() => { session.TransactionTimeout = 0; });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("IsSecureConnection = {0}", session.IsSecureConnection); });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("Authority = {0}", session.Authority); });
      Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("IsIdling = {0}", session.IsIdling); });
    }

    [Test]
    public void TestNotConnected_Dispose()
    {
      using (var session = new ImapSession()) {
        Assert.DoesNotThrow(() => session.Dispose());

        AssertSessionDisposed(session, null);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestNotConnected_Disconnect(bool logout)
    {
      using (var session = new ImapSession()) {
        Assert.DoesNotThrow(() => session.Disconnect(logout));

        AssertSessionDisposed(session, null);
      }
    }

    [Test]
    public void TestReceiveTimeout_SetZero()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.ReceiveTimeout = 0;

        Assert.AreEqual(0, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestReceiveTimeout_SetInfinite()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.ReceiveTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestSendTimeout_SetZero()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.SendTimeout = 0;

        Assert.AreEqual(0, session.SendTimeout);
      });
    }

    [Test]
    public void TestSendTimeout_SetInfinite()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.SendTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.SendTimeout);
      });
    }

    [Test]
    public void TestTransactionTimeout_SetZero()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.TransactionTimeout = 0;

        Assert.AreEqual(0, session.TransactionTimeout);
      });
    }

    [Test]
    public void TestTransactionTimeout_SetInfinite()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.TransactionTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.TransactionTimeout);
      });
    }

    private void Idle(Action<ImapSession> preprocessSession, Action<ImapSession> processSession)
    {
      TestCases.WithAuthenticatedSession(new[] {"IDLE"}, (session, server) => {
        preprocessSession(session);

        server.EnqueueResponse("IDLE\r\n",
                               "+ idling\r\n");

        using (var cts = new CancellationTokenSource()) {
          var t = session.IdleAsync(cts.Token);

          Assert.IsTrue(session.IsIdling);
          Assert.IsTrue(session.IsTransactionProceeding);

          processSession(session);

          server.EnqueueTaggedResponse("DONE\r\n",
                                       "$tag OK done\r\n");

          cts.Cancel();

          Assert.IsTrue(t.Result.IsSuccess);

          Assert.IsFalse(session.IsIdling);
          Assert.IsFalse(session.IsTransactionProceeding);
        }
      });
    }

    [Test]
    public void TestTransactionTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.TransactionTimeout = 1000;
        Assert.AreEqual(1000, session.TransactionTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.TransactionTimeout = 1000);
      });
    }

    [Test]
    public void TestSendTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.SendTimeout = 1000;
        Assert.AreEqual(1000, session.SendTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.SendTimeout = 1000);
      });
    }

    [Test]
    public void TestReceiveTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.ReceiveTimeout = 1000;
        Assert.AreEqual(1000, session.ReceiveTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.ReceiveTimeout = 1000);
      });
    }

    [Test]
    public void TestBeginNewTransaction_WhileTransactionProceeding_1()
    {
      Idle(delegate(ImapSession session) {
        Assert.IsFalse(session.IsTransactionProceeding);
      },
      delegate(ImapSession session) {
        Assert.IsTrue(session.IsTransactionProceeding);
        Assert.Throws<InvalidOperationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestBeginNewTransaction_WhileTransactionProceeding_2()
    {
      TestCases.WithConnectedSession((session, server) => {
        using (var transactionStartedEvent = new ManualResetEvent(false)) {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag ", // incomplete response
                                       null,
                                       () => transactionStartedEvent.Set());

          var noopTask = Task.Run(() => session.NoOp());

          transactionStartedEvent.WaitOne();

          Assert.IsTrue(session.IsTransactionProceeding);

          Assert.Throws<InvalidOperationException>(() => session.NoOp());

          server.EnqueuePushResponse("OK \r\n");

          Assert.DoesNotThrow(() => {
            if (!noopTask.Wait(TimeSpan.FromSeconds(1)))
              Assert.Fail("transaction not stopped");
          });
        }
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.HandlesIncapableAsException = false;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Binary));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = new ImapString[] {
          ImapLiteralString.Create("literal1", Encoding.ASCII, ImapLiteralOptions.Synchronizing),
          ImapLiteralString.Create("literal2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
          ImapLiteralString.Create("literal3", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
          ImapLiteralString.Create("literal4", Encoding.ASCII, ImapLiteralOptions.Literal),
          ImapLiteralString.Create("literal5", Encoding.ASCII, ImapLiteralOptions.Literal8),
          ImapLiteralString.Create("literal6", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable),
          ImapLiteralString.Create("literal7", Encoding.ASCII, ImapLiteralOptions.Synchronizing | ImapLiteralOptions.Literal8IfCapable),
          ImapLiteralString.Create("literal8", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable | ImapLiteralOptions.Literal8),
          ImapStringList.CreateParenthesized(new[] {
            ImapLiteralString.Create("nested1", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
            ImapLiteralString.Create("nested2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
            ImapLiteralString.Create("nested3", Encoding.ASCII, ImapLiteralOptions.Literal8),
            ImapLiteralString.Create("nested4", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable),
          }),
        };

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST " +
                        "{8}\r\nliteral1 " +
                        "{8+}\r\nliteral2 " +
                        "{8}\r\nliteral3 " +
                        "{8}\r\nliteral4 " +
                        "~{8}\r\nliteral5 " +
                        "{8}\r\nliteral6 " +
                        "{8}\r\nliteral7 " +
                        "~{8}\r\nliteral8 " +
                        "({7+}\r\nnested1 " +
                        "{7}\r\nnested2 " +
                        "~{7}\r\nnested3 " +
                        "{7}\r\nnested4)\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_LiteralNonSyncCapable()
    {
      TestCases.WithConnectedSession(new[] {"LITERAL+"}, (session, server) => {
        session.HandlesIncapableAsException = true;

        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = ImapStringList.CreateParenthesized(new ImapString[] {
          ImapLiteralString.Create("literal1", Encoding.ASCII, ImapLiteralOptions.Synchronizing),
          ImapLiteralString.Create("literal2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
          ImapLiteralString.Create("literal3", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
        });

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST (" +
                        "{8}\r\nliteral1 " +
                        "{8+}\r\nliteral2 " +
                        "{8+}\r\nliteral3)\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_LiteralNonSyncIncapable()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.HandlesIncapableAsException = true;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));

        Assert.Throws<ImapIncapableException>(delegate {
          session.GenericCommand("X-TEST", ImapLiteralString.Create("literal", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing));
        });
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_Literal8Capable()
    {
      TestCases.WithConnectedSession(new[] {"BINARY"}, (session, server) => {
        session.HandlesIncapableAsException = true;

        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Binary));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = ImapStringList.CreateParenthesized(new ImapString[] {
          ImapLiteralString.Create("literal1", Encoding.ASCII, ImapLiteralOptions.Literal),
          ImapLiteralString.Create("literal2", Encoding.ASCII, ImapLiteralOptions.Literal8),
          ImapLiteralString.Create("literal3", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable),
        });

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST (" +
                        "{8}\r\nliteral1 " +
                        "~{8}\r\nliteral2 " +
                        "~{8}\r\nliteral3)\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_Literal8Incapable()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.HandlesIncapableAsException = true;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Binary));

        Assert.Throws<ImapIncapableException>(delegate {
          session.GenericCommand("X-TEST", ImapLiteralString.Create("literal", Encoding.ASCII, ImapLiteralOptions.Literal8));
        });
      });
    }

    [Test]
    public void TestConnectAsync()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.IsFalse(session.IsIdling);
        Assert.IsNull(session.SelectedMailbox);
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        Assert.AreEqual(string.Format("imap://{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
        Assert.AreEqual(0, session.ServerCapabilities.Count);
      });
    }

    [Test]
    public void TestConnectAsync_ConnectionFailure_Cancellation_Connection()
    {
      var timeout = TimeSpan.FromMilliseconds(250);
      var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

      using (var session = new ImapSession()) {
        TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => {
          Assert.ThrowsAsync<OperationCanceledException>(async () => {
            using (var cts = new CancellationTokenSource(timeout)) {
              await session.ConnectAsync("imap.nonexistent.test", 143, cancellationToken: cts.Token);
            }
          });
        });

        AssertSessionNotConnected(session);
      }
    }

    [Test]
    public void TestConnectAsync_ConnectionFailure_SocketException()
    {
      using (var session = new ImapSession()) {
        var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await session.ConnectAsync("imap.invalid", 10143));

        Assert.IsInstanceOf<SocketException>(ex.InnerException);

        AssertSessionNotConnected(session);
      }
    }

    [Test]
    public void TestConnectAsync_TransactionFailure_Timeout_ServerNotAccepting()
    {
      var timeout = TimeSpan.FromMilliseconds(250);
      var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

      ImapPseudoServer.StartNotAccepting((server) => {
        using (var session = new ImapSession()) {
          session.TransactionTimeout = (int)timeout.TotalMilliseconds;

          TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => {
            var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

            Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
          });

          AssertSessionDisposed(session, server, disposedByTransactionTimeout: true);
        }
      });
    }

    [Test]
    public void TestConnectAsync_TransactionFailure_Timeout_NoGreetingResponse()
    {
      var timeout = TimeSpan.FromMilliseconds(250);
      var timeoutAcceptable = TimeSpan.FromMilliseconds(500);

      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse(null); // no greeting response

        using (var session = new ImapSession()) {
          session.TransactionTimeout = (int)timeout.TotalMilliseconds;

          TestUtils.Assert.Elapses(timeout, timeoutAcceptable, () => {
            var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

            Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
          });

          AssertSessionDisposed(session, server, disposedByTransactionTimeout: true);
        }
      });
    }

    [Test]
    public void TestConnectAsync_TransactionFailure_ByeResponse()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* BYE not available.\r\n",
                                   null, null, true);

        using (var session = new ImapSession()) {
          var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

          Assert.IsNull(ex.InnerException);
          StringAssert.Contains("not available.", ex.Message);

          AssertSessionDisposed(session, server);
        }
      });
    }

    [TestCase("tag OK ready\r\n")] // tagged status response
    [TestCase("tag NO not ready\r\n")] // tagged status response
    [TestCase("* NO no\r\n")] // untagged-status response with unexpected condition
    [TestCase("* BAD bad\r\n")] // untagged-status response with unexpected condition
    [TestCase("* X-DATA foo\r\n")] // data response
    [TestCase("+ continue\r\n")] // command continuation request
    public void TestConnectAsync_TransactionFailure_UnexpectedResponse(string response)
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse(response);

        using (var session = new ImapSession()) {
          var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<ImapUnexpectedResponseException>(ex.InnerException);
          StringAssert.Contains("session establishment failed (unexpected response)", ex.Message);

          var innerEx = ex.InnerException as ImapUnexpectedResponseException;

          Assert.IsNotNull(innerEx.CausedResponse);

          AssertSessionDisposed(session, server);
        }
      });
    }

    [TestCase("+OK pseudo POP server ready.\r\n")]
    [TestCase("+OK\r\n")]
    [TestCase("* OK [CAPABILITY\r\n")]
    [TestCase("* OK [CAPABILITY )] server ready.\r\n")]
    //[TestCase("* OK [CAPABILITY (IMAP4 IMAP4rev1)] server ready.\r\n")] // throws unwrapped ImapMalformedDataException
    public void TestConnectAsync_TransactionFailure_MalformedResponse(string response)
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse(response);

        using (var session = new ImapSession()) {
          var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await session.ConnectAsync(server.Host, server.Port));

          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<ImapMalformedResponseException>(ex.InnerException);
          StringAssert.Contains("session establishment failed (malformed or invalid response)", ex.Message);

          AssertSessionDisposed(session, server);
        }
      });
    }

    [Test]
    public void TestConnect_Ssl()
    {
      ImapPseudoServer.Start((server) => {
        var streamUpgraded = false;

        using (var session = new ImapSession(server.Host,
                                             server.Port,
                                             createAuthenticatedStream: (connection, baseStream, cancellationToken) => { streamUpgraded = true; return Task.FromResult<Stream>(baseStream); /* TODO: return SslStream */})) {
          Assert.IsFalse(session.IsIdling);
          Assert.IsNull(session.SelectedMailbox);
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(string.Format("imaps://{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
          Assert.IsTrue(streamUpgraded, "stream upgraded");
          Assert.IsTrue(session.IsSecureConnection);
        }
      });
    }

    [Test]
    public void TestConnectAsync_Ssl_ExceptionWhileUpgrading()
    {
      ImapPseudoServer.Start((server) => {
        using (var session = new ImapSession()) {
          var ex = Assert.ThrowsAsync<ImapSecureConnectionException>(async () => {
            await session.ConnectAsync(server.Host,
                                       server.Port,
                                       createAuthenticatedStream: (connection, baseStream, cancellationToken) => throw new System.Security.Authentication.AuthenticationException());
          });

          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);

          AssertSessionNotConnected(session);
        }
      });
    }

    [Test]
    public void TestConnectAsync_ResolveHostAddressCallback()
    {
      ImapPseudoServer.Start((server) => {
        using (var session = new ImapSession()) {
          const string pseudoServerHostName  = "imap.invalid";

          var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
            Assert.AreEqual(pseudoServerHostName, hostname);

            return Task.FromResult(server.ServerEndPoint.Address);
          });

          server.SetGreetingResponse("* OK ImapPseudoServer ready\r\n");

          Assert.DoesNotThrowAsync(async () => {
            await session.ConnectAsync(pseudoServerHostName,
                                       server.Port,
                                       resolveHostAddress: resolveHostAddress);
          });

          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual($"imap://{pseudoServerHostName}:{server.Port}/", session.Authority.AbsoluteUri);
          Assert.IsFalse(session.IsSecureConnection);
        }
      });
    }

    [Test]
    public void TestConnectAsync_ResolveHostAddressCallback_Exception()
    {
      using (var session = new ImapSession()) {
        const string pseudoServerHostName  = "imap.invalid";

        var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
          Assert.AreEqual(pseudoServerHostName, hostname);
#if NET45
          var tcs = new TaskCompletionSource<IPAddress>();
          tcs.SetException(new NotImplementedException());
          return tcs.Task;
#else
          return Task.FromException<IPAddress>(new NotImplementedException());
#endif
        });

        var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => {
          await session.ConnectAsync(pseudoServerHostName,
                                     resolveHostAddress: resolveHostAddress);
        });

        Assert.IsInstanceOf<NotImplementedException>(ex.InnerException);

        AssertSessionNotConnected(session);
      }
    }

    [Test]
    public void TestDispose() => DisconnectDispose(session => session.Dispose());

    [Test]
    public void TestDisconnect() => DisconnectDispose(session => session.Disconnect(logout: false));

    private void DisconnectDispose(Action<ImapSession> disposeSession)
    {
      TestCases.WithConnectedSession((session, server) => {
        disposeSession(session);

        AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestImap4Rev1Incapable()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY NONE\r\n" +
                                     "$tag OK ImapCapability completed\r\n");

        Assert.IsTrue(session.Capability(out var caps).IsSuccess);
        Assert.IsNotNull(caps);
        Assert.IsFalse(caps.Contains(ImapCapability.Imap4Rev1));

        session.HandlesIncapableAsException = true;

        var ex = Assert.Throws<ImapIncapableException>(() => session.NoOp());

        CollectionAssert.AreEqual(new[] { ImapCapability.Imap4Rev1 }, ex.RequiredCapabilities);
      });
    }

    [Test]
    public void TestGreeting_WithCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 AUTH=PLAIN CHILDREN THREAD=REFERENCES X-EXTENSION1 X-EXTENSION2] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(7, session.ServerCapabilities.Count);

          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Children));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.CRAMMD5));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

          Assert.IsTrue(session.ServerCapabilities.Contains("THREAD=REFERENCES"));
          Assert.IsTrue(session.ServerCapabilities.Contains("X-EXTENSION1"));
          Assert.IsTrue(session.ServerCapabilities.Contains("X-EXTENSION2"));
        }
      });
    }

    [Test]
    public void TestGreeting_WithNoCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(0, session.ServerCapabilities.Count);

          Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestGreeting_PreAuth_WithCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* PREAUTH [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(2, session.ServerCapabilities.Count);
          Assert.AreEqual(string.Format("imap://{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);

          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));
        }
      });
    }

    [Test]
    public void TestGreeting_WithNoImapCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY AUTH=PLAIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsTrue(!session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

          session.HandlesIncapableAsException = true;

          var ex = Assert.Throws<ImapIncapableException>(() => session.Login(new NetworkCredential("user", "pass")));

          CollectionAssert.AreEqual(new[] { ImapCapability.Imap4Rev1 }, ex.RequiredCapabilities);
        }
      });
    }

    [Test]
    public void TestGreeting_HomeServerReferral_AsException()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n",
                                   null, null, true);

        try {
          using (var session = new ImapSession(server.Host, server.Port, handlesReferralAsException: true)) {
            Assert.Fail("connected without exception");
          }
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Server not accepting connections.  Try SERVER2"));
          StringAssert.AreEqualIgnoringCase("IMAP://user;AUTH=*@SERVER2/", ex.ReferToUri.AbsoluteUri);

          TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate (ImapLoginReferralException deserialized) {
            Assert.IsNotNull(deserialized.ReferToUri);
            Assert.AreEqual(ex.ReferToUri, deserialized.ReferToUri);
          });
        }
      });
    }

    [Test]
    public void TestGreeting_HomeServerReferral_AsError()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n",
                                   null, null, true);

        var ex = Assert.Throws<ImapConnectionException>(delegate {
          using (var session = new ImapSession(server.Host, server.Port)) {
          }
        });

        Assert.IsNull(ex.InnerException);
        StringAssert.Contains("Server not accepting connections.  Try SERVER2", ex.Message);
      });
    }

    [Test]
    public void TestTransactionException_TransactionTimeout()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.TransactionTimeout = 250;
        session.SendTimeout = System.Threading.Timeout.Infinite;
        session.ReceiveTimeout = System.Threading.Timeout.Infinite;

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK"); // not respond NOOP response termination(CRLF)

        Assert.Throws<TimeoutException>(() => session.NoOp());

        AssertSessionDisposed(session, server, disposedByTransactionTimeout: true);
      });
    }

    [Test]
    public void TestTransactionException_SocketTimeout()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.TransactionTimeout = System.Threading.Timeout.Infinite;
        session.SendTimeout = 250;
        session.ReceiveTimeout = 250;

        // not respond NOOP
        //server.EnqueueTaggedResponse("$tag OK\r\n");

        Assert.Throws<TimeoutException>(() => session.NoOp());

        AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestTransactionException_InternalError()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "*\r\n" + // invalid response
                                     "$tag CAPABILITY Ok\r\n");

        Assert.Throws<ImapMalformedResponseException>(() => session.Capability());

        AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestDisconnectedFromServer_WithByeResponse()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

        server.EnqueueUntaggedByeResponse("* BYE Disconnected for inactivity.\r\n");

        Assert.IsTrue(session.NoOp().IsFinalResponseCondition(ImapResponseCondition.Bye));

        AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestDisconnectedFromServer_WithoutByeResponse()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [ALERT] will soon shutdown",
                                   null, null, true);

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.Throws<ImapConnectionException>(() => session.NoOp());

          AssertSessionDisposed(session, server);
        }
      });
    }

    [Test]
    public void TestReceiveResponse_UnexpectedResponse_CommandTagUnmatched()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("NOOP",
                               "0001 OK done\r\n");

        var ex = Assert.Throws<ImapUnexpectedResponseException>(() => session.NoOp());

        TestUtils.Assert.IsSerializableBinaryFormat(ex, deserialized => {
          Assert.IsNotNull(ex.CausedResponse);

          var taggedResponse = ex.CausedResponse as ImapTaggedStatusResponse;

          Assert.IsNotNull(taggedResponse);
          Assert.AreEqual(ImapResponseCondition.Ok, taggedResponse.Condition);
          Assert.AreEqual("0001", taggedResponse.Tag);
        });
      });
    }

    [Test]
    public void TestReceiveResponse_UnexpectedResponse_CommandContinuationRequest()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("NOOP",
                               "+ continue\r\n");

        var ex = Assert.Throws<ImapUnexpectedResponseException>(() => session.NoOp());

        TestUtils.Assert.IsSerializableBinaryFormat(ex, deserialized => {
          Assert.IsNotNull(ex.CausedResponse);

          var continuation = ex.CausedResponse as ImapCommandContinuationRequest;

          Assert.IsNotNull(continuation);
          Assert.AreEqual("continue", continuation.Text.ToString());
        });
      });
    }

    [Test]
    public void TestReceiveResponse_MalformedResponse_UndefinedCondition()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("NOOP",
                               "0000 GOOD done\r\n");

        var ex = Assert.Throws<ImapMalformedResponseException>(() => session.NoOp());

        TestUtils.Assert.IsSerializableBinaryFormat(ex, deserialized => {
          Assert.IsNotNull(ex.CausedResponse);
          Assert.AreEqual("0000 GOOD done\r\n", ex.CausedResponse);
        });
      });
    }

    [Test]
    public void TestListStatusSelect()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK authenticated\r\n");

        Assert.IsTrue(session.Login(new NetworkCredential("user", "pass")).IsSuccess);

        var expectedAuthority = string.Format("imap://user@{0}/", server.HostPort);

        Assert.AreEqual(expectedAuthority, session.Authority.AbsoluteUri);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        session.Capability();

        // LIST transaction
        server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                     "* LIST (\\HasChildren) \".\" \"INBOX\"\r\n" +
                                     "$tag OK LIST completed\r\n");

        Assert.IsTrue(session.List("*", out var mailboxes).IsSuccess);

        Assert.AreEqual(1, mailboxes.Count);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.HasChildren));
        Assert.AreEqual(expectedAuthority + "INBOX", mailboxes[0].Url.AbsoluteUri);

        // LSUB transaction
        server.EnqueueTaggedResponse("LSUB \"\" *\r\n",
                                     "* LSUB (\\HasChildren) \"/\" \"Trash\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \"/\" \"Trash/mail\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \".\" \"INBOX.&U,BTFw-\"\r\n" +
                                     "$tag OK LSUB completed\r\n");

        Assert.IsTrue(session.Lsub("*", out mailboxes).IsSuccess);

        Assert.AreEqual(4, mailboxes.Count);
        Assert.AreEqual("Trash", mailboxes[0].Name);
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.HasChildren));

        Assert.AreEqual("Trash/mail", mailboxes[1].Name);
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.HasNoChildren));
        Assert.AreEqual(expectedAuthority + "Trash/mail", mailboxes[1].Url.AbsoluteUri);

        Assert.AreEqual("INBOX.日本語", mailboxes[2].Name);
        Assert.AreEqual("INBOX.台北", mailboxes[3].Name);

        // STATUS transaction
        server.EnqueueTaggedResponse("STATUS Trash (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                     "* STATUS \"Trash\" (MESSAGES 3 RECENT 1 UIDVALIDITY 123456 UNSEEN 2 UIDNEXT 4)\r\n" +
                                     "$tag OK STATUS completed\r\n");

        Assert.IsTrue(session.Status(mailboxes[0], ImapStatusDataItems.StandardAll).IsSuccess);

        Assert.AreEqual(3, mailboxes[0].ExistsMessage);
        Assert.AreEqual(1, mailboxes[0].RecentMessage);
        Assert.AreEqual(123456, mailboxes[0].UidValidity);
        Assert.AreEqual(2, mailboxes[0].UnseenMessage);
        Assert.AreEqual(4, mailboxes[0].UidNext);

        // SELECT transaction
        Assert.IsNull(session.SelectedMailbox);

        server.EnqueueTaggedResponse("SELECT INBOX.&ZeVnLIqe-\r\n",
                                     "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                                     "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                                     "* 15 EXISTS\r\n" +
                                     "* 2 RECENT\r\n" +
                                     "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                                     "* OK [UIDNEXT 16]\r\n" +
                                     "* OK [UNSEEN 13]\r\n" +
                                     "$tag OK [READ-WRITE] SELECT completed\r\n");

        var mailbox = mailboxes[2];

        Assert.IsTrue(session.Select(mailbox).IsSuccess);

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.AreSame(mailbox, session.SelectedMailbox);

        Assert.AreEqual("INBOX.日本語", mailbox.Name);
        Assert.AreEqual(15, mailbox.ExistsMessage);
        Assert.AreEqual(2, mailbox.RecentMessage);
        Assert.AreEqual(0, mailbox.UnseenMessage);
        Assert.AreEqual(13, mailbox.FirstUnseen);
        Assert.AreEqual(1202674433, mailbox.UidValidity);
        Assert.AreEqual(16, mailbox.UidNext);
        Assert.IsFalse(mailbox.ReadOnly);
        Assert.IsTrue(mailbox.UidPersistent);
        Assert.AreEqual((new Uri(expectedAuthority + "INBOX.日本語;UIDVALIDITY=1202674433")).AbsoluteUri,
                        mailbox.Url.AbsoluteUri);

        /*
        Assert.AreEqual(selectedMailbox.Name, mailbox.Name);
        Assert.AreEqual(selectedMailbox.ExistsMessage, mailbox.ExistsMessage);
        Assert.AreEqual(selectedMailbox.RecentMessage, mailbox.RecentMessage);
        Assert.AreEqual(selectedMailbox.UnseenMessage, mailbox.UnseenMessage);
        Assert.AreEqual(selectedMailbox.FirstUnseen, mailbox.FirstUnseen);
        Assert.AreEqual(selectedMailbox.UidValidity, mailbox.UidValidity);
        Assert.AreEqual(selectedMailbox.UidNext, mailbox.UidNext);
        Assert.AreEqual(selectedMailbox.ReadOnly, mailbox.ReadOnly);
        Assert.AreEqual(selectedMailbox.ApplicableFlags.Count, mailbox.ApplicableFlags.Count);
        Assert.AreEqual(selectedMailbox.PermanentFlags.Count, mailbox.PermanentFlags.Count);
        */

        Assert.AreEqual(7, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("custom1"));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("custom2"));

        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.PermanentFlags.Contains("custom1"));

        // CLOSE transaction
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK CLOSE completed\r\n");

        Assert.IsTrue(session.Close().IsSuccess);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      });
    }

    [Test]
    public void TestGenericCommand_WithoutArguments()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        // generic command transaction
        server.EnqueueTaggedResponse("X-EXT-COMMAND\r\n",
                                     "* X-EXT-COMMAND data\r\n" +
                                     "$tag OK X-EXT-COMMAND completed\r\n");

        Assert.IsTrue(session.GenericCommand("X-EXT-COMMAND", out var dataResps).IsSuccess);

        Assert.AreEqual(1, dataResps.Count);
        Assert.AreEqual(2, dataResps[0].DataList.Count);
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].DataList[0].Format);
        Assert.AreEqual("X-EXT-COMMAND", dataResps[0].DataList[0].GetTextAsString());
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].DataList[1].Format);
        Assert.AreEqual("data", dataResps[0].DataList[1].GetTextAsString());
      });
    }

    [Test]
    public void TestGenericCommand_WithArguments()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        // generic command transaction
        server.EnqueueTaggedResponse("X-EXT-COMMAND ",
                                     "$tag OK X-EXT-COMMAND completed\r\n");

        var args = new[] {
          ImapStringList.CreateParenthesized(
            new ImapString[] {
              (ImapString)"arg1",
              (ImapString)"arg2",
              ImapStringList.CreateParenthesized(new[] { "arg3-1", "arg3-2" }),
              ImapLiteralString.Create("ascii text", OctetEncoding.EightBits)
            }
          ),
          ImapStringList.CreateParenthesized(
            new ImapString[] {
              ImapQuotedString.Create("quoted"),
              (ImapString)"nonquoted",
            }
          ),
        };

        Assert.IsTrue(session.GenericCommand("X-EXT-COMMAND", out var dataResps, args).IsSuccess);

        Assert.AreEqual("0002 X-EXT-COMMAND (arg1 arg2 (arg3-1 arg3-2) {10}\r\n" +
                        "ascii text) (\"quoted\" nonquoted)\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(0, dataResps.Count);
      });
    }

    [Test]
    public void TestCommand_Tag()
    {
      TestCases.WithConnectedSession((session, server) => {
        for (var n = 0; n < 0x10; n++) {
          server.EnqueueTaggedResponse($"{n:x4} NOOP\r\n", "$tag OK done\r\n");

          Assert.IsTrue(session.NoOp().IsSuccess);

          server.DequeueRequest();
        }
      });
    }

#if false
    [Test]
    public void TestCommand_Tag_NotOverflow()
    {
      const int commandTagMax = 0x10000;

      TestCases.WithConnectedSession((session, server) => {
        for (var n = 0; n < commandTagMax; n++) {
          server.EnqueueTaggedResponse($"{n:x4} NOOP\r\n", "$tag OK done\r\n");

          Assert.IsTrue(session.NoOp().IsSuccess);

          server.DequeueRequest();
        }

        server.EnqueueTaggedResponse("0000 NOOP\r\n", "$tag OK done\r\n");

        Assert.IsTrue(session.NoOp().IsSuccess);

        server.DequeueRequest();
      });
    }
#endif
  }
}
