using System;
using System.Collections.Generic;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionCommandsAnyStateTests {
    [Test]
    public void TestCapability()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(0, session.ServerCapabilities.Count);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 CHILDREN\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.Capability(out var capabilities).IsSuccess);

        Assert.IsNotNull(capabilities);
        Assert.AreEqual(2, capabilities.Count);
        Assert.IsTrue(capabilities.Contains(ImapCapability.Imap4Rev1));
        Assert.IsTrue(capabilities.Contains(ImapCapability.Children));

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(2, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Children));
      });
    }

    [Test]
    public void TestCapability_NotHandleIncapableAsException()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.HandlesIncapableAsException = true;

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

        // CAPABILITY transaction 1
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.Capability().IsSuccess);

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

        // CAPABILITY transaction 2
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.Capability().IsSuccess);

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
      });
    }

    [Test]
    public void TestNoOp_InNonAuthenticatedState()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

        // NOOP transaction
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.NoOp().IsSuccess);

        Assert.AreEqual("0000 NOOP\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestNoOp_InAuthenticatedState()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        // NOOP transaction
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.NoOp().IsSuccess);

        Assert.AreEqual("0002 NOOP\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestNoOp_InSelectedState()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.Selected, session.State);

        // NOOP transaction
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.NoOp().IsSuccess);

        Assert.AreEqual("0004 NOOP\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestNoOp_StatusUpdate()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // NOOP transaction
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 22 EXPUNGE\r\n" +
                                     "* 18 EXPUNGE\r\n" +
                                     "* 23 EXISTS\r\n" +
                                     "* 3 RECENT\r\n" +
                                     "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                                     "* 14 FETCH (FLAGS (\\Seen \\Deleted))\r\n" +
                                     "$tag OK NOOP completed\r\n");

        Assert.IsTrue(session.NoOp().IsSuccess);

        Assert.AreEqual(23L, session.SelectedMailbox.ExistsMessage, "selected mailbox exist message count");
        Assert.AreEqual(3L, session.SelectedMailbox.RecentMessage, "selected mailbox recent message count");
        Assert.IsNotNull(session.SelectedMailbox.ApplicableFlags, "selected mailbox applicable flags");
        Assert.AreEqual(5, session.SelectedMailbox.ApplicableFlags.Count);
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Draft));
      });
    }

    [Test]
    public void TestNoOp_StatusUpdate_ExistsBroken()
    {
      TestCases.WithMailboxSelectedSession(1, (session, server) => {
        Assert.AreEqual(1L, session.SelectedMailbox.ExistsMessage);

        // NOOP transaction
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 1 EXPUNGE\r\n" +
                                     "* 1 EXPUNGE\r\n" +
                                     "* 1 EXPUNGE\r\n" +
                                     "$tag OK NOOP completed\r\n");

        Assert.IsTrue(session.NoOp().IsSuccess);

        Assert.AreEqual(0L, session.SelectedMailbox.ExistsMessage);
      });
    }

    [Test]
    public void TestLogout_InNonAuthenticatedState()
    {
      ImapPseudoServer.Start((server) => {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "* BYE Logging out\r\n" +
                                     "$tag OK logged out.\r\n",
                                     isByeResponse: true);

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsTrue(session.Logout().IsFinalResponseCondition(ImapResponseCondition.Bye));

          ImapSessionTests.AssertSessionDisposed(session, server);
        }
      });
    }

    [Test]
    public void TestLogout_InAuthenticatedState()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "* BYE Logging out\r\n" + 
                                     "$tag OK logged out.\r\n");

        Assert.IsTrue(session.Logout().IsFinalResponseCondition(ImapResponseCondition.Bye));

        ImapSessionTests.AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestLogout_InSelectedState()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.IsNotNull(session.SelectedMailbox);

        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "* BYE Logging out\r\n" + 
                                     "$tag OK logged out.\r\n");

        Assert.IsTrue(session.Logout().IsFinalResponseCondition(ImapResponseCondition.Bye));

        ImapSessionTests.AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestLogout_AfterDisconnect() => LogoutAfterDisconnectOrDispose(session => session.Disconnect(logout: false));

    [Test]
    public void TestLogout_AfterDispose() => LogoutAfterDisconnectOrDispose(session => session.Dispose());

    private void LogoutAfterDisconnectOrDispose(Action<ImapSession> disposeAction)
    {
      TestCases.WithConnectedSession((session, server) => {
        disposeAction(session);

        Assert.Throws<ObjectDisposedException>(() => session.Logout());

        ImapSessionTests.AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestLogout_NotConnected()
    {
      using (var session = new ImapSession()) {
        Assert.Throws<InvalidOperationException>(() => session.Logout());
      }
    }

    [Test]
    public void TestID()
    {
      TestCases.WithAuthenticatedSession(new[] {"ID"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        // ID transaction
        server.EnqueueTaggedResponse("ID (\"name\" \"sodr\" \"version\" \"19.34\" \"vendor\" \"Pink Floyd Music Limited\")\r\n",
                                     "* ID NIL\r\n" +
                                     "$tag OK ID completed\r\n");

        Assert.IsTrue(session.ID(new Dictionary<string, string>() {
          {"name", "sodr"},
          {"version", "19.34"},
          {"vendor", "Pink Floyd Music Limited"},
        }, out var serverParameterList).IsSuccess);

        Assert.IsNotNull(serverParameterList);
        Assert.AreEqual(0, serverParameterList.Count);
        Assert.AreEqual(0, session.ServerID.Count);
      });
    }

    [Test]
    public void TestID_ParameterNil()
    {
      TestCases.WithAuthenticatedSession(new[] {"ID"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        // ID transaction
        server.EnqueueTaggedResponse("ID (\"name\" NIL \"version\" NIL \"vendor\" NIL)\r\n",
                                     "* ID (\"name\" \"Cyrus\" \"version\" \"1.5\")\r\n" +
                                     "$tag OK ID completed\r\n");

        Assert.IsTrue(session.ID(new Dictionary<string, string>() {
          {"name", null},
          {"version", null},
          {"vendor", null},
        }).IsSuccess);

        Assert.AreEqual(2, session.ServerID.Count);
        Assert.AreEqual("Cyrus", session.ServerID["name"]);
        Assert.AreEqual("Cyrus", session.ServerID["NAME"]);
        Assert.AreEqual("1.5", session.ServerID["version"]);
        Assert.AreEqual("1.5", session.ServerID["VERSION"]);
      });
    }

    [Test]
    public void TestID_TooMatchFieldValuePair()
    {
      TestCases.WithAuthenticatedSession(new[] {"ID"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        var paramList = new Dictionary<string, string>();

        for (var i = 0; i < 31; i++) {
          paramList.Add(string.Format("field{0}", i), string.Format("value{0}", i));
        }

        Assert.Throws<ArgumentException>(() => session.ID(paramList));
      });
    }

    [Test]
    public void TestID_TooLongField()
    {
      TestCases.WithAuthenticatedSession(new[] {"ID"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        Assert.Throws<ArgumentException>(() => session.ID(new Dictionary<string, string>() {
          {new string('x', 31), "value"},
        }));
      });
    }

    [Test]
    public void TestID_TooLongValue()
    {
      TestCases.WithAuthenticatedSession(new[] {"ID"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        Assert.Throws<ArgumentException>(() => session.ID(new Dictionary<string, string>() {
          {"field", new string('x', 1025)},
        }));
      });
    }

    [Test]
    public void TestID_ContaisNoAsciiChar()
    {
      TestCases.WithAuthenticatedSession(new[] {"ID"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        Assert.Throws<ArgumentException>(() => session.ID(new Dictionary<string, string>() {
          {"オワタ", "＼(＾o＾)／"},
        }));
      });
    }

    [Test]
    public void TestLanguage_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Language));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.LanguageSelectDefault());
      });
    }

    [Test]
    public void TestLanguage_ListSupported()
    {
      TestCases.WithAuthenticatedSession(new[] {"LANGUAGE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueTaggedResponse("LANGUAGE\r\n",
                                     "* LANGUAGE (EN DE IT i-default)\r\n" +
                                     "$tag OK Supported languages have been enumerated\r\n");

        Assert.IsTrue(session.LanguageListSupported(out var supportedLanguageTags).IsSuccess);

        Assert.AreEqual(4, supportedLanguageTags.Count);
        Assert.AreEqual("EN", supportedLanguageTags[0]);
        Assert.AreEqual("DE", supportedLanguageTags[1]);
        Assert.AreEqual("IT", supportedLanguageTags[2]);
        Assert.AreEqual("i-default", supportedLanguageTags[3]);
      });
    }

    [Test]
    public void TestLanguage_SelectDefault()
    {
      TestCases.WithAuthenticatedSession(new[] {"LANGUAGE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        server.EnqueueTaggedResponse("LANGUAGE default\r\n",
                                     "* LANGUAGE (DE)\r\n" +
                                     "$tag OK Sprachwechsel durch LANGUAGE-Befehl ausgeführt\r\n",
                                     () => server.SetCharset(Encoding.UTF8),
                                     null);

        var result = session.LanguageSelectDefault(out var selectedLanguageTag);

        Assert.IsTrue(result.IsSuccess);
        Assert.AreEqual("DE", selectedLanguageTag);
        Assert.AreEqual("Sprachwechsel durch LANGUAGE-Befehl ausgeführt", result.TaggedStatusResponse.ResponseText.Text);

        Assert.AreEqual("DE", session.SelectedLanguage);
      });
    }

    [Test]
    public void TestLanguage_SelectSupported()
    {
      TestCases.WithAuthenticatedSession(new[] {"LANGUAGE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueTaggedResponse("LANGUAGE FR-CA EN-CA\r\n",
                                     "* LANGUAGE (EN)\r\n" +
                                     "$tag OK Now speaking English\r\n");

        var result = session.Language(out var selectedLanguageTag, "FR-CA", "EN-CA");

        Assert.IsTrue(result.IsSuccess);
        Assert.AreEqual("EN", selectedLanguageTag);
        Assert.AreEqual("Now speaking English", result.TaggedStatusResponse.ResponseText.Text);

        Assert.AreEqual("EN", session.SelectedLanguage);
      });
    }

    [Test]
    public void TestLanguage_SelectWithCultureInfo()
    {
      TestCases.WithAuthenticatedSession(new[] {"LANGUAGE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueTaggedResponse("LANGUAGE ja-JP\r\n",
                                     "$tag NO Unsupported language\r\n");

        var result = session.Language(new System.Globalization.CultureInfo("ja-JP"));

        Assert.IsFalse(result.IsSuccess);
        Assert.AreEqual("Unsupported language", result.TaggedStatusResponse.ResponseText.Text);

        Assert.IsNull(session.SelectedLanguage);
      });
    }

    [Test]
    public void TestLanguage_NamespaceResponse()
    {
      TestCases.WithAuthenticatedSession(new[] {"LANGUAGE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        server.EnqueueTaggedResponse("LANGUAGE DE",
                                     "* LANGUAGE (DE)\r\n" +
                                     "* NAMESPACE ((\"\" \"/\")) ((\"Other Users/\" \"/\" \"TRANSLATION\"" +
                                     " (\"Andere Ben&APw-tzer/\"))) ((\"Public Folders/\" \"/\" " +
                                     " \"TRANSLATION\" (\"Gemeinsame Postf&AM8-cher/\")))\r\n" +
                                     "$tag OK Sprachwechsel durch LANGUAGE-Befehl ausgeführt\r\n",
                                     () => server.SetCharset(Encoding.UTF8),
                                     null);

        var result = session.Language("DE");

        Assert.IsTrue(result.IsSuccess);
        Assert.AreEqual("Sprachwechsel durch LANGUAGE-Befehl ausgeführt", result.TaggedStatusResponse.ResponseText.Text);

        Assert.AreEqual("DE", session.SelectedLanguage);

        Assert.AreEqual(1, session.Namespaces.PersonalNamespaces.Count);
        Assert.AreEqual(string.Empty, session.Namespaces.PersonalNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.PersonalNamespaces[0].HierarchyDelimiter);

        Assert.AreEqual(1, session.Namespaces.OtherUsersNamespaces.Count);
        Assert.AreEqual("Other Users/", session.Namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.OtherUsersNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual("Andere Benützer/", session.Namespaces.OtherUsersNamespaces[0].TranslatedPrefix);

        Assert.AreEqual(1, session.Namespaces.SharedNamespaces.Count);
        Assert.AreEqual("Public Folders/", session.Namespaces.SharedNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.SharedNamespaces[0].HierarchyDelimiter);
        // RFC error? (Gemeinsame Postfächer)
        Assert.AreEqual("Gemeinsame PostfÏcher/", session.Namespaces.SharedNamespaces[0].TranslatedPrefix);
      });
    }

    [Test]
    public void TestLanguage_ReceiveResponseAsUTF8()
    {
      TestCases.WithAuthenticatedSession(new[] {"LANGUAGE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        server.EnqueueTaggedResponse("LANGUAGE JA\r\n",
                                     "* LANGUAGE (JA)\r\n" +
                                     "$tag OK 完了\r\n",
                                     () => server.SetCharset(Encoding.UTF8),
                                     null);

        var result = session.Language("JA");

        Assert.IsTrue(result.IsSuccess);
        Assert.AreEqual("完了", result.TaggedStatusResponse.ResponseText.Text);

        // following command/response
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK 完了\r\n");

        session.NoOp();

        Assert.IsTrue(result.IsSuccess);
        Assert.AreEqual("完了", result.TaggedStatusResponse.ResponseText.Text);
      });
    }
  }
}
