using System;
using System.IO;
using System.Linq;
using System.Text;
using NUnit.Framework;

using Smdn.Text.Encodings;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionCommandsFetchTests {
    [Test]
    public void TestFetch_EmptySequenceSet()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ArgumentException>(() => session.Fetch(ImapSequenceSet.Empty(), ImapFetchDataItems.Fast, out ImapMessageFullAttributes[] _));
      });
    }

    [Test]
    public void TestFetch_EmptyUidSet()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ArgumentException>(() => session.Fetch(ImapUidSet.Empty(), ImapFetchDataItems.Fast, out ImapMessageFullAttributes[] _));
      });
    }

    [Test]
    public void TestFetch_ResultMustBeInReceivedOrder()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* 3 FETCH (UID 3)\r\n" +
                                     "* 2 FETCH (UID 2)\r\n" +
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Range(1, 3), ImapFetchDataItems.Uid, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(3L, messages[0].Uid);
        Assert.AreEqual(3L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=3", messages[0].Url.AbsoluteUri);

        Assert.AreEqual(2L, messages[1].Uid);
        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=2", messages[1].Url.AbsoluteUri);

        Assert.AreEqual(1L, messages[2].Uid);
        Assert.AreEqual(1L, messages[2].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=1", messages[2].Url.AbsoluteUri);
      });
    }

    [Test]
    public void TestFetch_UntaggedNoResponseMustBeIgnored1()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "* NO fetch message 2 failed\r\n" +
                                     "* 3 FETCH (UID 3)\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Range(1, 3), ImapFetchDataItems.Uid, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(2, messages.Length);

        Assert.AreEqual(1L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=1", messages[0].Url.AbsoluteUri);

        Assert.AreEqual(3L, messages[1].Uid);
        Assert.AreEqual(3L, messages[1].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=3", messages[1].Url.AbsoluteUri);
      });
    }

    [Test]
    public void TestFetch_UntaggedNoResponseMustBeIgnored2()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* NO fetch message 2 failed\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Range(1, 3), ImapFetchDataItems.Uid, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(0, messages.Length);
      });
    }

    [Test]
    public void TestFetch_MalformedResponseShouldBeIgnored1()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "* 2 FETCH boom\r\n" +
                                     "* 3 FETCH (UID 3)\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Range(1, 3), ImapFetchDataItems.Uid, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(2, messages.Length);

        Assert.AreEqual(1L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=1", messages[0].Url.AbsoluteUri);

        Assert.AreEqual(3L, messages[1].Uid);
        Assert.AreEqual(3L, messages[1].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=3", messages[1].Url.AbsoluteUri);
      });
    }

    [Test]
    public void TestFetch_MalformedResponseShouldBeIgnored2()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                     "* 1 FETCH boom\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Uid, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(0, messages.Length);
      });
    }

    [Test]
    public void TestFetch_Attributes()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (UID FLAGS INTERNALDATE RFC822.SIZE)\r\n",
                                     "* 1 FETCH (" +
                                     "UID 4 " + 
                                     "FLAGS (\\Answered custom1) " +
                                     "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                                     "RFC822.SIZE 1716" + 
                                     ")\r\n" + 
                                     "$tag OK FETCH completed\r\n");

        var dataItem =
          ImapFetchDataItems.Uid +
          ImapFetchDataItems.Flags +
          ImapFetchDataItems.InternalDate +
          ImapFetchDataItems.Rfc822Size;

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), dataItem, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        var message = messages[0];

        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(4, message.Uid);
        Assert.AreEqual(1, message.Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=4", message.Url.AbsoluteUri);

        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(message.Flags.Contains("custom1"));
        Assert.AreEqual(new DateTimeOffset(2008, 2, 19, 8, 15, 48, new TimeSpan(+9, 0, 0)), message.InternalDate);
        Assert.AreEqual(1716, message.Rfc822Size);
      });
    }

    [Test]
    public void TestFetch_Attributes_NotRetrieveResult()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (UID FLAGS INTERNALDATE RFC822.SIZE)\r\n",
                                     "* 1 FETCH (" +
                                     "UID 4 " + 
                                     "FLAGS (\\Answered custom1) " +
                                     "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                                     "RFC822.SIZE 1716" + 
                                     ")\r\n" + 
                                     "$tag OK FETCH completed\r\n");

        var dataItem =
          ImapFetchDataItems.Uid +
          ImapFetchDataItems.Flags +
          ImapFetchDataItems.InternalDate +
          ImapFetchDataItems.Rfc822Size;

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), dataItem).IsSuccess);
      });
    }

    [TestCase("invalid date and time")]
    [TestCase(" 9-Mar-10 00:58:29 -0500")]
    [TestCase("99-Mar-2010 00:58:29 -0500")]
    [TestCase("19-Mar-12010 00:58:29 -0500")]
    [TestCase(" 9-Mar-2010 00:58:29 0500")]
    public void TestFetch_Attributes_MalformedDateAndTime(string internalDate)
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (INTERNALDATE)\r\n",
                                     $"* 1 FETCH (INTERNALDATE \"{internalDate}\")\r\n" +
                                     "$tag OK FETCH completed\r\n");

        // invalid internal date will be ignored
        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.InternalDate, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(0, messages.Length);
      });
    }

    [Test]
    public void TestFetch_DynamicAttributes()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (FLAGS)\r\n",
                                     "* 1 FETCH (FLAGS (\\Answered custom1))\r\n" + 
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Flags, out ImapMessageDynamicAttributes[] messages).IsSuccess);

        Assert.IsNotNull(messages);
        Assert.AreEqual(1, messages.Length);

        var message = messages[0];

        Assert.AreEqual(1L, message.Sequence);

        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(message.Flags.Contains("custom1"));

        Assert.DoesNotThrow(() => message.ToString());
      });
    }

    [Test]
    public void TestFetch_DynamicAttributes_NotRetrieveResult()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (FLAGS)\r\n",
                                     "* 1 FETCH (FLAGS (\\Answered custom1))\r\n" + 
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Flags).IsSuccess);
      });
    }

    [Test]
    public void TestFetch_Envelope()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (ENVELOPE)\r\n",
                                     "* 1 FETCH (ENVELOPE (" + 
                                     "\"Tue, 19 Feb 2008 08:15:48 +0900\" " +
                                     "\"fetch test\" " + 
                                     "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                                     "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                                     "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                                     "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                                     "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                                     "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                                     "\"<in.reply.to@mail.example.com>\" " + 
                                     "\"<message.id@mail.example.com>\" " + 
                                     "))\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Envelope, out ImapMessageStaticAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        var message = messages[0];

        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(0L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", message.Url.AbsoluteUri);

        // envelope
        ImapEnvelope envelope = message.Envelope;
        Assert.AreEqual("Tue, 19 Feb 2008 08:15:48 +0900", envelope.Date);
        Assert.AreEqual("fetch test", envelope.Subject);
        Assert.AreEqual("fetch test", envelope.GetDecodedSubject());
        Assert.AreEqual(1, envelope.From.Count);
        Assert.AreEqual("from address", envelope.From[0].Name);
        Assert.AreEqual(1, envelope.Sender.Count);
        Assert.AreEqual("sender address", envelope.Sender[0].Name);
        Assert.AreEqual(1, envelope.ReplyTo.Count);
        Assert.AreEqual("reply-to address", envelope.ReplyTo[0].Name);
        Assert.AreEqual(1, envelope.To.Count);
        Assert.AreEqual("to address", envelope.To[0].Name);
        Assert.AreEqual(1, envelope.Cc.Count);
        Assert.AreEqual("cc address", envelope.Cc[0].Name);
        Assert.AreEqual(1, envelope.Bcc.Count);
        Assert.AreEqual("bcc address", envelope.Bcc[0].Name);
        Assert.AreEqual("<in.reply.to@mail.example.com>", envelope.InReplyTo);
        Assert.AreEqual("<message.id@mail.example.com>", envelope.MessageId);
      });
    }

    [Test]
    public void TestFetch_BodyStructure_SinglePart()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (BODY)\r\n",
                                     "* 1 FETCH (BODY (\"text\" \"plain\" (\"charset\" \"ISO-2022-JP\") NIL NIL \"7bit\" 22 1))\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Body, out ImapMessageStaticAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        var message = messages[0];

        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(0L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", message.Url.AbsoluteUri);

        // body structure
        Assert.IsNotNull(message.BodyStructure);
        Assert.IsNull(message.BodyStructure.ParentStructure);
        Assert.IsFalse(message.BodyStructure.IsMultiPart);
        Assert.IsFalse(message.BodyStructure.HasEncapsulatedMessage);
        Assert.IsTrue(message.BodyStructure is ImapSinglePartBodyStructure);

        var structure = message.BodyStructure as ImapSinglePartBodyStructure;

        Assert.AreEqual(string.Empty, structure.Section);
        Assert.AreEqual("text/plain", (string)structure.MediaType);
        Assert.AreEqual(1, structure.Parameters.Count);
        Assert.AreEqual("ISO-2022-JP", structure.Parameters["charset"]);
        Assert.IsNull(structure.Id);
        Assert.IsNull(structure.Description);
        Assert.AreEqual("7bit", structure.Encoding);
        Assert.AreEqual(22, structure.Size);
        Assert.AreEqual(1, structure.LineCount);

        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", message.Url.AbsoluteUri);
      });
    }

    [Test]
    public void TestFetch_BodyStructure_MultiPart()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (BODY)\r\n",
                                     "* 1 FETCH (BODY " +
                                     "((\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\") NIL NIL \"7BIT\" 1152 " +
                                     "23)(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\" \"NAME\" \"cc.diff\") " +
                                     "\"<960723163407.20117h@cac.washington.edu>\" \"Compiler diff\" " +
                                     "\"BASE64\" 4554 73) \"MIXED\")" +
                                     ")\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Body, out ImapMessageStaticAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        var message = messages[0];

        Assert.AreEqual(0L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", message.Url.AbsoluteUri);

        // body structure
        Assert.IsNotNull(message.BodyStructure);
        Assert.IsNull(message.BodyStructure.ParentStructure);
        Assert.IsTrue(message.BodyStructure.IsMultiPart);
        Assert.IsFalse(message.BodyStructure.HasEncapsulatedMessage);
        Assert.IsTrue(message.BodyStructure is ImapMultiPartBodyStructure);

        var structure = message.BodyStructure as ImapMultiPartBodyStructure;

        Assert.AreEqual(string.Empty, structure.Section);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", structure.Url.AbsoluteUri);

        var part1 = structure.NestedPartStructures[0] as ImapSinglePartBodyStructure;

        Assert.IsNotNull(part1);
        Assert.AreEqual("1", part1.Section);
        Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(part1.Url.AbsoluteUri));

        var part2 = structure.NestedPartStructures[1] as ImapSinglePartBodyStructure;

        Assert.IsNotNull(part2);
        Assert.AreEqual("2", part2.Section);
        Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(part2.Url.AbsoluteUri));
      });
    }

    [Test]
    public void TestFetch_BodyStructure_MultiPart_Uid()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (UID BODY)\r\n",
                                     "* 1 FETCH (UID 6 BODY "+
                                     "((\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\") NIL NIL \"7BIT\" 1152 " + 
                                     "23)(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\" \"NAME\" \"cc.diff\") " + 
                                     "\"<960723163407.20117h@cac.washington.edu>\" \"Compiler diff\" " + 
                                     "\"BASE64\" 4554 73) \"MIXED\")" + 
                                     ")\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Uid + ImapFetchDataItems.Body, out ImapMessageStaticAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        var message = messages[0];

        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(6L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=6", message.Url.AbsoluteUri);

        // body structure
        Assert.IsNotNull(message.BodyStructure);
        Assert.IsNull(message.BodyStructure.ParentStructure);
        Assert.IsTrue(message.BodyStructure.IsMultiPart);
        Assert.IsFalse(message.BodyStructure.HasEncapsulatedMessage);
        Assert.IsTrue(message.BodyStructure is ImapMultiPartBodyStructure);

        var structure = message.BodyStructure as ImapMultiPartBodyStructure;

        Assert.AreEqual("multipart/MIXED", (string)structure.MediaType);
        Assert.AreEqual(string.Empty, structure.Section);
        Assert.AreEqual(2, structure.NestedPartStructures.Count);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=6", structure.Url.AbsoluteUri);

        var part1 = structure.NestedPartStructures[0] as ImapSinglePartBodyStructure;

        Assert.IsNotNull(part1);
        Assert.AreSame(structure, part1.ParentStructure);
        Assert.AreEqual("1", part1.Section);
        Assert.AreEqual("TEXT/PLAIN", (string)part1.MediaType);
        Assert.AreEqual(1, part1.Parameters.Count);
        Assert.AreEqual("US-ASCII", part1.Parameters["CHARSET"]);
        Assert.IsNull(part1.Id);
        Assert.IsNull(part1.Description);
        Assert.AreEqual("7BIT", part1.Encoding);
        Assert.AreEqual(1152, part1.Size);
        Assert.AreEqual(23, part1.LineCount);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=6/;SECTION=1", part1.Url.AbsoluteUri);

        var part2 = structure.NestedPartStructures[1] as ImapSinglePartBodyStructure;

        Assert.IsNotNull(part2);
        Assert.AreSame(structure, part2.ParentStructure);
        Assert.AreEqual("2", part2.Section);
        Assert.AreEqual("TEXT/PLAIN", (string)part2.MediaType);
        Assert.AreEqual(2, part2.Parameters.Count);
        Assert.AreEqual("US-ASCII", part2.Parameters["CHARSET"]);
        Assert.AreEqual("cc.diff", part2.Parameters["NAME"]);
        Assert.AreEqual("<960723163407.20117h@cac.washington.edu>", part2.Id);
        Assert.AreEqual("Compiler diff", part2.Description);
        Assert.AreEqual("BASE64", part2.Encoding);
        Assert.AreEqual(4554, part2.Size);
        Assert.AreEqual(73, part2.LineCount);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=6/;SECTION=2", part2.Url.AbsoluteUri);
      });
    }

    [Test]
    public void TestFetch_Macro()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 FAST\r\n",
                                     "* 1 FETCH (" +
                                     "FLAGS (\\Answered custom1) " +
                                     "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                                     "RFC822.SIZE 1716" + 
                                     ")\r\n" + 
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.Fast, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);
        Assert.AreEqual(new DateTimeOffset(2008, 2, 19, 8, 15, 48, new TimeSpan(+9, 0, 0)), messages[0].InternalDate);
      });
    }

    private string ConvertMessageBodyToString(Stream stream)
    {
      using (var reader = new StreamReader(stream, OctetEncoding.SevenBits)) {
        return reader.ReadToEnd();
      }
    }

    [Test]
    public void TestFetch_Body()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1:2 (BODY.PEEK[])\r\n",
                                     "* 1 FETCH (BODY[] {12}\r\ntest message)\r\n" +
                                     "* 2 FETCH (BODY[] \"test message\")\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Range(1, 2), ImapFetchDataItems.BodyPeek(), out ImapMessageFullAttributes[] messages).IsSuccess);

        Assert.AreEqual(2, messages.Length);

        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", messages[0].Url.AbsoluteUri);

        Assert.IsNotNull(messages[0].BodySpecifiers);
        Assert.AreEqual(1, messages[0].BodySpecifiers.Count);
        Assert.IsTrue(messages[0].BodySpecifiers.Contains("BODY[]"));

        messages[0].ReadBodyOf("body[]", stream => {
          Assert.AreEqual("test message", ConvertMessageBodyToString(stream));
        });
        messages[0].ReadFirstBody(stream => {
          Assert.AreEqual("test message", ConvertMessageBodyToString(stream));
        });

        Assert.AreEqual(0L, messages[1].Uid);
        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?2", messages[1].Url.AbsoluteUri);

        Assert.IsNotNull(messages[1].BodySpecifiers);
        Assert.AreEqual(1, messages[1].BodySpecifiers.Count);
        Assert.IsTrue(messages[0].BodySpecifiers.Contains("BODY[]"));

        messages[1].ReadBodyOf("body[]", stream => {
          Assert.AreEqual("test message", ConvertMessageBodyToString(stream));
        });
        messages[1].ReadFirstBody(stream => {
          Assert.AreEqual("test message", ConvertMessageBodyToString(stream));
        });
      });
    }

    private string CreateLiteral8(string body, Encoding encoding, int maxoctet)
    {
      return CreateLiteral(body, encoding, true, maxoctet);
    }

    private string CreateLiteral(string body, Encoding encoding, int maxoctet)
    {
      return CreateLiteral(body, encoding, false, maxoctet);
    }

    private string CreateLiteral(string body, Encoding encoding, bool literal8, int maxoctet)
    {
      var literalizedBody = encoding.GetBytes(body);

      var octetBody = (maxoctet == 0) ? literalizedBody.Length : maxoctet;

      return string.Format("{0}{{{1}}}\r\n{2}",
                           literal8 ? "~" : string.Empty,
                           octetBody,
                           OctetEncoding.EightBits.GetString(literalizedBody, 0, octetBody));
    }

    [Test]
    public void TestFetch_Section()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var body = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 25 Feb 2008 15:01:12 +0900\r\n";

        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (BODY.PEEK[header.fields (from to subject date)])\r\n",
                                     "* 1 FETCH (BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)] " +
                                     CreateLiteral(body, Encoding.ASCII, 0) +
                                     ")\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.BodyPeek("header.fields (from to subject date)"), out ImapMessageBodyAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.IsNotNull(messages[0].BodySpecifiers);
        Assert.AreEqual(1, messages[0].BodySpecifiers.Count);
        Assert.IsTrue(messages[0].BodySpecifiers.Contains("BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)]"));

        messages[0].ReadBodyOf("body[header.fields (from to subject date)]", stream => {
          Assert.AreEqual(body, ConvertMessageBodyToString(stream));
        });
        messages[0].ReadFirstBody(stream => {
          Assert.AreEqual(body, ConvertMessageBodyToString(stream));
        });
      });
    }

    [Test]
    public void TestFetch_Partial()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var body = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 25 Feb 2008 15:01:12 +0900\r\n";
        var octet = 10;

        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (BODY[header.fields (from to subject date)]<0." + octet.ToString() + ">)\r\n",
                                     "* 1 FETCH (BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)] " +
                                     CreateLiteral(body, Encoding.ASCII, octet) +
                                     ")\r\n" +
                                     "$tag OK FETCH completed\r\n");

        var messageDataItems = ImapFetchDataItems.BodyText("header.fields (from to subject date)", 0, octet);

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), messageDataItems, out ImapMessageFullAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", messages[0].Url.AbsoluteUri);

        Assert.IsNotNull(messages[0].BodySpecifiers);
        Assert.AreEqual(1, messages[0].BodySpecifiers.Count);
        Assert.IsTrue(messages[0].BodySpecifiers.Contains("BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)]"));

        messages[0].ReadBodyOf("body[header.fields (from to subject date)]", stream => {
          Assert.AreEqual(body.Substring(0, octet), ConvertMessageBodyToString(stream));
        });
        messages[0].ReadFirstBody(stream => {
          Assert.AreEqual(body.Substring(0, octet), ConvertMessageBodyToString(stream));
        });
      });
    }

    [Test]
    public void TestFetch_ReturnedInSeparatedResponse()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 2 (UID FLAGS INTERNALDATE)\r\n",
                                     "* 2 FETCH (INTERNALDATE \"24-Jan-2010 23:08:54 +0900\")\r\n" +
                                     "* 2 FETCH (UID 399)\r\n" +
                                     "* 2 FETCH (FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        var messageDataItems = ImapFetchDataItems.Uid + ImapFetchDataItems.Flags + ImapFetchDataItems.InternalDate;

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(2), messageDataItems, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        Assert.AreEqual(399L, messages[0].Uid);
        Assert.AreEqual(2L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=399", messages[0].Url.AbsoluteUri);

        Assert.AreEqual(0, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].InternalDate.HasValue);
        Assert.AreEqual(new DateTimeOffset(2010, 1, 24, 23, 8, 54, new TimeSpan(9, 0, 0)), messages[0].InternalDate.Value);
      });
    }

    [Test]
    public void TestFetch_ModSeq()
    {
      TestCases.WithMailboxSelectedSession(new[] {"CONDSTORE"}, (session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1:3 (MODSEQ)\r\n",
                                     "* 1 FETCH (MODSEQ (624140003))\r\n" +
                                     "* 2 FETCH (MODSEQ (624140007))\r\n" +
                                     "* 3 FETCH (MODSEQ (624140005))\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Range(1, 3), ImapFetchDataItems.ModSeq, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", messages[0].Url.AbsoluteUri);
        Assert.AreEqual(624140003UL, messages[0].ModSeq);

        Assert.AreEqual(0L, messages[1].Uid);
        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?2", messages[1].Url.AbsoluteUri);
        Assert.AreEqual(624140007UL, messages[1].ModSeq);

        Assert.AreEqual(0L, messages[2].Uid);
        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?3", messages[2].Url.AbsoluteUri);
        Assert.AreEqual(624140005UL, messages[2].ModSeq);

        Assert.AreEqual(0L, messages[2].Uid);
        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?3", messages[2].Url.AbsoluteUri);
      });
    }

    [Test]
    public void TestFetch_ModSeq_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        session.HandlesIncapableAsException = true;

        // FETCH
        Assert.Throws<ImapIncapableException>(delegate {
          session.Fetch(ImapSequenceSet.Range(1, 3), ImapFetchDataItems.Fast + ImapFetchDataItems.ModSeq, out ImapMessageFullAttributes[] _);
        });
      });
    }

    [Test]
    public void TestFetch_ChangedSince()
    {
      TestCases.WithMailboxSelectedSession(new[] {"CONDSTORE"}, (session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("UID FETCH 1:* (FLAGS) (CHANGEDSINCE 12345)\r\n",
                                     "* 1 FETCH (UID 4 MODSEQ (65402) FLAGS (\\Seen))\r\n" +
                                     "* 2 FETCH (UID 6 MODSEQ (75403) FLAGS (\\Deleted))\r\n" +
                                     "* 4 FETCH (UID 8 MODSEQ (29738) FLAGS ($NoJunk $AutoJunk $MDNSent))\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.FetchChangedSince(ImapUidSet.From(1), ImapFetchDataItems.Flags, 12345UL, out ImapMessageAttributes[] messages).IsSuccess);

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(4L, messages[0].Uid);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=4", messages[0].Url.AbsoluteUri);
        Assert.AreEqual(65402UL, messages[0].ModSeq);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));

        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(6L, messages[1].Uid);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=6", messages[1].Url.AbsoluteUri);
        Assert.AreEqual(75403UL, messages[1].ModSeq);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));

        Assert.AreEqual(4L, messages[2].Sequence);
        Assert.AreEqual(8L, messages[2].Uid);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "/;UID=8", messages[2].Url.AbsoluteUri);
        Assert.AreEqual(29738UL, messages[2].ModSeq);
        Assert.IsTrue(messages[2].Flags.Contains("$NoJunk"));
        Assert.IsTrue(messages[2].Flags.Contains("$AutoJunk"));
        Assert.IsTrue(messages[2].Flags.Contains("$MDNSent"));
      });
    }

    [Test]
    public void TestFetch_ChangedSince_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        Assert.Throws<ImapIncapableException>(delegate {
          session.FetchChangedSince(ImapUidSet.From(1), ImapFetchDataItems.Flags, 12345UL, out ImapMessageFullAttributes[] _);
        });
      });
    }

    [Test]
    public void TestFetch_Binary()
    {
      TestCases.WithMailboxSelectedSession(new[] {"BINARY"}, (session, server) => {
        var body = "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f" +
          "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f";

        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (BINARY.PEEK[1.1])\r\n",
                                     "* 1 FETCH (BINARY[1.1] " +
                                     CreateLiteral8(body, Encoding.ASCII, 0) +
                                     ")\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.BinaryPeek("1.1"), out ImapMessageFullAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", messages[0].Url.AbsoluteUri);

        Assert.IsNotNull(messages[0].BodySpecifiers);
        Assert.AreEqual(1, messages[0].BodySpecifiers.Count);
        Assert.IsTrue(messages[0].BodySpecifiers.Contains("BINARY[1.1]"));

        messages[0].ReadBodyOf("binary[1.1]", stream => {
          Assert.AreEqual(body, ConvertMessageBodyToString(stream));
        });
        messages[0].ReadFirstBody(stream => {
          Assert.AreEqual(body, ConvertMessageBodyToString(stream));
        });
      });
    }

    [Test]
    public void TestFetch_BinarySize()
    {
      TestCases.WithMailboxSelectedSession(new[] {"BINARY"}, (session, server) => {
        // FETCH
        server.EnqueueTaggedResponse("FETCH 1 (BINARY.SIZE[1.1])\r\n",
                                     "* 1 FETCH (BINARY.SIZE[1.1] 32)\r\n" +
                                     "$tag OK FETCH completed\r\n");

        Assert.IsTrue(session.Fetch(ImapSequenceSet.Of(1), ImapFetchDataItems.BinarySize("1.1"), out ImapMessageStaticAttributes[] messages).IsSuccess);

        Assert.AreEqual(1, messages.Length);

        Assert.AreEqual(0L, messages[0].Uid);
        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(session.SelectedMailbox.Url.AbsoluteUri + "?1", messages[0].Url.AbsoluteUri);

        Assert.AreEqual(32L, messages[0].GetFirstBinarySize());
        Assert.AreEqual(32L, messages[0].GetBinarySizeOf("1.1"));
      });
    }

    [Test]
    public void TestFetch_Binary_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // FETCH
        Assert.Throws<ImapIncapableException>(delegate {
          session.Fetch(ImapUidSet.From(1), ImapFetchDataItems.Binary("1.1"), out ImapMessageFullAttributes[] _);
        });
      });
    }
  }
}
