using System;
using System.Collections.Generic;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Transmission;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionCommandsListTests {
    [Test]
    public void TestListRoot()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.HierarchyDelimiters.ContainsKey(""));
        Assert.IsFalse(session.HierarchyDelimiters.ContainsKey("#news.comp.mail.misc"));

        // LIST transaction 1
        server.EnqueueTaggedResponse("LIST \"\" \"\"\r\n",
                                     "* LIST (\\Noselect) \"/\" \"\"\r\n" +
                                     "$tag OK LIST Completed\r\n");

        ImapListedMailbox root;

        Assert.IsTrue(session.ListRoot(out root).IsSuccess);

        Assert.IsNotNull(root);
        Assert.IsEmpty(root.Name);
        Assert.AreEqual("/", root.HierarchyDelimiter);
        Assert.AreEqual(1, root.NameAttributes.Count);
        Assert.IsTrue(root.NameAttributes.Contains(ImapMailboxFlag.NoSelect));

        Assert.IsTrue(session.HierarchyDelimiters.ContainsKey(""));
        Assert.AreEqual("/", session.HierarchyDelimiters[""]);

        // LIST transaction 2
        server.EnqueueTaggedResponse("LIST #news.comp.mail.misc \"\"\r\n",
                                     "* LIST (\\Noselect) \".\" #news.\r\n" +
                                     "$tag OK LIST Completed\r\n");

        Assert.IsTrue(session.ListRoot("#news.comp.mail.misc", out root).IsSuccess);

        Assert.IsNotNull(root);
        Assert.AreEqual("#news.", root.Name);
        Assert.AreEqual(".", root.HierarchyDelimiter);
        Assert.AreEqual(1, root.NameAttributes.Count);
        Assert.IsTrue(root.NameAttributes.Contains(ImapMailboxFlag.NoSelect));

        Assert.IsTrue(session.HierarchyDelimiters.ContainsKey("#news.comp.mail.misc"));
        Assert.AreEqual(".", session.HierarchyDelimiters["#news.comp.mail.misc"]);
      });
    }

    [Test]
    public void TestListAll() => ListAllRListAll(false);

    [Test]
    public void TestRListAll() => ListAllRListAll(true);

    private void ListAllRListAll(bool mailboxReferral)
    {
      TestCases.WithAuthenticatedSession(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0], (session, server) => {
        // LIST transaction
        server.EnqueueTaggedResponse(mailboxReferral ? "RLIST \"\" *\r\n"
                                                     : "LIST \"\" *\r\n",
                                     "* LIST () \"/\" blurdybloop\r\n" +
                                     "* LIST (\\Noselect) \"/\" foo\r\n" +
                                     "* LIST () \"/\" foo/bar\r\n" +
                                     "* LIST (\\Noinferiors) \"\" Draft\r\n" +
                                     "$tag OK LIST Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (mailboxReferral)
          Assert.IsTrue(session.RListAll(out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session.ListAll(out mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(4, mailboxes.Count);

        Assert.AreEqual("blurdybloop", mailboxes[0].Name, "mailbox list #1 name");
        Assert.IsTrue(mailboxes[0].NameEquals("blurdybloop"), "mailbox list #1 name comparison");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(string.Empty, mailboxes[0].SuperiorName, "mailbox list #1 superior name");
        Assert.AreEqual("blurdybloop", mailboxes[0].LeafName, "mailbox list #1 leaf name");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");
        Assert.IsFalse(mailboxes[0].IsUnselectable, "mailbox list #1 IsUnselectable");

        Assert.AreEqual("foo", mailboxes[1].Name, "mailbox list #2 name");
        Assert.IsTrue(mailboxes[1].NameEquals("foo"), "mailbox list #2 name comparison");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual(string.Empty, mailboxes[1].SuperiorName, "mailbox list #2 superior name");
        Assert.AreEqual("foo", mailboxes[1].LeafName, "mailbox list #2 leaf name");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox list #2 flag count");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.NoSelect), "mailbox list #2 flags");
        Assert.IsTrue(mailboxes[1].IsUnselectable, "mailbox list #2 IsUnselectable");

        Assert.AreEqual("foo/bar", mailboxes[2].Name, "mailbox list #3 name");
        Assert.IsTrue(mailboxes[2].NameEquals("foo/bar"), "mailbox list #3 name comparison");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox list #3 delimiter");
        Assert.AreEqual("foo", mailboxes[2].SuperiorName, "mailbox list #3 superior name");
        Assert.AreEqual("bar", mailboxes[2].LeafName, "mailbox list #3 leaf name");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox list #3 flag count");

        Assert.AreEqual("Draft", mailboxes[3].Name, "mailbox list #4 name");
        Assert.IsTrue(mailboxes[3].NameEquals("Draft"), "mailbox list #4 name comparison");
        Assert.AreEqual("", mailboxes[3].HierarchyDelimiter, "mailbox list #4 delimiter");
        Assert.AreEqual("", mailboxes[3].SuperiorName, "mailbox list #4 superior name");
        Assert.AreEqual("Draft", mailboxes[3].LeafName, "mailbox list #4 leaf name");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox list #4 flag count");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.NoInferiors));
        Assert.IsTrue(mailboxes[3].IsNonHierarchical, "mailbox list #4 IsNonHierarchical");
      });
    }

    [Test]
    public void TestList_NonEmptyReferenceName_EmptyMailboxName()
    {
      ListRListNonEmptyReferenceNameAndEmptyMailboxName(false);
    }

    [Test]
    public void TestRList_NonEmptyReferenceName_EmptyMailboxName()
    {
      ListRListNonEmptyReferenceNameAndEmptyMailboxName(true);
    }

    private void ListRListNonEmptyReferenceNameAndEmptyMailboxName(bool mailboxReferral)
    {
      TestCases.WithAuthenticatedSession(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0], (session, server) => {
        // LIST transaction
        server.EnqueueTaggedResponse(mailboxReferral ? "RLIST #news.comp.mail.misc \"\"\r\n"
                                                     : "LIST #news.comp.mail.misc \"\"\r\n",
                                     "* LIST (\\Noselect) \".\" #news.\r\n" +
                                     "$tag OK LIST Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (mailboxReferral)
          Assert.IsTrue(session.RList("#news.comp.mail.misc", string.Empty, out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session.List("#news.comp.mail.misc", string.Empty, out mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(1, mailboxes.Count);

        Assert.AreEqual("#news.", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual("#news", mailboxes[0].SuperiorName, "mailbox list #1 superior name");
        Assert.AreEqual(string.Empty, mailboxes[0].LeafName, "mailbox list #1 leaf name");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox list #1 flag count");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.NoSelect), "mailbox list #1 flags");
      });
    }

    [Test]
    public void TestList_NonEmptyReferenceName_WildcardMailboxName()
    {
      ListRListNonEmptyReferenceNameAndWildcardMailboxName(false);
    }

    [Test]
    public void TestRList_NonEmptyReferenceName_WildcardMailboxName()
    {
      ListRListNonEmptyReferenceNameAndWildcardMailboxName(true);
    }

    private void ListRListNonEmptyReferenceNameAndWildcardMailboxName(bool mailboxReferral)
    {
      TestCases.WithAuthenticatedSession(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0], (session, server) => {
        // LIST transaction
        server.EnqueueTaggedResponse(mailboxReferral ? "RLIST ~/Mail/ %\r\n"
                                                     : "LIST ~/Mail/ %\r\n",
                                     "* LIST (\\Noselect) \"/\" ~/Mail/foo\r\n" +
                                     "* LIST () \"/\" ~/Mail/meetings\r\n" +
                                     "$tag OK LIST Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (mailboxReferral)
          Assert.IsTrue(session.RList("~/Mail/", "%", out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session.List("~/Mail/", "%", out mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Count);

        Assert.AreEqual("~/Mail/foo", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual("~/Mail", mailboxes[0].SuperiorName, "mailbox list #1 superior name");
        Assert.AreEqual("foo", mailboxes[0].LeafName, "mailbox list #1 leaf name");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox list #1 flag count");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.NoSelect), "mailbox list #1 flags");

        Assert.AreEqual("~/Mail/meetings", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual("~/Mail", mailboxes[1].SuperiorName, "mailbox list #2 superior name");
        Assert.AreEqual("meetings", mailboxes[1].LeafName, "mailbox list #2 leaf name");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");
      });
    }

    [Test]
    public void TestLsubAll() => LsubAllRLsubAll(false);

    [Test]
    public void TestRLsubAll() => LsubAllRLsubAll(true);

    private void LsubAllRLsubAll(bool mailboxReferral)
    {
      TestCases.WithAuthenticatedSession(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0], (session, server) => {
        // LSUB transaction
        server.EnqueueTaggedResponse(mailboxReferral ? "RLSUB \"\" *\r\n"
                                                     : "LSUB \"\" *\r\n",
                                     "* LSUB () \"/\" INBOX\r\n" +
                                     "* LSUB () \"/\" INBOX/Sent\r\n" +
                                     "* LSUB () \"/\" INBOX/Drafts\r\n" +
                                     "* LSUB () \"/\" INBOX/Junk\r\n" +
                                      "$tag OK LSUB Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (mailboxReferral)
          Assert.IsTrue(session.RLsubAll(out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session.LsubAll(out mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(4, mailboxes.Count);

        Assert.AreEqual("INBOX", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual(string.Empty, mailboxes[0].SuperiorName, "mailbox list #1 superior name");
        Assert.AreEqual("INBOX", mailboxes[0].LeafName, "mailbox list #1 leaf name");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");

        Assert.AreEqual("INBOX/Sent", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual("INBOX", mailboxes[1].SuperiorName, "mailbox list #2 superior name");
        Assert.AreEqual("Sent", mailboxes[1].LeafName, "mailbox list #2 leaf name");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");

        Assert.AreEqual("INBOX/Drafts", mailboxes[2].Name, "mailbox list #3 name");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox list #3 delimiter");
        Assert.AreEqual("INBOX", mailboxes[2].SuperiorName, "mailbox list #3 superior name");
        Assert.AreEqual("Drafts", mailboxes[2].LeafName, "mailbox list #3 leaf name");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox list #3 flag count");

        Assert.AreEqual("INBOX/Junk", mailboxes[3].Name, "mailbox list #4 name");
        Assert.AreEqual("/", mailboxes[3].HierarchyDelimiter, "mailbox list #4 delimiter");
        Assert.AreEqual("INBOX", mailboxes[3].SuperiorName, "mailbox list #4 superior name");
        Assert.AreEqual("Junk", mailboxes[3].LeafName, "mailbox list #4 leaf name");
        Assert.AreEqual(0, mailboxes[3].Flags.Count, "mailbox list #4 flag count");
      });
    }

    [Test]
    public void TestLsub_EmptyReferenceName_NonEmptyMailboxName()
    {
      LsubRLsubEmptyReferenceNameAndNonEmptyMailboxName(false);
    }

    [Test]
    public void TestRLsub_EmptyReferenceName_NonEmptyMailboxName()
    {
      LsubRLsubEmptyReferenceNameAndNonEmptyMailboxName(true);
    }

    private void LsubRLsubEmptyReferenceNameAndNonEmptyMailboxName(bool mailboxReferral)
    {
      TestCases.WithAuthenticatedSession(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0], (session, server) => {
        // LSUB transaction
        server.EnqueueTaggedResponse(mailboxReferral ? "RLSUB \"\" INBOX/*\r\n"
                                                     : "LSUB \"\" INBOX/*\r\n",
                                     "* LSUB () \"/\" INBOX/Sent\r\n" +
                                     "* LSUB () \"/\" INBOX/Drafts\r\n" +
                                     "* LSUB () \"/\" INBOX/Junk\r\n" +
                                      "$tag OK LSUB Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (mailboxReferral)
          Assert.IsTrue(session.RLsub("INBOX/*", out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session.Lsub("INBOX/*", out mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Count);

        Assert.AreEqual("INBOX/Sent", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual("INBOX", mailboxes[0].SuperiorName, "mailbox list #1 superior name");
        Assert.AreEqual("Sent", mailboxes[0].LeafName, "mailbox list #1 leaf name");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");

        Assert.AreEqual("INBOX/Drafts", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual("INBOX", mailboxes[1].SuperiorName, "mailbox list #2 superior name");
        Assert.AreEqual("Drafts", mailboxes[1].LeafName, "mailbox list #2 leaf name");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");

        Assert.AreEqual("INBOX/Junk", mailboxes[2].Name, "mailbox list #3 name");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox list #3 delimiter");
        Assert.AreEqual("INBOX", mailboxes[2].SuperiorName, "mailbox list #3 superior name");
        Assert.AreEqual("Junk", mailboxes[2].LeafName, "mailbox list #3 leaf name");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox list #3 flag count");
      });
    }

    [Test]
    public void TestLsub_NonEmptyReferenceName_WildcardMailboxName()
    {
      LsubRLsubNonEmptyReferenceNameAndWildcardMailboxName(false);
    }

    [Test]
    public void TestRLsub_NonEmptyReferenceName_WildcardMailboxName()
    {
      LsubRLsubNonEmptyReferenceNameAndWildcardMailboxName(true);
    }

    private void LsubRLsubNonEmptyReferenceNameAndWildcardMailboxName(bool mailboxReferral)
    {
      TestCases.WithAuthenticatedSession(mailboxReferral ? new[] {"MAILBOX-REFERRAL"} : new string[0], (session, server) => {
        // LSUB transaction
        server.EnqueueTaggedResponse(mailboxReferral ? "RLSUB #news. comp.mail.*\r\n"
                                                     : "LSUB #news. comp.mail.*\r\n",
                                     "* LSUB () \".\" #news.comp.mail.mime\r\n" +
                                     "* LSUB () \".\" #news.comp.mail.misc\r\n" +
                                     "$tag OK LSUB Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (mailboxReferral)
          Assert.IsTrue(session.RLsub("#news.", "comp.mail.*", out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session.Lsub("#news.", "comp.mail.*", out mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Count);

        Assert.AreEqual("#news.comp.mail.mime", mailboxes[0].Name, "mailbox list #1 name");
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter, "mailbox list #1 delimiter");
        Assert.AreEqual("#news.comp.mail", mailboxes[0].SuperiorName, "mailbox list #1 superior name");
        Assert.AreEqual("mime", mailboxes[0].LeafName, "mailbox list #1 leaf name");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox list #1 flag count");

        Assert.AreEqual("#news.comp.mail.misc", mailboxes[1].Name, "mailbox list #2 name");
        Assert.AreEqual(".", mailboxes[1].HierarchyDelimiter, "mailbox list #2 delimiter");
        Assert.AreEqual("#news.comp.mail", mailboxes[1].SuperiorName, "mailbox list #2 superior name");
        Assert.AreEqual("misc", mailboxes[1].LeafName, "mailbox list #2 leaf name");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox list #2 flag count");
      });
    }

    [Test]
    public void TestList_NonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("LIST");
    }

    [Test]
    public void TestLsub_NonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("LSUB");
    }

    [Test]
    public void TestRList_NonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("RLIST");
    }

    [Test]
    public void TestRLsub_NonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("RLSUB");
    }

    [Test]
    public void TestXList_NonAsciiMailboxName()
    {
      ListNonAsciiMailboxName("XLIST");
    }

    private void ListNonAsciiMailboxName(string command)
    {
      TestCases.WithAuthenticatedSession(new[] {"XLIST", "MAILBOX-REFERRAL"}, (session, server) => {
        // list transaction
        server.EnqueueTaggedResponse(command + " \"\" INBOX.&MOEw,DDrMNwwwzCvMLk-.*\r\n",
                                     "$tag OK completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes = null;
        ImapCommandResult result = null;

        switch (command) {
          case "LIST":  result = session.List("INBOX.メールボックス.*", out mailboxes); break;
          case "LSUB":  result = session.Lsub("INBOX.メールボックス.*", out mailboxes); break;
          case "RLIST": result = session.RList("INBOX.メールボックス.*", out mailboxes); break;
          case "RLSUB": result = session.RLsub("INBOX.メールボックス.*", out mailboxes); break;
          default:
            return;
        }

        Assert.IsTrue(result.IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(0, mailboxes.Count);
      });
    }


    [Test] public void TestList_ListedMailbox() => ListLsub_ListedMailbox(true);
    [Test] public void TestLsub_ListedMailbox() => ListLsub_ListedMailbox(false);

    private void ListLsub_ListedMailbox(bool listForTrue)
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        if (listForTrue)
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \".\" INBOX\r\n" +
                                       "$tag OK LIST Completed\r\n");
        else
          server.EnqueueTaggedResponse("LSUB \"\" INBOX\r\n",
                                       "* LSUB () \".\" INBOX\r\n" +
                                       "$tag OK LSUB Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (listForTrue)
          Assert.IsTrue(session.List("INBOX", out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session.Lsub("INBOX", out mailboxes).IsSuccess);

        Assert.AreEqual(1, mailboxes.Count);

        var inbox = mailboxes[0];

        Assert.IsTrue(inbox.IsInbox);
        Assert.AreEqual(0, inbox.Flags.Count);

        // LIST/LSUB again
        if (listForTrue)
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST (\\Marked) \".\" INBOX\r\n" + // flags changed
                                       "$tag OK LIST Completed\r\n");
        else
          server.EnqueueTaggedResponse("LSUB \"\" INBOX\r\n",
                                       "* LSUB (\\Marked) \".\" INBOX\r\n" + // flags changed
                                       "$tag OK LSUB Completed\r\n");

        if (listForTrue)
          Assert.IsTrue(session.List(inbox).IsSuccess);
        else
          Assert.IsTrue(session.Lsub(inbox).IsSuccess);

        Assert.IsTrue(inbox.IsInbox);
        Assert.AreEqual(1, inbox.Flags.Count);
        Assert.IsTrue(inbox.Flags.Contains(ImapMailboxFlag.Marked));
      });
    }

    [Test] public void TestList_ListedMailbox_OfDifferentSession() => ListLsub_ListedMailbox_OfDifferentSession(true);
    [Test] public void TestLsub_ListedMailbox_OfDifferentSession() => ListLsub_ListedMailbox_OfDifferentSession(false);

    private void ListLsub_ListedMailbox_OfDifferentSession(bool listForTrue)
    {
      TestCases.WithAuthenticatedSession((session1, server1) => {
        if (listForTrue)
          server1.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                        "* LIST () \".\" INBOX\r\n" +
                                        "$tag OK LIST Completed\r\n");
        else
          server1.EnqueueTaggedResponse("LSUB \"\" INBOX\r\n",
                                        "* LSUB () \".\" INBOX\r\n" +
                                        "$tag OK LSUB Completed\r\n");

        IReadOnlyList<ImapMailboxStatus> mailboxes;

        if (listForTrue)
          Assert.IsTrue(session1.List("INBOX", out mailboxes).IsSuccess);
        else
          Assert.IsTrue(session1.Lsub("INBOX", out mailboxes).IsSuccess);

        Assert.AreEqual(1, mailboxes.Count);

        var mailboxOfDifferentSession = mailboxes[0];

        TestCases.WithAuthenticatedSession((session2, server2) => {
          Assert.AreEqual(ImapSessionState.Authenticated, session2.State);

          var ex = listForTrue
            ? Assert.Throws<ArgumentException>(() => session2.List(mailbox: mailboxOfDifferentSession))
            : Assert.Throws<ArgumentException>(() => session2.Lsub(mailbox: mailboxOfDifferentSession));

          Assert.AreEqual("mailbox", ex.ParamName);

          Assert.AreEqual(ImapSessionState.Authenticated, session2.State);
        });
      });
    }

    [Test] public void TestList_ArgumentNull_MailboxName1() => List_ArgumentNull_MailboxName(session => session.List(mailboxName: null, out _));
    [Test] public void TestLsub_ArgumentNull_MailboxName1() => List_ArgumentNull_MailboxName(session => session.Lsub(mailboxName: null, out _));
    [Test] public void TestRList_ArgumentNull_MailboxName1() => List_ArgumentNull_MailboxName(session => session.RList(mailboxName: null, out _));
    [Test] public void TestRLsub_ArgumentNull_MailboxName1() => List_ArgumentNull_MailboxName(session => session.RLsub(mailboxName: null, out _));

    [Test] public void TestList_ArgumentNull_MailboxName2() => List_ArgumentNull_MailboxName(session => session.List(string.Empty, mailboxName: null, out _));
    [Test] public void TestLsub_ArgumentNull_MailboxName2() => List_ArgumentNull_MailboxName(session => session.Lsub(string.Empty, mailboxName: null, out _));
    [Test] public void TestRList_ArgumentNull_MailboxName2() => List_ArgumentNull_MailboxName(session => session.RList(string.Empty, mailboxName: null, out _));
    [Test] public void TestRLsub_ArgumentNull_MailboxName2() => List_ArgumentNull_MailboxName(session => session.RLsub(string.Empty, mailboxName: null, out _));

    private void List_ArgumentNull_MailboxName(Action<ImapSession> testAction)
    {
      TestCases.WithAuthenticatedSession(new[] { "MAILBOX-REFERRAL" }, (session, server) => {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        var ex = Assert.Throws<ArgumentNullException>(() => testAction(session));

        Assert.AreEqual("mailboxName", ex.ParamName);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      });
    }

    [Test] public void TestList_ArgumentNull_ReferenceName() => List_ArgumentNull_ReferenceName(session => session.List(referenceName: null, string.Empty, out _));
    [Test] public void TestLsub_ArgumentNull_ReferenceName() => List_ArgumentNull_ReferenceName(session => session.Lsub(referenceName: null, string.Empty, out _));
    [Test] public void TestRList_ArgumentNull_ReferenceName() => List_ArgumentNull_ReferenceName(session => session.RList(referenceName: null, string.Empty, out _));
    [Test] public void TestRLsub_ArgumentNull_ReferenceName() => List_ArgumentNull_ReferenceName(session => session.RLsub(referenceName: null, string.Empty, out _));

    private void List_ArgumentNull_ReferenceName(Action<ImapSession> testAction)
    {
      TestCases.WithAuthenticatedSession(new[] { "MAILBOX-REFERRAL" }, (session, server) => {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        var ex = Assert.Throws<ArgumentNullException>(() => testAction(session));

        Assert.AreEqual("referenceName", ex.ParamName);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      });
    }

    [Test] public void TestList_ArgumentNull_Mailbox() => List_ArgumentNull_Mailbox(session => session.List(mailbox: null));
    [Test] public void TestLsub_ArgumentNull_Mailbox() => List_ArgumentNull_Mailbox(session => session.Lsub(mailbox: null));

    private void List_ArgumentNull_Mailbox(Action<ImapSession> testAction)
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        var ex = Assert.Throws<ArgumentNullException>(() => testAction(session));

        Assert.AreEqual("mailbox", ex.ParamName);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      });
    }

    [Test]
    public void TestListExtended_MailboxPattern_NonAscii()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST (SUBSCRIBED) \"\" INBOX.&MOEw,DDrMNwwwzCvMLk-.*\r\n",
                                     "* LIST (\\Subscribed) \".\" \"INBOX.&MOEw,DDrMNwwwzCvMLk-\"\r\n" +
                                     "* LIST (\\Subscribed \\NonExistent) \".\" \"INBOX.&MOEw,DDrMNwwwzCvMLk-.1\"\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("INBOX.メールボックス.*",
                                           ImapListSelectionOptions.Subscribed,
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Count);

        Assert.AreEqual("INBOX.メールボックス", mailboxes[0].Name, "mailbox #1");
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter, "mailbox #1 delimiter");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #1 flags");
        Assert.IsFalse(mailboxes[0].IsUnselectable, "mailbox #1 IsUnselectable");

        Assert.AreEqual("INBOX.メールボックス.1", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(".", mailboxes[1].HierarchyDelimiter, "mailbox #2 delimiter");
        Assert.AreEqual(2, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #2 flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.NonExistent), "mailbox #2 flags");
        Assert.IsTrue(mailboxes[1].IsUnselectable, "mailbox #2 IsUnselectable");
      });
    }

    [Test]
    public void TestListExtended_SelectSubscribed()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST (SUBSCRIBED) \"\" *\r\n",
                                     "* LIST (\\Marked \\NoInferiors \\Subscribed) \"/\" \"inbox\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"Fruit/Banana\"\r\n" +
                                     "* LIST (\\Subscribed \\NonExistent) \"/\" \"Fruit/Peach\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"Vegetable\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"Vegetable/Broccoli\"\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("*",
                                           ImapListSelectionOptions.Subscribed,
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(5, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.IsTrue(mailboxes[0].NameEquals("inbox"), "mailbox list #1 name comparison");
        Assert.IsTrue(mailboxes[0].NameEquals("INBOX"), "mailbox list #1 name comparison (ignore case)");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox #1 delimiter");
        Assert.AreEqual(3, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #1 flags");

        Assert.AreEqual("Fruit/Banana", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox #2 delimiter");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #2 flags");
        Assert.IsFalse(mailboxes[1].IsUnselectable, "mailbox #2 IsUnselectable");

        Assert.AreEqual("Fruit/Peach", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox #3 delimiter");
        Assert.AreEqual(2, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #3 flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.NonExistent), "mailbox #3 flags");
        Assert.IsTrue(mailboxes[2].IsUnselectable, "mailbox #3 IsUnselectable");

        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual("/", mailboxes[3].HierarchyDelimiter, "mailbox #4 delimiter");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #4 flags");

        Assert.AreEqual("Vegetable/Broccoli", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual("/", mailboxes[4].HierarchyDelimiter, "mailbox #5 delimiter");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #5 flags");
      });
    }

    [Test]
    public void TestListExtended_ReturnChildren()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST \"\" % RETURN (CHILDREN)\r\n",
                                     "* LIST (\\Marked \\NoInferiors) \"/\" \"inbox\"\r\n" +
                                     "* LIST (\\HasChildren) \"/\" \"Fruit\"\r\n" +
                                     "* LIST (\\HasNoChildren) \"/\" \"Tofu\"\r\n" +
                                     "* LIST (\\HasChildren) \"/\" \"Vegetable\"\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("%",
                                           ImapListReturnOptions.Children,
                                           out var mailboxes).IsSuccess);

        //Assert.AreEqual("0002 LIST () \"\" \"%\" RETURN (CHILDREN)\r\n",

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(4, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual("/", mailboxes[0].HierarchyDelimiter, "mailbox #1 delimiter");
        Assert.AreEqual(2, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");

        Assert.AreEqual("Fruit", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual("/", mailboxes[1].HierarchyDelimiter, "mailbox #2 delimiter");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.HasChildren), "mailbox #2 flags");

        Assert.AreEqual("Tofu", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual("/", mailboxes[2].HierarchyDelimiter, "mailbox #3 delimiter");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.HasNoChildren), "mailbox #3 flags");

        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual("/", mailboxes[3].HierarchyDelimiter, "mailbox #4 delimiter");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.HasChildren), "mailbox #4 flags");
      });
    }

    [Test]
    public void TestListExtended_SelectRemote_ReturnChildren()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST (REMOTE) \"\" % RETURN (CHILDREN)\r\n",
                                     "* LIST (\\Marked \\NoInferiors) \"/\" \"inbox\"\r\n" +
                                     "* LIST (\\HasChildren) \"/\" \"Fruit\"\r\n" +
                                     "* LIST (\\HasNoChildren) \"/\" \"Tofu\"\r\n" +
                                     "* LIST (\\HasChildren) \"/\" \"Vegetable\"\r\n" +
                                     "* LIST (\\Remote) \"/\" \"Bread\"\r\n" +
                                     "* LIST (\\HasChildren \\Remote) \"/\" \"Meat\"\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("%",
                                           ImapListSelectionOptions.Remote,
                                           ImapListReturnOptions.Children,
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(6, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(2, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");

        Assert.AreEqual("Fruit", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.HasChildren), "mailbox #2 flags");

        Assert.AreEqual("Tofu", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.HasNoChildren), "mailbox #3 flags");

        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.HasChildren), "mailbox #4 flags");

        Assert.AreEqual("Bread", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Contains(ImapMailboxFlag.Remote), "mailbox #5 flags");

        Assert.AreEqual("Meat", mailboxes[5].Name, "mailbox #6");
        Assert.AreEqual(2, mailboxes[5].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[5].Flags.Contains(ImapMailboxFlag.HasChildren), "mailbox #6 flags");
        Assert.IsTrue(mailboxes[5].Flags.Contains(ImapMailboxFlag.Remote), "mailbox #6 flags");
      });
    }

    [Test]
    public void TestListExtended_SelectRemote_Subscribed()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST (REMOTE SUBSCRIBED) \"\" *\r\n",
                                     "* LIST (\\Marked \\NoInferiors \\Subscribed) \"/\" \"inbox\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"Fruit/Banana\"\r\n" +
                                     "* LIST (\\Subscribed \\NonExistent) \"/\" \"Fruit/Peach\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"Vegetable\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"Vegetable/Broccoli\"\r\n" +
                                     "* LIST (\\Remote \\Subscribed) \"/\" \"Bread\"\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("*",
                                           ImapListSelectionOptions.Remote + ImapListSelectionOptions.Subscribed,
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(6, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(3, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.NoInferiors), "mailbox #1 flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #1 flags");

        Assert.AreEqual("Fruit/Banana", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #2 flags");

        Assert.AreEqual("Fruit/Peach", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(2, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #3 flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.NonExistent), "mailbox #3 flags");
        
        Assert.AreEqual("Vegetable", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #4 flags");

        Assert.AreEqual("Vegetable/Broccoli", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #5 flags");

        Assert.AreEqual("Bread", mailboxes[5].Name, "mailbox #6");
        Assert.AreEqual(2, mailboxes[5].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[5].Flags.Contains(ImapMailboxFlag.Remote), "mailbox #6 flags");
        Assert.IsTrue(mailboxes[5].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #6 flags");
      });
    }

    [Test]
    public void TestListExtended_SelectRecursiveMatchSubscribed()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST (RECURSIVEMATCH SUBSCRIBED) \"\" *2\r\n",
                                     "* LIST () \"/\" \"foo2\" (\"CHILDINFO\" (\"SUBSCRIBED\"))\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"foo2/bar2\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"baz2/bar2\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"baz2/bar22\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"baz2/bar222\"\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"eps2\" (\"CHILDINFO\" (\"SUBSCRIBED\"))\r\n" +
                                     "* LIST (\\Subscribed) \"/\" \"qux2/bar2\"\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("*2",
                                           ImapListSelectionOptions.RecursiveMatch + ImapListSelectionOptions.Subscribed,
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(7, mailboxes.Count);

        Assert.AreEqual("foo2", mailboxes[0].Name, "mailbox #1");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        // TODO: child info

        Assert.AreEqual("foo2/bar2", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #2 flags");

        Assert.AreEqual("baz2/bar2", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #3 flags");

        Assert.AreEqual("baz2/bar22", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #4 flags");

        Assert.AreEqual("baz2/bar222", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(1, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #5 flags");

        Assert.AreEqual("eps2", mailboxes[5].Name, "mailbox #6");
        Assert.AreEqual(1, mailboxes[5].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[5].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #6 flags");
        // TODO: child info

        Assert.AreEqual("qux2/bar2", mailboxes[6].Name, "mailbox #6");
        Assert.AreEqual(1, mailboxes[6].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[6].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #6 flags");
      });
    }

    [Test]
    public void TestListExtended_MailboxPattern_Multiple()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST \"\" (INBOX Drafts Sent/%)\r\n",
                                     "* LIST () \"/\" \"INBOX\"\r\n" +
                                     "* LIST (\\NoInferiors) \"/\" \"Drafts\"\r\n" +
                                     "* LIST () \"/\" \"Sent/March2004\"\r\n" +
                                     "* LIST (\\Marked) \"/\" \"Sent/December2003\"\r\n" +
                                     "* LIST () \"/\" \"Sent/August2004\"\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended(out var mailboxes,
                                           "INBOX",
                                           "Drafts",
                                           "Sent/%").IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(5, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox #1 count of flags");

        Assert.AreEqual("Drafts", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(1, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.NoInferiors), "mailbox #2 flags");

        Assert.AreEqual("Sent/March2004", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox #3 count of flags");

        Assert.AreEqual("Sent/December2003", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #4 flags");

        Assert.AreEqual("Sent/August2004", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(0, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
      });
    }

    [Test]
    public void TestListExtended_ReturnStatus_Incapable()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.ListStatus));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.ListExtended("%", ImapListReturnOptions.StatusDataItems(ImapStatusDataItems.Messages), out _));
      });
    }

    [Test]
    public void TestListExtended_ReturnStatus()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED", "LIST-STATUS"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListStatus));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST \"\" % RETURN (STATUS (MESSAGES UNSEEN))\r\n",
                                     "* LIST () \".\"  \"inbox\"\r\n" + // test case insensitivity
                                     "* STATUS \"inbox\" (messages 17 unseen 16)\r\n" + // test case insensitivity
                                     "* LIST () \".\" \"foo\"\r\n" +
                                     "* STATUS \"foo\" (MESSAGES 30 UNSEEN 29)\r\n" +
                                     "* LIST (\\NoSelect) \".\" \"bar\"\r\n" +
                                     "$tag OK List completed.\r\n");

        Assert.IsTrue(session.ListExtended("%",
                                           ImapListReturnOptions.StatusDataItems(ImapStatusDataItems.Messages +
                                                                                 ImapStatusDataItems.Unseen),
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(0, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.AreEqual(17L, mailboxes[0].ExistsMessage, "mailbox #1 exists messages");
        Assert.AreEqual(16L, mailboxes[0].UnseenMessage, "mailbox #1 unseen messages");

        Assert.AreEqual("foo", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.AreEqual(30L, mailboxes[1].ExistsMessage, "mailbox #2 exists messages");
        Assert.AreEqual(29L, mailboxes[1].UnseenMessage, "mailbox #2 unseen messages");

        Assert.AreEqual("bar", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.NoSelect), "mailbox #3 flags");
        Assert.AreEqual(0L, mailboxes[2].ExistsMessage, "mailbox #3 exists messages");
        Assert.AreEqual(0L, mailboxes[2].UnseenMessage, "mailbox #3 unseen messages");
      });
    }

    [Test]
    public void TestListExtended_ReturnStatus_SelectSubscribedRecursiveMatch()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED", "LIST-STATUS"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListStatus));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST (SUBSCRIBED RECURSIVEMATCH) \"\" % RETURN (STATUS (MESSAGES))\r\n",
                                     "* LIST (\\Subscribed) \".\"  \"INBOX\"\r\n" +
                                     "* STATUS \"INBOX\" (MESSAGES 17)\r\n" +
                                     "* LIST () \".\" \"foo\" (CHILDINFO (\"SUBSCRIBED\"))\r\n" +
                                     "$tag OK List completed.\r\n");

        Assert.IsTrue(session.ListExtended("%",
                                           ImapListSelectionOptions.Subscribed + ImapListSelectionOptions.RecursiveMatch,
                                           ImapListReturnOptions.StatusDataItems(ImapStatusDataItems.Messages),
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(2, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Subscribed), "mailbox #1 flags");
        Assert.AreEqual(17L, mailboxes[0].ExistsMessage, "mailbox #1 exists messages");

        Assert.AreEqual("foo", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        // child info
      });
    }

    [Test]
    public void TestListExtended_ReturnSpecialUse()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED", "SPECIAL-USE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.SpecialUse));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST \"\" % RETURN (SPECIAL-USE)\r\n",
                                     "* LIST (\\Marked) \"/\" Inbox\r\n" +
                                     "* LIST () \"/\" ToDo\r\n" +
                                     "* LIST () \"/\" Projects\r\n" +
                                     "* LIST (\\Sent) \"/\" SentMail\r\n" +
                                     "* LIST (\\Marked \\Drafts) \"/\" MyDrafts\r\n" +
                                     "* LIST (\\Trash) \"/\" Trash\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("%",
                                           ImapListReturnOptions.SpecialUse,
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(6, mailboxes.Count);

        Assert.IsTrue(mailboxes[0].IsInbox, "mailbox #1 (inbox)");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #1 \\Marked");

        Assert.AreEqual("ToDo", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(0, mailboxes[1].Flags.Count, "mailbox #2 count of flags");

        Assert.AreEqual("Projects", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(0, mailboxes[2].Flags.Count, "mailbox #3 count of flags");

        Assert.AreEqual("SentMail", mailboxes[3].Name, "mailbox #4");
        Assert.AreEqual(1, mailboxes[3].Flags.Count, "mailbox #4 count of flags");
        Assert.IsTrue(mailboxes[3].Flags.Contains(ImapMailboxFlag.Sent), "mailbox #4 \\Sent");

        Assert.AreEqual("MyDrafts", mailboxes[4].Name, "mailbox #5");
        Assert.AreEqual(2, mailboxes[4].Flags.Count, "mailbox #5 count of flags");
        Assert.IsTrue(mailboxes[4].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #5 \\Marked");
        Assert.IsTrue(mailboxes[4].Flags.Contains(ImapMailboxFlag.Drafts), "mailbox #5 \\Drafts");

        Assert.AreEqual("Trash", mailboxes[5].Name, "mailbox #6");
        Assert.AreEqual(1, mailboxes[5].Flags.Count, "mailbox #6 count of flags");
        Assert.IsTrue(mailboxes[5].Flags.Contains(ImapMailboxFlag.Trash), "mailbox #6 \\Trash");
      });
    }

    [Test]
    public void TestListExtended_ReturnSpecialUse_Incapable()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.SpecialUse));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.ListExtended("%", ImapListReturnOptions.SpecialUse, out _));
      });
    }

    [Test]
    public void TestListExtended_SelectSpecialUse()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED", "SPECIAL-USE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.SpecialUse));

        // LIST transaction
        server.EnqueueTaggedResponse("LIST (SPECIAL-USE) \"\" *\r\n",
                                     "* LIST (\\Sent) \"/\" SentMail\r\n" +
                                     "* LIST (\\Marked \\Drafts) \"/\" MyDrafts\r\n" +
                                     "* LIST (\\Trash) \"/\" Trash\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(session.ListExtended("*",
                                           ImapListSelectionOptions.SpecialUse,
                                           out var mailboxes).IsSuccess);

        Assert.IsNotNull(mailboxes);
        Assert.AreEqual(3, mailboxes.Count);

        Assert.AreEqual("SentMail", mailboxes[0].Name, "mailbox #1");
        Assert.AreEqual(1, mailboxes[0].Flags.Count, "mailbox #1 count of flags");
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.Sent), "mailbox #1 \\Sent");

        Assert.AreEqual("MyDrafts", mailboxes[1].Name, "mailbox #2");
        Assert.AreEqual(2, mailboxes[1].Flags.Count, "mailbox #2 count of flags");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.Marked), "mailbox #1 \\Marked");
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.Drafts), "mailbox #1 \\Drafts");

        Assert.AreEqual("Trash", mailboxes[2].Name, "mailbox #3");
        Assert.AreEqual(1, mailboxes[2].Flags.Count, "mailbox #3 count of flags");
        Assert.IsTrue(mailboxes[2].Flags.Contains(ImapMailboxFlag.Trash), "mailbox #1 \\Trash");
      });
    }

    [Test]
    public void TestListExtended_SelectSpecialUse_Incapable()
    {
      TestCases.WithAuthenticatedSession(new[] {"LIST-EXTENDED"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ListExtended));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.SpecialUse));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.ListExtended("*", ImapListSelectionOptions.SpecialUse, out _));
      });
    }
  }
}

