using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Formats;
using Smdn.Net.Connection;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Urls;
using Smdn.Security.Authentication.Sasl;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionCreationTests {
    private class Profile : IImapSessionProfile {
      public Uri Authority {
        get; set;
      }

      public int Timeout {
        get; set;
      }

      public int SendTimeout {
        get; set;
      }

      public int ReceiveTimeout {
        get; set;
      }

      public ICredentialsByHost Credentials {
        get; set;
      }

      public bool UseDeflateIfAvailable {
        get { return false; }
      }

      public IReadOnlyList<string> UsingSaslMechanisms {
        get; set;
      }

      public bool UseTlsIfAvailable {
        get; set;
      }

      public bool AllowInsecureLogin {
        get; set;
      }

      public Profile(NetworkCredential cred, string userinfo, string hostPort)
        : this(cred, ImapUri.UriSchemeImap, userinfo, hostPort)
      {
      }

      public Profile(NetworkCredential cred, string scheme, string userinfo, string hostPort)
      {
        this.UseTlsIfAvailable = true;
        this.Timeout = 1000;
        this.SendTimeout = -1;
        this.ReceiveTimeout = -1;
        this.Credentials = cred;
        this.AllowInsecureLogin = false;

        if (userinfo == null)
          this.Authority = new Uri(string.Format("{0}://{1}/", scheme, hostPort));
        else
          this.Authority = new Uri(string.Format("{0}://{1}@{2}/", scheme, userinfo, hostPort));
      }
    }

    [Test]
    public void TestCreateAsync_ArgumentException_ProfileNull()
    {
      Assert.Throws<ArgumentNullException>(() => ImapSession.CreateAsync(null));
    }

    [Test]
    public void TestCreateAsync_ArgumentException_CreateSslStreamNull()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "IMAPS", "user", "imap.invalid");

      Assert.Throws<ArgumentNullException>(() => ImapSession.CreateAsync(prof));
    }

    [Test]
    public void TestCreateAsync_ConnectionRefusedByTimeout()
    {
      ImapPseudoServer.StartNotAccepting((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.Timeout = 500;

        var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
      });
    }

    [Test]
    public void TestCreateAsync_ConnectionRefusedByDnsError()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "user", "imap.invalid");

      prof.Timeout = -1;

      var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await ImapSession.CreateAsync(prof));

      Assert.IsInstanceOf<System.Net.Sockets.SocketException>(ex.InnerException);
    }

    [Test]
    public void TestCreateAsync_Imap4Rev1Incapable()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4\r\n" +
                                     "$tag OK done\r\n");

        Assert.ThrowsAsync<ImapIncapableException>(async () => await ImapSession.CreateAsync(prof));
      });
    }

    [Test]
    public void TestCreateAsync_ConnectSecurePort()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imaps", "user", server.HostPort);

        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          var streamUpgraded = false;

          using (var session = await ImapSession.CreateAsync(prof, createSslStream: (connection, baseStream, cancellationToken) => {
            streamUpgraded = true;
            return Task.FromResult<Stream>(baseStream); // TODO: return SSL stream
          })) {
            Assert.IsTrue(streamUpgraded, "stream upgraded");

            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsTrue(session.IsSecureConnection);
          }
        });
      });
    }

    [Test]
    public void TestCreateAsync_ConnectSecurePort_Exception()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imaps", "user", server.HostPort);

        var ex = Assert.ThrowsAsync<ImapSecureConnectionException>(async () => {
          using (var session = await ImapSession.CreateAsync(prof, createSslStream: (connection, baseStream, cancellationToken) => {
            throw new System.Security.Authentication.AuthenticationException();
          })) { }
        });

        Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);
      });
    }

    [Test]
    public void TestCreateAsync_StartTls_UseTlsIfAvailable()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          var streamUpgraded = false;

          using (var session = await ImapSession.CreateAsync(prof, createSslStream: (connection, baseStream, cancellationToken) => {
            streamUpgraded = true;
            return Task.FromResult<Stream>(baseStream); // TODO: return SSL stream
          })) {
            Assert.IsTrue(streamUpgraded, "stream upgraded");

            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsTrue(session.IsSecureConnection);
          }
        });
      });
    }

    [Test]
    public void TestCreateAsync_StartTls_DontUseTlsWhetherAvailable()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.AllowInsecureLogin = true;
        prof.UseTlsIfAvailable = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          var streamUpgraded = false;

          using (var session = await ImapSession.CreateAsync(prof, createSslStream: (connection, baseStream, cancellationToken) => {
            streamUpgraded = true;
            return Task.FromResult<Stream>(baseStream); // TODO: return SSL stream
          })) {
            Assert.IsFalse(streamUpgraded, "stream upgraded");

            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsFalse(session.IsSecureConnection);
          }
        });
      });
    }

    [Test]
    public void TestCreateAsync_StartTls_Failure()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag BAD done\r\n");

        var streamUpgraded = false;

        var ex = Assert.ThrowsAsync<ImapSecureConnectionException>(async () => {
          using (var session = await ImapSession.CreateAsync(prof, createSslStream: (connection, baseStream, cancellationToken) => {
            streamUpgraded = true;
            return Task.FromResult<Stream>(baseStream); // TODO: return SSL stream
          })) { }
        });

        Assert.IsNull(ex.InnerException);

        Assert.IsFalse(streamUpgraded, "stream upgraded");
      });
    }

    [Test]
    public void TestCreateAsync_StartTls_Exception()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK done\r\n");

        var ex = Assert.ThrowsAsync<ImapSecureConnectionException>(async () => {
          using (var session = await ImapSession.CreateAsync(prof, createSslStream: (connection, baseStream, cancellationToken) => {
            throw new System.Security.Authentication.AuthenticationException();
          })) { }
        });

        Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);
      });
    }

    [Test]
    public void TestCreateAsync_ProfileTimeoutMustBeApplied()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", server.HostPort);

        prof.AllowInsecureLogin = true;
        prof.Timeout = 100 * 1000;
        prof.SendTimeout = 300 * 1000;
        prof.ReceiveTimeout = 500 * 1000;

        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsFalse(session.IsSecureConnection);

            Assert.AreEqual(100 * 1000, session.TransactionTimeout);
            Assert.AreEqual(300 * 1000, session.SendTimeout);
            Assert.AreEqual(500 * 1000, session.ReceiveTimeout);
          }
        });
      });
    }

    [Test]
    public void TestCreateAsync_CapabilityAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsFalse(session.IsSecureConnection);
          }
        });
      });
    }

    [Test]
    public void TestCreateAsync_CapabilityNotAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsFalse(session.IsSecureConnection);
          }
        });
      });
    }

    [Test]
    public void TestCreateAsync_ResolveHostAddressCallback()
    {
      ImapPseudoServer.Start((server) => {
        const string pseudoServerHostName = "imap.invalid";

        var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
          Assert.AreEqual(pseudoServerHostName, hostname);
          return Task.FromResult(server.ServerEndPoint.Address);
        });
        var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", $"{pseudoServerHostName}:{server.Port}");

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof, resolveHostAddress: resolveHostAddress)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsFalse(session.IsSecureConnection);
          }
        });
      });
    }

    [Test]
    public void TestCreateAsync_ResolveHostAddressCallback_Exception()
    {
      const string pseudoServerHostName = "imap.invalid";

      var resolveHostAddress = new ResolveHostAddressCallback((hostname, cancellationToken) => {
        Assert.AreEqual(pseudoServerHostName, hostname);
#if NET45
        var tcs = new TaskCompletionSource<IPAddress>();
        tcs.SetException(new NotImplementedException());
        return tcs.Task;
#else
        return Task.FromException<IPAddress>(new NotImplementedException());
#endif
      });
      var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", $"{pseudoServerHostName}:10143");

      var ex = Assert.ThrowsAsync<ImapConnectionException>(async () => await ImapSession.CreateAsync(prof, resolveHostAddress: resolveHostAddress));

      Assert.IsInstanceOf<NotImplementedException>(ex.InnerException);
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeCreateSession()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", "imap.invalid");

      prof.AllowInsecureLogin = true;

      using (var cts = new CancellationTokenSource()) {
        cts.Cancel();

        Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                    Throws.InstanceOf<OperationCanceledException>());
      }
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeQueryingServerCapability()
    {
      // connect
      //   <--cancellation
      // CAPABILITY

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        using (var cts = new CancellationTokenSource()) {
          server.SetGreetingResponse("* OK ready\r\n",
                                     () => cts.Cancel(),
                                     null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        Assert.IsNull(server.DequeueRequest()); // CAPABILITY not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeUpgradeStream()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // STARTTLS

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                       "$tag OK done\r\n",
                                       () => cts.Cancel(),
                                       null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // STARTTLS not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeSendSaslClientResponse()
    {
      // connect
      // CAPABILITY
      // AUTHENTICATE(SaslMechanism)
      // <--cancellation
      // SASL server challenge

      ImapPseudoServer.Start((server) => {
        var credential = new NetworkCredential("imap-user", (string)null);
        var prof = new Profile(null, "imap", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5" };
        prof.AllowInsecureLogin = true;

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                                 "+ \r\n",
                                 () => cts.Cancel(),
                                 null);
          server.EnqueueTaggedResponse("$tag BAD authentication cancelled\r\n");

          using (var authMechanism = new SaslPseudoMechanism(1)) {
            authMechanism.Credential = credential;

            Assert.ThrowsAsync<ImapAuthenticationException>(async () => await ImapSession.CreateAsync(prof, authMechanism: authMechanism, cancellationToken: cts.Token));
          }
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        StringAssert.Contains(" AUTHENTICATE X-PSEUDO-MECHANISM", server.DequeueRequest());
        Assert.AreEqual("*\r\n", server.DequeueRequest()); // client response not sent
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeAuth_WithSpecificMechanism()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(SaslMechanism)

      ImapPseudoServer.Start((server) => {
        var credential = new NetworkCredential("imap-user", (string)null);
        var prof = new Profile(null, "imap", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5" };
        prof.AllowInsecureLogin = true;

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                         "* CAPABILITY IMAP4rev1\r\n" +
                                         "$tag OK done\r\n",
                                         () => cts.Cancel(),
                                         null);

          using (var authMechanism = new SaslPseudoMechanism(1)) {
            authMechanism.Credential = credential;

            Assert.That(async () => await ImapSession.CreateAsync(prof, authMechanism: authMechanism, cancellationToken: cts.Token),
                        Throws.InstanceOf<OperationCanceledException>());
          }
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeAuth_SelectAppropriateMechanism()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(AUTH=*)

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN" };
        prof.AllowInsecureLogin = true;

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                       "$tag OK done\r\n",
                                       () => cts.Cancel(),
                                       null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeAuth_SelectAppropriateMechanism_FallbackToNext()
    {
      // connect
      // CAPABILITY
      // AUTHENTICATE(AUTH=*)
      //   <--cancellation
      // AUTHENTICATE(AUTH=*)

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("AUTHENTICATE LOGIN\r\n",
                                       "$tag NO done\r\n",
                                       () => cts.Cancel(),
                                       null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        StringAssert.Contains(" AUTHENTICATE LOGIN", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeAuth_SelectAppropriateMechanism_FallbackToLogin()
    {
      // connect
      // CAPABILITY
      // AUTHENTICATE(AUTH=*)
      //   <--cancellation
      // LOGIN

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN" };
        prof.AllowInsecureLogin = true;

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("AUTHENTICATE LOGIN\r\n",
                                       "$tag NO done\r\n",
                                       () => cts.Cancel(),
                                       null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        StringAssert.Contains(" AUTHENTICATE LOGIN", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // LOGIN not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeAuth_MechanismSpecified()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(AUTH=XXX)

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=LOGIN", server.HostPort);

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                       "$tag OK done\r\n",
                                       () => cts.Cancel(),
                                       null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeAuth_Anonymous()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(AUTH=ANONYMOUS)

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(null, "user%40imap.example.net;auth=anonymous", server.HostPort);

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                       "$tag OK done\r\n",
                                       () => cts.Cancel(),
                                       null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      });
    }

    [Test]
    public void TestCreateAsync_CancellationRequested_BeforeAuth_Anonymous_FallbackToLogin()
    {
      // connect
      // CAPABILITY
      // AUTHENTICATE(AUTH=ANONYMOUS)
      //   <--cancellation
      // LOGIN

      ImapPseudoServer.Start((server) => {
        var prof = new Profile(null, null, server.HostPort);

        using (var cts = new CancellationTokenSource()) {
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("AUTHENTICATE ANONYMOUS\r\n",
                                       "$tag NO done\r\n",
                                       () => cts.Cancel(),
                                       null);

          Assert.That(async () => await ImapSession.CreateAsync(prof, cancellationToken: cts.Token),
                      Throws.InstanceOf<OperationCanceledException>());
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        StringAssert.Contains(" AUTHENTICATE ANONYMOUS", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // LOGIN not requested
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAuthMechanismSpecified()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAuthMechanismSpecified_AuthenticationFailure()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        var ex = Assert.ThrowsAsync<ImapAuthenticationException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsNull(ex.InnerException);
        Assert.IsNotNull(ex.Result);
        Assert.IsTrue(ex.Result.IsFinalResponseCondition(ImapResponseCondition.No));
      });
    }

    [Test]
    public void TestAuthenticate_NoUserAndAuthMechanismSpecified()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK ready\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            var b = new ImapUriBuilder(prof.Authority);

            b.UserName = "user";

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_NoUserAndAuthMechanismSpecified_AuthenticationFailure()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        var ex = Assert.ThrowsAsync<ImapAuthenticationException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsNull(ex.InnerException);
        Assert.IsNotNull(ex.Result);
        Assert.IsTrue(ex.Result.IsFinalResponseCondition(ImapResponseCondition.No));
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAuthMechanismSpecified_MechanismNotSupported()
      => AuthenticateWithSpecifiedButUnsupportedMechanism("user;auth=X-UNKNOWN");

    [Test]
    public void TestAuthenticate_NoUserAndAuthMechanismSpecified_MechanismNotSupported()
      => AuthenticateWithSpecifiedButUnsupportedMechanism(";auth=X-UNKNOWN");

    private void AuthenticateWithSpecifiedButUnsupportedMechanism(string userinfo)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN AUTH=X-UNKNOWN\r\n" +
                                     "$tag OK done\r\n");

        var ex = Assert.ThrowsAsync<SaslMechanismNotSupportedException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsNull(ex.InnerException);
        Assert.AreEqual("X-UNKNOWN", ex.MechanismName);
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanism(string userinfo)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] { "ANONYMOUS", "LOGIN", "PLAIN" };
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE LOGIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            var b = new ImapUriBuilder(prof.Authority);

            b.AuthType = ImapAuthenticationMechanism.Login;

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_ExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_ExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user");
    }

    private void AuthenticateExceptionWhileAuthentication(string userinfo)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] { "DIGEST-MD5", "PLAIN" };

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE DIGEST-MD5\r\n",
                               "+ xxx-invalid-response-xxx\r\n");

        Assert.ThrowsAsync<FormatException>(async () =>  await ImapSession.CreateAsync(prof));
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(string userinfo)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] { "ANONYMOUS", "LOGIN", "PLAIN" };
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE LOGIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            var b = new ImapUriBuilder(prof.Authority);

            b.AuthType = ImapAuthenticationMechanism.Plain;

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToLogin()
    {
      AuthenticateWithAppropriateMechanismFallbackToLogin("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToLogin()
    {
      AuthenticateWithAppropriateMechanismFallbackToLogin("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToLogin(string userinfo)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            var b = new ImapUriBuilder(prof.Authority);

            b.AuthType = null;

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToLogin_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToLogin_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure(string userinfo)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag NO done\r\n");

        var ex = Assert.ThrowsAsync<ImapAuthenticationException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsNull(ex.InnerException);
        Assert.IsNotNull(ex.Result);
        Assert.IsTrue(ex.Result.IsFinalResponseCondition(ImapResponseCondition.No));
      });
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound("user");
    }

    private void AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(string userinfo)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");

        Assert.ThrowsAsync<ImapNoAppropriateAuthMechanismException>(async () => await ImapSession.CreateAsync(prof));
      });
    }

    [Test]
    public void TestAuthenticate_AnonymousMechanismSpecified()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(null, "user%40imap.example.net;auth=anonymous", server.HostPort);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE ANONYMOUS\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // AUTHENTICATE
        StringAssert.Contains("user@imap.example.net", Base64.GetDecodedString(server.DequeueRequest()));
      });
    }

    [Test]
    public void TestAuthenticate_AnonymousMechanismSpecified_AuthenticationFailure()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(null, "user%40imap.example.net;auth=anonymous", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE ANONYMOUS\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        var ex = Assert.ThrowsAsync<ImapAuthenticationException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsNull(ex.InnerException);
        Assert.IsNotNull(ex.Result);
        Assert.IsTrue(ex.Result.IsFinalResponseCondition(ImapResponseCondition.No));

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // AUTHENTICATE ANONYMOUS
        StringAssert.Contains("user@imap.example.net", Base64.GetDecodedString(server.DequeueRequest()));
      });
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(null, null, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE ANONYMOUS\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueTaggedResponse("LOGIN anonymous anonymous@\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            var b = new ImapUriBuilder(prof.Authority);

            b.UserName = "anonymous";

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // AUTHENTICATE ANONYMOUS
        StringAssert.Contains("anonymous@", Base64.GetDecodedString(server.DequeueRequest()));
      });
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified_AuthenticationFailure()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(null, null, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN anonymous anonymous@\r\n",
                                     "$tag NO done\r\n");

        var ex = Assert.ThrowsAsync<ImapAuthenticationException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsNull(ex.InnerException);
        Assert.IsNotNull(ex.Result);
        Assert.IsTrue(ex.Result.IsFinalResponseCondition(ImapResponseCondition.No));
      });
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(null, null, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");

        Assert.ThrowsAsync<ImapNoAppropriateAuthMechanismException>(async () => await ImapSession.CreateAsync(prof));
      });
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN", "CRAM-MD5" };
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            var b = new ImapUriBuilder(prof.Authority);

            b.AuthType = ImapAuthenticationMechanism.CRAMMD5;

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_SecureSession()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "ANONYMOUS", "LOGIN", "PLAIN" };
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 SASL-IR AUTH=LOGIN AUTH=PLAIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("AUTHENTICATE LOGIN ",
                                     "$tag NO done\r\n");
        server.EnqueueTaggedResponse("AUTHENTICATE PLAIN ",
                                     "$tag NO done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof, createSslStream: (connection, baseStream, cancellationToken) => {
            return Task.FromResult<Stream>(baseStream); // TODO: return SSL stream
          })) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);
            StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
            Assert.IsTrue(session.IsSecureConnection);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_FallbackToNextAppropriateMechanism()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5" };
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE DIGEST-MD5\r\n",
                               "+ cmVhbG09ImVsd29vZC5pbm5vc29mdC5jb20iLG5vbmNlPSJPQTZNRzl0" +
                               "RVFHbTJoaCIscW9wPSJhdXRoIixhbGdvcml0aG09bWQ1LXNlc3MsY2hh" +
                               "cnNldD11dGYtOA==\r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueResponse("AUTHENTICATE CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        Assert.DoesNotThrowAsync(async () => {
          using (var session = await ImapSession.CreateAsync(prof)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            var b = new ImapUriBuilder(prof.Authority);

            b.AuthType = ImapAuthenticationMechanism.CRAMMD5;

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
          }
        });
      });
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_FallbackToNextAppropriateMechanism_AuthenticationFailure()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5" };
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE DIGEST-MD5\r\n",
                               "+ cmVhbG09ImVsd29vZC5pbm5vc29mdC5jb20iLG5vbmNlPSJPQTZNRzl0" +
                               "RVFHbTJoaCIscW9wPSJhdXRoIixhbGdvcml0aG09bWQ1LXNlc3MsY2hh" +
                               "cnNldD11dGYtOA==\r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueResponse("AUTHENTICATE CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        var ex = Assert.ThrowsAsync<ImapAuthenticationException>(async () => await ImapSession.CreateAsync(prof));

        Assert.IsNull(ex.InnerException);
        Assert.IsNotNull(ex.Result);
        Assert.IsTrue(ex.Result.IsFinalResponseCondition(ImapResponseCondition.No));
      });
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_AppropriateMechanismNotFound()
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "ANONYMOUS", "LOGIN", "PLAIN" };
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");

        Assert.ThrowsAsync<ImapNoAppropriateAuthMechanismException>(async () => await ImapSession.CreateAsync(prof));
      });
    }

    [Test]
    public void TestAuthenticate_WithSpecificMechanism()
    {
      AuthenticateSpecificMechanism(new NetworkCredential("sasl-user", (string)null));
    }

    [Test]
    public void TestAuthenticate_WithSpecificMechanism_CredentialNotSet()
    {
      AuthenticateSpecificMechanism(null);
    }

    private void AuthenticateSpecificMechanism(NetworkCredential credential)
    {
      ImapPseudoServer.Start((server) => {
        var prof = new Profile(credential, "imap", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5" };
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var authMechanism = new SaslPseudoMechanism(1)) {
          authMechanism.Credential = credential;

          Assert.DoesNotThrowAsync(async () => {
            using (var session = await ImapSession.CreateAsync(prof, authMechanism: authMechanism)) {
              Assert.AreEqual(ImapSessionState.Authenticated, session.State);

              if (credential == null)
                StringAssert.AreEqualIgnoringCase(string.Format("imap://;AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort),
                                                  session.Authority.AbsoluteUri);
              else
                StringAssert.AreEqualIgnoringCase(string.Format("imap://sasl-user;AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort),
                                                  session.Authority.AbsoluteUri);
            }
          });

          Assert.AreSame(credential,
                         authMechanism.Credential,
                         "credential must be kept");

          Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);
        }
      });
    }

    [Test]
    public void TestAuthenticate_WithSpecificMechanism_CapabilityNotAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        var credential = new NetworkCredential("sasl-user", (string)null);
        var prof = new Profile(credential, "imap", server.HostPort);

        prof.UsingSaslMechanisms = new[] { "LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5" };
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var authMechanism = new SaslPseudoMechanism(1)) {
          authMechanism.Credential = credential;

          Assert.DoesNotThrowAsync(async () => {
            using (var session = await ImapSession.CreateAsync(prof, authMechanism: authMechanism)) {
              Assert.AreEqual(ImapSessionState.Authenticated, session.State);

              StringAssert.AreEqualIgnoringCase(string.Format("imap://sasl-user;AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort),
                                                session.Authority.AbsoluteUri);
            }
          });

          Assert.AreSame(credential,
                         authMechanism.Credential,
                         "credential must be kept");

          Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);
        }
      });
    }
  }
}
