using System;
using System.Net;
using System.Runtime.CompilerServices;
using NUnit.Framework;

using Smdn.Net.Imap4.Urls;

#if NET45
using Array = Smdn.ArrayExtensions; // for Empty<T>()
#endif

namespace Smdn.Net.Imap4.Protocol.Session {
  internal static class TestCases {
    public static void WithConnectedSession(Action<ImapSession> action,
                                            [CallerMemberName] string callerMemberName = "",
                                            [CallerFilePath] string sourceFilePath = "",
                                            [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithConnectedSession(null, (session, server) => {
        action(session);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithConnectedSession(Action<ImapSession, ImapPseudoServer> action,
                                            [CallerMemberName] string callerMemberName = "",
                                            [CallerFilePath] string sourceFilePath = "",
                                            [CallerLineNumber] int sourceLineNumber = 0)

    {
      WithConnectedSession(null, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithConnectedSession(string[] capabilities, Action<ImapSession, ImapPseudoServer> action,
                                            [CallerMemberName] string callerMemberName = "",
                                            [CallerFilePath] string sourceFilePath = "",
                                            [CallerLineNumber] int sourceLineNumber = 0)
    {
      ImapPseudoServer.Start((server) => {
        var capa = (capabilities == null || capabilities.Length == 0)
                   ? string.Empty
                   : string.Format(" [CAPABILITY IMAP4REV1 {0}]", string.Join(" ", capabilities));

        server.SetGreetingResponse(string.Format("* OK{0} ImapPseudoServer ready\r\n", capa));

        using (var session = new ImapSession(server.Host, server.Port)) {
          action(session, server);
        }
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithAuthenticatedSession(Action<ImapSession, ImapPseudoServer> action,
                                                [CallerMemberName] string callerMemberName = "",
                                                [CallerFilePath] string sourceFilePath = "",
                                                [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedSession((string[])null, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithAuthenticatedSession(string[] additionalCapabilities, Action<ImapSession, ImapPseudoServer> action,
                                                [CallerMemberName] string callerMemberName = "",
                                                [CallerFilePath] string sourceFilePath = "",
                                                [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedSession(additionalCapabilities, (session, server, expectedAuthority) => {
        action(session, server);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithAuthenticatedSession(Action<ImapSession, ImapPseudoServer, Uri> action,
                                                [CallerMemberName] string callerMemberName = "",
                                                [CallerFilePath] string sourceFilePath = "",
                                                [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedSession((string[])null, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithAuthenticatedSession(string[] additionalCapabilities, Action<ImapSession, ImapPseudoServer, Uri> action,
                                                [CallerMemberName] string callerMemberName = "",
                                                [CallerFilePath] string sourceFilePath = "",
                                                [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithConnectedSession((string[])null, (session, server) => {
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK authenticated\r\n");

        session.Login(new NetworkCredential("user", "pass"));

        server.DequeueRequest(); // LOGIN

        if (additionalCapabilities == null)
          additionalCapabilities = Array.Empty<string>();

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     string.Format("* CAPABILITY {0} IMAP4rev1\r\n", string.Join(" ", additionalCapabilities)) +
                                     "$tag OK CAPABILITY completed\r\n");

        session.Capability();

        server.DequeueRequest(); // CAPABILITY

        var expectedAuthority = new Uri(string.Format("{0}://{1}@{2}/",
                                                      ImapUri.UriSchemeImap,
                                                      "user", // XXX
                                                      server.HostPort));

        action(session, server, expectedAuthority);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    private const int defaultExistingMessageCount = 15;

    public static void WithMailboxSelectedSession(Action<ImapSession, ImapPseudoServer> action,
                                                  [CallerMemberName] string callerMemberName = "",
                                                  [CallerFilePath] string sourceFilePath = "",
                                                  [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailboxSelectedSession(defaultExistingMessageCount, (string[])null, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailboxSelectedSession(int existingMessageCount, Action<ImapSession, ImapPseudoServer> action,
                                                  [CallerMemberName] string callerMemberName = "",
                                                  [CallerFilePath] string sourceFilePath = "",
                                                  [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailboxSelectedSession(existingMessageCount, (string[])null, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailboxSelectedSession(string[] additionalCapabilities, Action<ImapSession, ImapPseudoServer> action,
                                                  [CallerMemberName] string callerMemberName = "",
                                                  [CallerFilePath] string sourceFilePath = "",
                                                  [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailboxSelectedSession(defaultExistingMessageCount, additionalCapabilities, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailboxSelectedSession(int existingMessageCount,
                                                  string[] additionalCapabilities,
                                                  Action<ImapSession, ImapPseudoServer> action,
                                                  [CallerMemberName] string callerMemberName = "",
                                                  [CallerFilePath] string sourceFilePath = "",
                                                  [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailboxSelectedSession(existingMessageCount,
                                 additionalCapabilities,
                                 (session, server, expectedAuthority) => {
        action(session, server);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailboxSelectedSession(Action<ImapSession, ImapPseudoServer, Uri> action,
                                                  [CallerMemberName] string callerMemberName = "",
                                                  [CallerFilePath] string sourceFilePath = "",
                                                  [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailboxSelectedSession(defaultExistingMessageCount, null, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailboxSelectedSession(string[] additionalCapabilities,
                                                  Action<ImapSession, ImapPseudoServer, Uri> action,
                                                  [CallerMemberName] string callerMemberName = "",
                                                  [CallerFilePath] string sourceFilePath = "",
                                                  [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailboxSelectedSession(defaultExistingMessageCount, additionalCapabilities, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailboxSelectedSession(int existingMessageCount,
                                                  string[] additionalCapabilities,
                                                  Action<ImapSession, ImapPseudoServer, Uri> action,
                                                  [CallerMemberName] string callerMemberName = "",
                                                  [CallerFilePath] string sourceFilePath = "",
                                                  [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedSession(additionalCapabilities, (session, server) => {
        server.EnqueueTaggedResponse("LIST",
                                     "* LIST () \".\" \"INBOX\"\r\n" +
                                     "$tag OK LIST completed\r\n");

        session.ListAll(out var mailboxes);

        server.DequeueRequest();

        Assert.IsNull(session.SelectedMailbox);

        server.EnqueueTaggedResponse("SELECT",
                                     "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                                     "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                                     string.Format("* {0} EXISTS\r\n", existingMessageCount) +
                                     "* 2 RECENT\r\n" +
                                     "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                                     "* OK [UIDNEXT 16]\r\n" +
                                     "* OK [UNSEEN 13]\r\n" +
                                     "$tag OK [READ-WRITE] SELECT completed\r\n");

        session.Select(mailboxes[0]);

        Assert.AreEqual(mailboxes[0], session.SelectedMailbox);

        server.DequeueRequest();

        session.HandlesIncapableAsException = true;

        var expectedAuthority = new Uri(string.Format("{0}://{1}@{2}/",
                                                      ImapUri.UriSchemeImap,
                                                      "user", // XXX
                                                      server.HostPort));

        action(session, server, expectedAuthority);

        if (session.State == ImapSessionState.Selected) {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK CLOSE completed\r\n");

          Assert.IsTrue(session.Close().IsSuccess);
        }
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }
  }
}
