using System;
using System.Net;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.Session;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  [TestFixture]
  public class ImapCommandResultTests {
    [Test]
    public void TestSerializeBinary_FinalStatus_TaggedOk()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        var result = session.Capability();

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsFalse(result.IsConcludedByClientSide);
          Assert.IsTrue(result.IsSuccess);

          Assert.AreEqual("\"CAPABILITY completed\"", deserialized.Description);

          Assert.IsNotNull(deserialized.FinalStatusResponse);
          Assert.AreEqual(ImapResponseCondition.Ok, deserialized.FinalStatusResponse.Condition);

          Assert.IsNotNull(deserialized.FinalStatusResponse.ResponseText);
          Assert.AreEqual("CAPABILITY completed", deserialized.FinalStatusResponse.ResponseText.Text);

          Assert.IsNotNull(deserialized.TaggedStatusResponse);
          Assert.AreEqual("0002", deserialized.TaggedStatusResponse.Tag);
          Assert.AreEqual(ImapResponseCondition.Ok, deserialized.TaggedStatusResponse.Condition);
          Assert.IsNull(deserialized.TaggedStatusResponse.ResponseText.Code);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          Assert.AreEqual(2, deserialized.ReceivedResponses.Count);
          Assert.IsInstanceOf<ImapServerDataResponse>(deserialized.ReceivedResponses[0]);
          Assert.IsInstanceOf<ImapTaggedStatusResponse>(deserialized.ReceivedResponses[1]);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_FinalStatus_TaggedNo()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag NO incorrect\r\n");

        var result = session.Login(new NetworkCredential("user", "pass"));

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsFalse(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);

          Assert.AreEqual("\"incorrect\"", deserialized.Description);

          Assert.IsNotNull(deserialized.FinalStatusResponse);
          Assert.AreEqual(ImapResponseCondition.No, deserialized.FinalStatusResponse.Condition);

          Assert.IsNotNull(deserialized.FinalStatusResponse.ResponseText);
          Assert.AreEqual("incorrect", deserialized.FinalStatusResponse.ResponseText.Text);

          Assert.IsNotNull(deserialized.TaggedStatusResponse);
          Assert.AreEqual("0000", deserialized.TaggedStatusResponse.Tag);
          Assert.AreEqual(ImapResponseCondition.No, deserialized.TaggedStatusResponse.Condition);
          Assert.IsNull(deserialized.TaggedStatusResponse.ResponseText.Code);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          Assert.AreEqual(1, deserialized.ReceivedResponses.Count);
          Assert.IsInstanceOf<ImapTaggedStatusResponse>(deserialized.ReceivedResponses[0]);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_FinalStatus_UntaggedBye()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueUntaggedByeResponse("* BYE Disconnected for inactivity.\r\n");

        var result = session.NoOp();

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsFalse(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);

          Assert.AreEqual("\"Disconnected for inactivity.\"", deserialized.Description);

          Assert.IsNotNull(deserialized.FinalStatusResponse);
          Assert.AreEqual(ImapResponseCondition.Bye, deserialized.FinalStatusResponse.Condition);

          Assert.IsNotNull(deserialized.FinalStatusResponse.ResponseText);
          Assert.AreEqual("Disconnected for inactivity.", deserialized.FinalStatusResponse.ResponseText.Text);

          Assert.IsNull(deserialized.TaggedStatusResponse);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          Assert.AreEqual(1, deserialized.ReceivedResponses.Count);
          Assert.IsInstanceOf<ImapUntaggedStatusResponse>(deserialized.ReceivedResponses[0]);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_ConcludedByClientSide_Success()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        var result = session.Login(new NetworkCredential("user", "pass")); // already authenticated

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsTrue(result.IsConcludedByClientSide);
          Assert.IsTrue(result.IsSuccess);

          StringAssert.StartsWith("<", deserialized.Description);
          StringAssert.EndsWith(">", deserialized.Description);

          Assert.IsNull(deserialized.FinalStatusResponse);
          Assert.IsNull(deserialized.TaggedStatusResponse);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          CollectionAssert.IsEmpty(deserialized.ReceivedResponses);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_ConcludedByClientSide_Failure()
    {
      TestCases.WithConnectedSession(delegate (ImapSession session) {
        var result = session.Login(new NullCredential());

        TestUtils.Assert.IsSerializableBinaryFormat(result, deserialized => {
          Assert.IsTrue(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);

          StringAssert.StartsWith("<", deserialized.Description);
          StringAssert.EndsWith(">", deserialized.Description);

          Assert.IsNull(deserialized.FinalStatusResponse);
          Assert.IsNull(deserialized.TaggedStatusResponse);

          Assert.IsNotNull(deserialized.ReceivedResponses);
          CollectionAssert.IsEmpty(deserialized.ReceivedResponses);
        });
      });
    }
  }
}

