using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.Collections;
using Smdn.IO; // Stream.Close
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  [TestFixture]
  public class ImapDataTests {
    private IReadOnlyList<ImapData> Create(string text)
    {
      using (var receiver = new ImapPseudoResponseReceiver("* X-DUMMY " + text + "\r\n")) {
        for (; ; ) {
          if (receiver.ReceiveResponse() is ImapServerDataResponse data)
            return data.DataList.Slice(1);
        }
      }
    }

    private static readonly Action<Stream> emptyReadAction = s => { };

    [Test]
    public void TestNil()
    {
      var nil = Create("NIL")[0];

      Assert.AreEqual(ImapDataFormat.Nil, nil.Format);
      Assert.AreSame(Create("NIL")[0], nil);

      Assert.Throws<InvalidOperationException>(() => Assert.IsNull(nil.List));

      Assert.Throws<InvalidOperationException>(() => nil.GetTextAsString());
      Assert.Throws<InvalidOperationException>(() => nil.GetTextAsByteString());
      Assert.Throws<InvalidOperationException>(() => nil.ReadText(emptyReadAction));
      Assert.Throws<InvalidOperationException>(() => nil.GetTextLength());
    }

    [Test]
    public void TestList()
    {
      var list = Create("(NIL NIL text)")[0];

      Assert.AreEqual(ImapDataFormat.List, list.Format);
      Assert.IsNotNull(list.List);
      Assert.AreEqual(3, list.List.Count);
      Assert.AreEqual(ImapDataFormat.Nil, list.List[0].Format);
      Assert.AreEqual(ImapDataFormat.Nil, list.List[1].Format);
      Assert.AreEqual(ImapDataFormat.Text, list.List[2].Format);
      Assert.AreEqual("text", list.List[2].GetTextAsString());

      Assert.Throws<InvalidOperationException>(() => list.GetTextAsString());
      Assert.Throws<InvalidOperationException>(() => list.GetTextAsByteString());
      Assert.Throws<InvalidOperationException>(() => list.ReadText(emptyReadAction));
      Assert.Throws<InvalidOperationException>(() => list.GetTextLength());
    }

    [Test]
    public void TestList_Empty()
    {
      var list = Create("()")[0];

      Assert.AreEqual(ImapDataFormat.List, list.Format);
      Assert.IsNotNull(list.List);
      Assert.AreEqual(0, list.List.Count);
    }

    [Test]
    public void TestText_Array()
    {
      const string input = "string";
      var text = Create(input)[0];

      Assert.AreEqual(ImapDataFormat.Text, text.Format);
      Assert.AreEqual(input.Length, text.GetTextLength());
      Assert.AreEqual(ByteString.CreateImmutable(input), text.GetTextAsByteString());
      Assert.AreSame(text.GetTextAsByteString(), text.GetTextAsByteString());
      Assert.AreEqual(input, text.GetTextAsString());
      Assert.Throws<ArgumentNullException>(() => text.ReadText(null));

      Assert.Throws<InvalidOperationException>(() => Assert.IsNull(text.List));

      var str = ByteString.CreateImmutable(input);

      text.ReadText(stream => {
        Assert.IsTrue(stream.CanRead, nameof(stream.CanRead));
        Assert.IsTrue(stream.CanSeek, nameof(stream.CanSeek));
        Assert.IsFalse(stream.CanWrite, nameof(stream.CanWrite));
        Assert.Throws<NotSupportedException>(() => stream.WriteByte(0), nameof(stream.WriteByte));
        Assert.Throws<NotSupportedException>(() => stream.Write(new byte[4], 0, 4), nameof(stream.Write));
#if NETCOREAPP2_0 || NETCOREAPP2_1
        Assert.Throws<NotSupportedException>(() => stream.BeginWrite(new byte[4], 0, 4, null, null), nameof(stream.BeginWrite));
#endif
        Assert.ThrowsAsync<NotSupportedException>(async () => await stream.WriteAsync(new byte[4], 0, 4), nameof(stream.WriteAsync));
        Assert.Throws<NotSupportedException>(() => stream.SetLength(0L), nameof(stream.SetLength));

        FileAssert.AreEqual(new MemoryStream(str.Segment.Array, str.Segment.Offset, str.Segment.Count, false),
                            stream);

        Assert.DoesNotThrow(stream.Close, "Close");
        Assert.DoesNotThrow(stream.Dispose, "Dispose");
      });

      // read again
      text.ReadText(stream => {
        FileAssert.AreEqual(new MemoryStream(str.Segment.Array, str.Segment.Offset, str.Segment.Count, false),
                            stream);
      });
    }

    [Test]
    public void TestText_Stream()
    {
      var input = "string" + new string('x', 80 * 1024 /* 80k octets*/); // > literalAllocationThreshold (40k)
      var text = Create($"~{{{input.Length}}}\r\n" + input)[0];

      Assert.AreEqual(ImapDataFormat.Text, text.Format);
      Assert.AreEqual(input.Length, text.GetTextLength());
      Assert.AreEqual(ByteString.CreateImmutable(input), text.GetTextAsByteString());
      Assert.AreNotSame(text.GetTextAsByteString(), text.GetTextAsByteString());
      Assert.AreEqual(input, text.GetTextAsString());
      Assert.Throws<ArgumentNullException>(() => text.ReadText(null));

      Assert.Throws<InvalidOperationException>(() => Assert.IsNull(text.List));

      text.ReadText(stream => {
        Assert.IsTrue(stream.CanRead, nameof(stream.CanRead));
        Assert.IsTrue(stream.CanSeek, nameof(stream.CanSeek));
        Assert.IsFalse(stream.CanWrite, nameof(stream.CanWrite));
        Assert.Throws<NotSupportedException>(() => stream.WriteByte(0), nameof(stream.WriteByte));
        Assert.Throws<NotSupportedException>(() => stream.Write(new byte[4], 0, 4), nameof(stream.Write));
#if NETCOREAPP2_0 || NETCOREAPP2_1
        Assert.Throws<NotSupportedException>(() => stream.BeginWrite(new byte[4], 0, 4, null, null), nameof(stream.BeginWrite));
#endif
        Assert.ThrowsAsync<NotSupportedException>(async () => await stream.WriteAsync(new byte[4], 0, 4), nameof(stream.WriteAsync));
        Assert.Throws<NotSupportedException>(() => stream.SetLength(0L), nameof(stream.SetLength));

        FileAssert.AreEqual(new MemoryStream(Encoding.ASCII.GetBytes(input), false),
                            stream);

        Assert.DoesNotThrow(stream.Close, "Close");
        Assert.DoesNotThrow(stream.Dispose, "Dispose");
      });

      // read again
      text.ReadText(stream => {
        FileAssert.AreEqual(new MemoryStream(Encoding.ASCII.GetBytes(input), false),
                            stream);
      });
    }

    [Test]
    public void TestGetTextAsNumber()
    {
      Assert.AreEqual(0UL, Create("0")[0].GetTextAsByteString().ToUInt64());
      Assert.AreEqual(1UL, Create("1")[0].GetTextAsByteString().ToUInt64());
      Assert.AreEqual(5UL, Create("005")[0].GetTextAsByteString().ToUInt64());
      Assert.AreEqual(123UL, Create("123")[0].GetTextAsByteString().ToUInt64());

      Assert.Throws<OverflowException>(() => Create("18446744073709551616")[0].GetTextAsByteString().ToUInt64());
    }

    [Test]
    public void TestSerializeBinary_Nil()
    {
      var nil = Create("NIL")[0];

      TestUtils.Assert.IsSerializableBinaryFormat(nil, deserialized => {
        Assert.AreEqual(ImapDataFormat.Nil, deserialized.Format);
        Assert.Throws<InvalidOperationException>(() => Assert.IsNull(deserialized.List));

        Assert.AreNotSame(nil, deserialized);
      });
    }

    [Test]
    public void TestSerializeBinary_List()
    {
      var list = Create("(NIL NIL text)")[0];

      TestUtils.Assert.IsSerializableBinaryFormat(list, deserialized => {
        Assert.AreEqual(ImapDataFormat.List, deserialized.Format);
        Assert.IsNotNull(deserialized.List);
        Assert.AreEqual(3, deserialized.List.Count);
        Assert.AreEqual(ImapDataFormat.Nil, deserialized.List[0].Format);
        Assert.AreEqual(ImapDataFormat.Nil, deserialized.List[1].Format);
        Assert.AreEqual(ImapDataFormat.Text, deserialized.List[2].Format);
        Assert.AreEqual("text", deserialized.List[2].GetTextAsString());
      });
    }

    [Test]
    public void TestSerializeBinary_List_Empty()
    {
      var list = Create("()")[0];

      TestUtils.Assert.IsSerializableBinaryFormat(list, deserialized => {
        Assert.AreEqual(ImapDataFormat.List, deserialized.Format);
        Assert.IsNotNull(deserialized.List);
        Assert.AreEqual(0, deserialized.List.Count);
      });
    }

    [Test]
    public void TestSerializeBinary_Text_Array()
    {
      var input = "string";
      var text = Create(input)[0];

      TestUtils.Assert.IsSerializableBinaryFormat(text, deserialized => {
        Assert.AreEqual(ImapDataFormat.Text, deserialized.Format);
        Assert.AreEqual(input.Length, deserialized.GetTextLength());
        Assert.AreNotSame(text.GetTextAsByteString(), deserialized.GetTextAsByteString());
        Assert.AreEqual(input, deserialized.GetTextAsString());

        Assert.Throws<InvalidOperationException>(() => Assert.IsNull(deserialized.List));
      });
    }

    [Test]
    public void TestSerializeBinary_Text_Stream()
    {
      var input = "string" + new string('x', 80 * 1024 /* 80k octets*/); // > literalAllocationThreshold (40k)
      var text = Create($"~{{{input.Length}}}\r\n" + input)[0];

      TestUtils.Assert.IsSerializableBinaryFormat(text, deserialized => {
        Assert.AreEqual(ImapDataFormat.Text, deserialized.Format);
        Assert.AreEqual(input.Length, deserialized.GetTextLength());

        deserialized.ReadText(stream => {
          Assert.IsTrue(stream.CanRead, nameof(stream.CanRead));
          Assert.IsTrue(stream.CanSeek, nameof(stream.CanSeek));
          Assert.IsFalse(stream.CanWrite, nameof(stream.CanWrite));
          Assert.Throws<NotSupportedException>(() => stream.WriteByte(0), nameof(stream.WriteByte));
          Assert.Throws<NotSupportedException>(() => stream.Write(new byte[4], 0, 4), nameof(stream.Write));
#if NETCOREAPP2_0 || NETCOREAPP2_1
          Assert.Throws<NotSupportedException>(() => stream.BeginWrite(new byte[4], 0, 4, null, null), nameof(stream.BeginWrite));
#endif
          Assert.ThrowsAsync<NotSupportedException>(async () => await stream.WriteAsync(new byte[4], 0, 4), nameof(stream.WriteAsync));
          Assert.Throws<NotSupportedException>(() => stream.SetLength(0L), nameof(stream.SetLength));

          Assert.DoesNotThrow(stream.Close, "Close");
          Assert.DoesNotThrow(stream.Dispose, "Dispose");
        });

        text.ReadText(originalStream => {
          deserialized.ReadText(deserializedStream => {
            FileAssert.AreEqual(originalStream, deserializedStream);
          });
        });

        Assert.Throws<InvalidOperationException>(() => Assert.IsNull(deserialized.List));
      });
    }
  }
}
