using System;
using System.Linq;
using System.IO;

using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  [TestFixture]
  public class ImapServerDataResponseParserTests {
    private ImapServerDataResponse GetSingleDataResponse(string response)
    {
      using (var receiver = new ImapPseudoResponseReceiver(response)) {
        for (; ; ) {
          if (receiver.ReceiveResponse() is ImapServerDataResponse data)
            return data;
        }
      }
    }

    [Test]
    public void TestCapabilityResponse()
    {
      var response =
        GetSingleDataResponse("* CAPABILITY IMAP4rev1 STARTTLS AUTH=GSSAPI XPIG-LATIN\r\n");

      var caps = ImapServerDataResponseParser.FromCapability(response);

      Assert.AreEqual(4, caps.Count, "capability count");
      Assert.IsTrue(caps.Contains(ImapCapability.Imap4Rev1));
      Assert.IsTrue(caps.Contains(ImapCapability.StartTls));
      Assert.IsTrue(caps.Contains("IMAP4rev1"));
      Assert.IsTrue(caps.Contains("STARTTLS"));
      Assert.IsTrue(caps.Contains("AUTH=GSSAPI"));
      Assert.IsTrue(caps.Contains("XPIG-LATIN"));
    }

    [Test]
    public void TestCapabilityResponse_Duplicated()
    {
      var response =
        GetSingleDataResponse("* CAPABILITY IMAP4rev1 IMAP4rev1 XPIG-LATIN XPIG-LATIN\r\n");

      var caps = ImapServerDataResponseParser.FromCapability(response);

      Assert.AreEqual(2, caps.Count, "capability count");
      Assert.IsTrue(caps.Contains(ImapCapability.Imap4Rev1));
      Assert.IsTrue(caps.Contains("XPIG-LATIN"));
    }

    [Test]
    public void TestListResponse()
    {
      var response =
        GetSingleDataResponse("* LIST (\\Noselect) \"/\" ~/Mail/foo\r\n");

      var mailboxList = ImapServerDataResponseParser.FromList(response);

      Assert.IsTrue(mailboxList.NameAttributes.Contains(ImapMailboxFlag.NoSelect));
      Assert.IsTrue(mailboxList.NameAttributes.Contains(@"\Noselect"));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("~/Mail/foo", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);

      Assert.DoesNotThrow(() => mailboxList.ToString());
    }

    [Test]
    public void TestListResponse_Duplicated()
    {
      var response =
        GetSingleDataResponse("* LIST (\\Noselect \\Noselect) \"/\" ~/Mail/foo\r\n");

      var mailboxList = ImapServerDataResponseParser.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Contains(ImapMailboxFlag.NoSelect));

      Assert.IsNull(mailboxList.ChildInfo);
    }

    [Test]
    public void TestListExtendedResponse_WithEmptyExtendedData()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" ()\r\n");

      var mailboxList = ImapServerDataResponseParser.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Contains(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);

      Assert.DoesNotThrow(() => mailboxList.ToString());
    }

    [Test]
    public void TestListExtendedResponse_WithUnknownExtension()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" (\"X-EXTENSION\" ())\r\n");

      var mailboxList = ImapServerDataResponseParser.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Contains(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);
    }

    [Test]
    public void TestListExtendedResponse_WithUnknownExtension_Duplicated()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" (\"X-EXTENSION\" () \"X-EXTENSION\" ())\r\n");

      var mailboxList = ImapServerDataResponseParser.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Contains(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);

      Assert.DoesNotThrow(() => mailboxList.ToString());
    }

    [TestCase("* LIST (\\NonExistent) \"/\" \"Foo\" (\"CHILDINFO\" (\"SUBSCRIBED\"))\r\n")]
    [TestCase("* list (\\nonexistent) \"/\" \"Foo\" (\"childinfo\" (\"subscribed\"))\r\n")]
    public void TestListExtendedResponse_WithChildInfo(string resp)
    {
      var response = GetSingleDataResponse(resp);

      var mailboxList = ImapServerDataResponseParser.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Contains(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNotNull(mailboxList.ChildInfo);
      Assert.IsTrue(mailboxList.ChildInfo.Subscribed);

      Assert.DoesNotThrow(() => mailboxList.ToString());
    }

    [Test]
    public void TestListExtendedResponse_WithUnknownChildInfoExtension()
    {
      var response =
        GetSingleDataResponse("* LIST (\\NonExistent) \"/\" \"Foo\" (\"CHILDINFO\" (\"X-EXTENSION\"))\r\n");

      var mailboxList = ImapServerDataResponseParser.FromList(response);

      Assert.AreEqual(1, mailboxList.NameAttributes.Count);
      Assert.IsTrue(mailboxList.NameAttributes.Contains(ImapMailboxFlag.NonExistent));
      Assert.AreEqual("/", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("Foo", mailboxList.Name);

      Assert.IsNotNull(mailboxList.ChildInfo);
      Assert.IsFalse(mailboxList.ChildInfo.Subscribed);

      Assert.DoesNotThrow(() => mailboxList.ToString());
    }

    [Test]
    public void TestLsubResponse()
    {
      var response =
        GetSingleDataResponse("* LSUB () \".\" #news.comp.mail.misc\r\n");

      var mailboxList = ImapServerDataResponseParser.FromLsub(response);

      Assert.AreEqual(0, mailboxList.NameAttributes.Count);
      Assert.AreEqual(".", mailboxList.HierarchyDelimiter);
      Assert.AreEqual("#news.comp.mail.misc", mailboxList.Name);

      Assert.IsNull(mailboxList.ChildInfo);

      Assert.DoesNotThrow(() => mailboxList.ToString());
    }

    [Test]
    public void TestStatusResponse()
    {
      var response =
        GetSingleDataResponse("* STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292)\r\n");

      var (name, statusAttributes) = ImapServerDataResponseParser.FromStatus(response);

      Assert.AreEqual("blurdybloop", name);

      Assert.IsTrue(statusAttributes.ContainsKey("MESSAGES"));
      Assert.AreEqual("231", statusAttributes["MESSAGES"].GetTextAsString());

      Assert.IsTrue(statusAttributes.ContainsKey("UIDNEXT"));
      Assert.AreEqual("44292", statusAttributes["UIDNEXT"].GetTextAsString());

      Assert.IsFalse(statusAttributes.ContainsKey("RECENT"));
      Assert.IsFalse(statusAttributes.ContainsKey("UNSEEN"));
      Assert.IsFalse(statusAttributes.ContainsKey("UIDVALIDITY"));
    }

    [TestCase("* STATUS INBOX (MESSAGES 1 RECENT 2 UIDNEXT 3 UIDVALIDITY 4 UNSEEN 5 HIGHESTMODSEQ 6)\r\n")]
    [TestCase("* status inbox (messages 1 recent 2 uidnext 3 uidvalidity 4 unseen 5 highestmodseq 6)\r\n")]
    public void TestStatusResponse_CaseInsensitivity(string resp)
    {
      var response = GetSingleDataResponse(resp);

      var (name, statusAttributes) = ImapServerDataResponseParser.FromStatus(response);

      StringAssert.AreEqualIgnoringCase("INBOX", name);

      Assert.IsTrue(statusAttributes.ContainsKey("MESSAGES"));
      Assert.AreEqual("1", statusAttributes["MESSAGES"].GetTextAsString());

      Assert.IsTrue(statusAttributes.ContainsKey("RECENT"));
      Assert.AreEqual("2", statusAttributes["RECENT"].GetTextAsString());

      Assert.IsTrue(statusAttributes.ContainsKey("UIDNEXT"));
      Assert.AreEqual("3", statusAttributes["UIDNEXT"].GetTextAsString());

      Assert.IsTrue(statusAttributes.ContainsKey("UIDVALIDITY"));
      Assert.AreEqual("4", statusAttributes["UIDVALIDITY"].GetTextAsString());

      Assert.IsTrue(statusAttributes.ContainsKey("UNSEEN"));
      Assert.AreEqual("5", statusAttributes["UNSEEN"].GetTextAsString());

      Assert.IsTrue(statusAttributes.ContainsKey("HIGHESTMODSEQ"));
      Assert.AreEqual("6", statusAttributes["HIGHESTMODSEQ"].GetTextAsString());
    }

    [Test]
    public void TestSearchResponse()
    {
      var response =
        GetSingleDataResponse("* SEARCH 2 3 6\r\n");

      var messages = ImapServerDataResponseParser.FromSearch(response, false);

      Assert.IsNull(messages.HighestModSeq);

      CollectionAssert.AreEqual(new long[] { 2, 3, 6 }, messages);
    }

    [Test]
    public void TestSortResponse()
    {
      var response =
        GetSingleDataResponse("* SORT 2 84 882\r\n");

      var messages = ImapServerDataResponseParser.FromSort(response, false);

      Assert.IsNull(messages.HighestModSeq);

      CollectionAssert.AreEqual(new long[] { 2, 84, 882 }, messages);
    }

    [Test]
    public void TestSearchResponse_NoMatch()
    {
      var response =
        GetSingleDataResponse("* SEARCH\r\n");

      var messages = ImapServerDataResponseParser.FromSearch(response, false);

      Assert.IsTrue(messages.IsEmpty);
    }

    [Test]
    public void TestSortResponse_NoMatch()
    {
      var response =
        GetSingleDataResponse("* SORT\r\n");

      var messages = ImapServerDataResponseParser.FromSort(response, false);

      Assert.IsTrue(messages.IsEmpty);
    }

    [Test]
    public void TestSearchResponse_WithModSeq()
    {
      var response =
        GetSingleDataResponse("* SEARCH 2 5 6 7 11 12 18 19 20 23 (MODSEQ 917162500)\r\n");

      var messages = ImapServerDataResponseParser.FromSearch(response, false);

      Assert.AreEqual(917162500UL, messages.HighestModSeq);

      CollectionAssert.AreEqual(new long[] { 2, 5, 6, 7, 11, 12, 18, 19, 20, 23 }, messages);
    }

    [Test]
    public void TestSearchResponse_YahooImapWorkaround()
    {
      var response =
        GetSingleDataResponse("* SEARCH 2 5 6 \r\n"); // ends with extra SP

      var messages = ImapServerDataResponseParser.FromSearch(response, false);

      Assert.IsNull(messages.HighestModSeq);

      CollectionAssert.AreEqual(new long[] { 2, 5, 6 }, messages);
    }

    [Test]
    public void TestSearchResponse_YahooImapWorkaround_NoMatch()
    {
      var response =
        GetSingleDataResponse("* SEARCH \r\n"); // ends with extra SP

      var messages = ImapServerDataResponseParser.FromSearch(response, false);

      Assert.IsNull(messages.HighestModSeq);

      Assert.IsTrue(messages.IsEmpty);
    }

    [Test]
    public void TestThreadResponse_1()
    {
      var response =
        GetSingleDataResponse("* THREAD (2)(3 6 (4 23)(44 7 96))\r\n");

      var threadList = ImapServerDataResponseParser.FromThread(response, false);

      Assert.AreEqual(2, threadList.Children.Count);

      var traversed   = new[] {2, 3, 6, 4, 23, 44, 7, 96};
      var childCounts = new[] {0, 1, 2, 1, 0,  1,  1, 0};
      var index = 0;

      threadList.Traverse(delegate(ImapThreadList list) {
        Assert.AreEqual(traversed[index], list.Number, "traversed {0}", index);
        Assert.AreEqual(childCounts[index], list.Children.Count, "children {0}", index);

        index++;
      });

      Assert.AreEqual(8, index, "count");

      CollectionAssert.AreEqual(traversed, threadList.ToSequenceSet());

      Assert.DoesNotThrow(() => threadList.ToString());
    }

    [Test]
    public void TestThreadResponse_2()
    {
      var response =
        GetSingleDataResponse("* THREAD (166)(167)(168)(169)(172)(170)(171)(173)(174 (175)(176)(178)(181)(180))(179)(177 (183)(182)(188)(184)" +
                              "(185)(186)(187)(189))(190)(191)(192)(193)(194 195)(196 (197)(198))(199)(200 202)(201)(203)(204)(205)(206 207)(208)\r\n");

      var threadList = ImapServerDataResponseParser.FromThread(response, false);

      var traversed   = new[] {
        166, 167, 168, 169, 172, 170, 171, 173, 174, 175, 176, 178, 181, 180, 179, 177, 183, 182, 188, 184,
        185, 186, 187, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 202, 201, 203, 204, 205, 206, 207, 208,
      };
      var childCounts = new[] {
          0,   0,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   8,   0,   0,   0,   0,
          0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,
      };
      var index = 0;

      threadList.Traverse(delegate(ImapThreadList list) {
        Assert.AreEqual(traversed[index], list.Number, "traversed {0}", index);
        Assert.AreEqual(childCounts[index], list.Children.Count, "children {0}", index);

        index++;
      });

      Assert.AreEqual(43, index, "count");

      CollectionAssert.AreEqual(traversed, threadList.ToSequenceSet());

      Assert.DoesNotThrow(() => threadList.ToString());
    }

    [Test]
    public void TestThreadResponse_NoMatch1()
    {
      var response =
        GetSingleDataResponse("* THREAD\r\n");

      var threadList = ImapServerDataResponseParser.FromThread(response, false);

      Assert.AreEqual(0, threadList.Children.Count);

      Assert.DoesNotThrow(() => threadList.ToString());
    }

    [Test]
    public void TestThreadResponse_NoMatch2()
    {
      var response =
        GetSingleDataResponse("* THREAD \r\n"); // Dovecot 2.2.3 bug?

      var threadList = ImapServerDataResponseParser.FromThread(response, false);

      Assert.AreEqual(0, threadList.Children.Count);

      Assert.DoesNotThrow(() => threadList.ToString());
    }

    [Test]
    public void TestFlagsResponse()
    {
      var response =
        GetSingleDataResponse("* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n");

      var flags = ImapServerDataResponseParser.FromFlags(response);

      Assert.AreEqual(5, flags.Count);
      Assert.IsTrue(flags.Contains(ImapMessageFlag.Answered));
      Assert.IsTrue(flags.Contains(ImapMessageFlag.Flagged));
      Assert.IsTrue(flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(flags.Contains(ImapMessageFlag.Seen));
      Assert.IsTrue(flags.Contains(ImapMessageFlag.Draft));
    }

    [Test]
    public void TestFlagsResponse_Duplicated()
    {
      var response =
        GetSingleDataResponse("* FLAGS (\\Answered \\Answered $label2 $label2)\r\n");

      var flags = ImapServerDataResponseParser.FromFlags(response);

      Assert.AreEqual(2, flags.Count);
      Assert.IsTrue(flags.Contains(ImapMessageFlag.Answered));
      Assert.IsTrue(flags.Contains("$label2"));
    }

    [Test]
    public void TestExistsResponse()
    {
      var response =
        GetSingleDataResponse("* 23 EXISTS\r\n");

      Assert.AreEqual(23, ImapServerDataResponseParser.FromExists(response));
    }

    [Test]
    public void TestRecentResponse()
    {
      var response =
        GetSingleDataResponse("* 5 RECENT\r\n");

      Assert.AreEqual(5, ImapServerDataResponseParser.FromRecent(response));
    }

    [Test]
    public void TestExpungeResponse()
    {
      var response =
        GetSingleDataResponse("* 44 EXPUNGE\r\n");

      Assert.AreEqual(44, ImapServerDataResponseParser.FromExpunge(response));
    }

    private static readonly Action<Stream> emptyReadAction = s => { };

    [Test]
    public void TestFetchResponse()
    {
      var response =
        GetSingleDataResponse("* 23 FETCH (FLAGS (\\Seen) RFC822.SIZE 44827)\r\n");

      var message = ImapServerDataResponseParser.FromFetch<ImapMessageFullAttributes>(response);

      Assert.AreEqual(23, message.Sequence);
      Assert.AreEqual(1, message.Flags.Count);
      Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
      Assert.AreEqual(44827, message.Rfc822Size);

      Assert.IsNotNull(message.BodySpecifiers);
      Assert.AreEqual(0, message.BodySpecifiers.Count);

      Assert.IsNull(message.BodyStructure);
      Assert.IsNull(message.Envelope);
      Assert.AreEqual(0, message.Uid);
      Assert.IsNull(message.InternalDate);

      Assert.Throws<InvalidOperationException>(() => message.ReadFirstBody(emptyReadAction));
      Assert.Throws<InvalidOperationException>(() => message.GetFirstBodyLength());

      Assert.Throws<ArgumentNullException>(() => message.ReadBodyOf(null, emptyReadAction));
      Assert.Throws<ArgumentException>(() => message.ReadBodyOf("non-existent section", emptyReadAction));

      Assert.Throws<ArgumentNullException>(() => message.GetBodyLengthOf(null));
      Assert.Throws<ArgumentException>(() => message.GetBodyLengthOf("non-existent section"));

      Assert.DoesNotThrow(() => message.ToString());
    }

    [Test]
    public void TestFetchResponse_DynamicAttributes_OfMessageFullAttributes()
             => TestFetchResponse_DynamicAttributes<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_DynamicAttributes_OfMessageAttributes()
             => TestFetchResponse_DynamicAttributes<ImapMessageAttributes>();

    [Test]
    public void TestFetchResponse_DynamicAttributes_OfMessageDynamicAttributes()
             => TestFetchResponse_DynamicAttributes<ImapMessageDynamicAttributes>();

    private void TestFetchResponse_DynamicAttributes<TMessageAttributes>()
      where TMessageAttributes : ImapMessageAttributesBase, IImapMessageDynamicAttributes
    {
      var response =
        GetSingleDataResponse("* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Seen))\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(4, message.Sequence);
      Assert.AreEqual(3, message.Flags.Count);
      Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Flagged));
      Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));

      if (message is IImapMessageStaticAttributes staticAtt) {
        Assert.AreEqual(0L, staticAtt.Uid);
        Assert.IsNull(staticAtt.InternalDate);
        Assert.AreEqual(0, staticAtt.BinarySizeSpecifiers.Count);
        Assert.Throws<ArgumentException>(() => staticAtt.GetBinarySizeOf(string.Empty));
      }

      if (message is ImapMessageBodyAttributes msgBody) {
        Assert.AreEqual(0, msgBody.BodySpecifiers.Count);

        Assert.Throws<InvalidOperationException>(() => msgBody.GetFirstBodyLength());
        Assert.Throws<InvalidOperationException>(() => msgBody.ReadFirstBody(emptyReadAction));

        Assert.Throws<ArgumentException>(() => msgBody.GetBodyLengthOf(string.Empty));
        Assert.Throws<ArgumentException>(() => msgBody.ReadBodyOf(string.Empty, emptyReadAction));
      }

      Assert.DoesNotThrow(() => message.ToString());
    }

        [Test]
    public void TestFetchResponse_CaseInsensitivity_StaticAttributes_OfMessageFullAttributes()
             => TestFetchResponse_CaseInsensitivity_StaticAttributes<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_CaseInsensitivity_StaticAttributes_OfMessageAttributes()
             => TestFetchResponse_CaseInsensitivity_StaticAttributes<ImapMessageAttributes>();

    [Test]
    public void TestFetchResponse_CaseInsensitivity_StaticAttributes_OfMessageStaticAttributes()
             => TestFetchResponse_CaseInsensitivity_StaticAttributes<ImapMessageStaticAttributes>();

    private void TestFetchResponse_CaseInsensitivity_StaticAttributes<TMessageAttributes>()
      where TMessageAttributes : ImapMessageAttributesBase, IImapMessageStaticAttributes
    {
      var bodystructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL)";
      var envelope = "(" +
                       "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
                       "\"test message\" " +
                       "((\"from address\" NIL \"from\" \"mail.example.com\")) " +
                       "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " +
                       "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " +
                       "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                       "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                       "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                       "\"<in.reply.to@mail.example.com>\" " +
                       "\"<message.id@mail.example.com>\" " +
                       ")";
      var internaldate = "\"14-Jan-2019 20:28:04 +0900\"";
      var response =
        GetSingleDataResponse($"* 4 fetch (uid 4 envelope {envelope} internaldate {internaldate} rfc822.size 2109 bodystructure {bodystructure} binary.size[1.1] 5)\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(4, message.Sequence);

      Assert.AreEqual(4, message.Uid);

      Assert.IsNotNull(message.Envelope);
      Assert.AreEqual("test message", message.Envelope.Subject);

      Assert.IsNotNull(message.InternalDate);
      Assert.AreEqual(new DateTimeOffset(2019, 1, 14, 20, 28, 04, TimeSpan.FromHours(+9.0)), message.InternalDate);

      Assert.AreEqual(2109, message.Rfc822Size);

      Assert.IsNotNull(message.BodyStructure);
      Assert.AreEqual(MimeType.TextPlain, message.BodyStructure.MediaType);

      Assert.AreEqual(1, message.BinarySizeSpecifiers.Count);
      Assert.IsTrue(message.BinarySizeSpecifiers.Contains("1.1"));
      Assert.AreEqual(5, message.GetBinarySizeOf("1.1"));

      Assert.DoesNotThrow(() => message.ToString());
    }

    [Test]
    public void TestFetchResponse_CaseInsensitivity_DynamicAttributes_OfMessageFullAttributes()
             => TestFetchResponse_CaseInsensitivity_DynamicAttributes<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_CaseInsensitivity_DynamicAttributes_OfMessageAttributes()
             => TestFetchResponse_CaseInsensitivity_DynamicAttributes<ImapMessageAttributes>();

    [Test]
    public void TestFetchResponse_CaseInsensitivity_DynamicAttributes_OfMessageDynamicAttributes()
             => TestFetchResponse_CaseInsensitivity_DynamicAttributes<ImapMessageDynamicAttributes>();

    private void TestFetchResponse_CaseInsensitivity_DynamicAttributes<TMessageAttributes>()
      where TMessageAttributes : ImapMessageAttributesBase, IImapMessageDynamicAttributes
    {
      var response =
        GetSingleDataResponse("* 4 fetch (flags (\\Deleted) modseq (9))\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(4, message.Sequence);

      Assert.AreEqual(1, message.Flags.Count);
      Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Deleted));

      Assert.AreEqual(9UL, message.ModSeq);

      Assert.DoesNotThrow(() => message.ToString());
    }

    [Test]
    public void TestFetchResponse_CaseInsensitivity_Body_OfMessageFullAttributes()
             => TestFetchResponse_CaseInsensitivity_Body<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_CaseInsensitivity_Body_OfMessageBodyAttributes()
             => TestFetchResponse_CaseInsensitivity_Body<ImapMessageBodyAttributes>();

    private void TestFetchResponse_CaseInsensitivity_Body<TMessageAttributes>()
      where TMessageAttributes : ImapMessageBodyAttributes
    {
      var response =
        GetSingleDataResponse("* 10 fetch (" +
                              "body[] {4}\r\nbody " +
                              "body[1.1] {7}\r\nbody1.1 " +
                              "binary[2.1] {9}\r\nbinary2.1 " +
                              "rfc822 {6}\r\nrfc822 " +
                              "rfc822.header {13}\r\nrfc822.header " +
                              "rfc822.text {11}\r\nrfc822.text" +
                              ")\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(10L, message.Sequence);

      Assert.IsNotNull(message.BodySpecifiers);
      Assert.AreEqual(6, message.BodySpecifiers.Count);
      Assert.IsTrue(message.BodySpecifiers.Contains("BODY[]"));
      Assert.IsTrue(message.BodySpecifiers.Contains("BODY[1.1]"));
      Assert.IsTrue(message.BodySpecifiers.Contains("BINARY[2.1]"));
      Assert.IsTrue(message.BodySpecifiers.Contains("RFC822"));
      Assert.IsTrue(message.BodySpecifiers.Contains("RFC822.HEADER"));
      Assert.IsTrue(message.BodySpecifiers.Contains("RFC822.TEXT"));

      message.ReadBodyOf("BODY[]", stream => Assert.AreEqual("body", ToString(stream)));
      message.ReadBodyOf("BODY[1.1]", stream => Assert.AreEqual("body1.1", ToString(stream)));
      message.ReadBodyOf("BINARY[2.1]", stream => Assert.AreEqual("binary2.1", ToString(stream)));
      message.ReadBodyOf("RFC822", stream => Assert.AreEqual("rfc822", ToString(stream)));
      message.ReadBodyOf("RFC822.HEADER", stream => Assert.AreEqual("rfc822.header", ToString(stream)));
      message.ReadBodyOf("RFC822.TEXT", stream => Assert.AreEqual("rfc822.text", ToString(stream)));

      string ToString(Stream stream)
      {
        using (var reader = new StreamReader(stream, OctetEncoding.SevenBits)) {
          return reader.ReadToEnd();
        }
      }
    }

    [Test]
    public void TestFetchResponse_BodyStructure_Singlepart()
    {
      var response =
        GetSingleDataResponse("* 7 FETCH (UID 23 BODYSTRUCTURE (\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL))\r\n");

      var message = ImapServerDataResponseParser.FromFetch<ImapMessageAttributes>(response);

      Assert.AreEqual(7L, message.Sequence);
      Assert.AreEqual(23L, message.Uid);

      Assert.DoesNotThrow(() => message.ToString());

      var bodystructure = message.BodyStructure as ImapSinglePartBodyStructure;

      Assert.IsNotNull(bodystructure);
      Assert.IsNull(bodystructure.ParentStructure);
      Assert.IsFalse(bodystructure.IsMultiPart);
      Assert.IsFalse(bodystructure.HasEncapsulatedMessage);
      Assert.AreEqual(string.Empty, bodystructure.Section);
      Assert.AreEqual("text/plain", (string)bodystructure.MediaType);
      Assert.IsNotNull(bodystructure.Parameters);
      Assert.AreEqual(0, bodystructure.Parameters.Count);
      Assert.IsNull(bodystructure.Id);
      Assert.IsNull(bodystructure.Description);
      Assert.AreEqual("7bit", bodystructure.Encoding);
      Assert.AreEqual(2109, bodystructure.Size);
      Assert.AreEqual(54, bodystructure.LineCount);

      Assert.DoesNotThrow(() => bodystructure.ToString());

      var extBodyStructure = bodystructure as ImapExtendedSinglePartBodyStructure;

      Assert.IsNotNull(extBodyStructure);
      Assert.IsNull(extBodyStructure.MD5);
      Assert.IsNull(extBodyStructure.Disposition);
      Assert.IsNotNull(extBodyStructure.Languages);
      Assert.IsEmpty(extBodyStructure.Languages);
      Assert.IsNull(extBodyStructure.Location);
      Assert.IsNotNull(extBodyStructure.Extensions);
      Assert.IsEmpty(extBodyStructure.Extensions);

      Assert.DoesNotThrow(() => extBodyStructure.ToString());
    }

    [Test]
    public void TestFetchResponse_BodyStructure_Multipart()
    {
      var response =
        GetSingleDataResponse("* 7 FETCH (UID 23 BODYSTRUCTURE (" +
                              "(\"text\" \"plain\" (\"format\" \"flowed\" \"charset\" \"iso-2022-jp\" \"reply-type\" \"original\" \"x-param\" \"foo\" \"x-param\" \"bar\") "+
                              "NIL NIL \"7bit\" 75 3 NIL NIL NIL NIL)" +
                              "(\"application\" \"x-zip-compressed\" (\"name\" \"tmp.zip\") " +
                              "NIL NIL \"base64\" 8200 NIL (\"attachment\" (\"filename\" \"tmp.zip\" \"x-param\" \"hoge\" \"x-param\" \"huga\")) NIL NIL) " +
                              "\"mixed\" (\"boundary\" \"----=_NextPart_000_0018_01CA37E0.F7885730\") NIL NIL NIL))\r\n");

      var message = ImapServerDataResponseParser.FromFetch<ImapMessageAttributes>(response);

      Assert.AreEqual(7L, message.Sequence);
      Assert.AreEqual(23L, message.Uid);

      Assert.DoesNotThrow(() => message.ToString());

      var bodystructure = message.BodyStructure as ImapMultiPartBodyStructure;

      Assert.IsNotNull(bodystructure);
      Assert.IsNull(bodystructure.ParentStructure);
      Assert.IsTrue(bodystructure.IsMultiPart);
      Assert.IsFalse(bodystructure.HasEncapsulatedMessage);
      Assert.AreEqual(2, bodystructure.NestedPartStructures.Count);
      Assert.AreEqual(string.Empty, bodystructure.Section);
      Assert.AreEqual("multipart/mixed", (string)bodystructure.MediaType);

      Assert.DoesNotThrow(() => bodystructure.ToString());

      var extBodyStructure = bodystructure as ImapExtendedMultiPartBodyStructure;

      Assert.IsNotNull(extBodyStructure);
      Assert.AreEqual(1, extBodyStructure.Parameters.Count);
      Assert.AreEqual("----=_NextPart_000_0018_01CA37E0.F7885730", extBodyStructure.Parameters["boundary"]);
      Assert.IsNull(extBodyStructure.Disposition);
      Assert.IsNotNull(extBodyStructure.Languages);
      Assert.IsEmpty(extBodyStructure.Languages);
      Assert.IsNull(extBodyStructure.Location);
      Assert.IsNotNull(extBodyStructure.Extensions);
      Assert.IsEmpty(extBodyStructure.Extensions);

      Assert.DoesNotThrow(() => extBodyStructure.ToString());

      var nested = bodystructure.NestedPartStructures;

      var part1 = nested[0] as ImapSinglePartBodyStructure;

      Assert.IsNotNull(part1);
      Assert.AreSame(bodystructure, part1.ParentStructure);
      Assert.IsFalse(part1.IsMultiPart);
      Assert.IsFalse(part1.HasEncapsulatedMessage);
      Assert.AreEqual("1", part1.Section);
      Assert.AreEqual("text/plain", (string)part1.MediaType);
      Assert.AreEqual(4, part1.Parameters.Count);
      Assert.AreEqual("flowed", part1.Parameters["format"]);
      Assert.AreEqual("iso-2022-jp", part1.Parameters["charset"]);
      Assert.AreEqual("original", part1.Parameters["reply-type"]);
      Assert.AreEqual("bar", part1.Parameters["x-param"]);
      Assert.IsNull(part1.Id);
      Assert.IsNull(part1.Description);
      Assert.AreEqual("7bit", part1.Encoding);
      Assert.AreEqual(75L, part1.Size);
      Assert.AreEqual(3L, part1.LineCount);

      var part1ext = part1 as ImapExtendedSinglePartBodyStructure;

      Assert.IsNotNull(part1ext);
      Assert.IsNull(part1ext.MD5);
      Assert.IsNull(part1ext.Disposition);
      Assert.IsNotNull(part1ext.Languages);
      Assert.IsEmpty(part1ext.Languages);
      Assert.IsNull(part1ext.Location);
      Assert.IsNotNull(part1ext.Extensions);
      Assert.IsEmpty(part1ext.Extensions);

      var part2 = nested[1] as ImapSinglePartBodyStructure;

      Assert.IsNotNull(part2);
      Assert.AreSame(bodystructure, part2.ParentStructure);
      Assert.IsFalse(part2.IsMultiPart);
      Assert.IsFalse(part2.HasEncapsulatedMessage);
      Assert.AreEqual("2", part2.Section);
      Assert.AreEqual("application/x-zip-compressed", (string)part2.MediaType);
      Assert.AreEqual(1, part2.Parameters.Count);
      Assert.AreEqual("tmp.zip", part2.Parameters["name"]);
      Assert.IsNull(part2.Id);
      Assert.IsNull(part2.Description);
      Assert.AreEqual("base64", part2.Encoding);
      Assert.AreEqual(8200, part2.Size);
      Assert.AreEqual(0L, part2.LineCount);

      var part2ext = part2 as ImapExtendedSinglePartBodyStructure;

      Assert.IsNotNull(part2ext);
      Assert.IsNull(part2ext.MD5);
      Assert.IsNotNull(part2ext.Disposition);
      Assert.AreEqual("attachment", part2ext.Disposition.Type);
      Assert.IsTrue(part2ext.Disposition.IsAttachment);
      Assert.IsFalse(part2ext.Disposition.IsInline);
      Assert.AreEqual(2, part2ext.Disposition.Parameters.Count);
      Assert.AreEqual("tmp.zip", part2ext.Disposition.Parameters["filename"]);
      Assert.AreEqual("tmp.zip", part2ext.Disposition.FileName);
      Assert.AreEqual("huga", part2ext.Disposition.Parameters["x-param"]);
      Assert.IsNotNull(part2ext.Languages);
      Assert.IsEmpty(part2ext.Languages);
      Assert.IsNull(part2ext.Location);
      Assert.IsNotNull(part2ext.Extensions);
      Assert.IsEmpty(part2ext.Extensions);

      /*
       * IEnumerable
       */
      var expectingEnumeratedStructures = new IImapBodyStructure[] {
        part1,
        part2,
      };
      var index = 0;

      ImapBodyStructureUtils.Traverse(bodystructure, (part) => {
        Assert.AreSame(expectingEnumeratedStructures[index++], part, "enumerated instances are same #{0}", index);
      });

      /*
       * ImapBodyStructureUtils
       */
      var expectingTraversedStructures = new IImapBodyStructure[] {
        part1,
        part2,
      };

      index = 0;

      bodystructure.Traverse(delegate(IImapBodyStructure part) {
        Assert.AreSame(expectingTraversedStructures[index++], part, "traversed instances are same #{0}", index);
      });
    }

    [Test]
    public void TestFetchResponse_BodyStructure_MessageRfc822()
    {
      var response =
        GetSingleDataResponse("* 14 FETCH (UID 30 BODYSTRUCTURE (" +
                              "(\"text\" \"plain\" (\"charset\" \"ISO-2022-JP\") NIL NIL \"7bit\" 6 1 NIL NIL NIL NIL)" +
                              "(\"message\" \"rfc822\" (\"name\" \"=?ISO-2022-JP?B?GyRCRTpJVSVhJUMlOyE8JTgbKEI=?=\") " +
                              "NIL NIL \"7bit\" 188 (NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL) " +
                              "(\"text\" \"plain\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" 121 6 NIL NIL NIL NIL) 10 NIL " +
                              "(\"inline\" (\"filename\" \"ISO-2022-JP''%1B%24%42%45%3A%49%55%25%61%25%43%25%3B%21%3C%25%38%1B%28%42\")) NIL NIL) " +
                              "\"mixed\" (\"boundary\" \"------------040401080108050302040809\") NIL NIL NIL))\r\n");
      var message = ImapServerDataResponseParser.FromFetch<ImapMessageStaticAttributes>(response);

      Assert.AreEqual(14L, message.Sequence);
      Assert.AreEqual(30L, message.Uid);

      Assert.DoesNotThrow(() => message.ToString());

      var bodystructure = message.BodyStructure as ImapMultiPartBodyStructure;

      Assert.IsNotNull(bodystructure);
      Assert.IsNull(bodystructure.ParentStructure);
      Assert.IsTrue(bodystructure.IsMultiPart);
      Assert.IsFalse(bodystructure.HasEncapsulatedMessage);
      Assert.AreEqual(2, bodystructure.NestedPartStructures.Count);
      Assert.AreEqual(string.Empty, bodystructure.Section);
      Assert.AreEqual("multipart/mixed", (string)bodystructure.MediaType);

      Assert.DoesNotThrow(() => bodystructure.ToString());

      var extBodyStructure = bodystructure as ImapExtendedMultiPartBodyStructure;

      Assert.IsNotNull(extBodyStructure);
      Assert.AreEqual(1, extBodyStructure.Parameters.Count);
      Assert.AreEqual("------------040401080108050302040809", extBodyStructure.Parameters["boundary"]);
      Assert.IsNull(extBodyStructure.Disposition);
      Assert.IsNotNull(extBodyStructure.Languages);
      Assert.IsEmpty(extBodyStructure.Languages);
      Assert.IsNull(extBodyStructure.Location);
      Assert.IsNotNull(extBodyStructure.Extensions);
      Assert.IsEmpty(extBodyStructure.Extensions);

      Assert.DoesNotThrow(() => extBodyStructure.ToString());

      var part1 = bodystructure.NestedPartStructures[0] as ImapSinglePartBodyStructure;

      Assert.IsNotNull(part1);
      Assert.AreSame(bodystructure, part1.ParentStructure);
      Assert.IsFalse(part1.IsMultiPart);
      Assert.IsFalse(part1.HasEncapsulatedMessage);
      Assert.AreEqual("1", part1.Section);
      Assert.AreEqual("text/plain", (string)part1.MediaType);
      Assert.AreEqual(1, part1.Parameters.Count);
      Assert.AreEqual("ISO-2022-JP", part1.Parameters["charset"]);
      Assert.IsNull(part1.Id);
      Assert.IsNull(part1.Description);
      Assert.AreEqual("7bit", part1.Encoding);
      Assert.AreEqual(6L, part1.Size);
      Assert.AreEqual(1L, part1.LineCount);

      var part1ext = part1 as ImapExtendedSinglePartBodyStructure;

      Assert.IsNotNull(part1ext);
      Assert.IsNull(part1ext.MD5);
      Assert.IsNull(part1ext.Disposition);
      Assert.IsNotNull(part1ext.Languages);
      Assert.IsEmpty(part1ext.Languages);
      Assert.IsNull(part1ext.Location);
      Assert.IsNotNull(part1ext.Extensions);
      Assert.IsEmpty(part1ext.Extensions);

      var part2 = bodystructure.NestedPartStructures[1] as ImapMessageRfc822BodyStructure;

      Assert.IsNotNull(part2);
      Assert.AreSame(bodystructure, part1.ParentStructure);
      Assert.IsFalse(part2.IsMultiPart);
      Assert.IsTrue(part2.HasEncapsulatedMessage);
      Assert.AreEqual("2", part2.Section);
      Assert.AreEqual("message/rfc822", (string)part2.MediaType);
      Assert.AreEqual(1, part2.Parameters.Count);
      Assert.AreEqual("=?ISO-2022-JP?B?GyRCRTpJVSVhJUMlOyE8JTgbKEI=?=", part2.Parameters["name"]);
      Assert.IsNull(part2.Id);
      Assert.IsNull(part2.Description);
      Assert.AreEqual("7bit", part2.Encoding);
      Assert.AreEqual(188L, part2.Size);

      Assert.IsNotNull(part2.Envelope);
      Assert.IsNull(part2.Envelope.Date);
      Assert.AreEqual("test mail", part2.Envelope.Subject);
      Assert.IsEmpty(part2.Envelope.From);
      Assert.IsEmpty(part2.Envelope.Sender);
      Assert.IsEmpty(part2.Envelope.ReplyTo);
      Assert.IsEmpty(part2.Envelope.To);
      Assert.IsEmpty(part2.Envelope.Cc);
      Assert.IsEmpty(part2.Envelope.Bcc);
      Assert.IsNull(part2.Envelope.InReplyTo);
      Assert.IsNull(part2.Envelope.MessageId);

      var part2encapsulated = part2.EncapsulatedMessageStructure as ImapSinglePartBodyStructure;

      Assert.IsNotNull(part2encapsulated);
      Assert.AreSame(part2, part2encapsulated.ParentStructure);
      Assert.IsFalse(part2encapsulated.IsMultiPart);
      Assert.IsFalse(part2encapsulated.HasEncapsulatedMessage);
      Assert.AreEqual("2.1", part2encapsulated.Section);
      Assert.AreEqual("text/plain", (string)part2encapsulated.MediaType);
      Assert.AreEqual(1, part2encapsulated.Parameters.Count);
      Assert.AreEqual("us-ascii", part2encapsulated.Parameters["charset"]);
      Assert.IsNull(part2encapsulated.Id);
      Assert.IsNull(part2encapsulated.Description);
      Assert.AreEqual("7bit", part2encapsulated.Encoding);
      Assert.AreEqual(121L, part2encapsulated.Size);
      Assert.AreEqual(6L, part2encapsulated.LineCount);

      var part2encapsulatedExt = part2encapsulated as ImapExtendedSinglePartBodyStructure;

      Assert.IsNotNull(part2encapsulatedExt);
      Assert.IsNull(part2encapsulatedExt.MD5);
      Assert.IsNull(part2encapsulatedExt.Disposition);
      Assert.IsNotNull(part2encapsulatedExt.Languages);
      Assert.IsEmpty(part2encapsulatedExt.Languages);
      Assert.IsNull(part2encapsulatedExt.Location);
      Assert.IsNotNull(part2encapsulatedExt.Extensions);
      Assert.IsEmpty(part2encapsulatedExt.Extensions);

      Assert.AreEqual(10L, part2.LineCount);

      var part2ext = part2 as ImapExtendedMessageRfc822BodyStructure;

      Assert.IsNotNull(part2ext);
      Assert.IsNull(part2ext.MD5);
      Assert.IsNotNull(part2ext.Disposition);
      Assert.AreEqual("inline", part2ext.Disposition.Type);
      Assert.IsTrue(part2ext.Disposition.IsInline);
      Assert.AreEqual(1, part2ext.Disposition.Parameters.Count);
      Assert.AreEqual("ISO-2022-JP''%1B%24%42%45%3A%49%55%25%61%25%43%25%3B%21%3C%25%38%1B%28%42",
                      part2ext.Disposition.Parameters["filename"]);
      Assert.AreEqual("ISO-2022-JP''%1B%24%42%45%3A%49%55%25%61%25%43%25%3B%21%3C%25%38%1B%28%42",
                      part2ext.Disposition.FileName);
      Assert.IsNotNull(part2ext.Languages);
      Assert.IsEmpty(part2ext.Languages);
      Assert.IsNull(part2ext.Location);
      Assert.IsNotNull(part2ext.Extensions);
      Assert.IsEmpty(part2ext.Extensions);

      /*
       * IEnumerable
       */
      var expectingEnumeratedStructures = new IImapBodyStructure[] {
        part1,
        part2,
        part2encapsulated,
      };
      var index = 0;

      ImapBodyStructureUtils.Traverse(bodystructure, (part) => {
        Assert.AreSame(expectingEnumeratedStructures[index++], part, "enumerated instances are same #{0}", index);
      });

      var expectingEnumeratedStructuresInPart2 = new IImapBodyStructure[] {
        part2encapsulated,
      };

      index = 0;

      ImapBodyStructureUtils.Traverse(part2, (part) => {
        Assert.AreSame(expectingEnumeratedStructuresInPart2[index++], part, "enumerated instances are same #{0}", index);
      });

      /*
       * ImapBodyStructureUtils
       */
      var expectingTraversedStructures = new IImapBodyStructure[] {
        part1,
        part2,
        part2encapsulated,
      };

      index = 0;

      bodystructure.Traverse(delegate(IImapBodyStructure part) {
        Assert.AreSame(expectingTraversedStructures[index++], part, "traversed instances are same #{0}", index);
      });
    }

    [Test]
    public void TestFetchResponse_Body_Single_OfMessageFullAttributes()
             => TestFetchResponse_Body_Single<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_Body_Single_OfMessageBodyAttributes()
             => TestFetchResponse_Body_Single<ImapMessageBodyAttributes>();

    private void TestFetchResponse_Body_Single<TMessageAttributes>()
      where TMessageAttributes : ImapMessageBodyAttributes
    {
      var response =
        GetSingleDataResponse("* 10 FETCH (UID 26 " +
                              "BODY[] {4}\r\nbody" +
                              ")\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(10L, message.Sequence);

      if (message is IImapMessageStaticAttributes staticAtt)
        Assert.AreEqual(26L, staticAtt.Uid);
      if (message is IImapMessageDynamicAttributes dynamicAtt)
        Assert.AreEqual(0, dynamicAtt.Flags.Count);

      Assert.IsNotNull(message.BodySpecifiers);
      Assert.AreEqual(1, message.BodySpecifiers.Count);
      Assert.IsTrue(message.BodySpecifiers.Contains("BODY[]"));

      Assert.DoesNotThrow(() => message.ReadFirstBody(emptyReadAction));
      Assert.DoesNotThrow(() => message.GetFirstBodyLength());

      Assert.DoesNotThrow(() => message.ToString());

      Assert.AreEqual("body".Length, message.GetBodyLengthOf("body[]"));

      message.ReadBodyOf("BODY[]", stream => Assert.AreEqual("body", ToString(stream)));
      message.ReadBodyOf("body[]", stream => Assert.AreEqual("body", ToString(stream)));

      Assert.AreEqual(message.GetBodyLengthOf("BODY[]"), message.GetFirstBodyLength());

      // XXX: this will fail with ImapTextStreamData
      message.ReadFirstBody(firstBodyStream => {
        message.ReadBodyOf("BODY[]", specificBodyStream => {
          FileAssert.AreEqual(specificBodyStream, firstBodyStream);
        });
      });

      Assert.Throws<ArgumentException>(() => message.ReadBodyOf("non-existent section", emptyReadAction));
      Assert.Throws<ArgumentException>(() => message.GetBodyLengthOf("non-existent section"));

      string ToString(Stream stream)
      {
        using (var reader = new StreamReader(stream, OctetEncoding.SevenBits)) {
          return reader.ReadToEnd();
        }
      }
    }

    [Test]
    public void TestFetchResponse_Body_Multiple_OfMessageFullAttributes()
             => TestFetchResponse_Body_Multiple<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_Body_Multiple_OfMessageBodyAttributes()
             => TestFetchResponse_Body_Multiple<ImapMessageBodyAttributes>();

    private void TestFetchResponse_Body_Multiple<TMessageAttributes>()
      where TMessageAttributes : ImapMessageBodyAttributes
    {
      var response =
        GetSingleDataResponse("* 10 FETCH (UID 26 " +
                              "BODY[1.1] {10}\r\nsection1.1 " +
                              "BODY[1.2] {10}\r\nsection1.2 " +
                              "BINARY[2.1] {10}\r\nsection2.1 " +
                              "RFC822 {6}\r\nrfc822 " +
                              "RFC822.HEADER {6}\r\nheader " +
                              "RFC822.TEXT {4}\r\ntext" +
                              ")\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(10L, message.Sequence);

      if (message is IImapMessageStaticAttributes staticAtt)
        Assert.AreEqual(26L, staticAtt.Uid);
      if (message is IImapMessageDynamicAttributes dynamicAtt)
        Assert.AreEqual(0, dynamicAtt.Flags.Count);

      Assert.IsNotNull(message.BodySpecifiers);
      Assert.AreEqual(6, message.BodySpecifiers.Count);
      Assert.IsTrue(message.BodySpecifiers.Contains("BODY[1.1]"));
      Assert.IsTrue(message.BodySpecifiers.Contains("BODY[1.2]"));
      Assert.IsTrue(message.BodySpecifiers.Contains("BINARY[2.1]"));
      Assert.IsTrue(message.BodySpecifiers.Contains("RFC822"));
      Assert.IsTrue(message.BodySpecifiers.Contains("RFC822.HEADER"));
      Assert.IsTrue(message.BodySpecifiers.Contains("RFC822.TEXT"));

      Assert.DoesNotThrow(() => message.ReadFirstBody(emptyReadAction));
      Assert.DoesNotThrow(() => message.GetFirstBodyLength());

      Assert.DoesNotThrow(() => message.ToString());

      // ---
      Assert.AreEqual("section1.1".Length, message.GetBodyLengthOf("body[1.1]"));

      message.ReadBodyOf("BODY[1.1]", stream => Assert.AreEqual("section1.1", ToString(stream)));
      message.ReadBodyOf("body[1.1]", stream => Assert.AreEqual("section1.1", ToString(stream)));

      // ---
      Assert.AreEqual("section1.2".Length, message.GetBodyLengthOf("body[1.2]"));

      message.ReadBodyOf("BODY[1.2]", stream => Assert.AreEqual("section1.2", ToString(stream)));
      message.ReadBodyOf("body[1.2]", stream => Assert.AreEqual("section1.2", ToString(stream)));

      // ---
      Assert.AreEqual("section2.1".Length, message.GetBodyLengthOf("binary[2.1]"));

      message.ReadBodyOf("BINARY[2.1]", stream => Assert.AreEqual("section2.1", ToString(stream)));
      message.ReadBodyOf("binary[2.1]", stream => Assert.AreEqual("section2.1", ToString(stream)));

      // ---
      Assert.AreEqual("rfc822".Length, message.GetBodyLengthOf("rfc822"));

      message.ReadBodyOf("RFC822", stream => Assert.AreEqual("rfc822", ToString(stream)));
      message.ReadBodyOf("rfc822", stream => Assert.AreEqual("rfc822", ToString(stream)));

      // ---
      Assert.AreEqual("header".Length, message.GetBodyLengthOf("rfc822.header"));

      message.ReadBodyOf("RFC822.HEADER", stream => Assert.AreEqual("header", ToString(stream)));
      message.ReadBodyOf("rfc822.header", stream => Assert.AreEqual("header", ToString(stream)));

      // ---
      Assert.AreEqual("text".Length, message.GetBodyLengthOf("rfc822.text"));

      message.ReadBodyOf("RFC822.TEXT", stream => Assert.AreEqual("text", ToString(stream)));
      message.ReadBodyOf("rfc822.text", stream => Assert.AreEqual("text", ToString(stream)));

      // ---
      Assert.Throws<ArgumentException>(() => message.ReadBodyOf("non-existent section", emptyReadAction));
      Assert.Throws<ArgumentException>(() => message.GetBodyLengthOf("non-existent section"));

      string ToString(Stream stream)
      {
        using (var reader = new StreamReader(stream, OctetEncoding.SevenBits)) {
          return reader.ReadToEnd();
        }
      }
    }

    [Test]
    public void TestFetchResponse_Body_Nil_OfMessageFullAttributes()
             => TestFetchResponse_Body_Nil<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_Body_Nil_OfMessageBodyAttributes()
             => TestFetchResponse_Body_Nil<ImapMessageBodyAttributes>();

    private void TestFetchResponse_Body_Nil<TMessageAttributes>()
      where TMessageAttributes : ImapMessageBodyAttributes
    {
      var response =
        GetSingleDataResponse("* 10 FETCH (UID 26 BODY[] NIL)\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(10L, message.Sequence);

      if (message is IImapMessageStaticAttributes staticAtt)
        Assert.AreEqual(26L, staticAtt.Uid);
      if (message is IImapMessageDynamicAttributes dynamicAtt)
        Assert.AreEqual(0, dynamicAtt.Flags.Count);

      Assert.IsNotNull(message.BodySpecifiers);
      Assert.AreEqual(0, message.BodySpecifiers.Count);
      Assert.IsFalse(message.BodySpecifiers.Contains("BODY[]"));

      Assert.Throws<InvalidOperationException>(() => message.ReadFirstBody(emptyReadAction));
      Assert.Throws<InvalidOperationException>(() => message.GetFirstBodyLength());

      Assert.Throws<ArgumentException>(() => message.ReadBodyOf("BODY[]", emptyReadAction));
      Assert.Throws<ArgumentException>(() => message.GetBodyLengthOf("BODY[]"));

      Assert.Throws<ArgumentException>(() => message.ReadBodyOf("non-existent section", emptyReadAction));
      Assert.Throws<ArgumentException>(() => message.GetBodyLengthOf("non-existent section"));

      Assert.DoesNotThrow(() => message.ToString());
    }

    [Test]
    public void TestFetchResponse_Body_Empty_OfMessageFullAttributes()
             => TestFetchResponse_Body_Empty<ImapMessageFullAttributes>();

    [Test]
    public void TestFetchResponse_Body_Empty_OfMessageBodyAttributes()
             => TestFetchResponse_Body_Empty<ImapMessageBodyAttributes>();

    private void TestFetchResponse_Body_Empty<TMessageAttributes>()
      where TMessageAttributes : ImapMessageBodyAttributes
    {
      var response =
        GetSingleDataResponse("* 10 FETCH (UID 26)\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(10L, message.Sequence);

      if (message is IImapMessageStaticAttributes staticAtt)
        Assert.AreEqual(26L, staticAtt.Uid);
      if (message is IImapMessageDynamicAttributes dynamicAtt)
        Assert.AreEqual(0, dynamicAtt.Flags.Count);

      Assert.IsNotNull(message.BodySpecifiers);
      Assert.AreEqual(0, message.BodySpecifiers.Count);

      Assert.Throws<InvalidOperationException>(() => message.ReadFirstBody(emptyReadAction));
      Assert.Throws<InvalidOperationException>(() => message.GetFirstBodyLength());

      Assert.Throws<ArgumentException>(() => message.ReadBodyOf("non-existent section", emptyReadAction));
      Assert.Throws<ArgumentException>(() => message.GetBodyLengthOf("non-existent section"));

      Assert.DoesNotThrow(() => message.ToString());
    }

    [Test]
    public void TestFetchResponse_BinarySize_OfMessageStaticAttributes()
             => TestFetchResponse_BinarySize<ImapMessageStaticAttributes>();

    [Test]
    public void TestFetchResponse_BinarySize_OfMessageAttributes()
             => TestFetchResponse_BinarySize<ImapMessageAttributes>();

    [Test]
    public void TestFetchResponse_BinarySize_OfMessageFullAttributes()
             => TestFetchResponse_BinarySize<ImapMessageFullAttributes>();

    private void TestFetchResponse_BinarySize<TMessageAttributes>()
      where TMessageAttributes : ImapMessageAttributesBase, IImapMessageStaticAttributes
    {
      var response =
        GetSingleDataResponse("* 10 FETCH (UID 26 BINARY.SIZE[1.1] 12 BINARY.SIZE[1.2] 456)\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(10L, message.Sequence);
      Assert.AreEqual(26L, message.Uid);

      Assert.AreEqual(2, message.BinarySizeSpecifiers.Count);
      Assert.IsTrue(message.BinarySizeSpecifiers.Contains("1.1"));
      Assert.IsTrue(message.BinarySizeSpecifiers.Contains("1.2"));

      Assert.DoesNotThrow(() => {
        switch (message) {
          case ImapMessageFullAttributes msgAttFull: msgAttFull.GetFirstBinarySize(); break;
          case ImapMessageAttributes msgAtt: msgAtt.GetFirstBinarySize(); break;
          case ImapMessageStaticAttributes msgAttStatic: msgAttStatic.GetFirstBinarySize(); break;
        }
      });

      Assert.AreEqual(12L, message.GetBinarySizeOf("1.1"));
      Assert.AreEqual(456L, message.GetBinarySizeOf("1.2"));

      Assert.Throws<ArgumentNullException>(() => message.GetBinarySizeOf(null));
      Assert.Throws<ArgumentException>(() => message.GetBinarySizeOf("non-existent section"));

      Assert.DoesNotThrow(() => message.ToString());
    }

    [Test]
    public void TestFetchResponse_BinarySize_Empty_OfMessageStaticAttributes()
             => TestFetchResponse_BinarySize_Empty<ImapMessageStaticAttributes>();

    [Test]
    public void TestFetchResponse_BinarySize_Empty_OfMessageAttributes()
             => TestFetchResponse_BinarySize_Empty<ImapMessageAttributes>();

    [Test]
    public void TestFetchResponse_BinarySize_Empty_OfMessageFullAttributes()
             => TestFetchResponse_BinarySize_Empty<ImapMessageFullAttributes>();

    private void TestFetchResponse_BinarySize_Empty<TMessageAttributes>()
      where TMessageAttributes : ImapMessageAttributesBase, IImapMessageStaticAttributes
    {
      var response =
        GetSingleDataResponse("* 10 FETCH (UID 26)\r\n");

      var message = ImapServerDataResponseParser.FromFetch<TMessageAttributes>(response);

      Assert.AreEqual(10L, message.Sequence);
      Assert.AreEqual(26L, message.Uid);

      Assert.AreEqual(0, message.BinarySizeSpecifiers.Count);
      Assert.IsFalse(message.BinarySizeSpecifiers.Contains("1.1"));
      Assert.IsFalse(message.BinarySizeSpecifiers.Contains("non-existent section"));

      Assert.Throws<InvalidOperationException>(() => {
        switch (message) {
          case ImapMessageFullAttributes msgAttFull: msgAttFull.GetFirstBinarySize(); break;
          case ImapMessageAttributes msgAtt: msgAtt.GetFirstBinarySize(); break;
          case ImapMessageStaticAttributes msgAttStatic: msgAttStatic.GetFirstBinarySize(); break;
        }
      });

      Assert.Throws<ArgumentException>(() => message.GetBinarySizeOf("1.1"));
      Assert.Throws<ArgumentException>(() => message.GetBinarySizeOf("non-existent section"));
      Assert.Throws<ArgumentNullException>(() => message.GetBinarySizeOf(null));

      Assert.DoesNotThrow(() => message.ToString());
    }

    [Test]
    public void TestNamespaceResponse()
    {
      var response =
        GetSingleDataResponse("* NAMESPACE ((\"\" \"/\")) NIL ((\"Public Folders/\" \"/\"))\r\n");

      var namespaces = ImapServerDataResponseParser.FromNamespace(response);

      Assert.AreEqual(1, namespaces.PersonalNamespaces.Count);
      Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);

      Assert.AreEqual(0, namespaces.OtherUsersNamespaces.Count);

      Assert.AreEqual(1, namespaces.SharedNamespaces.Count);
      Assert.AreEqual("Public Folders/", namespaces.SharedNamespaces[0].Prefix);
      Assert.AreEqual("/", namespaces.SharedNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.SharedNamespaces[0].Extensions.Count);
    }

    [Test]
    public void TestNamespaceResponse_DelimiterNil()
    {
      var response =
        GetSingleDataResponse("* NAMESPACE ((\"\" NIL)) NIL NIL\r\n");

      var namespaces = ImapServerDataResponseParser.FromNamespace(response);

      Assert.AreEqual(1, namespaces.PersonalNamespaces.Count);
      Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
      Assert.IsNull(namespaces.PersonalNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);

      Assert.AreEqual(0, namespaces.OtherUsersNamespaces.Count);

      Assert.AreEqual(0, namespaces.SharedNamespaces.Count);
    }

    [Test]
    public void TestNamespaceResponse_ContainsExtension()
    {
      var response =
        GetSingleDataResponse("* NAMESPACE ((\"\" \"/\")(\"#mh/\" \"/\" \"X-PARAM\" (\"FLAG1\" \"FLAG2\"))) NIL NIL\r\n");

      var namespaces = ImapServerDataResponseParser.FromNamespace(response);

      Assert.AreEqual(2, namespaces.PersonalNamespaces.Count);
      Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);
      Assert.AreEqual("#mh/", namespaces.PersonalNamespaces[1].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[1].HierarchyDelimiter);
      Assert.AreEqual(1, namespaces.PersonalNamespaces[1].Extensions.Count);
      Assert.IsTrue(namespaces.PersonalNamespaces[1].Extensions.ContainsKey("X-PARAM"));
      Assert.AreEqual("FLAG1", namespaces.PersonalNamespaces[1].Extensions["X-PARAM"][0]);
      Assert.AreEqual("FLAG2", namespaces.PersonalNamespaces[1].Extensions["X-PARAM"][1]);

      Assert.AreEqual(0, namespaces.OtherUsersNamespaces.Count);
      Assert.AreEqual(0, namespaces.SharedNamespaces.Count);
    }

    [Test]
    public void TestNamespaceResponse_ContainsExtension_Duplicated()
    {
      var response =
        GetSingleDataResponse("* NAMESPACE ((\"\" \"/\")(\"#mh/\" \"/\" \"X-PARAM\" (\"foo\") \"X-PARAM\" (\"bar\"))) NIL NIL\r\n");

      var namespaces = ImapServerDataResponseParser.FromNamespace(response);

      Assert.AreEqual(2, namespaces.PersonalNamespaces.Count);
      Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
      Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);
      Assert.AreEqual("#mh/", namespaces.PersonalNamespaces[1].Prefix);
      Assert.AreEqual("/", namespaces.PersonalNamespaces[1].HierarchyDelimiter);
      Assert.AreEqual(1, namespaces.PersonalNamespaces[1].Extensions.Count);
      Assert.IsTrue(namespaces.PersonalNamespaces[1].Extensions.ContainsKey("X-PARAM"));
      Assert.AreEqual("bar", namespaces.PersonalNamespaces[1].Extensions["X-PARAM"][0]);

      Assert.AreEqual(0, namespaces.OtherUsersNamespaces.Count);
      Assert.AreEqual(0, namespaces.SharedNamespaces.Count);
    }

    [Test]
    public void TestEnabledResponse()
    {
      var response =
        GetSingleDataResponse("* ENABLED CONDSTORE X-GOOD-IDEA\r\n");

      var caps = ImapServerDataResponseParser.FromEnabled(response);

      Assert.AreEqual(2, caps.Count, "capability count");
      Assert.IsTrue(caps.Contains("CONDSTORE"));
      Assert.IsTrue(caps.Contains("X-GOOD-IDEA"));
    }

    [Test]
    public void TestIdResponse()
    {
      var response =
        GetSingleDataResponse("* ID (\"name\" \"Cyrus\" \"version\" \"1.5\" \"os\" \"sunos\" \"os-version\" \"5.5\" \"support-url\" \"mailto:cyrus-bugs+@andrew.cmu.edu\")\r\n");

      var paramList = ImapServerDataResponseParser.FromId(response);

      Assert.AreEqual(5, paramList.Count, "param count");
      Assert.AreEqual("Cyrus", paramList["name"]);
      Assert.AreEqual("1.5", paramList["version"]);
      Assert.AreEqual("sunos", paramList["os"]);
      Assert.AreEqual("5.5", paramList["os-version"]);
      Assert.AreEqual("mailto:cyrus-bugs+@andrew.cmu.edu", paramList["support-url"]);
    }

    [Test]
    public void TestIdResponse_NilValue()
    {
      var response =
        GetSingleDataResponse("* ID (\"name\" NIL)\r\n");

      var paramList = ImapServerDataResponseParser.FromId(response);

      Assert.AreEqual(1, paramList.Count, "param count");
      Assert.IsNull(paramList["name"]);
    }

    [Test]
    public void TestIdResponse_DuplicatedField()
    {
      var response =
        GetSingleDataResponse("* ID (\"name\" \"foo\" \"name\" \"bar\")\r\n");

      var paramList = ImapServerDataResponseParser.FromId(response);

      Assert.AreEqual(1, paramList.Count, "param count");
      Assert.AreEqual("bar", paramList["name"]);
    }

    [Test]
    public void TestIdResponse_Nil()
    {
      var response =
        GetSingleDataResponse("* ID NIL\r\n");

      var paramList = ImapServerDataResponseParser.FromId(response);

      Assert.AreEqual(0, paramList.Count, "param count");
    }

    [TestCase("* ESEARCH UID COUNT 5 ALL 4:19,21,28\r\n")]
    [TestCase("* esearch uid count 5 all 4:19,21,28\r\n")]
    public void TestESearchResponse_1(string resp)
    {
      var response = GetSingleDataResponse(resp);

      var matched = ImapServerDataResponseParser.FromESearch(response);

      Assert.IsTrue(matched.IsUidSet);
      Assert.IsNull(matched.Max);
      Assert.IsNull(matched.Min);
      Assert.IsNull(matched.HighestModSeq);

      Assert.IsNotNull(matched.Count);
      Assert.AreEqual(5, matched.Count);

      CollectionAssert.AreEqual(new long[] { 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 21, 28 },
                                matched);
    }

    [TestCase("* ESEARCH (TAG \"a567\") UID COUNT 5 ALL 4:19,21,28\r\n")]
    [TestCase("* esearch (tag \"a567\") uid count 5 all 4:19,21,28\r\n")]
    public void TestESearchResponse_2(string resp)
    {
      var response = GetSingleDataResponse(resp);

      var matched = ImapServerDataResponseParser.FromESearch(response);

      Assert.IsTrue(matched.IsUidSet);
      Assert.IsNull(matched.Max);
      Assert.IsNull(matched.Min);
      Assert.IsNull(matched.HighestModSeq);

      Assert.IsNotNull(matched.Count);
      Assert.AreEqual(5, matched.Count);
      Assert.AreEqual("a567", matched.Tag);
    }

    [TestCase("* ESEARCH COUNT 5 ALL 1:17,21\r\n")]
    [TestCase("* esearch count 5 all 1:17,21\r\n")]
    public void TestESearchResponse_3(string resp)
    {
      var response = GetSingleDataResponse(resp);

      var matched = ImapServerDataResponseParser.FromESearch(response);

      Assert.IsFalse(matched.IsUidSet);
      Assert.IsNull(matched.Max);
      Assert.IsNull(matched.Min);
      Assert.IsNull(matched.HighestModSeq);
      Assert.IsNull(matched.Tag);

      CollectionAssert.AreEqual(new long[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 21 },
                                matched);
    }

    [TestCase("* ESEARCH (TAG \"a3\") MIN 2 MAX 23 MODSEQ 917162488\r\n")]
    [TestCase("* esearch (tag \"a3\") min 2 max 23 modseq 917162488\r\n")]
    public void TestESearchResponse_4(string resp)
    {
      var response = GetSingleDataResponse(resp);

      var matched = ImapServerDataResponseParser.FromESearch(response);

      Assert.IsFalse(matched.IsUidSet);
      Assert.AreEqual("a3", matched.Tag);
      Assert.AreEqual(2, matched.Min);
      Assert.AreEqual(23, matched.Max);
      Assert.AreEqual(917162488, matched.HighestModSeq);
      Assert.IsNull(matched.Count);

      CollectionAssert.IsEmpty(matched);
    }

    [Test]
    public void TestQuotaResponse()
    {
      var response =
        GetSingleDataResponse("* QUOTA \"\" (STORAGE 10 512)\r\n");

      var quota = ImapServerDataResponseParser.FromQuota(response);

      Assert.IsNotNull(quota);
      Assert.AreEqual(string.Empty, quota.Root);
      Assert.AreEqual(1, quota.Resources.Count);

      Assert.IsNotNull(quota.Resources[0]);
      Assert.AreEqual("STORAGE", quota.Resources[0].Name);
      Assert.AreEqual(10L, quota.Resources[0].Usage);
      Assert.AreEqual(512L, quota.Resources[0].Limit);
    }

    [Test]
    public void TestQuotaRootResponse_1()
    {
      var response =
        GetSingleDataResponse("* QUOTAROOT INBOX \"\"\r\n");

      var mailbox = ImapServerDataResponseParser.FromQuotaRoot(response, out var quotaRootNames);

      Assert.AreEqual("INBOX", mailbox);
      Assert.IsNotNull(quotaRootNames);
      Assert.AreEqual(1, quotaRootNames.Count);
      Assert.AreEqual(string.Empty, quotaRootNames[0]);
    }

    [Test]
    public void TestQuotaRootResponse_2()
    {
      var response =
        GetSingleDataResponse("* QUOTAROOT comp.mail.mime\r\n");

      var mailbox = ImapServerDataResponseParser.FromQuotaRoot(response, out var quotaRootNames);

      Assert.AreEqual("comp.mail.mime", mailbox);
      Assert.IsNotNull(quotaRootNames);
      Assert.AreEqual(0, quotaRootNames.Count);
    }

    [Test]
    public void TestLanguageResponse_1()
    {
      var response =
        GetSingleDataResponse("* LANGUAGE (DE)\r\n");

      var langTagList = ImapServerDataResponseParser.FromLanguage(response);

      Assert.IsNotNull(langTagList);
      Assert.AreEqual(1, langTagList.Count);
      Assert.AreEqual("DE", langTagList[0]);
    }

    [Test]
    public void TestLanguageResponse_2()
    {
      var response =
        GetSingleDataResponse("* LANGUAGE (EN DE IT i-default)\r\n");

      var langTagList = ImapServerDataResponseParser.FromLanguage(response);

      Assert.IsNotNull(langTagList);
      Assert.AreEqual(4, langTagList.Count);
      Assert.AreEqual("EN", langTagList[0]);
      Assert.AreEqual("DE", langTagList[1]);
      Assert.AreEqual("IT", langTagList[2]);
      Assert.AreEqual("i-default", langTagList[3]);
    }

    [Test]
    public void TestComparatorResponse()
    {
      var response =
        GetSingleDataResponse("* COMPARATOR i;basic\r\n");

      var activeComparator = ImapServerDataResponseParser.FromComparator(response, out var matchingComparators);

      Assert.AreEqual(new ImapCollationAlgorithm("i;basic"), activeComparator);
      Assert.IsEmpty(matchingComparators);
    }

    [Test]
    public void TestComparatorResponse_WithMatchingComparators()
    {
      var response =
        GetSingleDataResponse("* COMPARATOR i;unicode-casemap (i;unicode-casemap i;ascii-casemap)\r\n");

      var activeComparator = ImapServerDataResponseParser.FromComparator(response, out var matchingComparators);

      Assert.AreEqual(ImapCollationAlgorithm.UnicodeCasemap, activeComparator);

      Assert.IsNotNull(matchingComparators);
      Assert.AreEqual(2, matchingComparators.Count);
      Assert.AreEqual(ImapCollationAlgorithm.UnicodeCasemap, matchingComparators[0]);
      Assert.AreEqual(ImapCollationAlgorithm.AsciiCasemap, matchingComparators[1]);
    }

    [Test]
    public void TestMetadataResponse_WithValues1()
    {
      var response =
        GetSingleDataResponse("* METADATA \"\" (/shared/comment \"My comment\")\r\n");

      string mailbox;
      var metadata = ImapServerDataResponseParser.FromMetadataEntryValues(response, out mailbox);

      Assert.AreEqual(string.Empty, mailbox);
      Assert.AreEqual(1, metadata.Count);
      Assert.IsTrue(metadata[0].IsShared);
      Assert.AreEqual("/shared/comment", metadata[0].EntryName);
      //Assert.AreEqual("My comment", metadata[0].Value);
      Assert.AreEqual("My comment", metadata[0].Value.ToString());
    }

    [Test]
    public void TestMetadataResponse_WithValues2()
    {
      var response =
        GetSingleDataResponse("* METADATA \"INBOX\" (/private/comment \"My comment\"" +
                              "/shared/comment \"Its sunny outside!\")\r\n");

      string mailbox;
      var metadata = ImapServerDataResponseParser.FromMetadataEntryValues(response, out mailbox);

      Assert.AreEqual("INBOX", mailbox);
      Assert.AreEqual(2, metadata.Count);

      Assert.IsTrue(metadata[0].IsPrivate);
      Assert.AreEqual("/private/comment", metadata[0].EntryName);
      //Assert.AreEqual("My comment", metadata[0].Value);
      Assert.AreEqual("My comment", metadata[0].Value.ToString());

      Assert.IsTrue(metadata[1].IsShared);
      Assert.AreEqual("/shared/comment", metadata[1].EntryName);
      //Assert.AreEqual("Its sunny outside!", metadata[1].Value);
      Assert.AreEqual("Its sunny outside!", metadata[1].Value.ToString());
    }

    [Test]
    public void TestMetadataResponse_WithoutValues1()
    {
      var response =
        GetSingleDataResponse("* METADATA \"\" /shared/comment\r\n");

      string mailbox;
      var list = ImapServerDataResponseParser.FromMetadataEntryList(response, out mailbox);

      Assert.AreEqual(string.Empty, mailbox);

      Assert.AreEqual(1, list.Count);
      Assert.AreEqual("/shared/comment", list[0]);
    }

    [Test]
    public void TestMetadataResponse_WithoutValues2()
    {
      var response =
        GetSingleDataResponse("* METADATA \"INBOX\" /shared/comment /private/comment\r\n");

      string mailbox;
      var list = ImapServerDataResponseParser.FromMetadataEntryList(response, out mailbox);

      Assert.AreEqual("INBOX", mailbox);

      Assert.AreEqual(2, list.Count);
      Assert.AreEqual("/shared/comment", list[0]);
      Assert.AreEqual("/private/comment", list[1]);
    }
  }
}
