using System;
using NUnit.Framework;

using Smdn.Text;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;

namespace Smdn.Net.Imap4.Protocol {
  [TestFixture]
  public class ImapResponseCodeTests {
    private ImapResponseCode CreateFromResponse(string response)
    {
      using (var receiver = new ImapPseudoResponseReceiver(response)) {
        return (receiver.ReceiveResponse() as ImapUntaggedStatusResponse).ResponseText.Code;
      }
    }

    [Test]
    public void TestEquals()
    {
      var x = ImapResponseCode.Alert;

      Assert.IsTrue(x.Equals(ImapResponseCode.Alert));
      Assert.IsTrue(x.Equals((object)"ALERT"));
      Assert.IsTrue(x.Equals((object)"alert"));
      Assert.IsTrue(x.Equals((object)ImapResponseCode.Alert));
      Assert.IsTrue(x.Equals("ALERT"));
      Assert.IsTrue(x.Equals("alert"));
      Assert.IsTrue(x.Equals(ImapResponseCode.Alert.ToString()));

      Assert.IsFalse(x.Equals(ImapResponseCode.BadCharset));
      Assert.IsFalse(x.Equals((ImapString)ImapResponseCode.Alert.Value));
      Assert.IsFalse(x.Equals(new ImapCapability("ALERT")));
      Assert.IsFalse(x.Equals((object)new ImapCapability("ALERT")));
    }

    [Test]
    public void TestOpEquality()
    {
      var x = ImapResponseCode.Alert;

      Assert.IsTrue(x == ImapResponseCode.Alert);
      Assert.IsFalse(x != ImapResponseCode.Alert);

      Assert.IsFalse(x == ImapResponseCode.ClientBug);
      Assert.IsTrue(x != ImapResponseCode.ClientBug);
    }

    [Test]
    public void TestEquals_WithSubCode()
    {
      var x = ImapResponseCode.MetadataTooMany;

      Assert.IsTrue(x.Equals(ImapResponseCode.MetadataTooMany));
      Assert.IsTrue(x.Equals(CreateFromResponse("* OK [METADATA TOOMANY] text\r\n")));
      Assert.IsFalse(x.Equals(ImapResponseCode.Metadata));
      Assert.IsFalse(x.Equals(ImapResponseCode.MetadataNoPrivate));
    }

    [Test]
    public void TestOpEquality_WithSubCode()
    {
      var x = ImapResponseCode.MetadataTooMany;

      Assert.IsTrue(x == ImapResponseCode.MetadataTooMany);
      Assert.IsFalse(x != ImapResponseCode.MetadataTooMany);
      Assert.IsTrue(x == CreateFromResponse("* OK [METADATA TOOMANY] text\r\n"));
      Assert.IsFalse(x != CreateFromResponse("* OK [METADATA TOOMANY] text\r\n"));

      Assert.IsFalse(x == ImapResponseCode.Metadata);
      Assert.IsTrue(x != ImapResponseCode.Metadata);
      Assert.IsFalse(x == ImapResponseCode.MetadataNoPrivate);
      Assert.IsTrue(x != ImapResponseCode.MetadataNoPrivate);
    }

    [Test]
    public void TestEquals_HasArguments()
    {
      var x = ImapResponseCode.UidValidity;

      Assert.IsTrue(x.Equals(CreateFromResponse("* OK [UIDVALIDITY 1] text\r\n")));
      Assert.IsTrue(x.Equals(CreateFromResponse("* OK [UIDVALIDITY 2] text\r\n")));

      Assert.IsFalse(x.Equals(CreateFromResponse("* OK [UIDNEXT 1] text\r\n")));
      Assert.IsFalse(x.Equals(CreateFromResponse("* OK [UIDNEXT 2] text\r\n")));
    }

    [Test]
    public void TestOpEquality_HasArguments()
    {
      var x = ImapResponseCode.UidValidity;

      Assert.IsTrue(x == CreateFromResponse("* OK [UIDVALIDITY 1] text\r\n"));
      Assert.IsTrue(x == CreateFromResponse("* OK [UIDVALIDITY 2] text\r\n"));
      Assert.IsFalse(x != CreateFromResponse("* OK [UIDVALIDITY 1] text\r\n"));
      Assert.IsFalse(x != CreateFromResponse("* OK [UIDVALIDITY 2] text\r\n"));

      Assert.IsFalse(x == CreateFromResponse("* OK [UIDNEXT 1] text\r\n"));
      Assert.IsFalse(x == CreateFromResponse("* OK [UIDNEXT 2] text\r\n"));
      Assert.IsTrue(x != CreateFromResponse("* OK [UIDNEXT 1] text\r\n"));
      Assert.IsTrue(x != CreateFromResponse("* OK [UIDNEXT 2] text\r\n"));
    }

    [Test]
    public void TestGetKnownOrCreate()
    {
      Assert.AreSame(ImapResponseCode.Alert, CreateFromResponse("* OK [ALERT] text\r\n"));
    }

    [Test]
    public void TestGetKnownOrCreate_HasArguments()
    {
      Assert.AreSame(ImapResponseCode.UidValidity, CreateFromResponse("* OK [UIDVALIDITY 1] text\r\n"));
    }

    [Test]
    public void TestGetKnownOrCreate_WithSubCode()
    {
      Assert.AreSame(ImapResponseCode.MetadataMaxSize, CreateFromResponse("* OK [METADATA MAXSIZE 1024] text\r\n"));
    }

    [Test]
    public void TestGetKnownOrCreate_Unknown()
    {
      Assert.AreEqual("X-UNKNOWN", CreateFromResponse("* OK [X-UNKNOWN] text\r\n"));
    }

    [Test]
    public void TestGetKnownOrCreate_Unknown_HasArguments()
    {
      Assert.AreEqual("X-UNKNOWN", CreateFromResponse("* OK [X-UNKNOWN ARG1 ARG2] text\r\n"));
    }

    [Test]
    public void TestToString()
    {
      Assert.AreEqual("ALERT", ImapResponseCode.Alert.ToString());
      Assert.AreEqual("METADATA LONGENTRIES", ImapResponseCode.MetadataLongEntries.ToString());
    }

    [Test]
    public void TestSerializeBinary()
    {
      TestUtils.Assert.IsSerializableBinaryFormat(ImapResponseCode.Alert, delegate(ImapResponseCode deserialized) {
        Assert.AreEqual(ImapResponseCode.Alert, deserialized);
        Assert.AreNotSame(ImapResponseCode.Alert, deserialized);
      });
    }

    [Test]
    public void TestSerializeBinary_WithSubCode()
    {
      TestUtils.Assert.IsSerializableBinaryFormat(ImapResponseCode.MetadataTooMany, delegate(ImapResponseCode deserialized) {
        Assert.AreEqual(ImapResponseCode.MetadataTooMany, deserialized);
        Assert.AreNotSame(ImapResponseCode.MetadataTooMany, deserialized);
      });
    }

    [Test]
    public void TestSerializeBinary_HasArguments()
    {
      var respCode = CreateFromResponse("* OK [UIDVALIDITY 1] text\r\n");

      TestUtils.Assert.IsSerializableBinaryFormat(respCode, delegate(ImapResponseCode deserialized) {
        Assert.AreEqual(ImapResponseCode.UidValidity, deserialized);
        Assert.AreNotSame(ImapResponseCode.UidValidity, deserialized);
      });
    }
  }
}
