using System;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.DataStructures;

namespace Smdn.Net.Imap4.Protocol {
  [TestFixture]
  public class ImapSearchCriteriaTests : ImapCombinableDataItemsTestsBase {
    [Test]
    public void TestCreate_DateCriteria()
    {
      Assert.AreEqual("BEFORE 25-Feb-2008", Send(ImapSearchCriteria.Before(new DateTime(2008, 2, 25))));
      Assert.AreEqual("ON 25-Feb-2008", Send(ImapSearchCriteria.On(new DateTime(2008, 2, 25))));
      Assert.AreEqual("SINCE 25-Feb-2008", Send(ImapSearchCriteria.Since(new DateTime(2008, 2, 25))));

      Assert.AreEqual("SENTBEFORE 25-Feb-2008", Send(ImapSearchCriteria.SentBefore(new DateTime(2008, 2, 25))));
      Assert.AreEqual("SENTON 25-Feb-2008", Send(ImapSearchCriteria.SentOn(new DateTime(2008, 2, 25))));
      Assert.AreEqual("SENTSINCE 25-Feb-2008", Send(ImapSearchCriteria.SentSince(new DateTime(2008, 2, 25))));
    }

    [Test]
    public void TestCreate_KeywordCriteria()
    {
      Assert.AreEqual("KEYWORD \\Answered", Send(ImapSearchCriteria.Keyword(ImapMessageFlag.Answered)));
      Assert.AreEqual("UNKEYWORD \\Answered", Send(ImapSearchCriteria.Unkeyword(ImapMessageFlag.Answered)));

      Assert.AreEqual("KEYWORD $label1", Send(ImapSearchCriteria.Keyword("$label1")));
      Assert.AreEqual("UNKEYWORD $label1", Send(ImapSearchCriteria.Unkeyword("$label1")));
    }

    [Test]
    public void TestCreate_KeywordCriteria_ArgumentNull()
    {
      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.Keyword((ImapMessageFlag)null), "#1");
      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.Unkeyword((ImapMessageFlag)null), "#2");
      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.Keyword((string)null), "#3");
      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.Unkeyword((string)null), "#4");
    }

    [Test]
    public void TestCreate_KeywordCriteria_ArgumentEmpty()
    {
      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.Keyword(string.Empty), "#1");
      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.Unkeyword(string.Empty), "#2");
    }

    [Test]
    public void TestCreate_KeywordCriteria_ArgumentInvalid()
    {
      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.Keyword("(invalid keyword)"), "#1");
      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.Unkeyword("(invalid keyword)"), "#2");
    }

    [Test]
    public void TestCreate_SizeCriteria()
    {
      Assert.AreEqual("LARGER 72", Send(ImapSearchCriteria.Larger(72)));
      Assert.AreEqual("SMALLER 72", Send(ImapSearchCriteria.Smaller(72)));

      Assert.Throws<ArgumentOutOfRangeException>(() => ImapSearchCriteria.Larger(-1));

      Assert.Throws<ArgumentOutOfRangeException>(() => ImapSearchCriteria.Smaller(-1));
    }

    private delegate ImapSearchCriteria CreateCriteriaMethod(string val);

    [Test]
    public void TestCreate_StringCriteria()
    {
      foreach (var test in new[] {
        new {Criterion = "BCC",     Method = (CreateCriteriaMethod)ImapSearchCriteria.Bcc},
        new {Criterion = "BODY",    Method = (CreateCriteriaMethod)ImapSearchCriteria.Body},
        new {Criterion = "CC",      Method = (CreateCriteriaMethod)ImapSearchCriteria.Cc},
        new {Criterion = "FROM",    Method = (CreateCriteriaMethod)ImapSearchCriteria.From},
        new {Criterion = "SUBJECT", Method = (CreateCriteriaMethod)ImapSearchCriteria.Subject},
        new {Criterion = "TEXT",    Method = (CreateCriteriaMethod)ImapSearchCriteria.Text},
        new {Criterion = "TO",      Method = (CreateCriteriaMethod)ImapSearchCriteria.To},
      }) {
        var val = "query";
        var criteria = test.Method(val);

        Assert.AreEqual($"{test.Criterion} {{5}}\r\n{val}", Send(criteria), test.Criterion);

        Assert.Throws<ArgumentNullException>(() => test.Method(null), test.Criterion);
      }
    }

    [Test]
    public void TestHeader()
    {
      Assert.AreEqual("HEADER \"Message-ID\" {4}\r\nxxxx", Send(ImapSearchCriteria.Header("Message-ID", "xxxx")));
    }

    [Test]
    public void TestHeader_FieldNameNull()
    {
      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.Header(null, "value"));
    }

    [Test]
    public void TestHeader_ValueNull()
    {
      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.Header("field name", null));
    }

    [Test]
    public void TestSequenceSet()
    {
      var criteria = ImapSearchCriteria.SequenceSet(ImapSequenceSet.Of(1, 4, 5));

      Assert.AreEqual("1,4,5", Send(criteria));

      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.SequenceSet(ImapUidSet.Of(1, 4, 5)));

      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.SequenceSet(ImapSequenceSet.Empty()));

      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.SequenceSet(null));
    }

    [Test]
    public void TestUid()
    {
      var criteria = ImapSearchCriteria.Uid(ImapUidSet.Of(1, 4, 5));

      Assert.AreEqual("UID 1,4,5", Send(criteria));

      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.Uid(ImapSequenceSet.Of(1, 4, 5)));

      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.Uid(ImapUidSet.Empty()));

      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.Uid(null));
    }

    private ImapMatchedSequenceSet GetSavedResult()
    {
      ImapMatchedSequenceSet matched = null;

      Session.TestCases.WithMailboxSelectedSession(new[] { "ESEARCH", "SEARCHRES" }, (session, server) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) UNSEEN\r\n",
                                     "$tag OK SEARCH completed, result saved\r\n");

        session.ESearch(out matched, ImapSearchCriteria.Unseen, ImapSearchResultOptions.Save);
      });

      return matched;
    }

    [Test]
    public void TestSequenceSet_FromSavedResult()
    {
      var matched = GetSavedResult();

      Assert.AreEqual("$", matched.Expression);
      Assert.AreEqual("$", Send(ImapSearchCriteria.SequenceSet(matched)));
    }

    [Test]
    public void TestUid_FromSavedResult()
    {
      var matched = GetSavedResult();

      Assert.AreEqual("$", matched.Expression);
      Assert.AreEqual("UID $", Send(ImapSearchCriteria.Uid(matched)));
    }

    [Test]
    public void TestSequenceOrUidSet()
    {
      Assert.AreEqual("1,4,5", Send(ImapSearchCriteria.SequenceOrUidSet(ImapSequenceSet.Of(1, 4, 5))));
      Assert.AreEqual("UID 1,4,5", Send(ImapSearchCriteria.SequenceOrUidSet(ImapUidSet.Of(1, 4, 5))));

      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.SequenceOrUidSet(ImapSequenceSet.Empty()));

      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.SequenceOrUidSet(ImapUidSet.Empty()));

      Assert.Throws<ArgumentNullException>(() => ImapSearchCriteria.SequenceOrUidSet(null));
    }

    [Test]
    public void TestAnd_1()
    {
      var criteria1 = ImapSearchCriteria.SentBefore(new DateTime(2010, 01, 30));
      var criteria2 = ImapSearchCriteria.Recent;
      var expected = "SENTBEFORE 30-Jan-2010 RECENT";

      Assert.AreEqual(expected, Send(criteria1 & criteria2));
      Assert.AreEqual(expected, Send(criteria1.And(criteria2)));
      Assert.AreEqual(expected, Send(ImapSearchCriteria.And(criteria1, criteria2)));
    }

    [Test]
    public void TestAnd_2()
    {
      var criteria1 = ImapSearchCriteria.SentBefore(new DateTime(2010, 01, 30));
      var criteria2 = ImapSearchCriteria.Keyword("label1");

      Assert.AreEqual("SENTBEFORE 30-Jan-2010 KEYWORD label1", Send(criteria1 & criteria2));
      Assert.AreEqual("KEYWORD label1 SENTBEFORE 30-Jan-2010", Send(criteria2 & criteria1));
    }

    [Test]
    public void TestAnd_3()
    {
      var criteria = ImapSearchCriteria.Flagged & ImapSearchCriteria.Undeleted;

      Assert.AreEqual("FLAGGED UNDELETED", Send(criteria));
    }

    [Test]
    public void TestOr_1()
    {
      var criteria1 = ImapSearchCriteria.Recent;
      var criteria2 = ImapSearchCriteria.Keyword("label1");
      var expected = "OR (RECENT) (KEYWORD label1)";

      Assert.AreEqual(expected, Send(criteria1 | criteria2));
      Assert.AreEqual(expected, Send(criteria1.Or(criteria2)));
      Assert.AreEqual(expected, Send(ImapSearchCriteria.Or(criteria1, criteria2)));
    }

    [Test]
    public void TestOr_2()
    {
      var criteria1 = ImapSearchCriteria.SentBefore(new DateTime(2010, 01, 30)) & ImapSearchCriteria.Keyword("label1");
      var criteria2 = ImapSearchCriteria.Recent;

      Assert.AreEqual("OR (SENTBEFORE 30-Jan-2010 KEYWORD label1) (RECENT)", Send(criteria1 | criteria2));
      Assert.AreEqual("OR (RECENT) (SENTBEFORE 30-Jan-2010 KEYWORD label1)", Send(criteria2 | criteria1));
    }

    [Test]
    public void TestNot_1()
    {
      var criteria = ImapSearchCriteria.Recent;
      var expected = "NOT (RECENT)";

      Assert.AreEqual(expected, Send(!criteria));
      Assert.AreEqual(expected, Send(criteria.Not()));
      Assert.AreEqual(expected, Send(ImapSearchCriteria.Not(criteria)));
    }

    [Test]
    public void TestAndOrNot_1()
    {
      var criteria1 = ImapSearchCriteria.SentBefore(new DateTime(2010, 01, 30)) & ImapSearchCriteria.Keyword("label1");
      var criteria2 = ImapSearchCriteria.Recent;
      var criteria = !(criteria1 | criteria2);

      Assert.AreEqual("NOT (OR (SENTBEFORE 30-Jan-2010 KEYWORD label1) (RECENT))", Send(criteria));
    }

    [Test]
    public void TestAndOrNot_2()
    {
      var criteria1 = ImapSearchCriteria.SentBefore(new DateTime(2010, 01, 30)) | ImapSearchCriteria.Keyword("label1");
      var criteria2 = ImapSearchCriteria.Recent;
      var criteria = criteria1 & !criteria2;

      Assert.AreEqual("OR (SENTBEFORE 30-Jan-2010) (KEYWORD label1) NOT (RECENT)", Send(criteria));
    }

    [Test]
    public void TestCombine_ContainsRequiredCapability()
    {
      var criteria1 = ImapSearchCriteria.ModSeq(720162338UL);
      var criteria2 = ImapSearchCriteria.Larger(50000);

      var criteria = !criteria1 | criteria2;

      Assert.AreEqual("OR (NOT (MODSEQ 720162338)) (LARGER 50000)", Send(criteria));

      var capabilitiesCondStore = new ImapPseudoCapabilitySet(ImapCapability.CondStore);

      Assert.IsFalse(ImapPseudoCapabilitySet.Empty.IsCapable(criteria));
      Assert.IsTrue(capabilitiesCondStore.IsCapable(criteria));
    }

    [Test]
    public void TestCreate_WithinCriteria()
    {
      Assert.AreEqual("YOUNGER 17", Send(ImapSearchCriteria.Younger(17L)));
      Assert.AreEqual("OLDER 17", Send(ImapSearchCriteria.Older(17L)));

      Assert.AreEqual("YOUNGER 3600", Send(ImapSearchCriteria.Younger(TimeSpan.FromHours(1.0))));
      Assert.AreEqual("OLDER 3600", Send(ImapSearchCriteria.Older(TimeSpan.FromHours(1.0))));

      Assert.Throws<ArgumentOutOfRangeException>(() => ImapSearchCriteria.Younger(0L));

      Assert.Throws<ArgumentOutOfRangeException>(() => ImapSearchCriteria.Older(0L));
    }
  }
}
