using System;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Urls;

namespace Smdn.Net.Imap4.Protocol {
  [TestFixture]
  public class ImapSearchCriteriaImapUrlTests : ImapCombinableDataItemsTestsBase {
    [Test]
    public void TestFromUri_1()
    {
      //   imap://<iserver>/<enc_mailbox>[uidvalidity][?<enc_search>]
      var uri = new Uri("imap://;AUTH=*@minbari.org/gray%20council?SUBJECT%20shadows");
      var criteria = ImapSearchCriteria.FromUri(uri);

      Assert.AreEqual("gray council", ImapStyleUriParser.GetMailbox(uri));
      Assert.AreEqual("?SUBJECT%20shadows", uri.Query);
      Assert.AreEqual(ImapUriForm.SearchMessages, ImapStyleUriParser.GetUriForm(uri));

      Assert.AreEqual("SUBJECT shadows", Send(criteria));
    }

    [Test]
    public void TestFromUri_2()
    {
      var uri = new Uri("imap://john;AUTH=*@minbari.example.org/babylon5/personel?charset%20UTF-8%20SUBJECT%20%7B14+%7D%0D%0A%D0%98%D0%B2%D0%B0%D0%BD%D0%BE%D0%B2%D0%B0");
      var criteria = ImapSearchCriteria.FromUri(uri);

      Assert.AreEqual(ImapUriForm.SearchMessages, ImapStyleUriParser.GetUriForm(uri));
      Assert.AreEqual("charset UTF-8 SUBJECT {14+}\r\n\xD0\x98\xD0\xB2\xD0\xB0\xD0\xBD\xD0\xBE\xD0\xB2\xD0\xB0", Send(criteria));
    }

    [Test]
    public void TestFromUri_QueryNotPresent()
    {
      var uri = new Uri("imap://localhost/INBOX");
      var criteria = ImapSearchCriteria.FromUri(uri);

      Assert.AreEqual(ImapUriForm.ListMessages, ImapStyleUriParser.GetUriForm(uri));
      Assert.IsNull(criteria);
    }

    [Test]
    public void TestFromUri_EmptyQuery()
    {
      var uri = new Uri("imap://localhost/INBOX?");
      var criteria = ImapSearchCriteria.FromUri(uri);

      Assert.AreEqual(ImapUriForm.SearchMessages, ImapStyleUriParser.GetUriForm(uri));
      Assert.IsNotNull(criteria);
      Assert.AreEqual(string.Empty, Send(criteria));
    }

    [TestCase("SUBJECT {3+}\r\nabc", "SUBJECT {3}\r\nabc")]
    [TestCase("SUBJECT {3+}\r\nabc", "SUBJECT {3+}\r\nabc")]
    [TestCase("FROM {3+}\r\nabc SUBJECT \"{3}\"", "FROM {3}\r\nabc SUBJECT \"{3}\"")]
    [TestCase("FROM {3+}\r\nabc SUBJECT \"{3}\"", "FROM {3+}\r\nabc SUBJECT \"{3}\"")]
    public void TestFromUri_ConvertToNonSynchronized(string expected, string input)
    {
      var baseUri = new Uri("imap://localhost/INBOX");
      var criteria = ImapSearchCriteria.FromUri(new Uri(baseUri, "?" + input));

      Assert.IsNotNull(criteria);
      Assert.AreEqual(expected, Send(criteria));
    }

#if false
    [Test, Ignore("disabled feature")]
    public void TestFromUriConvertToSynchronized()
    {
      var baseUri = new Uri("imap://localhost/INBOX");

      foreach (var test in new[] {
        new {Expected = "SUBJECT {3}\r\nabc",   Input = "SUBJECT {3}\r\nabc"},
        new {Expected = "SUBJECT {3}\r\nabc",   Input = "SUBJECT {3+}\r\nabc"},
        new {Expected = "FROM {3}\r\nabc SUBJECT \"{3+}\"",   Input = "FROM {3}\r\nabc SUBJECT \"{3+}\""},
        new {Expected = "FROM {3}\r\nabc SUBJECT \"{3+}\"",   Input = "FROM {3+}\r\nabc SUBJECT \"{3+}\""},
      }) {
        try {
          var criteria = ImapSearchCriteria.FromUri(new Uri(baseUri, "?" + test.Input)/*, true, true*/);

          Assert.IsNotNull(criteria);
          Assert.AreEqual(test.Expected, criteria.ToString());
        }
        catch (ArgumentException ex) {
          Assert.Fail("ArgumentException thrown: {0} {1}", test.Input, ex);
        }
      }
    }
#endif

    [Test]
    public void TestFromUri_SplitCharset()
    {
      string charset;
      ImapSearchCriteria criteria;

      criteria = ImapSearchCriteria.FromUri(new Uri("imap://;AUTH=*@minbari.org/gray%20council?SUBJECT%20shadows"),
                                            true, out _, out charset);

      Assert.IsNull(charset);
      Assert.AreEqual("SUBJECT shadows", Send(criteria));

      criteria = ImapSearchCriteria.FromUri(new Uri("imap://john;AUTH=*@minbari.example.org/babylon5/personel?charset%20UTF-8%20SUBJECT%20%7B14+%7D%0D%0A%D0%98%D0%B2%D0%B0%D0%BD%D0%BE%D0%B2%D0%B0"),
                                            true, out _, out charset);

      Assert.AreEqual("UTF-8", charset);
      Assert.AreEqual("SUBJECT {14+}\r\n\xD0\x98\xD0\xB2\xD0\xB0\xD0\xBD\xD0\xBE\xD0\xB2\xD0\xB0", Send(criteria));
    }

    [TestCase("SUBJECT {1+}\r\n\"")]
    [TestCase("SUBJECT {2+}\r\n\"\\")]
    [TestCase("SUBJECT {3+}\r\n\"\\x")]
    [TestCase("SUBJECT {5+}\r\n\"\\\"x\\")]
    [TestCase("SUBJECT {6+}\r\n\"\\\"x\\\"")]
    [TestCase("SUBJECT \"\"")]
    [TestCase("SUBJECT \"x\"")]
    [TestCase("SUBJECT \"\\\"x\\\"\"")]
    [TestCase("SUBJECT \"{\"")]
    [TestCase("SUBJECT \"{a\"")]
    [TestCase("SUBJECT \"{0\"")]
    [TestCase("SUBJECT \"{1a\"")]
    [TestCase("SUBJECT \"{1-\"")]
    [TestCase("SUBJECT \"{1+\"")]
    [TestCase("SUBJECT \"{1+1\"")]
    [TestCase("SUBJECT \"{3}x\"")]
    [TestCase("SUBJECT \"{3}\\\\rx\"")]
    public void TestFromUri_ContainsValidQuotedStringAndLiteral(string query)
    {
      var baseUri = new Uri("imap://localhost/INBOX");

      Assert.AreEqual(query, Send(ImapSearchCriteria.FromUri(new Uri(baseUri, "?" + query))));
    }

    [Test]
    public void TestFromUri_ContainsInvalidQuotedString()
    {
      var baseUri = new Uri("imap://localhost/INBOX");

      foreach (var uri in new[] {
        new Uri(baseUri, "?SUBJECT \""),
        new Uri(baseUri, "?SUBJECT \"\\"),
        new Uri(baseUri, "?SUBJECT \"\\x"),
        new Uri(baseUri, "?SUBJECT \"\\\"x\\"),
        new Uri(baseUri, "?SUBJECT \"\\\"x\\\""),
      }) {
        Assert.Throws<ArgumentException>(() => ImapSearchCriteria.FromUri(uri));
      }
    }

    [TestCase("?SUBJECT {")]
    [TestCase("?SUBJECT {a")]
    [TestCase("?SUBJECT {0")]
    [TestCase("?SUBJECT {1a")]
    [TestCase("?SUBJECT {1-")]
    [TestCase("?SUBJECT {1+")]
    [TestCase("?SUBJECT {1+1")]
    [TestCase("?SUBJECT {3}x")]
    [TestCase("?SUBJECT {3}\rx")]
    [TestCase("?SUBJECT {1}\r\n")]
    [TestCase("?SUBJECT {9223372036854775808}\r\n")]
    //
    [TestCase("?SUBJECT {3}\r\nabc{")]
    [TestCase("?SUBJECT {3}\r\nabc{a")]
    [TestCase("?SUBJECT {3}\r\nabc{0")]
    [TestCase("?SUBJECT {3}\r\nabc{1a")]
    [TestCase("?SUBJECT {3}\r\nabc{1-")]
    [TestCase("?SUBJECT {3}\r\nabc{1+")]
    [TestCase("?SUBJECT {3}\r\nabc{1+1")]
    [TestCase("?SUBJECT {3}\r\nabc{3}x")]
    [TestCase("?SUBJECT {3}\r\nabc{3}\rx")]
    [TestCase("?SUBJECT {3}\r\nabc{1}\r\n")]
    [TestCase("?SUBJECT {3}\r\nabc{9223372036854775808}\r\n")]
    public void TestFromUri_ContainsInvalidLiteral(string query)
    {
      var baseUri = new Uri("imap://localhost/INBOX");
      var uri = new Uri(baseUri, query);

      Assert.Throws<ArgumentException>(() => ImapSearchCriteria.FromUri(uri));
    }

#if false
    [Test, Ignore("disabled feature")]
    public void TestFromUriRejectNonSynchronizedLiteral()
    {
      var baseUri = new Uri("imap://localhost/INBOX");

      foreach (var uri in new[] {
        new Uri(baseUri, "?SUBJECT {1}\r\na"),
        new Uri(baseUri, "?SUBJECT {3}\r\nabc"),
        new Uri(baseUri, "?SUBJECT \"{1}\" FROM {1}\r\na"),
        new Uri(baseUri, "?SUBJECT \"{1}\" FROM {3}\r\nabc"),
      }) {
        Assert.Throws<ArgumentException>(() => ImapSearchCriteria.FromUri(uri));
      }
    }
#endif

    [Test]
    public void TestIImapUrlSearchQuery_BuildUrl_ListsOfMessagesFormUri_SearchCriteria()
    {
      var b = new ImapUriBuilder("imap://localhost/", "INBOX", 12345L, ImapSearchCriteria.All);

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345?ALL", b.Uri.AbsoluteUri);

      b.Charset = Encoding.UTF8;

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345?ALL", b.Uri.AbsoluteUri);

      b.SearchCriteria = null;

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345", b.Uri.AbsoluteUri);

      b.Charset = null;

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345", b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestIImapUrlSearchQuery_BuildUrl_ListsOfMessagesFormUri_SearchCriteria_ContainsLiterals()
    {
      var b = new ImapUriBuilder("imap://localhost/", "INBOX", 12345L, ImapSearchCriteria.From("差出人"));

      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(b.Uri));

      b.Charset = Encoding.UTF8;

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345?CHARSET%20utf-8%20FROM%20%7B9+%7D%0D%0A%E5%B7%AE%E5%87%BA%E4%BA%BA",
                      b.Uri.AbsoluteUri);

      Assert.DoesNotThrow(() => ImapSearchCriteria.FromUri(b.Uri));

      // change charset
      b.Charset = Encoding.GetEncoding("shift_jis");

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345?CHARSET%20shift_jis%20FROM%20%7B6+%7D%0D%0A%8D%B7%8Fo%90l",
                      b.Uri.AbsoluteUri);

      Assert.DoesNotThrow(() => ImapSearchCriteria.FromUri(b.Uri));

      b.SearchCriteria = null;

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345",
                      b.Uri.AbsoluteUri);

      b.Charset = null;

      Assert.AreEqual("imap://localhost/INBOX;UIDVALIDITY=12345",
                      b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestConstructImapUriBuilder()
    {
      var uri = "imap://localhost/INBOX?FROM {12+}\r\nfrom address";
      var b = new ImapUriBuilder(uri, ImapSearchCriteria.FromUri);

      Assert.IsNotNull(b.SearchCriteria);
      Assert.IsInstanceOf<ImapSearchCriteria>(b.SearchCriteria);
      Assert.AreEqual("imap://localhost/INBOX?FROM%20%7B12+%7D%0D%0Afrom%20address", b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestIImapUrlSearchQuery_BuildUrl_Example1()
    {
      var b = new ImapUriBuilder();

      b.Scheme = "imap";
      b.Host = "minbari.example.org";
      b.AuthType = ImapAuthenticationMechanism.SelectAppropriate;
      b.Mailbox = "gray council";
      b.SearchCriteria = ImapSearchCriteria.Subject("shadows");

      /*
       Assert.AreEqual("imap://;AUTH=*@minbari.example.org/gray%20council?SUBJECT%20shadows",
                       b.Uri.AbsoluteUri);
      */
      Assert.AreEqual("imap://;AUTH=*@minbari.example.org/gray%20council?SUBJECT%20%7B7+%7D%0D%0Ashadows",
                      b.Uri.AbsoluteUri);

      Assert.DoesNotThrow(() => ImapSearchCriteria.FromUri(b.Uri));
    }

    [Test]
    public void TestIImapUrlSearchQuery_BuildUrl_Example2()
    {
      var b = new ImapUriBuilder();

      b.Scheme = "imap";
      b.Host = "minbari.example.org";
      b.UserName = "john";
      b.AuthType = ImapAuthenticationMechanism.SelectAppropriate;
      b.Mailbox = "babylon5/personel";
      b.Charset = Encoding.UTF8;
      b.SearchCriteria = ImapSearchCriteria.Subject("Иванова");

      /*
      Assert.AreEqual("imap://john;AUTH=*@minbari.example.org/babylon5/personel?" +
                      "charset%20UTF-8%20SUBJECT%20%7B14+%7D%0D%0A%D0%98%D0%B2%" +
                      "D0%B0%D0%BD%D0%BE%D0%B2%D0%B0",
                      b.Uri.AbsoluteUri);
                      */
      Assert.AreEqual("imap://john;AUTH=*@minbari.example.org/babylon5/personel?" +
                      "CHARSET%20utf-8%20SUBJECT%20%7B14+%7D%0D%0A%D0%98%D0%B2%" +
                      "D0%B0%D0%BD%D0%BE%D0%B2%D0%B0",
                      b.Uri.AbsoluteUri);

      Assert.DoesNotThrow(() => ImapSearchCriteria.FromUri(b.Uri));
    }
  }
}
