using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  [TestFixture]
  public class ImapStoreDataItemsTests {
    [Test]
    public void TestCreateReplaceFlags()
    {
      Assert.AreEqual("FLAGS (\\Answered \\Deleted)",
                      ImapStoreDataItems.ReplaceFlags(ImapMessageFlag.Answered, ImapMessageFlag.Deleted).ToString());
      Assert.AreEqual("FLAGS ($label1 $label2)",
                      ImapStoreDataItems.ReplaceFlags("$label1", "$label2").ToString());
      Assert.AreEqual("FLAGS (\\Flagged $label3)",
                      ImapStoreDataItems.ReplaceFlags(new[] { "$label3" }, new[] { ImapMessageFlag.Flagged }).ToString());
      Assert.AreEqual("FLAGS (\\Seen \\Draft $label1 $label2)",
                      ImapStoreDataItems.ReplaceFlags(new ImapPseudoMessageFlagSet(new[] { "$label1", "$label2" }, new[] { ImapMessageFlag.Seen, ImapMessageFlag.Draft })).ToString());
    }

    [Test]
    public void TestCreateReplaceFlagsSilent()
    {
      Assert.AreEqual("FLAGS.SILENT (\\Answered \\Deleted)",
                      ImapStoreDataItems.ReplaceFlagsSilent(ImapMessageFlag.Answered, ImapMessageFlag.Deleted).ToString());
      Assert.AreEqual("FLAGS.SILENT ($label1 $label2)",
                      ImapStoreDataItems.ReplaceFlagsSilent("$label1", "$label2").ToString());
      Assert.AreEqual("FLAGS.SILENT (\\Flagged $label3)",
                      ImapStoreDataItems.ReplaceFlagsSilent(new[] { "$label3" }, new[] { ImapMessageFlag.Flagged }).ToString());
      Assert.AreEqual("FLAGS.SILENT (\\Seen \\Draft $label1 $label2)",
                      ImapStoreDataItems.ReplaceFlagsSilent(new ImapPseudoMessageFlagSet(new[] { "$label1", "$label2" }, new[] { ImapMessageFlag.Seen, ImapMessageFlag.Draft })).ToString());
    }

    [Test]
    public void TestCreateAddFlags()
    {
      Assert.AreEqual("+FLAGS (\\Answered \\Deleted)",
                      ImapStoreDataItems.AddFlags(ImapMessageFlag.Answered, ImapMessageFlag.Deleted).ToString());
      Assert.AreEqual("+FLAGS ($label1 $label2)",
                      ImapStoreDataItems.AddFlags("$label1", "$label2").ToString());
      Assert.AreEqual("+FLAGS (\\Flagged $label3)",
                      ImapStoreDataItems.AddFlags(new[] { "$label3" }, new[] { ImapMessageFlag.Flagged }).ToString());
      Assert.AreEqual("+FLAGS (\\Seen \\Draft $label1 $label2)",
                      ImapStoreDataItems.AddFlags(new ImapPseudoMessageFlagSet(new[] { "$label1", "$label2" }, new[] { ImapMessageFlag.Seen, ImapMessageFlag.Draft })).ToString());
    }

    [Test]
    public void TestCreateAddFlagsSilent()
    {
      Assert.AreEqual("+FLAGS.SILENT (\\Answered \\Deleted)",
                      ImapStoreDataItems.AddFlagsSilent(ImapMessageFlag.Answered, ImapMessageFlag.Deleted).ToString());
      Assert.AreEqual("+FLAGS.SILENT ($label1 $label2)",
                      ImapStoreDataItems.AddFlagsSilent("$label1", "$label2").ToString());
      Assert.AreEqual("+FLAGS.SILENT (\\Flagged $label3)",
                      ImapStoreDataItems.AddFlagsSilent(new[] { "$label3" }, new[] { ImapMessageFlag.Flagged }).ToString());
      Assert.AreEqual("+FLAGS.SILENT (\\Seen \\Draft $label1 $label2)",
                      ImapStoreDataItems.AddFlagsSilent(new ImapPseudoMessageFlagSet(new[] { "$label1", "$label2" }, new[] { ImapMessageFlag.Seen, ImapMessageFlag.Draft })).ToString());
    }

    [Test]
    public void TestCreateRemoveFlags()
    {
      Assert.AreEqual("-FLAGS (\\Answered \\Deleted)",
                      ImapStoreDataItems.RemoveFlags(ImapMessageFlag.Answered, ImapMessageFlag.Deleted).ToString());
      Assert.AreEqual("-FLAGS ($label1 $label2)",
                      ImapStoreDataItems.RemoveFlags("$label1", "$label2").ToString());
      Assert.AreEqual("-FLAGS (\\Flagged $label3)",
                      ImapStoreDataItems.RemoveFlags(new[] { "$label3" }, new[] { ImapMessageFlag.Flagged }).ToString());
      Assert.AreEqual("-FLAGS (\\Seen \\Draft $label1 $label2)",
                      ImapStoreDataItems.RemoveFlags(new ImapPseudoMessageFlagSet(new[] { "$label1", "$label2" }, new[] { ImapMessageFlag.Seen, ImapMessageFlag.Draft })).ToString());
    }

    [Test]
    public void TestCreateRemoveFlagsSilent()
    {
      Assert.AreEqual("-FLAGS.SILENT (\\Answered \\Deleted)",
                      ImapStoreDataItems.RemoveFlagsSilent(ImapMessageFlag.Answered, ImapMessageFlag.Deleted).ToString());
      Assert.AreEqual("-FLAGS.SILENT ($label1 $label2)",
                      ImapStoreDataItems.RemoveFlagsSilent("$label1", "$label2").ToString());
      Assert.AreEqual("-FLAGS.SILENT (\\Flagged $label3)",
                      ImapStoreDataItems.RemoveFlagsSilent(new[] { "$label3" }, new[] { ImapMessageFlag.Flagged }).ToString());
      Assert.AreEqual("-FLAGS.SILENT (\\Seen \\Draft $label1 $label2)",
                      ImapStoreDataItems.RemoveFlagsSilent(new ImapPseudoMessageFlagSet(new[] { "$label1", "$label2" }, new[] { ImapMessageFlag.Seen, ImapMessageFlag.Draft })).ToString());
    }

    [Test]
    public void TestCreate_Empty()
    {
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags(new ImapPseudoMessageFlagSet()));
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags(new string[] {}, new ImapMessageFlag[] {}));
    }

    [Test]
    public void TestCreate_ApplicableFlagsEmpty()
    {
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags(new string[] {}, new ImapMessageFlag[] {ImapMessageFlag.Recent, ImapMessageFlag.AllowedCreateKeywords}));
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags(new ImapPseudoMessageFlagSet(new string[0], new[] { ImapMessageFlag.Recent, ImapMessageFlag.AllowedCreateKeywords })));
    }

    [Test]
    public void TestCreate_ContainsNullKeyword()
    {
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags((string)null));
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags("$label1", null, "$label2"));
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags(new[] { "$label1", null }, new ImapMessageFlag[0]));
    }

    [Test]
    public void TestCreate_ContainsNullFlag()
    {
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags((ImapMessageFlag)null));
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags(ImapMessageFlag.Flagged, null, ImapMessageFlag.Draft));
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlags(new string[0], new[] { ImapMessageFlag.Flagged, null }));
    }

    [Test]
    public void TestCreate_RemoveNonApplicableFlags()
    {
      Assert.AreEqual("FLAGS ($label2)",
                      ImapStoreDataItems.ReplaceFlags(new[] { "$label2" }, new[] { ImapMessageFlag.Recent }).ToString());
      Assert.AreEqual("FLAGS ($label2)",
                      ImapStoreDataItems.ReplaceFlags(new[] { "$label2" }, new[] { ImapMessageFlag.AllowedCreateKeywords }).ToString());
    }

    [Test]
    public void TestCreate_InvalidKeyword()
    {
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlagsSilent("$la bel1"), "#1");
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlagsSilent("$label1\r"), "#2");
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlagsSilent("\0$label1\r"), "#3");
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlagsSilent("[label1]\r"), "#4");
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlagsSilent("(label1)\r"), "#5");
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlagsSilent("label:)\r"), "#6");
      Assert.Throws<ArgumentException>(() => ImapStoreDataItems.ReplaceFlagsSilent("label%\r"), "#7");
    }
  }
}
