// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.IO;
using System.Linq;
using System.Reflection; // RuntimeReflectionExtensions.GetMethodInfo
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.MessageAccessProtocols.Diagnostics;
using Smdn.Text;

namespace Smdn.Net.Connection {
  public abstract class ConnectionBase : IDisposable, IConnectionInfo {
    /*
     * class members
     */
    public static SslProtocols SslProtocolsDefault { get; } =
#if NETFRAMEWORK || NETSTANDARD2_0
      SslProtocols.Default;
#else
      SslProtocols.Tls12;
#endif

    public static async Task<Stream> CreateClientSslStreamAsync(ConnectionBase connection,
                                                                Stream baseStream,
                                                                X509Certificate2Collection clientCertificates,
                                                                SslProtocols enabledSslProtocols,
                                                                RemoteCertificateValidationCallback serverCertificateValidationCallback,
                                                                LocalCertificateSelectionCallback clientCertificateSelectionCallback,
                                                                CancellationToken cancellationToken)
    {
      var sslStream = new SslStream(baseStream,
                                    false,
                                    serverCertificateValidationCallback,
                                    clientCertificateSelectionCallback);

      var taskAuthenticate = sslStream.AuthenticateAsClientAsync(
        connection.host,
        clientCertificates,
        enabledSslProtocols,
        true
      );

      if (await Task.WhenAny(taskAuthenticate, Task.Delay(Timeout.Infinite, cancellationToken)).ConfigureAwait(false) != taskAuthenticate)
        throw new OperationCanceledException(cancellationToken);

      await taskAuthenticate.ConfigureAwait(false);

      ConnectionTrace.Log(connection,
                          "TLS started: {0}, {1} {2}, {3} {4}, {5} {6}{7}",
                          sslStream.SslProtocol,
                          sslStream.CipherAlgorithm,
                          sslStream.CipherStrength,
                          sslStream.HashAlgorithm,
                          sslStream.HashStrength,
                          sslStream.KeyExchangeAlgorithm,
                          sslStream.KeyExchangeStrength,
                          sslStream.IsMutuallyAuthenticated ? ", mutually authenticated" : string.Empty);

      return sslStream;
    }

    private static async Task<IPAddress> DefaultResolveAddress(string hostname, CancellationToken cancellationToken)
    {
      if (IPAddress.TryParse(hostname, out var address))
        return address;

      var addresses = await Dns.GetHostAddressesAsync(hostname).ConfigureAwait(false);

      return addresses.FirstOrDefault(addr => addr.AddressFamily == AddressFamily.InterNetworkV6 ||
                                              addr.AddressFamily == AddressFamily.InterNetwork);
    }

    private static async Task<TcpClient> ConnectClientAsync(string hostname,
                                                            int port,
                                                            ResolveHostAddressCallback resolveHostAddress,
                                                            Func<string, Exception> createResolveHostException,
                                                            CancellationToken cancellationToken)
    {
      var taskResolveHostAddress = resolveHostAddress(hostname, cancellationToken);

      if (await Task.WhenAny(taskResolveHostAddress, Task.Delay(Timeout.Infinite, cancellationToken)).ConfigureAwait(false) != taskResolveHostAddress)
        throw new OperationCanceledException(cancellationToken);

      var address = await taskResolveHostAddress.ConfigureAwait(false);

      if (address == null)
        throw createResolveHostException($"could not resolve hostname: {hostname}");

      var c = new TcpClient();
      var taskConnect = c.ConnectAsync(address, port);

      if (await Task.WhenAny(taskConnect, Task.Delay(Timeout.Infinite, cancellationToken)).ConfigureAwait(false) != taskConnect)
        throw new OperationCanceledException(cancellationToken);

      await taskConnect.ConfigureAwait(false);

      return c;
    }

    /*
     * instance members
     */
    internal protected virtual TraceSource TraceSource => null;

    public bool IsConnected => client != null;

    public int Id { get; }

    public string Host {
      get { ThrowIfNotConnected(); return host; }
    }

    public int Port => RemoteEndPoint.Port;

    public IPEndPoint RemoteEndPoint {
      get { ThrowIfNotConnected(); return client.Client.RemoteEndPoint as IPEndPoint; }
    }

    public IPEndPoint LocalEndPoint {
      get { ThrowIfNotConnected(); return client.Client.LocalEndPoint as IPEndPoint; }
    }

    public int SendTimeout {
      get => sendTimeout;
      set {
        if (value < -1)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(SendTimeout), value);

        sendTimeout = value;

        if (IsConnected)
          SetSendTimeout(value);
      }
    }

    public int ReceiveTimeout {
      get => receiveTimeout;
      set {
        if (value < -1)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(ReceiveTimeout), value);

        receiveTimeout = value;

        if (IsConnected)
          SetReceiveTimeout(value);
      }
    }

    // -1: infinite time-out period
    //  0: minimal time-out period
    private void SetSendTimeout(int newValue) => client.Client.SendTimeout = (newValue == 0) ? 1 : newValue;
    private void SetReceiveTimeout(int newValue) => client.Client.ReceiveTimeout = (newValue == 0) ? 1 : newValue;

    public bool IsSecurePortConnection {
      get { ThrowIfNotConnected(); return isSecurePortConnection; }
    }

    public bool IsSecureConnection {
      get { ThrowIfNotConnected(); return isSecureConnection; }
    }

    protected Stream Stream {
      get { ThrowIfNotConnected(); return connectionStream; }
    }

    private static int connectionIdCount = 0;

    protected ConnectionBase()
    {
      Id = Interlocked.Increment(ref connectionIdCount);

      Interlocked.CompareExchange(ref connectionIdCount, 0, int.MaxValue);
    }

    ~ConnectionBase()
    {
      Dispose(false);
    }

    void IDisposable.Dispose()
    {
      Close();
    }

    public void Close()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    protected virtual void Dispose(bool disposing)
    {
      if (disposing) {
        if (connectionStream != null) {
          connectionStream.Dispose();
          connectionStream = null;
        }

        if (client != null) {
#if NETFRAMEWORK || NETSTANDARD2_0
          client.Close();
#else
          client.Dispose();
#endif
          client = null;

          ConnectionTrace.Log(this, "disconnected from {0}", host);
        }
      }
    }

    protected virtual int GetDefaultPort(bool asSecurePortConnection) => -1;

    public Task ConnectAsync(string hostname,
                             int port,
                             ResolveHostAddressCallback resolveHostAddress = null,
                             UpgradeConnectionStreamCallback createAuthenticatedStream = null,
                             CancellationToken cancellationToken = default)
    {
      if (port == -1)
        port = GetDefaultPort(createAuthenticatedStream != null);

      if (hostname == null)
        throw new ArgumentNullException(nameof(hostname));
      if (port < IPEndPoint.MinPort || IPEndPoint.MaxPort < port)
        throw ExceptionUtils.CreateArgumentMustBeInRange(IPEndPoint.MinPort, IPEndPoint.MaxPort, nameof(port), port);

      if (client != null)
        throw new InvalidOperationException("already connected");

      return ConnectAsyncImpl();

      async Task ConnectAsyncImpl()
      {
        try {
          ConnectionTrace.Log(this, "connecting to {0}:{1}", hostname, port);

          host = hostname;
          client = await ConnectClientAsync(hostname,
                                            port,
                                            resolveHostAddress ?? DefaultResolveAddress,
                                            message => CreateConnectException(message, null) ?? new InvalidOperationException(message),
                                            cancellationToken).ConfigureAwait(false);
          connectionStream = client.GetStream();

          SetSendTimeout(sendTimeout);
          SetReceiveTimeout(receiveTimeout);

          OnConnectionStreamChanged(ConnectionTrace.CreateTracingStream(this, connectionStream));

          ConnectionTrace.Log(this, "connected: {0} ({1})", this.client.Client.RemoteEndPoint, hostname);
        }
        catch (Exception ex) {
          ConnectionTrace.Log(this, ex);

          Close();

          var convertedException = ex is OperationCanceledException ? null : CreateConnectException("connection failed", ex);

          if (convertedException == null)
            throw;

          throw convertedException;
        }

        if (createAuthenticatedStream != null) {
          await UpgradeToSecureConnectionAsync(createAuthenticatedStream).ConfigureAwait(false);

          isSecurePortConnection = true;
        }
      }
    }

    public void Connect(string hostname,
                        int port,
                        ResolveHostAddressCallback resolveHostAddress = null,
                        UpgradeConnectionStreamCallback createAuthenticatedStream = null,
                        CancellationToken cancellationToken = default)
    {
      ConnectAsync(
        hostname,
        port,
        resolveHostAddress,
        createAuthenticatedStream,
        cancellationToken
      ).GetAwaiter().GetResult();
    }

    public Task UpgradeToSecureConnectionAsync(UpgradeConnectionStreamCallback upgradeStream,
                                               CancellationToken cancellationToken = default)
    {
      if (isSecurePortConnection || isSecureConnection)
        throw new InvalidOperationException("already upgraded");
      if (upgradeStream == null)
        throw new ArgumentNullException(nameof(upgradeStream));

      return UpgradeToSecureConnectionImpl();

      async Task UpgradeToSecureConnectionImpl()
      {
        await UpgradeStreamAsync(
          upgradeStream,
          baseException => CreateUpgradeToSecureConnectionException($"upgrading to secure connection failed (callback: {upgradeStream.GetMethodInfo()})", baseException),
          cancellationToken
        ).ConfigureAwait(false);

        isSecureConnection = true;
      }
    }

    public Task UpgradeConnectionAsync(UpgradeConnectionStreamCallback upgradeStream,
                                       CancellationToken cancellationToken = default)
    {
      return UpgradeStreamAsync(
        upgradeStream ?? throw new ArgumentNullException(nameof(upgradeStream)),
        baseException => CreateUpgradeConnectionException($"upgrading stream failed (callback: {upgradeStream.GetMethodInfo()})", baseException),
        cancellationToken
      );
    }

    private async Task UpgradeStreamAsync(UpgradeConnectionStreamCallback upgradeStream,
                                          Converter<Exception, Exception> convertUpgradeException,
                                          CancellationToken cancellationToken)
    {
      try {
        connectionStream = await upgradeStream(this, connectionStream, cancellationToken).ConfigureAwait(false);
      }
      catch (Exception ex) {
        Close();

        ConnectionTrace.Log(this, ex);

        var convertedException = ex is OperationCanceledException ? null : convertUpgradeException(ex);

        if (convertedException == null)
          throw;

        throw convertedException;
      }

      OnConnectionStreamChanged(ConnectionTrace.CreateTracingStream(this, connectionStream));
    }

    protected virtual void OnConnectionStreamChanged(Stream stream) { }
    protected virtual Exception CreateConnectException(string message, Exception baseException) => null;
    protected virtual Exception CreateUpgradeToSecureConnectionException(string message, Exception baseException) => null;
    protected virtual Exception CreateUpgradeConnectionException(string message, Exception baseException) => null;

    public override string ToString()
    {
      if (IsConnected)
        return StringConversion.ToString(GetType(), new (string, object)[] {
          (nameof(Id), Id),
          (nameof(Host), Host),
          (nameof(RemoteEndPoint), RemoteEndPoint),
          (nameof(LocalEndPoint), LocalEndPoint),
        });
      else
        return StringConversion.ToString(GetType(), null);
    }

    protected void ThrowIfNotConnected()
    {
      if (!IsConnected)
        throw new InvalidOperationException("not connected");
    }

    private string host = null;
    private int sendTimeout = Timeout.Infinite;
    private int receiveTimeout = Timeout.Infinite;
    private bool isSecurePortConnection;
    private bool isSecureConnection;
    private TcpClient client = null;
    private Stream connectionStream = null;
  }
}
